/* TeamSCOPE - Software for Collaborative Project Environments.
   Copyright (C) 1999, 2000 Michigan State University.
   Written by Ben Pfaff <pfaffben@msu.edu>.
 
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.
 
   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.
 
   To contact the TeamSCOPE team, send email to scope@cscw.msu.edu, or
   write to Charles Steinfield, Room 436 Communication Arts Building,
   Michigan State University, Department of Telecommunication, East
   Lansing MI 48824-1212. */

import java.applet.*;
import java.awt.*;
import java.awt.event.*;
import java.util.Vector;
import java.io.*;
import java.net.*;

public class ScopeChat
    extends Applet 
    implements ActionListener
{
    static final int historyLen = 100;

    ChatWindow chatWindow;
    Button chatButton;

    NetworkThread networkThread;
  
    Thread blinkThread;
    boolean blinking = false;
    boolean alwaysOnTop = true;

    String status;

    public void init () {
        setBackground (new Color (1.0f, 1.0f, 1.0f));
        setLayout (new BorderLayout ());

        chatButton = new Button ("Chat!");
        add (chatButton, "Center");
	chatButton.addActionListener (this);

        networkThread = new NetworkThread ();
        blinkThread = new BlinkThread ();
    }

    public void popup (boolean up) {
        if (up) {
            if (chatWindow == null)
                new ChatWindow ();
            else
                chatWindow.show ();
            setStatus (status);
        } else if (!up && chatWindow != null) {
            chatWindow.dispose ();
            networkThread.putRequest ("POPU 0", "sending popdown notification",
                                      null);
        }
    }

    public void blink (boolean on) {
        //     System.out.println ("blinking " + on);
        blinking = on;
    }

    public void actionPerformed (ActionEvent e) {
	if (e.getSource () == chatButton) {
	    networkThread.putRequest ("POPU 1", "sending popup request", null);
	}
    }

    public void setStatus (String s) {
        status = s;
        if (chatWindow != null) {
            if (s != null)
                chatWindow.setTitle ("TeamSCOPE Chat - " + s);
            else
                chatWindow.setTitle ("TeamSCOPE Chat");
        }
    }

    public void addLine (String s) {
        ChatWindow cw = chatWindow;
        if (cw != null) {
            HistoryCanvas hc = cw.historyCanvas;
            if (hc != null)
                hc.addLine (s);
        }
    }

    public void enableMenuBar (boolean cond) {
        ChatWindow cw = chatWindow;
        if (cw != null)
            cw.enableMenuBar (cond);
    }

    static final int port_lo = 6555;
    static final int port_hi = 6560;

    class Response {
        int code;
        String remark;
        String content;
    }

    abstract class ProcessThread extends Thread {
        Response response;

        ProcessThread () {
            super ("processor");

            setDaemon (true);
        }
    }

    class NetworkThread
        extends Thread {
        Socket sock = null;
        BufferedReader in;
        PrintWriter out;
        Queue requests;
        Queue responses;
        NetworkReader networkReader;
        boolean deadReader;

        class Request {
            String send;
            String statusMsg;
            ProcessThread processor;
        }

        class NetworkReader
            extends Thread {
            NetworkReader () {
                super ("network-reader");
                setDaemon (true);

                start ();
            }

            public void run () {
                try {
                    /* Read and discard server greeting message. */
                    in.readLine ();

                    if (chatWindow != null)
                        networkThread.putRequest ("POPU 2",
                                                  "sending popup request",
                                                  null);
                    blink (false);

                    for (;;) {
                        Response response = readResponse ();

                        switch (response.code) {
                        case 600:
                            addLine (response.remark);
                            break;

                        case 601:
                            if (chatWindow != null)
                                chatWindow.getUserList ();
                            break;

                        case 602:
                            popup (true);
                            break;

                        case 603:
                            blink (true);
                            break;

                        case 604:
                            blink (false);
                            break;

                        default:
                            putResponse (response);
                            break;
                        }
                    }
                } catch (Exception x) {
                    deadReader = true;

                    synchronized (requests) {
                        requests.notifyAll ();
                    }

                    synchronized (responses) {
                        responses.notifyAll ();
                    }
                }
            }

            String readLine () throws Exception {
                return in.readLine ();
            }

            Response readResponse () throws Exception {
                Response response = new Response ();
                String line = readLine ();

		//                 System.out.println (">> " + line);
                response.code = 0;
                for (int i = 0; i < 3; i++) {
                    int c = line.charAt (i);

                    if (c < '0' || c > '9')
                        throw new Exception ("syntax error");
                    response.code = response.code * 10 + c - '0';
                }

                if (line.charAt (3) == ' ') {
                    response.content = new String ();
                } else if (line.charAt (3) == '-') {
                    String lookFor = line.substring (0, 3) + " ";

                    response.content = new String ();
                    for (;;) {
                        line = readLine ();
                        if (line.substring (0, 4).equals (lookFor))
                            break;

                        if (line.charAt (0) == ' ')
                            response.content += line.substring (1) + "\n";
                        else
                            response.content += line + "\n";
                    }
                } else throw new Exception ("syntax error");

                response.remark = line.substring (4);

                return response;
            }
        }

        NetworkThread () {
            super ("network");
            setDaemon (true);

            start ();
        }

        public void run () {
            int port = port_lo;
	    int amt = 200;

            requests = new Queue ();
            responses = new Queue ();

            for (;;) {
                try {
                    connect (port);
                } catch (Exception x) {
		    if (++port > port_hi) {
                        port = port_lo;

			// Exponential backoff to reduce annoyance
			// to paranoid firewall admins.
			amt *= 2;
			if (amt > 6400)
			  amt = 6400;
		    }
		    
                    try {
                        sleep (amt);
                    } catch (InterruptedException y) {
                        /* Nothing to do. */
                    }

                    continue;
                }

		amt = 200;
                synchronized (this) {
                    flushQueue ();
                    putRequest ("SESS " + getParameter ("session"),
                                "logging in", null);
                }

                for (;;) {
                    try {
                        Request request = getRequest ();
                        handleRequest (request);
                    } catch (Exception x) {
                        break;
                    }

                    if (deadReader)
                        break;
                }
            }
        }

        synchronized public void flushQueue () {
            /* FIXME? May need to clean up after some of the requests
              in the queue. */
            requests.empty ();
        }

        public void putRequest (String send, String statusMsg,
                                ProcessThread processor) {
            synchronized (requests) {
                enableMenuBar (false);

                Request request = new Request ();
                request.send = send;
                request.statusMsg = statusMsg;
                request.processor = processor;
                requests.put (request);

                requests.notifyAll ();
            }
        }

        public Request getRequest () throws Exception {
            synchronized (requests) {
                while (requests.count () == 0 && !deadReader) {
                    enableMenuBar (true);
                    setStatus (null);

                    try {
                        requests.wait ();
                    } catch (InterruptedException x) {
                        /* Ignore. */
                    }
                }

                if (deadReader)
                    throw new Exception ("reader died");

                return (Request) requests.get ();
            }
        }

        public void putResponse (Response response) {
            synchronized (responses) {
                responses.put (response);

                responses.notifyAll ();
            }
        }

        public Response getResponse () throws Exception {
            synchronized (responses) {
                while (responses.count () == 0 && !deadReader) {
                    try {
                        responses.wait ();
                    } catch (InterruptedException x) {
                        /* Ignore. */
                    }
                }

                if (deadReader)
                    throw new Exception ("reader died");

                return (Response) responses.get ();
            }
        }

        public void handleRequest (Request request)
        throws Exception {
            setStatus (request.statusMsg);
            out.print (request.send + "\r\n");
	    out.flush ();
            //       System.out.println ("<< " + request.send);
            if (out.checkError ())
                throw new IOException ("error writing socket");

            Response response = getResponse ();
            if (request.processor != null) {
                request.processor.response = response;
                request.processor.start ();
            }
        }

        public void connect (int port) throws Exception {
            if (sock != null) {
                sock.close ();
                sock = null;
            }

            networkReader = null;

            setStatus ("connecting");

            sock = new Socket (getCodeBase ().getHost (), port);
            out = new PrintWriter (sock.getOutputStream (), true);
            in = new BufferedReader (new InputStreamReader
                                     (sock.getInputStream ()));

            deadReader = false;
            networkReader = new NetworkReader ();
        }
    }

    class ChatWindow
        extends Frame
        implements WindowListener, ItemListener {
        /* Components. */
        HistoryCanvas historyCanvas;
        Scrollbar sb;
        TextField inputLine;
        Label statusBar;
        String lastUserList;

        CheckboxMenuItem onTopMI;
        MenuItem closeMI;
        //	MenuItem usersMI;

        ChatWindow() {
            super ();

            MenuBar mb = new MenuBar ();

            Menu m = new Menu ("Chat");

            //	    m.add ("Instant message...");
            //	    m.add ("Change workgroup...");
            //	    usersMI = m.add (new MenuItem ("User list..."));
            //	    m.addSeparator ();

	    onTopMI = new CheckboxMenuItem ("Always on top");
	    onTopMI.setState (alwaysOnTop);
	    onTopMI.addItemListener (this);
	    m.add (onTopMI);

            closeMI = m.add (new MenuItem ("Close"));

            mb.add (m);

            //	    m = new Menu ("Status");
            //	    CheckboxMenuItem cbmi = new CheckboxMenuItem ("Free to chat");
            //	    cbmi.setState (true);
            //	    m.add (cbmi);
            //	    m.add (new CheckboxMenuItem ("Away"));
            //	    m.add (new CheckboxMenuItem ("Busy"));
            //	    m.add (new CheckboxMenuItem ("Do not disturb"));
            //	    mb.add (m);
            //
            //	    m = new Menu ("Network");
            //	    m.add ("Reconnect");
            //	    mb.add (m);

            setMenuBar (mb);

            sb = new Scrollbar (Scrollbar.VERTICAL, 0, 1, 0, historyLen);
            add (sb, BorderLayout.EAST);

            historyCanvas = new HistoryCanvas (this, sb);
            add (historyCanvas, BorderLayout.CENTER);

            inputLine = new TextField ();
            add (inputLine, BorderLayout.SOUTH);

            enableMenuBar (false);

            pack ();
            show ();

	    addWindowListener (this);

            chatWindow = this;
	    lastUserList = "";
            getUserList ();
        }

        public void itemStateChanged (ItemEvent e) {
            alwaysOnTop = onTopMI.getState ();
        }

        // Ye gods, JDK 1.1 really sucks.
        public void windowClosing (WindowEvent e) {
	    popup (false);
	}
        public void windowClosed (WindowEvent e) {}
        public void windowOpened (WindowEvent e) {}
        public void windowIconified (WindowEvent e) {}
        public void windowDeiconified (WindowEvent e) {}
        public void windowActivated (WindowEvent e) {}
        public void windowDeactivated (WindowEvent e) {}

        public synchronized void enableMenuBar (boolean cond) {
            MenuBar mb = getMenuBar ();

            for (int i = 0; i < mb.countMenus (); i++) {
                Menu m = mb.getMenu (i);

                for (int j = 0; j < m.countItems (); j++) {
                    MenuItem mi = m.getItem (j);

                    if (mi != closeMI)
                        mi.enable (cond);
                }
            }
        }

        public void dispose () {
            chatWindow = null;
            super.dispose ();
        }

        public void getUserList () {
            ProcessThread t;

            t = new ProcessThread () {
                    public void run () {
                        BufferedReader r;
                        String s = "Chatting: ";
                        int count = 0;

                        r = new BufferedReader (new StringReader
                                                (response.content));

                        try {
                            for (;;) {
                                String line = r.readLine ();
                                if (line == null)
                                    break;

                                if (count++ > 0)
                                    s += ", ";
                                s += "\2" + line + "\1";
                            }
                        } catch (Exception x) {
                            /* Nothing to do. */
                        }

                        if (count == 0)
                            s += "(none)";

			if (s != lastUserList) {
			    addLine (s);
			    lastUserList = s;
			}
                    }
                };

            networkThread.putRequest ("WHOM", "getting user list", t);
        }

        public boolean action (Event e, Object what) {
            if (e.target == closeMI)
                popup (false);
            //	    else if (e.target == usersMI)
            //	    showUserList ();
            else if (e.target == inputLine) {
                networkThread.putRequest ("POST " + what, "sending line",
                                          null);
                inputLine.setText ("");
            }
	    else return false;

            return true;
        }

        public Dimension getPreferredSize () {
            return new Dimension (300, 300);
        }
    }

    class UserList
        extends Frame {
    }

    class HistoryCanvas
                extends Canvas
        implements AdjustmentListener, ComponentListener {
        Vector history = new Vector (historyLen);

        ChatWindow parent;
        Scrollbar sb;

        class FontInfo {
            Font font;
            int widths[];
            int height;

            FontInfo (FontMetrics fm) {
                font = fm.getFont ();
                widths = fm.getWidths ();
                height = fm.getHeight ();
            }
        }

        int lnMap[];

        FontInfo plainFont, boldFont;

        HistoryCanvas (ChatWindow p, Scrollbar s) {
            super ();

            parent = p;
            sb = s;

            sb.addAdjustmentListener (this);
            addComponentListener (this);

            plainFont = new FontInfo (getFontMetrics (new Font ("Serif",
                                      Font.PLAIN, 14)));
            boldFont = new FontInfo (getFontMetrics (new Font ("Serif",
                                     Font.BOLD, 14)));

            lnMap = new int[historyLen + 1];
            lnUpdateMap ();
        }

        public synchronized void addLine (String s) {
            if (s.length () == 0)
                return;

            history.addElement (s);
            while (history.size () > historyLen)
                history.removeElementAt (0);

            lnUpdateMap ();
            sb.setValues (lnCountPhys () - lnPhysVisible (), lnPhysVisible (),
                          0, lnCountPhys ());
            repaint ();
        }

        public synchronized String lnGetVirt (int i) {
            if (i >= 0 && i < history.size ())
                return (String) history.elementAt (i);
            else
                return "";
        }

        public synchronized int lnCountVirt () {
            return history.size ();
        }

        public synchronized void lnUpdateMap () {
            int ofs = 0;
            int i;

            for (i = 0; i < historyLen; i++) {
                lnMap[i] = ofs;
                ofs += drawLine ((Graphics) null, 0, lnGetVirt (i), false);
            }
            lnMap[i] = ofs;
        }

        public synchronized int lnCountPhys () {
            return lnVirtToPhys (lnCountVirt ());
        }

        public synchronized int lnPhysToVirt (int ln) {
            for (int i = 0; i <= lnCountVirt (); i++)
                if (ln >= lnMap[i])
                    return i;
            return lnCountVirt ();
        }

        public synchronized int lnVirtToPhys (int ln) {
            return lnMap[ln];
        }

        public synchronized int lnPhysVisible () {
            return getSize ().height / plainFont.height;
        }

        public synchronized void paint (Graphics g) {
            int ln = lnPhysToVirt (sb.getValue ());
            int y = (lnVirtToPhys (ln) - sb.getValue () + 1) * plainFont.height;

            while (y < getSize ().height)
                y += plainFont.height * drawLine (g, y, lnGetVirt (ln++), true);
        }

        public synchronized int drawLine (Graphics g, int y, String s,
                                          boolean draw) {
            int maxWidth = getSize ().width;
            int n = 0;
            int w = 0;

            int start = 0;
            FontInfo startFont = plainFont;

            int lastSpace = -1;
            FontInfo lastSpaceFont = null;

            FontInfo f = plainFont;

            for (int i = 0; i < s.length (); i++) {
                char c = s.charAt (i);

                if (c == ' ') {
                    lastSpace = i;
                    lastSpaceFont = f;
                } else if (c == '\1') {
                    f = plainFont;
                    continue;
                } else if (c == '\2') {
                    f = boldFont;
                    continue;
                }

                if (w + f.widths[c] <= maxWidth) {
                    w += f.widths[c];
                    continue;
                }

                if (lastSpace != -1) {
                    i = lastSpace;
                    f = lastSpaceFont;
                } else
                    i--;

                if (draw)
                    reallyDraw (g, y, s, start, i, startFont);
                n++;
                start = i + 1;
                startFont = f;
                lastSpace = -1;
                y += f.height;
                w = 0;
            }

            if (draw)
                reallyDraw (g, y, s, start, s.length (), f);
            n++;

            return n;
        }

        public synchronized void reallyDraw (Graphics g, int y, String s,
                                             int start, int end, FontInfo f) {
            int b = start;
            int sx = 0;
            int x = 0;

            g.setFont (f.font);
            for (int i = start; i < end; i++) {
                char c = s.charAt (i);

                if (c != '\1' && c != '\2') {
                    x += f.widths[c];
                    continue;
                }

                if (y > 0)
                    g.drawString (s.substring (b, i), sx, y);
                sx = x;

                if (c == '\1')
                    f = plainFont;
                else if (c == '\2')
                    f = boldFont;
                g.setFont (f.font);
                b = i + 1;
            }

            if (y > 0)
                g.drawString (s.substring (b, end), sx, y);
        }

        public synchronized void adjustmentValueChanged (AdjustmentEvent e) {
            repaint ();
        }

        public synchronized void componentResized (ComponentEvent e) {
            lnUpdateMap ();
            sb.setVisibleAmount (lnPhysVisible ());
            sb.setMaximum (lnCountPhys ());
        }
        public void componentHidden (ComponentEvent e) {}

        public void componentMoved (ComponentEvent e) {}

        public void componentShown (ComponentEvent e) {
            componentResized (e);
        }
    }

    class BlinkThread
        extends Thread {
        BlinkThread () {
            super ("blink");
            setDaemon (true);

            start ();
        }

        void pause () {
            try {
                sleep (200);
            } catch (InterruptedException y) {
                /* Nothing to do. */
            }
        }

        public void run () {
            for (;;) {
                if (blinking && chatWindow == null && chatButton != null)
                    chatButton.setLabel ("");
                pause ();
                if (chatButton != null)
                    chatButton.setLabel ("Chat!");
                pause ();

		if (alwaysOnTop && chatWindow != null)
                    chatWindow.toFront ();
            }
        }
    }
}

/* Why in the hell doesn't Java have a queue type? */
class Queue
    extends Vector {
    public synchronized int count () {
        return elementCount;
    }

    public synchronized void empty () {
        removeAllElements ();
    }

    public synchronized Object peek () {
        return elementAt (0);
    }

    public synchronized Object get () {
        Object o = elementAt (0);
        removeElementAt (0);
        return o;
    }

    public synchronized void put (Object o) {
        addElement (o);
    }
}

/*
  Local Variables:
  compile-command: "javac ScopeChat.java"
  End:
*/
