/* TeamSCOPE - Software for Collaborative Project Environments.
   Copyright (C) 1999, 2000 Michigan State University.
   Written by Ben Pfaff <pfaffben@msu.edu>.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To contact the TeamSCOPE team, send email to scope@cscw.msu.edu, or
   write to Charles Steinfield, Room 436 Communication Arts Building,
   Michigan State University, Department of Telecommunication, East
   Lansing MI 48824-1212. */

#include <config.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include "alloc.h"
#include "buffer.h"
#include "conf.h"
#include "main.h"
#include "session.h"
#include "str.h"

/* See RFC 959 for the format of the TeamSCOPE chat responses. */

/* Command flags. */
enum
  {
    CF_ARG = 001,		/* Argument required. */
    CF_LOGIN = 002,		/* Must be logged in. */
  };

typedef void cmd_handler (struct session *, const char *arg);

/* Command definition. */
struct cmd
  {
    char name[5];		/* Command name. */
    cmd_handler *handler;	/* Callback. */
    int flags;			/* Any of CF_*. */
  };

#if __CHECKER__
static cmd_handler cmd_garb;
#endif
static cmd_handler cmd_help, cmd_post, cmd_quit, cmd_sess;
static cmd_handler cmd_whom, cmd_popu;

/* All the supported commands. */
struct cmd cmd_list[] = 
  {
#if __CHECKER__
    {"GARB", cmd_garb, 0},
#endif
    {"HELP", cmd_help, 0},
    {"POST", cmd_post, CF_LOGIN | CF_ARG},
    {"QUIT", cmd_quit, 0},
    {"SESS", cmd_sess, CF_ARG},
    {"WHOM", cmd_whom, CF_LOGIN},
    {"POPU", cmd_popu, CF_LOGIN | CF_ARG},
  };

static void post (struct session *s, int me_too, const char *string, ...);
static void update_poppage (const char *workgroup);

/* Performs a state transition on session S to NEW_STATE.
   Disallows invalid state transitions. */
void
transition (struct session *s, int new_state)
{
  if (new_state <= s->state)
    return;

  if (s->state == CLS_READY)
    {
      post (s, 0, "601 Userlist change notify\r\n");

      log ("Exit: %s, %s", s->username, s->workgroup);
    }

  s->state = new_state;
}

/* Handles input line LINE from session S. */
void
input (struct session *s, const char *line)
{
  unsigned i;
  
  for (i = 0; i < sizeof (cmd_list) / sizeof (cmd_list[0]); i++)
    if (!strncmp (line, cmd_list[i].name, 4))
      {
	const struct cmd *const cmd = &cmd_list[i];
	const char *arg = line + 4;
	
	if (*arg == ' ')
	  arg++;
	else
	  arg = NULL;

	if ((cmd->flags & CF_LOGIN) && s->state == CLS_INIT)
	  buf_write_string (s, "530 Not logged in\r\n");
	else if ((cmd->flags & CF_ARG) && arg == NULL)
	  buf_write_string (s, "501 Argument required\r\n");
	else if (!(cmd->flags & CF_ARG) && arg != NULL)
	  buf_write_string (s, "501 Command does not take an argument\r\n");
	else
	  cmd->handler (s, arg);

	return;
      }

  buf_write_string (s, "500 Command unrecognized; HELP for command list\r\n");
}

/* Sends STRING to all the users logged in for chat in the same
   workgroup as the user in session S.  If ME_TOO is nonzero then the
   post is sent to the posting user, otherwise it is not. */
static void
post (struct session *s, int me_too, const char *string, ...)
{
  struct line *line;
  int i;

  line = NULL;
  for (i = 0; i < n_sessions; i++)
    {
      struct session *const t = sessions[i];

      if ((t != s || me_too)
	  && t->state == CLS_READY
	  && !strcmp (s->workgroup, t->workgroup))
	{
	  if (line == NULL)
	    {
	      char buf[BUFSIZ];
	      va_list args;

	      va_start (args, string);
	      vsnprintf (buf, BUFSIZ, string, args);
	      va_end (args);

	      line = buf_create_line (xstrdup (buf));
	    }

	  buf_write_line (t, line);
	}
    }
}

/* Attempts to lock file descriptor FD with lock type TYPE and returns
   nonzero only if successful.  Note: this will block until the lock
   becomes available, so failure is really failure, not merely a
   temporary condition. */
int 
lock (int type, int fd)
{
  /* See _Advanced Programming in the UNIX Environment_,
     p. 368-370. */
  struct flock lock;

  lock.l_type = type;
  lock.l_start = 0;
  lock.l_whence = SEEK_SET;
  lock.l_len = 0;

  do
    {
      if (fcntl (fd, F_SETLKW, &lock) == 0)
	return 1;
    }
  while (errno == EINTR);

  return 0;
}

/* Command handlers. */

#if __CHECKER__
#include <checker_api.h>

/* Handle GARB command by asking Checker to run its garbage detector
   code. */
static void
cmd_garb (struct session *s unused, const char *arg unused)
{
  __chkr_garbage_detector ();
}
#endif /* __CHECKER__ */

/* Handle HELP command. */
static void
cmd_help (struct session *s, const char *arg unused)
{
  static const char *help =
    "214-The following commands are recognized.\r\n"
    " HELP   POST   SESS   QUIT   WHOM   POPU\r\n"
    "214 Direct comments to scope@cscw.msu.edu.\r\n";
  
  buf_write_string (s, help);
}

/* Handle POST command by posting a message to the other users. */
static void
cmd_post (struct session *s, const char *arg)
{
  buf_write_string (s, "250 Posted\r\n");
  if (strlen (arg) == 0)
    return;

  /* Post the message and log it. */
  post (s, 1, "600 \2%s>\1 %s\r\n", s->username, arg);
  log ("[%s] %s> %s", s->hostinfo, s->username, arg);
}

/* Handle QUIT command. */
static void 
cmd_quit (struct session *s, const char *arg unused)
{
  buf_write_string (s, "221 Goodbye\r\n");
  transition (s, CLS_CLOSING);
}

/* Handles the SESS (session) command, which takes an argument that is
   a TeamSCOPE session-id value.  Verifies that the session exists and
   reads the user and workgroup from it, authenticating the user. */
static void 
cmd_sess (struct session *s, const char *arg)
{
  char buf[BUFSIZ];
  FILE *f;

  /* Verify that the session-id is in the correct form. */
  {
    const char *cp;
    
    for (cp = arg; *cp; cp++)
      if (!isdigit ((unsigned char) *cp)
	  && (*cp < 'A' || *cp > 'F')
	  && *cp != '-')
	break;

    if (*cp || cp - arg != 41)
      {
	buf_write_string (s, "530 Malformed session ID\r\n");
	return;
      }
  }

  /* Open the session file. */
  snprintf (buf, BUFSIZ, "%s/%s", TRUSTED_STATE_DIR, arg);
  f = fopen (buf, "r");
  if (f == NULL)
    {
      if (errno == ENOENT)
	buf_write_string (s, "530 Session not found\r\n");
      else
	buf_write_printf (s, "530 fopen: %s\r\n", strerror (errno));
      
      return;
    }

  /* Lock the session file. */
  if (!lock (F_RDLCK, fileno (f)))
    {
      buf_write_string (s, "450 Couldn't lock session data\r\n");
      return;
    }
  
  /* Read the session file. */
  for (;;)
    {
      /* Read a line. */
      if (NULL == fgets (buf, BUFSIZ, f))
	{
	  if (feof (f))
	    buf_write_string (s, "530 Session not logged in\r\n");
	  else
	    buf_write_printf (s, "530 fgets: %s\r\n", strerror (errno));
	  fclose (f);

	  return;
	}

      /* Strip trailing newline. */
      {
	char *p = strchr (buf, '\n');
	if (p != NULL)
	  *p = '\0';
      }

      /* Parse the line for keys that we're interested in. */
      if (!strncmp (buf, "user=", 5))
	s->username = xstrdup (buf + 5);
      else if (!strncmp (buf, "workgroup=", 10))
	s->workgroup = xstrdup (buf + 10);

      if (s->username && s->workgroup)
	break;
    }

  if (fclose (f) == EOF)
    syslog (LOG_NOTICE, "fclose: %s", strerror (errno));

  buf_write_printf (s, "230 User %s logged into workgroup %s, proceed\r\n",
		    s->username, s->workgroup);

  /* Log user's entry and notify other users. */
  log ("Enter: %s, %s", s->username, s->workgroup);

  s->cookie = xstrdup (arg);

  /* User is now logged in. */
  transition (s, CLS_READY);

  update_poppage (s->workgroup);
}

/* Handle WHOM command by sending a list of users. */
static void
cmd_whom (struct session *s, const char *arg unused)
{
  struct string st;
  int count;
  int i;

  st_create (&st, "251-Sending user list\r\n");

  count = 0;
  for (i = 0; i < n_sessions; i++)
    {
      struct session *const t = sessions[i];
      
      if (t->state == CLS_READY
	  && (t->flags & FLG_POPPED)
	  && !strcmp (s->workgroup, t->workgroup))
	st_printf (&st, " %s\r\n", t->username);
    }
  st_concat (&st, "251 User list complete\r\n");
  buf_write_line (s, buf_create_line (st_value (&st)));
}

/* Sends a popup request to client session S. */
static void
popup (struct session *s)
{
  buf_write_string (s, "602 Popup\r\n");
}

/* Let clients in WORKGROUP know whether the other clients are popped
   up. */
static void
update_poppage (const char *workgroup)
{
  unsigned long all_popped = 0;
  int i, j;
  int n;
  
  for (i = j = 0; i < n_sessions; i++)
    {
      struct session *const t = sessions[i];
      
      if (t->state == CLS_READY && !strcmp (t->workgroup, workgroup))
	{
	  if (j > 31)
	    break;
	  if (t->flags & FLG_POPPED)
	    all_popped |= 1ul << j;
	  j++;
	}
    }
  n = j;
  
  for (i = j = 0; i < n_sessions; i++)
    {
      struct session *const t = sessions[i];
      
      if (t->state == CLS_READY && !strcmp (t->workgroup, workgroup))
	{
	  int are_others_popped = n > 31 || (all_popped & ~(1ul << j)) != 0;
	  int told_others_popped = (t->flags & FLG_OTHERS_POPPED) != 0;

	  if (are_others_popped && !told_others_popped)
	    {
	      buf_write_string (t, "603 Others chatting\r\n");
	      t->flags |= FLG_OTHERS_POPPED;
	    }
	  else if (!are_others_popped && told_others_popped)
	    {
	      buf_write_string (t, "604 Others not chatting\r\n");
	      t->flags &= ~FLG_OTHERS_POPPED;
	    }
	  j++;
	}
    }
}

/* Handle POPU command, which indicates the status of a client's chat
   window.  (Client sessions can often be logged in, to receive status
   information, without actually having a chat window popped up.) */
static void
cmd_popu (struct session *s, const char *arg)
{
  switch (stol (arg))
    {
    case 0:
      /* Chat window closed. */
      s->flags &= ~FLG_POPPED;
      break;
      
    case 1:
      /* `Chat' button clicked on.  The server will choose one of the
	 clients in this session to actually pop up, and send that one
	 a 602 message. */
      {
	int i;
	
	for (i = 0; ; i++)
	  if (i < n_sessions)
	    {
	      /* If there's a client that's already popped up, reuse it. */
	      struct session *const t = sessions[i];

	      if (s != t
		  && (t->flags & FLG_POPPED)
		  && !strcmp (s->cookie, t->cookie))
		{
		  /* Yes, it should already be popped up, but sending
		     it another notification can't hurt. */
		  popup (t);
		  break;
		}
	    }
	  else
	    {
	      /* Otherwise, pop up the current client. */
	      s->flags |= FLG_POPPED;
	      popup (s);
	      break;
	    }
      }
      break;

    case 2:
      /* Chat window is popped up.  Can be assumed if the client
	 received a 602 message, but otherwise needs to be sent.  For
	 instance, if the client-server connection dropped and the
	 client had to log in again, and the chat window was still
	 popped up, this should be sent. */
      s->flags |= FLG_POPPED;
      break;

    default:
      buf_write_string (s, "501 Bad argument value\r\n");
      return;
    }

  post (s, 0, "601 Userlist change notify\r\n");
  update_poppage (s->workgroup);
  buf_write_string (s, "200 POPU command okay\r\n");
}

