/* TeamSCOPE - Software for Collaborative Project Environments.
   Copyright (C) 1998-9, 2000 Free Software Foundation, Inc.
   Copyright (C) 1999, 2000 Michigan State University.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To contact the TeamSCOPE team, send email to scope@cscw.msu.edu, or
   write to Charles Steinfield, Room 436 Communication Arts Building,
   Michigan State University, Department of Telecommunication, East
   Lansing MI 48824-1212. */

#include <config.h>
#include <ctype.h>
#include <assert.h>
#include <limits.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "alloc.h"
#include "str.h"
#include "time.h"

/* Normal null-terminated string functions. */

/* Search for NEEDLE in HAYSTACK case-insensitively.  Returns a
   pointer to the leftmost instance of NEEDLE within HAYSTACK if
   successful; otherwise returns a null pointer. */
char *
blp_strcasestr (const char *haystack, const char *needle)
{
  long haystack_len = strlen (haystack);
  long needle_len = strlen (needle);
  int i;

  for (i = 0; i <= haystack_len - needle_len; i++)
    if (!strncasecmp (haystack + i, needle, needle_len))
      return (char *) haystack + i;

  return NULL;
}

/* Formats FMT into a dynamically allocated string as with printf().
   Returns the string. */
char *
blp_asprintf (const char *fmt, ...)
{
  char buf[64];
  char *s;
  va_list args;
  int n;
  
  va_start (args, fmt);
  n = vsnprintf (buf, 64, fmt, args);
  va_end (args);

  s = xmalloc (n + 1);
  if (n < 64)
    strcpy (s, buf);
  else
    vsprintf (s, fmt, args);
  return s;
}
  
/* Format FMT into a small static buffer and return the string.
   Note: don't use this function. */
const char *
format (const char *fmt, ...)
{
  static char buf[128];
  va_list args;

  va_start (args, fmt);
  vsnprintf (buf, 128, fmt, args);
  va_end (args);

  return buf;
}
/* Returns nonzero if STRING has PREFIX as a prefix. */
int
has_prefix (const char *string, const char *prefix)
{
  while (*prefix)
    if (*string++ != *prefix++)
      return 0;
  return 1;
}

/* st_* functions. */

/* Initializes ST with initial contents S. */
void
st_create (struct string *st, const char *s)
{
  st->len = strlen (s);
  st->size = 8 + st->len * 2;
  st->s = malloc (st->size + 1);
  strcpy (st->s, s);
}

/* Initializes ST, making room for at least SIZE characters. */
void
st_init (struct string *st, size_t size)
{
  st->len = 0;
  if (size > 8)
    st->size = size;
  else
    st->size = 8;
  st->s = malloc (st->size + 1);
}

/* Frees ST. */
void
st_destroy (struct string *st)
{
  free (st->s);
}

/* Truncates ST to zero length. */
void
st_clear (struct string *st)
{
  st->len = 0;
}

/* Ensures that ST can hold at least MIN_SIZE characters plus a null
   terminator. */
void
st_extend (struct string *st, size_t min_size)
{
  if (min_size > st->size)
    {
      st->size *= 2;
      if (st->size < min_size)
	st->size = min_size * 2;
      
      st->s = realloc (st->s, st->size + 1);
    }
}

/* Truncates ST to at most LENGTH characters long. */
void
st_truncate (struct string *st, size_t length)
{
  if (length >= st->len)
    return;
  st->len = length;
}

/* Returns the length of ST. */
size_t
st_length (const struct string *st)
{
  return st->len;
}

/* Returns the value of ST as a null-terminated string. */
char *
st_value (const struct string *st)
{
  ((char *) st->s)[st->len] = '\0';
  return st->s;
}

/* Concatenates S onto ST. */
void
st_concat (struct string *st, const char *s)
{
  size_t s_len = strlen (s);
  st_extend (st, st->len + s_len);
  strcpy (st->s + st->len, s);
  st->len += s_len;
}

/* Concatenates S onto ST, between quotes of type QUOTE ('
   or "), with instances of QUOTE, as well as \, in S quoted with a
   prefixed \. */
void
st_concat_quoted (struct string *st, const char *s, int quote)
{
  st_putchar (st, quote);
  for (; *s; s++)
    {
      if (*s == quote || *s == '\\')
	st_putchar (st, '\\');
      st_putchar (st, *s);
    }
  st_putchar (st, quote);
}

/* Appends S to ST, quoting characters which are reserved in URL query
   strings with %HH syntax. */
void
st_concat_url_query_quoted (struct string *st, const char *s)
{
  while (*s)
    {
      unsigned char ch = *s++;
      int quote = 0;
      
      switch (ch)
	{
	case ' ':
	  st_putchar (st, '+');
	  continue;
	  
	case '%': case '<': case '>': case '#': case '"':
	case '{': case '}': case '|': case '[': case ']':
	case '^': case '`': case ';': case '/': case '?':
	case ':': case '=': case ',': case '$': case '@':
	case '\\': case '&':
	  quote = 1;
	  break;

	default:
	  quote = ch < 0x20 || ch >= 0x7f;
	  break;
	}

      if (!quote)
	st_putchar (st, ch);
      else
	st_printf (st, "%%%02X", ch);
    }
}

/* Appends S to ST, escaping HTML-reserved characters in HTML
   fashion. */
void
st_concat_html_quoted (struct string *st, const char *s)
{
  for (; ; s++)
    switch (*s)
      {
      case '\0':
	return;

      case '<':
	st_concat (st, "&lt;");
	break;
	
      case '>':
	st_concat (st, "&gt;");
	break;

      case '&':
	st_concat (st, "&apos;");
	break;

      case '"':
	st_concat (st, "&quot;");
	break;

      default:
	st_putchar (st, *s);
	break;
      }
}

/* Concatenates LEN characters from BUF onto ST. */
void
st_concat_buffer (struct string *st, const char *buf, size_t len)
{
  st_extend (st, st->len + len);
  memcpy (st->s + st->len, buf, len);
  st->len += len;
}

/* Formats FORMAT as a printf string and appends the result to ST. */
void
st_printf (struct string *st, const char *format, ...)
{
  /* Fscking glibc silently changed behavior between 2.0 and 2.1.
     Fsck fsck fsck.  Before, it returned -1 on buffer overflow.  Now,
     it returns the number of characters (not bytes) that would have
     been written. */
  va_list args;

  int avail, needed;

  va_start (args, format);
  avail = st->size - st->len + 1;
  needed = vsnprintf (st->s + st->len, avail, format, args);
  va_end (args);

  if (needed >= avail)
    {
      st_extend (st, st->len + needed);
      
      va_start (args, format);
      vsprintf (st->s + st->len, format, args);
      va_end (args);
    }
  else
    while (needed == -1)
      {
	st_extend (st, (st->size + 1) * 2);
	avail = st->size - st->len + 1;
	
	va_start (args, format);
	needed = vsnprintf (st->s + st->len, avail, format, args);
	va_end (args);
      }

  st->len += needed;
}

/* Formats FORMAT as a strftime string for date-time DATE and appends
   the result to ST. */
void
st_strftime (struct string *st, const char *format, const struct tm *date)
{
  size_t needed;

  st_extend (st, st->len + 32);
  for (;;)
    {
      /* This assumes that you won't ever use a FORMAT string that
	 could result in a valid zero-length output.  If that ever
	 happens, you'll have to take another approach. */
      size_t avail = st->size - st->len + 1;
      needed = strftime (st->s + st->len, avail, format, date);
      if (needed)
	break;

      st_extend (st, st->size * 2);
    }

  st->len += needed;
}

/* Appends character CH to ST. */
void
st_putchar (struct string *st, int ch)
{
  if (st->len == st->size)
    st_extend (st, st->len + 1);
  st->s[st->len++] = ch;
}

