/* TeamSCOPE - Software for Collaborative Project Environments.
   Copyright (C) 1999, 2000, 2001 Michigan State University.
   Written by Ben Pfaff <pfaffben@msu.edu>.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To contact the TeamSCOPE team, send email to scope@cscw.msu.edu, or
   write to Charles Steinfield, Room 436 Communication Arts Building,
   Michigan State University, Department of Telecommunication, East
   Lansing MI 48824-1212. */

#include <config.h>
#include <assert.h>
#include <ctype.h>
#include <limits.h>
#include <stdlib.h>
#include "activity.h"
#include "activityP.h"
#include "alloc.h"
#include "cgi.h"
#include "db.h"
#include "error.h"
#include "html.h"
#include "identity.h"
#include "options.h"
#include "scope.cgi.h"
#include "std.h"
#include "str.h"

static int handle_reset (struct request *);

/* Handle the Recent Activity and activity search result screen by
   decoding the CGI parameters into an activity_query struct,
   running the query, and displaying the results. */
void
scr_activity (struct request *rq)
{
  struct activity_query query;

  if (rq->abbrev == 1)
    {
      html ("<P>%s</P>\n", _("Recent file activities:"));
  
      activity_abbreviated (rq, AC_CAT_FILE);
      return;
    }

  if (handle_reset (rq))
    return;

  /* Fill in query structure.
     For a search, we use the CGI parameters to set up the query.
     For recent activities, we use the webfilter settings. */
  activity_query_init (&query);
  if (std_contains (rq->cgi, "search"))
    {
      query.search = 1;
      activity_query_modify_from_cgi (&query, rq->cgi, rq->cx.usr);
    }
  else
    {
      activity_query_from_dict (&query,
				std_get_subdict (rq->cx.options, "webfilter"),
				rq->cx.usr);
      activity_query_set_last_read (&rq->cx, &query);
    }
  query.start = stoi (std_get (rq->cgi, "start"));

  html ("<H2 ALIGN=CENTER>%s</H2>\n",
	query.search ? _("Search Results") : _("Recent Activity"));

  html_flush ();
  
  /* Run query and present results. */
  activity_query_run (&rq->cx, &query);

  html (_("<P ALIGN=CENTER>Viewing matches %d - %d:</P>\n"),
	query.start + 1, query.start + query.count);
  
  {
    struct url url;

    url_create (&url, "activity");
    activity_query_present (&rq->cx, &query, &url, NULL);
    url_destroy (&url);
  }

  activity_query_destroy (&query);
}

/* Describes an entry in the activity selection list box
   for the activity search or webfilter or emailfilter. */
struct activity_selection
  {
    int level;			/* Indentation level. */
    const char *label;		/* String describing this group of options. */
    int cat;			/* Category of these options. */
    unsigned long types;	/* Bit string of included types. */
  };

/* The "easy" set of activity selections
   that try to group together related kinds of activities.
   For best results, selections must not overlap. */
static const struct activity_selection easy_sel[] =
  {
    {0, N_("All activities"), 0, 0},

    {1, N_("All file activities"), AC_CAT_FILE, ULONG_MAX},
    {2, N_("Uploads"), AC_CAT_FILE, (1 << AC_FILE_UPLOAD)},
    {2, N_("Comment posts"), AC_CAT_FILE, (1 << AC_FILE_POST)},
    {2, N_("File and comment modifications"), AC_CAT_FILE,
     (1 << AC_FILE_DELETE) | (1 << AC_FILE_COPY) | (1 << AC_FILE_RENAME)
     | (1 << AC_FILE_MKDIR) | (1 << AC_FILE_RMDIR) | (1 << AC_FILE_CHMOD)
     | (1 << AC_FILE_MOVE) | (1 << AC_FILE_POST_DELETE)},
    {2, N_("File and comment downloads and views"), AC_CAT_FILE,
     (1 << AC_FILE_DOWNLOAD) | (1 << AC_FILE_DOWNLOAD_ARCHIVE) 
     | (1 << AC_FILE_POST_READ) | (1 << AC_FILE_POST_MAILED)},

    {1, N_("All message activities"), AC_CAT_MSG_BOARD, ULONG_MAX},
    {2, N_("Message posts"), AC_CAT_MSG_BOARD, (1 << AC_MSG_POST)},
    {2, N_("Message deletions"), AC_CAT_MSG_BOARD, (1 << AC_MSG_DELETE)},
    {2, N_("Message views"), AC_CAT_MSG_BOARD, (1 << AC_MSG_READ) | (1 << AC_MSG_MAILED)},
     
    {1, N_("All calendar activities"), AC_CAT_EVENT, ULONG_MAX},
    {2, N_("New calendar events"), AC_CAT_EVENT, (1 << AC_EVENT_NEW)},
    {2, N_("Calendar event changes"), AC_CAT_EVENT, (1 << AC_EVENT_DELETE) | (1 << AC_EVENT_CHANGE)},
    {2, N_("Calendar event views"), AC_CAT_EVENT, (1 << AC_EVENT_VIEW)},
  };

/* Number of elements in easy_sel[]. */
#define EASY_SEL_CNT (sizeof easy_sel / sizeof *easy_sel)

/* The detailed activity selections.
   Do not group together more than one activity type on a line
   here, because it will break code. */
static const struct activity_selection detailed_sel[] =
  {
    {0, N_("All activities"), 0, 0},

    {1, N_("All file activities"), AC_CAT_FILE, ULONG_MAX},
    {2, N_("Download"), AC_CAT_FILE, (1 << AC_FILE_DOWNLOAD)},
    {2, N_("Download archive"), AC_CAT_FILE, (1 << AC_FILE_DOWNLOAD_ARCHIVE)},
    {2, N_("Upload"), AC_CAT_FILE, (1 << AC_FILE_UPLOAD)},
    {2, N_("Delete"), AC_CAT_FILE, (1 << AC_FILE_DELETE)},
    {2, N_("Copy"), AC_CAT_FILE, (1 << AC_FILE_COPY)},
    {2, N_("Rename"), AC_CAT_FILE, (1 << AC_FILE_RENAME)},
    {2, N_("New folder"), AC_CAT_FILE, (1 << AC_FILE_MKDIR)},
    {2, N_("Delete folder"), AC_CAT_FILE, (1 << AC_FILE_RMDIR)},
    {2, N_("Change access privileges"), AC_CAT_FILE, (1 << AC_FILE_CHMOD)},
    {2, N_("Move"), AC_CAT_FILE, (1 << AC_FILE_MOVE)},
    {2, N_("Post comment"), AC_CAT_FILE, (1 << AC_FILE_POST)},
    {2, N_("Delete comment"), AC_CAT_FILE, (1 << AC_FILE_POST_DELETE)},
    {2, N_("Read comment"), AC_CAT_FILE, (1 << AC_FILE_POST_READ)},
    {2, N_("Mail comment"), AC_CAT_FILE, (1 << AC_FILE_POST_MAILED)},

    {1, N_("All message activities"), AC_CAT_MSG_BOARD, ULONG_MAX},
    {2, N_("Post message"), AC_CAT_MSG_BOARD, (1 << AC_MSG_POST)},
    {2, N_("Delete message"), AC_CAT_MSG_BOARD, (1 << AC_MSG_DELETE)},
    {2, N_("Read message"), AC_CAT_MSG_BOARD, (1 << AC_MSG_READ)},
    {2, N_("Mail message"), AC_CAT_MSG_BOARD, (1 << AC_MSG_MAILED)},

    {1, N_("All calendar activities"), AC_CAT_EVENT, ULONG_MAX},
    {2, N_("New calendar event"), AC_CAT_EVENT, (1 << AC_EVENT_NEW)},
    {2, N_("Delete calendar event"), AC_CAT_EVENT, (1 << AC_EVENT_DELETE)},
    {2, N_("Change calendar event"), AC_CAT_EVENT, (1 << AC_EVENT_CHANGE)},
    {2, N_("View calendar event"), AC_CAT_EVENT, (1 << AC_EVENT_VIEW)},
  };

/* Number of elements in detailed_sel[]. */
#define DETAILED_SEL_CNT (sizeof detailed_sel / sizeof *detailed_sel)

/* Outputs JavaScript to set up for handling a hierarchical list box
   named NAME based on levels in NAMElvl.  Creates an ONCHANGE
   handler named NAME_change(). */
static void
setup_hier (const char *name)
{
  html ("%s = new Array(%slvl.length);\n"
	"for (var i = 0; i < %slvl.length; i++)\n"
	"	%s[i] = false;\n"
	"function %s_change() {\n"
	"	hierchg(%s,%slvl,document.forms[1].%s)\n"
	"}\n",
	name, name,
	name,
	name,
	name,
	name, name, name);
}

/* Outputs JavaScript to set up for handling two hierarchical list
   boxes, one for selecting activity types and the other for selecting
   users.  The N_CATEGORIES items in the first are described in
   CATEGORIES[].  The second has N_PEOPLE people preceded by a single
   "All Users" item. */
static void
show_script (const struct activity_selection *categories, size_t n_categories,
	     size_t n_people)
{
  size_t i;

  htmls
    ("<SCRIPT TYPE=\"text/javascript\" LANGUAGE=\"JavaScript\">\n"
     "<!--\n"
     "function hierchg (settings,levels,select) {\n"
     "	for (var i = 0; i < settings.length; i++) {\n"
     "		s = select.options[i].selected;\n"
     "		if (settings[i] != s) {\n"
     "			settings[i] = s;\n"
     "			list = new Array(0);\n"
     "			if (s == true) {\n"
     "				for (var j = i + 1; j < settings.length; j++) {\n"
     "					if (levels[j] <= levels[i]) break;\n"
     "					list[list.length] = j;\n"
     "				}\n"
     "			} else {\n"
     "				lvl = levels[i];\n"
     "				for (var j = i - 1; j >= 0; j--) {\n"
     "					if (levels[j] < lvl) {\n"
     "						lvl = levels[j];\n"
     "						list[list.length] = j;\n"
     "					}\n"
     "				}\n"
     "			}\n"
     "			for (var j = list.length - 1; j >= 0; j--) {\n"
     "				settings[list[j]] = s;\n"
     "				if (select.options[list[j]].selected != s)\n"
     "					select.options[list[j]].selected = s;\n"
     "			}\n"
     "		}\n"
     "	}\n"
     "}\n");

  html ("activitieslvl = new Array(");
  for (i = 0; i < n_categories; i++)
    {
      if (i != 0)
	htmlc (',');
      html ("%d", categories[i].level);
    }
  html (");\n");

  html ("peoplelvl = new Array(0");
  for (i = 0; i < n_people; i++)
    htmls (",1");
  html (");\n");

  setup_hier ("people");
  setup_hier ("activities");

  htmls ("// -->\n"
	 "</SCRIPT>\n");
}

/* Returns the name of activity category CAT. */
static const char *
cat_name (int cat)
{
  static const char *cat_names[AC_N_CATEGORIES] = 
    {
      NULL, N_("Files"), N_("Messages"), N_("Calendar events"),
    };

  assert (cat >= 1 && cat < AC_N_CATEGORIES);
  return gettext (cat_names[cat]);
}

/* Returns nonzero only if BITS, within activity category CAT, can be
   represented properly by the activity selections in SEL[], of which
   there are SEL_CNT.  This is always true if the user actually picked
   them out of SEL, and may or may not be true if the user did not. */
static int
bits_conform_to (unsigned long bits,
		 const struct activity_selection sel[], size_t sel_cnt,
		 int cat)
{
  for (; sel_cnt; sel++, sel_cnt--)
    if (sel->level == 2 && sel->cat == cat
	&& (sel->types & bits) != 0
	&& (sel->types & bits) != sel->types)
      return 0;

  return 1;
}

/* Outputs the text of the activity types defined in BITS for activity
   category CAT that are included in SEL[], which has SEL_CNT
   elements. */
static void
describe_per_cat_activities_as (unsigned long bits,
				const struct activity_selection sel[],
				size_t sel_cnt, int cat)
{
  int count = 0;

  for (; sel_cnt; sel++, sel_cnt--)
    if (sel->level == 2 && sel->cat == cat
	&& (sel->types & bits) == sel->types)
      {
	const char *gtlabel;
	
	if (count++ > 0)
	  htmls (", ");

	gtlabel = gettext (sel->label);
	htmlc (tolower ((unsigned char) gtlabel[0]));
	htmls (gtlabel + 1);
      }
}

/* Outputs HTML to describe the activities in bitmap BM. */
static void
describe_per_cat_activities (unsigned long bm[AC_N_CATEGORIES])
{
  int i;
  
  html ("%s <UL>\n", _("These kinds of activities <STRONG>only</STRONG>:"));
  for (i = 1; i < AC_N_CATEGORIES; i++)
    if ((bm[i] & ACTIVITY_MASK_FOR_WIDTH (activity_cat_cnt[i])) != 0)
      {
	html ("<LI>%s: ", cat_name (i));

	if (activity_all_bits_p (bm[i], i))
	  htmls (_("all activities"));
	else if (bits_conform_to (bm[i], easy_sel, EASY_SEL_CNT, i))
	  describe_per_cat_activities_as (bm[i], easy_sel, EASY_SEL_CNT, i);
	else
	  describe_per_cat_activities_as (bm[i], detailed_sel,
					  DETAILED_SEL_CNT, i);
	htmls (".\n");
      }
  htmls ("</UL>\n");
}

/* Outputs HTML describing the filter on people and activity types
   within activity query QUERY.
   USERS has been initialized with id_list_for_user()
   and INFO[] contains corresponding user information. */
static void
describe_people_filter (struct activity_query *query,
			struct id_list *users, struct user_info info[])
{
  int i;

  html ("<LI>%s\n"
	"<UL>\n",
	_("Activities of the specified types by team members listed below:"));

  for (i = 0; i < query->n_people; i++)
    {
      struct activity_person *p = &query->person[i];

      htmls ("<LI><STRONG>");
      if (id_list_equal (&p->usr, users))
	htmls (_("All users"));
      else if (p->usr.n < users->n * 2 / 3
	       || users->n - p->usr.n > 10 || p->usr.n < 5)
	{
	  size_t j;

	  for (j = 0; j < p->usr.n; j++)
	    {
	      int k;
	      
	      if (j > 0)
		htmls (", ");

	      k = id_list_find (users, p->usr.s[j]);
	      if (k != -1)
		htmls (info[k].fullname);
	      else
		htmls (p->usr.s[j]);
	    }
	}
      else
	{
	  size_t j;
	  int count;
	  
	  htmls (_("All users except "));
	  count = 0;
	  for (j = 0; j < users->n; j++)
	    if (id_list_find (&p->usr, users->s[j]) == -1)
	      {
		if (count++ > 0)
		  htmls (", ");
		
		htmls (info[j].fullname);
	      }
	}

      htmls (":</STRONG> ");

      if (activity_bm_zero_p (p->type))
	htmls (_("no activities\n"));
      else if (activity_bm_all_p (p->type))
	htmls (_("all activities\n"));
      else
	describe_per_cat_activities (p->type);
    }

  html ("<LI>%s\n",
	_("No other users' activities will be selected."));

  htmls ("</UL>\n");
}

/* Outputs HTML describing the filtering performed by QUERY.
   USERS has been initialized with id_list_for_user()
   and INFO[] contains corresponding user information. */
void
describe_filter (struct activity_query *query,
		 struct id_list *users, struct user_info info[])
{
  int count = 0;
  
  html ("<P>%s</P>\n"
	"<UL>", _("Only activities that match all of the criteria below will be selected:"));

  if (query->n_people > 0)
    {
      describe_people_filter (query, users, info);
      count++;
    }

  count++;
  if (query->begin_date != NULL && query->end_date != NULL)
    html (_("<LI>Activities between the dates `%s' and `%s'.\n"),
	  query->begin_date, query->end_date);
  else if (query->begin_date != NULL)
    html (_("<LI>Activities after the date `%s'.\n"), query->begin_date);
  else if (query->end_date != NULL)
    html (_("<LI>Activities before the date `%s'.\n"), query->end_date);
  else
    count--;

  if (query->filename != NULL)
    {
      html (_("<LI>Activities for files whose names contain `"));
      html_escape (query->filename);
      html (_("'.\n"));
      count++;
    }

  if (count == 0)
    html (_("<LI>All activities.\n"));

  htmls ("</UL>\n");
}

/* Display a form to allow an activity search.  Fill in the defaults
   of the form from QUERY.  BUTTON is the title for the button to
   submit the form.  A null pointer should be used for the activity
   search form. */
void 
activity_show_filter (struct request *rq, struct activity_query *query)
{
  static const char *selected[2] = {"", " SELECTED"};
  static const char *checked[2] = {"", " CHECKED"};
  
  /* Level of detail in selection of activities. */
  enum activity_detail 
    {
      AC_EASY,		/* Easy selection of groups of activity types. */
      AC_DETAILED	/* Detailed selection of individual activity types. */
    };

  enum activity_detail detail;
  const struct activity_selection *categories;
  size_t n_categories;

  struct id_list users;
  struct user_info *ui;

  if (query->search)
    html_hidden ("search", "yes");
  
  detail = (!strcmp (std_get (rq->cgi, "detail"), "detailed")
	    ? AC_DETAILED : AC_EASY);
  if (detail == AC_DETAILED)
    {
      categories = detailed_sel;
      n_categories = DETAILED_SEL_CNT;
    }
  else
    {
      categories = easy_sel;
      n_categories = EASY_SEL_CNT;
    }
  
  id_list_for_user (&users, rq->cx.usr);
  ui = xmalloc (sizeof *ui * users.n);
  get_user_info (&users, ui, UI_USR | UI_FULLNAME);
    
  show_script (categories, n_categories, users.n);

  if (!query->search)
    html ("<P>%s</P>\n"
	  "<P ALIGN=CENTER><A HREF=\"%s#current\">%s</A></P>\n",
	  _("By default, all activities are selected.\n"
	    "Filtering can be done on the basis of the type of activity\n"
	    "and the team member performing it.  To do so,\n"
	    "select one or more\n"
	    "team members and activity types, then click on\n"
	    "<STRONG>Add Criteria</STRONG>.  Repeat this step\n"
	    "to see more kinds of activities.\n"
	    "Use <STRONG>Remove Criteria</STRONG> in a similar way\n"
	    "to remove kinds of activities from the filter.\n"
	    "File name based filtering as well as order and limit\n"
	    "preferences may be set with <STRONG>Set Filter</STRONG>.\n"),
	  url_this (rq->screen, (char *) NULL), _("View Current Settings"));

  html
    ("<TABLE WIDTH=\"100%%\">\n"
     "<TR>\n"
     "  <TD COLSPAN=3><HR NOSHADE></TD>\n"
     "</TR>\n"
     "<TR>\n"
     "  <TH WIDTH=\"15%%\" ALIGN=RIGHT BGCOLOR=\"#C0C0C0\">Team Member and Activity Type</TH>\n"
     "  <TD WIDTH=\"60%%\">\n"
     "    <TABLE><TR VALIGN=TOP>\n"
     "      <TD ALIGN=CENTER NOWRAP>For these users:<BR>\n"
     "        <SELECT NAME=people SIZE=8 MULTIPLE "
     "ONCHANGE=\"people_change();\">\n"
     "          <OPTION VALUE=\"ALL*USERS\">All users\n");

  {
    size_t i;

    for (i = 0; i < users.n; i++)
      html ("        <OPTION VALUE=\"%s\"\">&nbsp;&nbsp;%s\n",
	    ui[i].usr, ui[i].fullname);
  }

  html ("        </SELECT>\n"
	"        <BR><SMALL>%s</SMALL>\n"
	"      </TD><TD ALIGN=CENTER NOWRAP>%s<BR>\n",
	_("Use Control or Command key to select multiple users."),
	_("Select only these kinds of activities:"));
  
  htmls ("        <SELECT NAME=activities SIZE=8 MULTIPLE ONCHANGE=\"activities_change();\">\n");

  {
    size_t i;
  
    for (i = 0; i < n_categories; i++)
      {
	const struct activity_selection *p = categories + i;
      
	html ("          <OPTION VALUE=\"%d:%08lx\">",
	      p->cat, (unsigned long) p->types);
	if (p->level > 1)
	  htmls ("&nbsp;&nbsp;");
	html ("%s</OPTION>\n", gettext (p->label));
      }
  }

  html ("        </SELECT><BR>\n"
	"        <SMALL>%s</SMALL>\n"
	"      <P></TD><TD ALIGN=LEFT NOWRAP>",
	_("Use Control or Command key to select multiple categories."));

  html ("      </TD>\n"
	"    </TR></TABLE>\n"
	"  </TD>"
	"  <TD WIDTH=\"15%%\">");
  
  if (!query->search)
    {
      html ("        <P>"
	    "<INPUT TYPE=submit NAME=addcrit VALUE=\"Add Criteria\"></P>"
	    "        <P>"
	    "<INPUT TYPE=submit NAME=rmcrit VALUE=\"Remove Criteria\"></P>");
    }

  html ("        <P>%s\n",
	_("Categories:"));

  if (detail == AC_EASY)
    html ("%s / <A HREF=\"%s\">%s</A>\n",
	  _("Easy"),
	  url_this (rq->screen, "detail", "detailed", (char *) NULL),
	  _("Detailed"));
  else
    html ("<A HREF=\"%s\">%s</A> / %s\n",
	  url_this (rq->screen, "detail", "easy", (char *) NULL), _("Easy"),
	  _("Detailed"));

  html ("        </P>\n"
	"        <P><INPUT TYPE=reset VALUE=\"%s\"></P>\n"
	"  </TD>\n"
	"</TR>\n"
	"<TR>\n"
	"  <TD COLSPAN=3><HR NOSHADE></TD>\n"
	"</TR>\n"
	"<TR>\n"
	"  <TH ALIGN=RIGHT BGCOLOR=\"#C0C0C0\">File Name</TH>\n"
	"  <TD>Enter part or all of a file name to accept:<BR>\n",
	_("Clear Form"));
  html_text_field ("filename", query->filename, 20);

  html ("      <BR>\n"
	"      <SMALL><STRONG>Only</STRONG> activities for matching files will be selected.\n"
	"  </TD>\n"
	"  <TD ROWSPAN=%d>\n"
	"    <INPUT TYPE=submit VALUE=\"%s\">\n"
	"  </TD>\n",
	query->search ? 4 : 3,
	query->search ? _("Search") : _("Set Filter"));

  if (query->search)
    {
      htmls (
	     "</TR><TR>\n"
	     "  <TH ALIGN=RIGHT BGCOLOR=\"#C0C0C0\">Date</TH>\n"
	     "  <TD>Select <STRONG>only</STRONG> activities within"
	     "this range of dates:<BR>\n");

      html ("    <INPUT TYPE=radio NAME=date VALUE=0%s>Any date<BR>\n"
	    "    <INPUT TYPE=radio NAME=date VALUE=1>The last 7 days<BR>\n"
	    "    <INPUT TYPE=radio NAME=date VALUE=2>The last 14 days<BR>\n"
	    "    <INPUT TYPE=radio NAME=date VALUE=3>The last 28 days<BR>\n"
	    "    <INPUT TYPE=radio NAME=date VALUE=4%s>Date within range:\n"
	    "    <TABLE BORDER=0 CELLSPACING=0>\n"
	    "      <TR>\n"
	    "        <TD><FONT SIZE=\"-1\">%s&nbsp;</FONT><BR>\n",
	    checked[query->begin_date == NULL && query->end_date == NULL],
	    checked[query->begin_date != NULL || query->end_date != NULL],
	    _("Example: Nov 1 1998"));
      
      html_text_field ("begin", query->begin_date, 10);
      
      html 
	("        <BR><FONT SIZE=\"-1\">%s</FONT></TD>\n"
	 "        <TD><FONT SIZE=\"-1\">%s</FONT><BR>\n",
	 _("from"),
	 _("Example: Nov 15 1998"));
      
      html_text_field ("end", query->end_date, 10);
      
      html
	("        <BR><FONT SIZE=\"-1\">%s</FONT></TD></TR>\n"
	 "      </TR>\n"
	 "    </TABLE>\n"
	 "  </TD>\n",
	 _("to"));
    }
    
  html ("</TR><TR>\n"
	"  <TH ALIGN=RIGHT VALIGN=BOTTOM BGCOLOR=\"#C0C0C0\">%s</TH>\n"
	"  <TD>\n"
	"    <SELECT NAME=sort>\n"
	"      <OPTION%s VALUE=filename>%s\n"
	"      <OPTION%s VALUE=date>%s\n"
	"      <OPTION%s VALUE=user>%s\n"
	"      <OPTION%s VALUE=activity>%s\n"
	"    </SELECT>\n"
	"  </TD>\n",
	_("Order results by"),
	selected[query->sort == AC_SORT_FILENAME], _("filename"),
	selected[query->sort == AC_SORT_DATE
		|| query->sort == AC_SORT_DEFAULT], _("date"),
	selected[query->sort == AC_SORT_USER], _("user"),
	selected[query->sort == AC_SORT_ACTIVITY], _("activity type"));

  html
    ("</TR><TR>\n"
     "  <TH ALIGN=RIGHT VALIGN=BOTTOM BGCOLOR=\"#C0C0C0\">%s</TH>\n"
     "  <TD>\n"
     "    <SELECT NAME=limit>\n"
     "      <OPTION%s VALUE=100>100\n"
     "      <OPTION%s VALUE=200>200\n"
     "      <OPTION%s VALUE=500>500\n"
     "      <OPTION%s VALUE=\"no limit\">%s\n"
     "    </SELECT>\n"
     "  </TD>\n",
     query->search ? _("Max results per page") : _("Max activities per email"),
     selected[query->limit == 100],
     selected[query->limit == 200],
     selected[query->limit == 500],
     selected[query->limit > 500],
     _("no limit"));

  html ("</TR>\n"
	"<TR>\n"
	"  <TD COLSPAN=3><HR NOSHADE></TD>\n"
	"</TR>\n"
	"</TABLE>\n");
     
  html_end_form ();

  html ("<H2 ALIGN=CENTER><A NAME=current>%s</A></H2>\n",
	_("Current Filter"));
  describe_filter (query, &users, ui);

  free_user_info (ui, users.n);
  free (ui);
  id_list_destroy (&users);
}

/* Returns a date string for N days before now. */
static char *
today_minus (int n)
{
  time_t t = cur_time - n * 86400;
  char *s = xstrdup (ctime (&t));
  char *p = strchr (s, '\n');
  if (p != NULL)
    *p = '\0';
  return s;
}

/* Destroys QUERY->people. */
static void
remove_all_people (struct activity_query *query)
{
  int i;

  for (i = 0; i < query->n_people; i++)
    id_list_destroy (&query->person[i].usr);
  free (query->person);

  query->person = NULL;
  query->n_people = 0;
}

/* Initializes BM[], which has an element for each username in PPL,
   setting the bits that correspond to activities enabled in QUERY. */
static void
make_crit_bitmap (unsigned long bm[][AC_N_CATEGORIES],
		  struct activity_query *query, struct id_list *ppl)
{
  size_t i;

  for (i = 0; i < ppl->n; i++)
    {
      int j;
      
      for (j = 0; j < AC_N_CATEGORIES; j++)
	bm[i][j] = 0;
    }

  for (i = 0; i < (size_t) query->n_people; i++)
    {
      size_t j;
      
      for (j = 0; j < query->person[i].usr.n; j++)
	{
	  int k = id_list_find (ppl, query->person[i].usr.s[j]);

	  if (k != -1)
	    {
	      int m;
		  
	      for (m = 0; m < AC_N_CATEGORIES; m++)
		bm[k][m] |= query->person[i].type[m];
	    }
	}
    }
}

/* Initializes activities bitmap NEW from the null-terminated list of
   strings in P.  Each string should be of the form
   <CAT>:<ACTIVITIES>, where <CAT> is a decimal category number and
   <ACTIVITIES> a hexadecimal bit mask. */
static void
read_crit (unsigned long new[AC_N_CATEGORIES], const char **p)
{
  activity_bm_zero (new);
  if (p == NULL)
    return;

  for (; *p != NULL; p++)
    {
      int ofs;
      unsigned long mask;

      if (sscanf (*p, "%d:%lx", &ofs, &mask) != 2
	  || ofs < 0 || ofs >= AC_N_CATEGORIES)
	continue;

      if (mask != 0)
	new[ofs] |= mask;
      else
	{
	  int i;
	  
	  for (i = 0; i < AC_N_CATEGORIES; i++)
	    new[i] = ULONG_MAX;
	}
    }
}

/* Applies operation OP ('+' for union, '-' for subtraction) and
   operands NEW and ADDPPL to BM and PPL.  That is, if OP is '+', then
   bits for activities in NEW[] are set in BM for the people listed in
   ADDPPL; if OP is '-', they are unset. */
static void
apply_crit (unsigned long (*bm)[AC_N_CATEGORIES],
	    const struct id_list *ppl, const struct id_list *addppl,
	    const unsigned long new[AC_N_CATEGORIES],
	    int op)
{
  size_t i;

  assert (op == '+' || op == '-');
  for (i = 0; i < addppl->n; i++)
    {
      int j = id_list_find (ppl, addppl->s[i]);

      if (j != -1)
	{
	  int k;

	  for (k = 0; k < AC_N_CATEGORIES; k++)
	    if (op == '+')
	      bm[j][k] |= new[k];
	    else
	      bm[j][k] &= ~new[k];
	}
    }
}

/* `Normalizes' the N sets of activity bitmaps in N
   by unsetting bits that don't correspond to real activities. */
static void
norm_crit (unsigned long bm[][AC_N_CATEGORIES], size_t n)
{
  size_t i;

  for (i = 0; i < n; i++)
    {
      int j;

      for (j = 0; j < AC_N_CATEGORIES; j++)
	bm[i][j] &= ACTIVITY_MASK_FOR_WIDTH (activity_cat_cnt[j]);
    }
}

/* Sets up QUERY->PEOPLE to correspond to the activity bitmaps in BM[]
   for PPL. */
static void
extract_people (struct activity_query *query,
		unsigned long bm[][AC_N_CATEGORIES],
		const struct id_list *ppl)
{
  size_t i;
  
  for (i = 0; i < ppl->n; i++)
    if (!activity_bm_zero_p (bm[i]))
      {
	struct activity_person *p;
	size_t j;

	query->person = xrealloc (query->person,
				  (sizeof *query->person
				   * (query->n_people + 1)));
	p = query->person + query->n_people++;
	activity_bm_copy (p->type, bm[i]);

	p->usr.n = 1;
	for (j = i + 1; j < ppl->n; j++)
	  if (activity_bm_equal (bm, i, j))
	    p->usr.n++;

	p->usr.s = xmalloc (sizeof *p->usr.s * p->usr.n);
	p->usr.s[0] = xstrdup (ppl->s[i]);
	p->usr.n = 1;
	
	for (j = i + 1; j < ppl->n; j++)
	  if (activity_bm_equal (bm, i, j))
	    {
	      p->usr.s[p->usr.n++] = xstrdup (ppl->s[j]);
	      activity_bm_zero (bm[j]);
	    }
	activity_bm_zero (bm[i]);
      }
}

#if 0
static void
log_one (unsigned long bm[])
{
  int j;

  for (j = 0; j < AC_N_CATEGORIES; j++)
    fprintf (stderr, " %08lx", bm[j]);
  fprintf (stderr, "\n");
}

static void
log_bm (struct id_list *ppl, unsigned long (*bm)[AC_N_CATEGORIES])
{
  size_t i;

  for (i = 0; i < ppl->n; i++)
    {
      fprintf (stderr, "%8s:", ppl->s[i]);
      log_one (bm[i]);
    }
}

static void
log_people (struct activity_query *query)
{
  int i;

  for (i = 0; i < query->n_people; i++)
    {
      size_t j;

      for (j = 0; j < query->person[i].usr.n; j++)
	{
	  if (j > 0)
	    fprintf (stderr, ", ");
	  fputs (query->person[i].usr.s[j], stderr);
	}
      fputs (": ", stderr);
      log_one (query->person[i].type);
    }
}
#endif

/* Modifies QUERY by USR according to CGI,
   which should be a result from activity_show_filter(). */
void
activity_query_modify_from_cgi (struct activity_query *query,
				struct std *cgi, const char *usr)
{
  if (std_contains (cgi, "addcrit") || std_contains (cgi, "rmcrit")
      || query->search != 0)
    {
      struct id_list add_ppl;
      unsigned long new[AC_N_CATEGORIES];

      read_crit (new, std_get_many (cgi, "activities", NULL));
      id_list_from_std (&add_ppl, cgi, "people", 0);
      if (add_ppl.n != 0 || !activity_bm_zero_p (new))
	{
	  struct id_list ppl;
	  unsigned long (*bm)[AC_N_CATEGORIES];

	  id_list_for_user (&ppl, usr);
	  bm = xmalloc (sizeof *bm * ppl.n);
	  make_crit_bitmap (bm, query, &ppl);

	  if (add_ppl.n == 0)
	    id_list_copy (&ppl, &add_ppl);
	  if (activity_bm_zero_p (new))
	    activity_bm_all (new);

	  apply_crit (bm, &ppl, &add_ppl, new,
		      std_contains (cgi, "rmcrit") ? '-' : '+');
	  norm_crit (bm, ppl.n);
	  remove_all_people (query);
	  extract_people (query, bm, &ppl);
	  id_list_destroy (&ppl);
	}
      id_list_destroy (&add_ppl);
    }

  query->filename = activity_dup_nonnull (std_get_nonempty (cgi,
							    "filename"));

  free (query->begin_date);
  query->begin_date = NULL;
  
  free (query->end_date);
  query->end_date = NULL;
  
  {
    int date = stoi (std_get (cgi, "date"));

    switch (date)
      {
      case 1:
	query->begin_date = today_minus (7);
	break;

      case 2:
	query->begin_date = today_minus (14);
	break;

      case 3:
	query->begin_date = today_minus (28);
	break;

      case 4:
	if (!empty (std_get (cgi, "begin")))
	  query->begin_date = xstrdup (std_get (cgi, "begin"));
	if (!empty (std_get (cgi, "end")))
	  query->end_date = xstrdup (std_get (cgi, "end"));
	break;
	
      default:
	break;
      }
  }
  
  query->limit = stoi (std_get (cgi, "limit"));
  if (query->limit < 1)
    query->limit = 100;
  else if (query->limit > 1000)
    query->limit = 1000;
  
  {
    const char *sort = std_get (cgi, "sort");
    
    if (!strcmp (sort, "filename"))
      query->sort = AC_SORT_FILENAME;
    else if (!strcmp (sort, "user"))
      query->sort = AC_SORT_USER;
    else if (!strcmp (sort, "activity"))
      query->sort = AC_SORT_ACTIVITY;
    else if (!strcmp (sort, "date"))
      query->sort = AC_SORT_DATE;
  }
}

/* Sets QUERY->last_read_seq to the appropriate value
   for the user represented in CX. */
void
activity_query_set_last_read (struct context *cx,
			      struct activity_query *query)
{
  db_result *r;

  db_query ("SELECT last_read_web FROM user_group "
	    "WHERE usr = '%sq' AND grp = '%sq'",
	    DB_FORMAT | DB_TUPLES, &r, cx->usr, cx->grp);

  query->last_read_seq = stoi (dbr_value (r, 0, 0));
      
  dbr_destroy (r);
}

/* Performs and reports a reset of the activity log if requested by
   the user.
   Returns nonzero only if a reset was performed. */
static int
handle_reset (struct request *rq)
{
  const char *reset_string;
  int reset;

  reset_string = std_get (rq->cgi, "reset");
  if (!strcmp (reset_string, "yes"))
    reset = get_seq ();
  else
    reset = stoi (reset_string);
  
  if (reset == 0)
    return 0;
  
  db_query ("UPDATE user_group SET last_read_web = %d "
	    "WHERE usr = '%sq' AND grp = '%sq' AND last_read_web < %d",
	    DB_FORMAT, NULL, reset, rq->cx.usr, rq->cx.grp, reset);

  if (!strcmp (std_get (rq->cgi, "webresetsmail"), "yes"))
    db_query ("UPDATE user_group SET last_read_mail = %d "
	      "WHERE usr = '%sq' AND grp = '%sq' AND last_read_web < %d",
	      DB_FORMAT, NULL, reset, rq->cx.usr, rq->cx.grp, reset);

  html ("<H2 ALIGN=CENTER>%s</H2>\n"
	"<P>%s</P>\n",
	_("Activity Reset"),
	_("The activity log reset is complete."));

  return 1;
}

