/* TeamSCOPE - Software for Collaborative Project Environments.
   Copyright (C) 1999, 2000, 2001 Michigan State University.
   Written by Ben Pfaff <pfaffben@msu.edu>.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To contact the TeamSCOPE team, send email to scope@cscw.msu.edu, or
   write to Charles Steinfield, Room 436 Communication Arts Building,
   Michigan State University, Department of Telecommunication, East
   Lansing MI 48824-1212. */

#include <config.h>
#include <ctype.h>
#include <assert.h>
#include <limits.h>
#include <stdlib.h>
#include "activity.h"
#include "activityP.h"
#include "alloc.h"
#include "cgi.h"
#include "datetime.h"
#include "db.h"
#include "error.h"
#include "html.h"
#include "identity.h"
#include "id-list.h"
#include "scope.cgi.h"
#include "std.h"
#include "str.h"

static void make_query_url (const struct activity_query *, struct url *);

/* Log an activity given S1 and S2, category CAT and type TYPE, and
   sequence number reference REF_SEQ. */
void 
log_activity (struct context *cx, int s1, int s2,
	      enum activity_cat cat, enum activity_type type, int ref_seq)
{
  db_query ("INSERT INTO activities VALUES"
	    "(datetime (timenow()), %d, '%sq', '%sq', %d, %d, %d, %d, "
	    "nextval('seq'))",
	    DB_FORMAT, NULL, s1, cx->usr, cx->grp, (int) cat, (int) type,
	    s2, ref_seq);
}

/* Log an activity given S1 and S2, category CAT and type TYPE, and
   sequence number reference REF_SEQ. */
void 
log_activity_by_name (struct context *cx, const char *s1, const char *s2,
		      enum activity_cat cat, enum activity_type type,
		      int ref_seq)
{
  log_activity (cx, get_file_seq (cx, s1), get_file_seq (cx, s2), cat, type,
		ref_seq);
}

/* Converts the activity query represented by QUERY into a new
   dictionary suitable for saving to a configuration file, and returns
   the dictionary. */
struct std *
activity_query_to_dict (const struct activity_query *query)
{
  struct std *d = std_create (1);

  if (query->sort == AC_SORT_USER)
    std_put (d, "sort", "user");
  else if (query->sort == AC_SORT_ACTIVITY)
    std_put (d, "sort", "activity");
  else if (query->sort == AC_SORT_FILENAME)
    std_put (d, "sort", "filename");
  else
    std_put (d, "sort", "date");

  if (query->begin_date)
    std_put (d, "begindate", query->begin_date);
  if (query->end_date)
    std_put (d, "enddate", query->end_date);
  
  {
    int i, j;
    struct st key, value;

    st_init (&key, 64);
    st_init (&value, 64);
    for (i = 0; i < query->n_people; i++)
      {
	for (j = 0; j < AC_N_CATEGORIES; j++)
	  if (query->person[i].type[j] != 0)
	    {
	      int k;
	    
	      st_clear (&key);
	      st_clear (&value);

	      st_printf (&key, "person-%d-type-%d", i, j);
	      for (k = 0; k < AC_MAX_ACTIVITIES; k++)
		if (query->person[i].type[j] & (1ul << k))
		  {
		    if (st_length (&value) > 0)
		      st_putchar (&value, ',');
		    st_printf (&value, "%d", k);
		  }

	      std_put (d, st_value (&key), st_value (&value));
	    }

	st_clear (&key);
	st_printf (&key, "person-%d-name", i);
	id_list_to_std (&query->person[i].usr, d, st_value (&key));
      }
    
    st_clear (&value);
    st_printf (&value, "%d", query->n_people);
    std_put (d, "n-people", st_value (&value));

    st_destroy (&key);
    st_destroy (&value);
  }
	    
  if (query->filename)
    std_put (d, "filename", query->filename);
  if (query->filename_exclude)
    std_put (d, "filetype", "exclude");
  
  {
    char n[INT_DIGITS + 1];
    snprintf (n, INT_DIGITS + 1, "%d", query->limit);
    std_put (d, "limit", n);
  }

  return d;
}

/* Converts a dictionary D saved by activity_query_to_dict() into
   activity query QUERY.  The dictionary corresponds to USR. */
void
activity_query_from_dict (struct activity_query *query, const struct std *d,
			  const char *usr)
{
  {
    struct st key;
    int i;
    
    query->n_people = stoi (std_get (d, "n-people"));
    if (query->n_people < 0)
      query->n_people = 0;
    else if (query->n_people > 32)
      query->n_people = 32;

    query->person = xmalloc (sizeof *query->person * query->n_people);

    st_init (&key, 64);
    for (i = 0; i < query->n_people; i++)
      {
	int j;
	
	st_clear (&key);
	st_printf (&key, "person-%d-name", i);
	id_list_from_std (&query->person[i].usr, d, st_value (&key), 0);

	for (j = 0; j < AC_N_CATEGORIES; j++)
	  {
	    const char *value;

	    query->person[i].type[j] = 0;
	    
	    st_clear (&key);
	    st_printf (&key, "person-%d-type-%d", i, j);
	    value = std_get (d, st_value (&key));
	    while (isdigit ((unsigned char) *value))
	      {
		int index = stoi (value);
		value += strspn (value, "0123456789");
		value += *value == ',';

		if (index >= 0 && index < AC_MAX_ACTIVITIES)
		  query->person[i].type[j] |= 1ul << index;
	      }
	  }
      }
  }

  /* Convert old-style inflexible specifications to the new format. */
  if (!std_contains (d, "n-people")
      && (std_contains (d, "cat") || std_contains (d, "people")))
    {
      query->person = xmalloc (sizeof *query->person);
      query->n_people = 1;
      if (std_contains (d, "people"))
	id_list_from_std (&query->person->usr, d, "people", 0);
      else
	id_list_for_user (&query->person->usr, usr);
      activity_bm_all (query->person->type);
      if (std_contains (d, "cat"))
	{
	  const char **keys = std_get_many (d, "cat", NULL);

	  if (keys)
	    {
	      activity_bm_zero (query->person->type);
	      for (; *keys; keys++)
		{
		  int index;
		  
		  if (!strcmp (*keys, "file"))
		    index = AC_CAT_FILE;
		  else if (!strcmp (*keys, "message"))
		    index = AC_CAT_MSG_BOARD;
		  else if (!strcmp (*keys, "event"))
		    index = AC_CAT_EVENT;
		  else
		    index = 0;

		  query->person->type[index]
		    |= ACTIVITY_MASK_FOR_WIDTH (activity_cat_cnt[index]);
		}
	    }
	}
    }
      
  if (!query->begin_date)
    query->begin_date = activity_dup_nonnull (std_get_null (d, "begindate"));
  if (!query->end_date)
    query->end_date = activity_dup_nonnull (std_get_null (d, "enddate"));
  if (!query->filename)
    {
      query->filename = activity_dup_nonnull (std_get_null (d, "filename"));
      if (!strcmp (std_get (d, "filetype"), "exclude"))
	query->filename_exclude = 1;
      else
	query->filename_exclude = 0;
    }
  if (std_contains (d, "limit"))
    query->limit = stoi (std_get (d, "limit"));

  if (query->sort == AC_SORT_DEFAULT)
    {
      const char *sort = std_get (d, "sort");

      if (!strcmp (sort, "user"))
	query->sort = AC_SORT_USER;
      else if (!strcmp (sort, "activity"))
	query->sort = AC_SORT_ACTIVITY;
      else if (!strcmp (sort, "filename"))
	query->sort = AC_SORT_FILENAME;
      else
	query->sort = AC_SORT_DATE;
    }
}

/* Initialize a query structure to its default contents, which cause
   all the records in the activity table to be retrieved. */
void
activity_query_init (struct activity_query *query)
{
  query->sort = AC_SORT_DEFAULT;

  query->begin_date = NULL;
  query->end_date = NULL;

  query->person = NULL;
  query->n_people = 0;
  
  query->filename = NULL;
  query->filename_exclude = 0;

  query->last_read_seq = 0;
  query->search = 0;

  query->start = 0;
  query->limit = 100;

  query->result = NULL;
  query->count = 0;
}

/* Destroy a query structure. */
void
activity_query_destroy (struct activity_query *query)
{
  int i;
  
  free (query->begin_date);
  query->begin_date = NULL;

  free (query->end_date);
  query->end_date = NULL;

  for (i = 0; i < query->n_people; i++)
    id_list_destroy (&query->person[i].usr);
  free (query->person);
  query->person = NULL;
  query->n_people = 0;

  free (query->filename);
  query->filename = NULL;

  if (query->result != NULL)
    {
      dbr_destroy (query->result);
      query->result = NULL;
    }
}

/* Producing SQL code for a query. */

/* A node in the SQL "parse tree" that we construct. */
struct query_node
  {
    int nonterm;			/* 1=nonterminal node, 0=terminal. */
    char *s;				/* Terminal: SQL expression;
					   nonterminal: SQL operator name. */
    struct query_node *child;		/* Nonterminal: first child node. */
    struct query_node *next;		/* Next node in child list. */
  };

/* Adds CHILD as a child of nonterminal node PARENT. */
static void
add_child (struct query_node *parent, struct query_node *child)
{
  assert (parent != NULL && child != NULL && parent->nonterm != 0);
  
  if (parent->child == NULL)
    parent->child = child;
  else
    {
      struct query_node *iter;

      for (iter = parent->child; iter->next != NULL; iter = iter->next)
	;
      iter->next = child;
    }
}

/* Creates and returns a nonterminal node for SQL operator OP.
   The new node is made a child of PARENT, if non-null. */
static struct query_node *
make_op (struct query_node *parent, char *op)
{
  struct query_node *new;

  assert (op != NULL);
  
  new = xmalloc (sizeof *new);
  new->nonterm = 1;
  new->s = op;
  new->child = new->next = NULL;

  if (parent != NULL)
    add_child (parent, new);

  return new;
}

/* Creates a new terminal node for SQL code TERM
   as a child of PARENT. */
static void 
make_term (struct query_node *parent, char *term)
{
  struct query_node *new;

  assert (parent != NULL && term != NULL);
  
  new = xmalloc (sizeof *new);
  new->nonterm = 0;
  new->s = term;
  new->child = new->next = NULL;
  add_child (parent, new);
}

/* Appends SQL code for NODE to S. */
static void
format_query_node (struct query_node *node, struct st *s)
{
  if (node->nonterm == 0)
    st_printf (s, "(%s)", node->s);
  else if (node->child == 0)
    st_putchar (s, '0');
  else if (node->child->next == NULL)
    format_query_node (node->child, s);
  else 
    {
      struct query_node *iter;
      
      st_putchar (s, '(');
      format_query_node (node->child, s);

      for (iter = node->child->next; iter != NULL; iter = iter->next)
	{
	  st_printf (s, " %s ", node->s);
	  format_query_node (iter, s);
	}

      st_putchar (s, ')');
    }
}

/* Destroys NODE and all its descendants. */
static void
free_query_node (struct query_node *node)
{
  assert (node != NULL && node->s != NULL);

  if (node->nonterm == 0)
    {
      free (node->s);
      free (node);
    }
  else 
    {
      struct query_node *iter, *next;

      for (iter = node->child; iter != NULL; iter = next)
	{
	  next = iter->next;
	  free_query_node (iter);
	}
      free (node);
    }
}

/* Run the query corresponding to QUERY and store the results back
   into QUERY. */
void
activity_query_run (struct context *cx, struct activity_query *query)
{
  struct st query_string;
  struct query_node *crit;

  crit = make_op (NULL, "AND");
  make_term (crit, xstrdup ("s1 = t1.seq AND s2 = t2.seq"));
  
  if (query->begin_date)
    make_term (crit, blp_asprintf ("date >= '%s'", query->begin_date));
  if (query->end_date)
    make_term (crit, blp_asprintf ("date <= '%s'", query->end_date));

  /* This chunk of code is the whole reason for the usage of explicit
     parse trees here.  It is ugly now, but you wouldn't believe how
     bad it was before. */
  if (query->n_people > 0)
    {
      struct query_node *people = make_op (crit, "OR");
      int i;

      for (i = 0; i < query->n_people; i++)
	{
	  struct query_node *person = make_op (people, "AND");
	  
	  if (query->person[i].usr.n > 0)
	    {
	      struct st s;
	      
	      st_create (&s, "usr IN ");
	      id_list_to_sql (&query->person[i].usr, &s);

	      make_term (person, st_value (&s));
	    }

	  if (!activity_bm_all_p (query->person[i].type))
	    {
	      struct query_node *activities = make_op (person, "OR");
	      int j;

	      for (j = 1; j < AC_N_CATEGORIES; j++)
		if (!activity_zero_bits_p (query->person[i].type[j], j))
		  {
		    struct query_node *category = make_op (activities, "AND");
		    
		    make_term (category, blp_asprintf ("cat = %d", j));
		    if (!activity_all_bits_p (query->person[i].type[j], j))
		      {
			struct st s;
			int k, count;

			st_create (&s, "type IN (");
			for (count = 0, k = 1; k < AC_MAX_ACTIVITIES; k++)
			  if ((query->person[i].type[j] & (1ul << k)) != 0)
			    {
			      if (count++ > 0)
				st_putchar (&s, ',');
			      st_printf (&s, "%d", k);
			    }
			st_putchar (&s, ')');

			make_term (category, st_value (&s));
		      }
		  }
	    }
	}
    }
  
  if (query->filename)
    {
      struct st filename;
      struct st regexp;

      /* Make the filename into a regular expression that will only
	 match in the basename, not the dirname. */
      st_init (&regexp, 64);
      {
	const char *cp;
	for (cp = query->filename; *cp; cp++)
	  {
	    if (strchr ("^.[$()|*+?{\\", *cp))
	      st_putchar (&regexp, '\\');
	    st_putchar (&regexp, *cp);
	  }
      }
      st_concat (&regexp, "[^/]*$");
      
      if (query->filename_exclude)
	st_create (&filename, "T1.string !~ ");
      else
	st_create (&filename, "T1.string ~ ");
      st_concat_quoted (&filename, st_value (&regexp), '\'');
      make_term (crit, st_value (&filename));

      st_destroy (&regexp);
    }

  if (query->last_read_seq)
    make_term (crit,
	       blp_asprintf ("activities.seq > %d", query->last_read_seq));

  make_term (crit, blp_asprintf ("((T1.grp = '%s' OR T2.grp = '%s') "
				 "AND cat = %d) "
				 "OR activities.grp = '%s'",
				 cx->grp, cx->grp,
				 AC_CAT_FILE,
				 cx->grp));
	    
  db_begin ();
  st_create (&query_string,
	     "DECLARE acts CURSOR FOR "
	     "SELECT date, T1.string AS string1, usr, cat, type, "
	     "T2.string AS string2, ref_seq, activities.seq "
	     "FROM activities, strings T1, strings T2 "
	     "WHERE ");
  format_query_node (crit, &query_string);
  free_query_node (crit);

  switch (query->sort)
    {
    case AC_SORT_FILENAME:
      st_concat (&query_string, " ORDER BY cat, string1, date");
      break;

    case AC_SORT_DEFAULT:
      query->sort = AC_SORT_DATE;
      /* fall through */

    case AC_SORT_DATE:
      /* Descending order is what users want, apparently.
	 Give it to them. */
      st_concat (&query_string, " ORDER BY date DESC");
      break;

    case AC_SORT_USER:
      st_concat (&query_string, " ORDER BY usr, date");
      break;

    case AC_SORT_ACTIVITY:
      st_concat (&query_string, " ORDER BY cat, type, date");
      break;

    default:
      assert (0);
    }

  /* FIXME: if the user gave us bad data, then the query might
     fail. */
  db_query (st_value (&query_string), 0, NULL);
  
  if (query->start)
    db_query ("MOVE %d IN acts", DB_FORMAT, NULL, query->start);

  st_clear (&query_string);
  st_printf (&query_string, "FETCH %d IN acts", query->limit);
  db_query (st_value (&query_string), DB_TUPLES, &query->result);
  query->count = dbr_nrows (query->result);

  db_query ("CLOSE acts", 0, NULL);
  db_end ();
  
  st_destroy (&query_string);
}

/* Return the activity name corresponding to CAT and TYPE. */
const char *
activity_type_name (int cat, int type)
{
  switch (cat)
    {
    case AC_CAT_FILE:
      {
	static const char *names[AC_FILE_N_ACTIVITIES] = 
	  {
	    NULL,
	    N_("downloaded"),
	    N_("downloaded as an archive"),
	    N_("uploaded"),
	    N_("deleted"),
	    N_("copied"),
	    N_("renamed"),
	    N_("folder created"),
	    N_("folder deleted"),
	    N_("permissions changed"),
	    N_("moved"),
	    N_("posted comment"),
	    N_("deleted comment"),
	    N_("read comment"),
	    N_("comment was mailed"),
	  };

	assert (type >= 1 && type < AC_FILE_N_ACTIVITIES);
	return gettext (names[type]);
      }

    case AC_CAT_MSG_BOARD:
      {
	static const char *names[AC_MSG_N_ACTIVITIES] = 
	  {
	    NULL,
	    N_("posted"),
	    N_("deleted"),
	    N_("read"),
	    N_("was mailed"),
	  };

	assert (type >= 1 && type < AC_MSG_N_ACTIVITIES);
	return gettext (names[type]);
      }

    case AC_CAT_EVENT:
      {
	static const char *names[AC_EVENT_N_ACTIVITIES] = 
	  {
	    NULL,
	    N_("created"),
	    N_("deleted"),
	    N_("changed"),
	    N_("viewed"),
	  };

	assert (type >= 1 && type < AC_EVENT_N_ACTIVITIES);
	return gettext (names[type]);
      }

    default:
      assert (0);
    }
}

/* Returns a human-readable name for activity category CAT. */
const char *
activity_category_name (int cat)
{
  switch (cat)
    {
    case AC_CAT_FILE:
      return _("File Activities");

    case AC_CAT_MSG_BOARD:
      return _("Message Board Activities");

    case AC_CAT_EVENT:
      return _("Calendar Activities");

    default:
      fail ("bad category value %d", cat);
    }
}

/* Display one page to query results for QUERY referencing url URL.
   Display the link as TAG.  The url won't be displayed if the current
   page is at offset START into the query results. */
static void
page (struct activity_query *query, struct url *url,
      const char *tag, int start)
{
  if (start < 0)
    start = 0;
  if (start == query->start)
    html ("%s\n", tag);
  else
    {
      url_release (url);
      url_arg_int (url, "start", start);
      html ("<A HREF=\"%s\">%s</A>\n", url_string (url), tag);
    }
}
      
/* Display a menu of pages of query results available, like this:
   << Prev 1 2 3 4 5 Next >> */
static void
page_menu (struct activity_query *query, const struct url *base_url)
{
  struct url url;
  
  if (query->start == 0 && query->count < query->limit)
    return;

  {
    int save_start = query->start;

    query->start = 0;
    url_copy (&url, base_url);
    make_query_url (query, &url);
    query->start = save_start;
  }

  htmls ("<P ALIGN=CENTER>");
  if (query->start != 0)
    page (query, &url, _("&lt;&lt; Prev"), query->start - query->limit);
  {
    char buf[INT_DIGITS + 1];
    int i;
    
    for (i = 0; i < 20 && i * query->limit <= query->start; i++)
      {
	sprintf (buf, "%d", i + 1);
	page (query, &url, buf, i * query->limit);
      }
  }
  if (query->count == query->limit)
    page (query, &url, _("Next &gt;&gt;"), query->start + query->limit);
  htmls ("</P>\n");

  url_destroy (&url);
}

/* Makes and returns a reset link string for QUERY. */
static char *
reset_link (struct activity_query *query, const struct url *base_url)
{
  struct url url;
  struct st s;

  url_copy (&url, base_url);
  make_query_url (query, &url);
  url_arg_int (&url, "reset", get_seq ());
  
  st_init (&s, 64);
  st_printf (&s, "<P ALIGN=CENTER>[ <A HREF=\"%s\">%s</A> ]</P>\n\n",
	     url_string (&url), _("reset"));
  url_destroy (&url);
  
  return st_value (&s);
}

/* Display the result of QUERY.
   Uses BASE_URL as the starting point for URLs within the activity
   screen.
   Adds contents of EXTRA as extra arguments to all URLs not based on
   BASE_URL. */
void
activity_query_present (struct context *cx, struct activity_query *query,
			const struct url *base_url,
			const struct std *extra)
{
  struct activity a, b;
  
  char *reset;
  int i;

  if (!query->count)
    {
      if (query->search)
	html ("<P>%s</P>\n", _("Activity search returned no results."));
      else
	html ("<P>%s</P>\n",
	      _("There are no new activities since your last visit."));

      return;
    }

  
  page_menu (query, base_url);
  
  if (!query->search)
    {
      reset = reset_link (query, base_url);
      
      htmls (reset);
      free (reset);
    }

  htmls ("<TABLE BGCOLOR=\"#FFFFFF\" WIDTH=\"100%%\" BORDER=1>\n"
	"<TR>\n");
  
  {
    int save_sort = query->sort;
    int sort;
    
    for (sort = 0; sort < AC_SORT_COUNT; sort++)
      {
	static const char *column_head[AC_SORT_COUNT]
	  = {N_("User"), N_("Activity"), N_("File"), N_("Date")};

	if (sort != save_sort)
	  {
	    struct url url;

	    query->sort = sort;
	    url_copy (&url, base_url);
	    make_query_url (query, &url);
	    html ("<TH><A HREF=\"%s\">%s</A>",
		  url_string (&url), gettext (column_head[sort]));
	    url_destroy (&url);
	  }
	else
	  html ("<TH>%s", gettext (column_head[sort]));

	if (sort == AC_SORT_DATE)
	  {
	    struct url url;

	    query->sort = save_sort;
	    url_copy (&url, base_url);
	    make_query_url (query, &url);
	    html ("\n<A NAME=\"datehead\" HREF=\"%s#datenote\">"
		  "<SUP>*</SUP></A>",
		  url_string (&url));
	    url_destroy (&url);
	  }

	htmls ("</TH>\n");
      }

    query->sort = save_sort;
  }

  htmls ("</TR>\n");
  html_flush ();

  read_user_list (&cx->wg, cx->grp);
  {
    struct url msg_url;
    struct url event_url;

    url_create (&msg_url, "messages");
    url_arg_std (&msg_url, extra);
    url_mark (&msg_url);

    url_create (&event_url, "calendar");
    url_arg_std (&event_url, extra);
    url_mark (&event_url);

    for (i = 0; i < query->count; i++)
      {
	/* Keep track of the current and the previous query results by
	   alternating between using `a' and `b' for storage and
	   alternating the pointers to them as well. */
	struct activity *cur, *prev;
	prev = i % 2 ? &a : &b;
	cur = i % 2 ? &b : &a;

	interpret_query (query, cur, i);

	if ((query->sort == AC_SORT_FILENAME
	     || query->sort == AC_SORT_ACTIVITY)
	    && (i == 0 || cur->cat != prev->cat))
	  html ("<TR><TH COLSPAN=4 ALIGN=CENTER>%s</TH></TR>\n",
		activity_category_name (cur->cat));

	html ("<TR><TD>%s</TD><TD>%s</TD><TD>",
	      username_to_fullname (cx->wg, cur->user),
	      activity_type_name (cur->cat, cur->type));
	switch (cur->cat)
	  {
	  case AC_CAT_FILE:
	    html_a_file (cx, cur->s1, extra);
	    if (cur->type == AC_FILE_COPY
		|| cur->type == AC_FILE_RENAME
		|| cur->type == AC_FILE_MOVE)
	      {
		htmls (_(" from "));
		html_a_file (cx, cur->s2, extra);
	      }
	    break;

	  case AC_CAT_MSG_BOARD:
	    url_release (&msg_url);
	    url_arg_int (&msg_url, "seq", cur->ref_seq);
	    html ("<A HREF=\"%s\">", url_string (&msg_url));
	    html_escape (cur->s2);
	    htmls ("</A>");
	    break;

	  case AC_CAT_EVENT:
	    url_release (&event_url);
	    url_arg_int (&event_url, "seq", cur->ref_seq);
	    html ("<A HREF=\"%s\">", url_string (&event_url));
	    html_escape (cur->s1);
	    htmls ("</A>");
	    break;

	  default:
	    assert (0);
	  }
	
	{
	  struct tm date;

	  if (datetime_parse (cur->date, 0, 0, &date))
	    {
	      const char *format_string;

	      if (date.tm_year == cur_tm.tm_year)
		format_string = "%b %d %H:%M";
	      else
		format_string = "%b %d %H:%M %Y";
	  
	      html ("</TD><TD>%s</TD></TR>\n",
		    datetime_format (format_string, &date));
	    }
	  else
	    html ("</TD><TD>%s</TD></TR>\n", cur->date);
	}
      }
    url_destroy (&event_url);
    url_destroy (&msg_url);
  }
  htmls ("</TABLE>\n");

  html ("<P><A NAME=\"datenote\">*</SUP></A>\n"
	"%s: <STRONG>%s</STRONG>\n"
	"</P>\n",
	_("Expressed in your currently selected time zone"),
	getenv ("TZ") ? getenv ("TZ") : tzname[0]);

  page_menu (query, base_url);
}

/* Adds arguments to URL that so that it becomes a search according to
   the specifications in QUERY. */
static void 
make_query_url (const struct activity_query *query, struct url *url)
{
  if (query->begin_date)
    url_arg (url, "begin", query->begin_date);
  if (query->end_date)
    url_arg (url, "end", query->end_date);
  if (query->n_people > 0)
    {
      /* Only the first person can be put into the query, but then
	 again only one person should ever be *in* the query:
	 we only transmit information this way on a search, 
	 and a search can only have one person. */
      size_t i;
      struct st s;

      st_init (&s, 64);
      for (i = 0; i < AC_N_CATEGORIES; i++)
	if (query->person->type[i] != 0)
	  {
	    st_clear (&s);
	    st_printf (&s, "%d:%08lx",
		       i, (unsigned long) query->person->type[i]);
	    url_arg (url, "activities", st_value (&s));
	  }

      for (i = 0; i < query->person->usr.n; i++)
	url_arg (url, "people", query->person->usr.s[i]);
    }
  if (query->filename)
    url_arg (url, "filename", query->filename);
  if (query->filename_exclude)
    url_arg (url, "filetype", "exclude");
  if (query->start)
    url_arg_int (url, "start", query->start);
  if (query->limit != 100)
    url_arg_int (url, "limit", query->limit);
  if (query->search)
    url_arg (url, "search", "yes");
  if (query->last_read_seq)
    url_arg_int (url, "since", query->last_read_seq);

  if (query->sort != AC_SORT_DATE && query->sort != AC_SORT_DEFAULT)
    {
      static const char *sort_types[AC_SORT_COUNT]
	= {"user", "activity", "filename", "date"};

      url_arg (url, "sort", sort_types[query->sort]);
    }
  
  url_mark (url);
}

/* Interprets row ROW of the activity database query result for QUERY
   into activity ACTIVITY. */
void
interpret_query (const struct activity_query *query,
		 struct activity *activity, int row)
{
  activity->date = dbr_value (query->result, row, 0);
  activity->s1 = dbr_value (query->result, row, 1);
  activity->user = dbr_value (query->result, row, 2);
  activity->cat = stoi (dbr_value (query->result, row, 3));
  activity->type = stoi (dbr_value (query->result, row, 4));
  activity->s2 = dbr_value (query->result, row, 5);
  activity->ref_seq = stoi (dbr_value (query->result, row, 6));
}

/* Displays an abbreviated version of activities in CATEGORY for the
   last week. */
void
activity_abbreviated (struct request *rq, int category)
{
  struct activity_query query;
  struct tm start;
  struct std *files;

  activity_query_init (&query);
  
  {
    time_t when = cur_time - 86400 * 7;

    query.begin_date = xstrdup (ctime (&when));
    *strchr (query.begin_date, '\n') = '\0';

    start = *localtime (&when);
  }

  query.limit = 10000;
  query.sort = AC_SORT_DATE;
  query.n_people = 1;
  query.person = xmalloc (sizeof *query.person);
  id_list_for_user (&query.person->usr, rq->cx.usr);
  activity_bm_zero (query.person->type);
  query.person->type[category]
    = ACTIVITY_MASK_FOR_WIDTH (activity_cat_cnt[category]);

  activity_query_run (&rq->cx, &query);
  free ((char *) query.begin_date);

  if (query.count == 0)
    {
      html (_("(No activity in past week)\n"));
      activity_query_destroy (&query);
      return;
    }

  {
    struct activity activity;
    int i;

    files = std_create (0);
    for (i = 0; i < query.count; i++)
      {
	interpret_query (&query, &activity, i);
	if (!std_put (files, activity.s1, "1"))
	  {
	    char count[INT_DIGITS + 1];
	    snprintf (count, INT_DIGITS + 1, "%ld",
		      stol (std_get (files, activity.s1)) + 1);
	    std_replace (files, activity.s1, count);
	  }
      }
  }
  
  {
    struct activity activity;
    struct url event_url;
    struct url activity_url;
    int n_shown;
    int i;

    url_create (&event_url, "calendar");
    url_mark (&event_url);

    url_create (&activity_url, "activity");
    url_arg (&activity_url, "sort", "date");
    url_arg (&activity_url, "search", "yes");
    url_mark (&activity_url);

    read_user_list (&rq->cx.wg, rq->cx.grp);
    for (i = n_shown = 0; i < query.count && n_shown < 10; i++)
      {
	int count;

	interpret_query (&query, &activity, i);
	count = stoi (std_get (files, activity.s1));
	if (!count)
	  continue;

	n_shown++;
	std_delete (files, activity.s1);

	html_list (1);
	htmls ("<STRONG>");
	if (activity.cat == AC_CAT_FILE)
	  html_a_file (&rq->cx, activity.s1, NULL);
	else
	  {
	    url_release (&event_url);
	    url_arg_int (&event_url, "seq", activity.ref_seq);
	    html ("<A HREF=\"%s\">", url_string (&event_url));
	    html_escape (activity.s1);
	    htmls ("</A>");
	  }
	htmls ("</STRONG>: ");
	
	if (count == 1)
	  {
	    struct tm date;

	    datetime_parse (activity.date, 0, 0, &date);
	    html ("%s %s %s %s %s\n",
		  activity_type_name (activity.cat, activity.type),
		  _("by"),
		  username_to_fullname (rq->cx.wg, activity.user),
		  _("at"),
		  datetime_format ("%d %b %H:%M", &date));
	  }
	else
	  {
	    url_release (&activity_url);
	    url_arg (&activity_url, "filename", activity.s1);
	    html ("%d <A HREF=\"%s\">%s</A> %s\n",
		  count, url_string (&activity_url), _("activities"),
		  _("in last week"));
	  }
      }
  }

  html_list (0);
  if (std_count (files))
    htmls (_("(For brevity, only the 10 most recently active "
	     "files or events are shown.)\n"));

  activity_query_destroy (&query);
}

/* Returns a new copy of string S if S is non-null; otherwise, returns
   a null pointer. */
char *
activity_dup_nonnull (const char *s)
{
  return s != NULL ? xstrdup (s) : NULL;
}

/* Activity category utilities. */

/* Users can select what activities they want to view on a pretty
   granular basis.  We represent a set of activity types as an array
   of `unsigned long's, one per activity category.  Each activity type
   is a single bit in its unsigned long.  Note that for simplicity
   (i.e., to avoid lots of addition and subtraction by 1 that would be
   confusing and troublesome) we don't use the first unsigned long or
   bit 0 in any of them. */

/* Number of activities in each category. */
const unsigned long activity_cat_cnt[AC_N_CATEGORIES] =
  {
    0,
    AC_FILE_N_ACTIVITIES,
    AC_MSG_N_ACTIVITIES,
    AC_EVENT_N_ACTIVITIES,
  };

/* Returns nonzero only if no bits are set in BITS for activity
   category CAT. */
int
activity_zero_bits_p (unsigned long bits, int cat)
{
  unsigned long mask = ACTIVITY_MASK_FOR_WIDTH (activity_cat_cnt[cat]);
  return (bits & mask) == 0;
}

/* Returns nonzero only if every bit is set in BITS for activity
   category CAT. */
int
activity_all_bits_p (unsigned long bits, int cat)
{
  unsigned long mask = ACTIVITY_MASK_FOR_WIDTH (activity_cat_cnt[cat]);
  return (bits & mask) == mask;
}

/* Returns nonzero only if no activities are selected in BM[]. */
int
activity_bm_zero_p (const unsigned long bm[AC_N_CATEGORIES])
{
  int i;

  for (i = 0; i < AC_N_CATEGORIES; i++)
    if (!activity_zero_bits_p (bm[i], i))
      return 0;
  return 1;
}

/* Returns nonzero only if every activity is selected in BM[]. */
int
activity_bm_all_p (const unsigned long bm[AC_N_CATEGORIES])
{
  int i;

  for (i = 0; i < AC_N_CATEGORIES; i++)
    if (!activity_all_bits_p (bm[i], i))
      return 0;
  
  return 1;
}

/* Zeros out all the bits in BM[]. */
void
activity_bm_zero (unsigned long bm[AC_N_CATEGORIES])
{
  int i;

  for (i = 0; i < AC_N_CATEGORIES; i++)
    bm[i] = 0;
}

/* Sets all of the bits in BM. */
void
activity_bm_all (unsigned long bm[AC_N_CATEGORIES])
{
  int i;

  for (i = 0; i < AC_N_CATEGORIES; i++)
    bm[i] = ULONG_MAX;
}

/* Returns nonzero only if the bitmaps at offsets A and B within BM
   have the same bits set.
   The bitmaps should have been normalized with norm_crit() to prevent
   false negatives. */
int
activity_bm_equal (unsigned long (*bm)[AC_N_CATEGORIES], int a, int b)
{
  int i;

  for (i = 0; i < AC_N_CATEGORIES; i++)
    if (bm[a][i] != bm[b][i])
      return 0;
  return 1;
}

/* Copies bitmap SRC into DST. */
void
activity_bm_copy (unsigned long dst[AC_N_CATEGORIES],
	 const unsigned long src[AC_N_CATEGORIES])
{
  int i;

  for (i = 0; i < AC_N_CATEGORIES; i++)
    dst[i] = src[i];
}
