/* TeamSCOPE - Software for Collaborative Project Environments.
   Copyright (C) 1999, 2000, 2001 Michigan State University.
   Written by Ben Pfaff <pfaffben@msu.edu>.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To contact the TeamSCOPE team, send email to scope@cscw.msu.edu, or
   write to Charles Steinfield, Room 436 Communication Arts Building,
   Michigan State University, Department of Telecommunication, East
   Lansing MI 48824-1212. */

#include <config.h>
#include <assert.h>
#include <ctype.h>
#include <limits.h>
#include <stdarg.h>
#include <stdlib.h>
#include <time.h>
#include "alloc.h"
#include "datetime.h"
#include "db.h"
#include "error.h"
#include "scope.cgi.h"
#include "std.h"
#include "str.h"

#undef DEBUGGING
/*#define DEBUGGING 1*/
/*#define VERIFY 1*/

/* A single date-time token extracted from a date-time string. */
struct dt_token
  {
    /* Token type:
       ':' - Time separator (:).
       ',' - Comma (,).
       '/' - Date separator (/, -, or .).
       'n' - Number.  Its value is in `value'.
       'w' - Weekday.  Its value (0 = Sunday) is in `value'.
       'm' - Month.  Its value (0 = January) is in `value'.
       'z' - Alphabetic string, presumably a timezone.
             Its contents are in `s'.
       '\0' - End of token stream. */
    short type;

    /* Used for token type 'n', 'w', 'm' only. */
    short value;

    /* Used for token type `z' only. */
    char s[16];
  };

/* When an error occurs, an appropriate error message is stored into
   this variable. */
const char *datetime_error;

static struct dt_token *tokenize_date (const char *s);
static int matches (struct dt_token **t, const char *format, ...);
static int recognize (const char *s, struct dt_token *tok);

#if VERIFY
static void verify_print (const char *ts, struct tm *tm);
#endif

/* Parse date-time string TS using OPTIONS.  If relative, it is taken
   with respect to date NOW, unless that is 0, in which case it is
   relative to the current date and time.  The results are returned as
   a time_t on success; 0 is returned on failure.  If successful and
   TM is non-NULL, the parsed date is also stored into *TM. */
time_t 
datetime_parse (const char *ts, int options, time_t now, struct tm *tm)
{
  struct dt_token *t, *st;
  int d, m, y;
  int H, M, S, mer;
  struct tm *now_tm;
  time_t rt;

  d = m = y = -1;
  H = M = S = -1;
  mer = -1;
  rt = 0;
  
#if DEBUGGING
  log ("parsing \"%s\" as date", ts);
#endif

  t = st = tokenize_date (ts);
  if (t == NULL)
    {
#if DEBUGGING
      log ("tokenizer failed");
#endif
      goto lossage;
    }

  if (!t->type)
    {
      datetime_error = _("Empty date string.");
      goto lossage;
    }
  
  if (!matches (&t, "w,dMy", &d, &m, &y)
      && !matches (&t, "w,dM", &d, &m)
      && !matches (&t, "w,dMy", &d, &m, &y)
      && !matches (&t, "w,Md,y", &m, &d, &y)
      && !matches (&t, "w,Md", &m, &d)
      && !matches (&t, "w,Mdy", &m, &d, &y)
      && !matches (&t, "w,Y/m/d", &y, &m, &d)
      && !matches (&t, "w,Ym", &y, &m))
    {
      if (options & DT_UK)
	(void) (matches (&t, "w,d/m/y", &d, &m, &y)
		|| matches (&t, "d/m/y", &d, &m, &y)
		|| matches (&t, "d/m", &d, &m));
      else
	(void) (matches (&t, "w,m/d/y", &m, &d, &y)
		|| matches (&t, "m/d/y", &m, &d, &y)
		|| matches (&t, "m/y", &m, &y));
    }

  /* FIXME: timezones. */
  if (t->type == 'z')
    t++;

  (void) (matches (&t, "h:S:Sa", &H, &M, &S, &mer)
	  || matches (&t, "h:Sa", &H, &M, &mer)
	  || matches (&t, "H:S:S", &H, &M, &S)
	  || matches (&t, "H:S", &H, &M));

  if (y == -1)
    matches (&t, "Y", &y);

  /* FIXME: timezones. */
  if (t->type == 'z')
    t++;

  if (t->type != 0)
    {
#if DEBUGGING
      log ("expected end of string");
#endif
      datetime_error = _("Syntax error in date.");
      goto lossage;
    }

#if DEBUGGING
  log ("SUCCESS!");
#endif

  if (now == 0)
    now = cur_time;
  now_tm = localtime (&now);
  if (y == -1)
    y = now_tm->tm_year;
  else if (y < 30)
    y += 100;
  else if (y >= 100)
    y -= 1900;

  if (m == -1)
    m = now_tm->tm_mon;
  else
    {
      if (m < 0 || m > 11)
	{
	  datetime_error = _("Month out of range.");
#if DEBUGGING
	  log ("bad month");
#endif
	  goto lossage;
	}
    }
  
  if (d == -1)
    {
      if (options & DT_DATE_REQUIRED)
	{
	  datetime_error = _("Date required but not supplied.");
	  goto lossage;
	}
      
      d = now_tm->tm_mday;
    }
  else if (d < 1 || d > datetime_month_days (y, m))
    {
      datetime_error = _("Day out of range for month.");
      goto lossage;
    }

  if (H == -1)
    {
      if (options & DT_TIME_REQUIRED)
	{
	  datetime_error = _("Time required but not supplied.");
	  goto lossage;
	}

      H = M = S = 0;
    }
  else
    {
      if (mer == 'p')
	H += 12;
      if (S == -1)
	S = 0;
    }

  {
    struct tm temp_tm;

    if (tm == NULL)
      tm = &temp_tm;

    tm->tm_sec = S;
    tm->tm_min = M;
    tm->tm_hour = H;
    tm->tm_mday = d;
    tm->tm_mon = m;
    tm->tm_year = y;
    tm->tm_isdst = -1;

    rt = mktime (tm);

#if VERIFY
    verify_print (ts, tm);
#endif
  }

 lossage:
  free (st);
  return rt;
}

/* Format WHEN into a static buffer using strftime() format TEMPLATE,
   and return the results. */
const char *
datetime_format (const char *template, const struct tm *when)
{
  static char buf[64];
  if (0 == strftime (buf, 64, template, when))
    fail ("date/time too long");
  return buf;
}

/* Returns the number of days in month MONTH (0...11) of year (YEAR +
   1900). */
int
datetime_month_days (int year, int month)
{
  assert (month >= 0 && month <= 11);
  
  if (month == 1 
      && ((year + 1900) % 4 == 0
	  && ((year + 1900) % 100 != 0 || (year + 1900) % 400 == 0)))
    return 29;
  else
    {
      static const int days[12] = {31,28,31,30,31,30,31,31,30,31,30,31};
      return days[month];
    }
}

/* Set timezone based on user preference. */
void
datetime_init_timezone (struct context *cx)
{
  const char *timezone = std_get_null (cx->options, "timezone");
  if (timezone == NULL)
    return;

  /* putenv() retains a reference to its string argument, so we must
     preserve it.  We can't use a struct st because then the string's
     location might change, leading to undefined behavior when
     putenv() looks through the environment. */
  {
    static char s[64];

    sprintf (s, "TZ=%.32s", timezone);
    putenv (s);
  }
  
  tzset ();

  db_query ("SET TIMEZONE = '%sq'", DB_FORMAT, NULL, timezone);
}

/* Private functions. */

/* Divide string S into date-time tokens and return a pointer to a
   dynamically allocated copy of the token stream. */
static struct dt_token *
tokenize_date (const char *s)
{
#define MAX_TOK 16
  struct dt_token tok[MAX_TOK];
  int n_tok;

  n_tok = 0;
  for (; *s; )
    {
      if (n_tok == MAX_TOK)
	{
	  datetime_error = _("String too long.");
	  return NULL;
	}

#if __CHECKER__
      memset (tok[n_tok].s, ' ', sizeof tok[n_tok].s);
      tok[n_tok].value = 0;
#endif
      switch (*s)
	{
	case ':':
	  tok[n_tok++].type = ':';
	  s++;
	  break;

	case ',':
	  tok[n_tok++].type = ',';
	  s++;
	  break;

	case ' ':
	case '\t':
	case '\r':
	case '\n':
	  /* Ignored. */
	  s++;
	  break;

	case '/':
	case '-':
	case '.':
	  tok[n_tok++].type = '/';
	  s++;
	  break;

	case '0': case '1': case '2': case '3': case '4':
	case '5': case '6': case '7': case '8': case '9':
	  {
	    tok[n_tok].value = 0;
	    while (isdigit ((unsigned char) *s))
	      tok[n_tok].value = tok[n_tok].value * 10 + *s++ - '0';

	    if (!strncasecmp (s, "st", 2)
		|| !strncasecmp (s, "nd", 2)
		|| !strncasecmp (s, "rd", 2)
		|| !strncasecmp (s, "th", 2))
	      s += 2;

	    tok[n_tok++].type = 'n';
	  }
	  break;

	default:
	  if (isalpha ((unsigned char) *s))
	    {
	      char buf[16];
	      char *p;

	      p = buf;
	      while (isalpha ((unsigned char) *s) && p < buf + 15)
		*p++ = *s++;
	      *p = '\0';
	      
	      if (!recognize (buf, &tok[n_tok]))
		{
		  tok[n_tok].type = 'z';
		  strcpy (tok[n_tok].s, buf);
		}
	      n_tok++;
	    }
	  else
	    {
	      datetime_error = "Bad date format.";
	      return NULL;
	    }
	  break;
	  
	}
    }
  
#if __CHECKER__
  memset (tok[n_tok].s, ' ', sizeof tok[n_tok].s);
  tok[n_tok].value = 0;
#endif
  tok[n_tok++].type = 0;

#if DEBUGGING
  {
    int i;
    char buf[MAX_TOK];
    
    for (i = 0; i < n_tok; i++)
      buf[i] = tok[i].type;
    log ("parsed tokens: \"%s\"", buf);
  }
#endif /* DEBUGGING */
  
  {
    struct dt_token *t;

    t = malloc (sizeof *t * n_tok);
    if (t == NULL)
      {
	datetime_error = _("Virtual memory exhausted.");
	return NULL;
      }
    memcpy (t, tok, sizeof *t * n_tok);

    return t;
  }
}

/* Compares a prefix of the token stream beginning at *T_ORIG to the
   pattern given by FORMAT.  If there is a match, stores the numeric
   values in the pattern into the provided varargs arguments, advances
   *T_ORIG beyond the match, and returns nonzero.  If there is no
   match, returns 0.

   FORMAT is a string of single-character match types, as follows.
   The token types that cause a value to be stored into a varargs int
   * argument are suffixed by (*).

   'w' - skip optional weekday name.
   ',' - skip optional comma.
   'd' - day number between 1 and 31 (*).
   'M' - month name, stored as 0...11 (*).
   'm' - month number between 1 and 12, stored as 0...11 (*).
   'y' - number of any value, representing a year (*).
   'Y' - four-digit year number (*).
   '/' - required date separator.
   ':' - required time separator.
   'h' - hour number between 1 and 12 (*).
   'H' - hour number between 0 and 23 (*).
   'S' - second or minute number between 0 and 59 (*).
   'a' - meridian: 'a', 'am' stored as 'a', 'p', 'pm' stored as 'p' (*).

   In addition, a token string is rejected as a match if it is
   followed by a separator, because that indicates that a full
   ``part'' of the date string wasn't matched.  */
static int 
matches (struct dt_token **t_orig, const char *format, ...)
{
#if DEBUGGING
  const char *orig_format = format;
  char tok_types[MAX_TOK];
#endif
  struct dt_token *t = *t_orig;
  int values[MAX_TOK];
  int *vp;
  va_list args;
  
#if DEBUGGING
  {
    int i;

    for (i = 0; ; i++)
      {
	tok_types[i] = (*t_orig)[i].type;
	if (tok_types[i] == 0)
	  break;
      }
  }
#endif

  vp = values;
  while (*format)
    switch (*format++)
      {
      case 'w':
	if (t->type == 'w')
	  t++;
	break;
	
      case ',':
	if (t->type == ',')
	  t++;
	break;

      case 'd':
	if (t->type != 'n' || t->value < 1 || t->value > 31)
	  goto fail;
	*vp++ = t->value;
	t++;
	break;

      case 'M':
	if (t->type != 'm')
	  goto fail;
	*vp++ = t->value;
	t++;
	break;

      case 'm':
	if (t->type != 'n' || t->value < 1 || t->value > 12)
	  goto fail;
	*vp++ = t->value - 1;
	t++;
	break;

      case 'y':
	if (t->type != 'n')
	  goto fail;
	*vp++ = t->value;
	t++;
	break;

      case 'Y':
	if (t->type != 'n' || t->value < 1000)
	  goto fail;
	*vp++ = t->value;
	t++;
	break;

      case '/':
	if (t->type != '/')
	  goto fail;
	t++;
	break;

      case ':':
	if (t->type != ':')
	  goto fail;
	t++;
	break;

      case 'h':
	if (t->type != 'n' || t->value < 1 || t->value > 12)
	  goto fail;
	*vp++ = t->value;
	t++;
	break;

      case 'H':
	if (t->type != 'n' || t->value < 0 || t->value > 23)
	  goto fail;
	*vp++ = t->value;
	t++;
	break;

      case 'S':
	if (t->type != 'n' || t->value < 0 || t->value > 59)
	  goto fail;
	*vp++ = t->value;
	t++;
	break;

      case 'a':
	if (t->type != 'z')
	  goto fail;
	if (!strcasecmp (t->s, "a") || !strcasecmp (t->s, "am"))
	  *vp++ = 'a';
	else if (!strcasecmp (t->s, "p") || !strcasecmp (t->s, "pm"))
	  *vp++ = 'p';
	else
	  goto fail;
	t++;
	break;

      default:
#if DEBUGGING
	log ("unknown format type %c", (unsigned char) format[-1]);
#endif
	assert (0);
      }

  /* If this part is followed by a separator, then it's not really a
     full part. */
  if (t->type == ':' || t->type == ',' || t->type == '/')
    goto fail;
  
#if DEBUGGING
  log ("\"%s\" matches \"%s\"", orig_format, tok_types);
#endif

  {
    int *v;

    va_start (args, format);
    for (v = values; v < vp; v++)
      *va_arg (args, int *) = *v;
    va_end (args);
  }

  *t_orig = t;
  return 1;

fail:
  va_end (args);
#if DEBUGGING
  log ("\"%s\" doesn't match \"%s\", failed at `%c'",
       orig_format, tok_types, (unsigned char) format[-1]);
#endif
  return 0;
}

static int try_recognize (const char *s, struct dt_token *tok, int type, ...);

/* Attempts to recognize string S as a day or month name.  If
   successful, sets *TOK appropriately and returns nonzero; otherwise,
   returns zero. */
static int
recognize (const char *s, struct dt_token *tok)
{
  return
    try_recognize (s, tok, 'm',
		   "January", "February", "March",
		   "April", "May", "June",
		   "July", "August", "September",
		   "October", "November", "December",
		   (char *) 0)
    || try_recognize (s, tok, 'w',
		      "Sunday", "Monday", "Tuesday", "Wednesday",
		      "Thursday", "Friday", "Saturday",
		      (char *) 0);
}

/* Tries to recognize S as one of the enumeral strings passed as
   varargs arguments and terminated with a null pointer.  If
   successful, stores TYPE as TOK->TYPE and the 0-based index of the
   enumeral as TOK->VALUE, and returns 1; otherwise, returns 1. */
static int 
try_recognize (const char *s, struct dt_token *tok, int type, ...)
{
  va_list args;
  int index;

  va_start (args, type);
  for (index = 0; ; index++)
    {
      const char *kw;

      kw = va_arg (args, const char *);
      if (kw == NULL)
	return 0;

      /* S matches KW if it at least the first three characters are
         the same. */
      {
	const char *cp;
	int n;

	for (n = 0, cp = s; *cp && *kw; cp++, kw++, n++)
	  if (tolower ((unsigned char) *cp) != tolower ((unsigned char) *kw))
	    goto no_match;
	if ((*kw == '\0' && *cp) || n < 3)
	  goto no_match;
      }
	
      tok->type = type;
      tok->value = index;
      va_end (args);
      return 1;

    no_match: ;
    }
  va_end (args);
}

#if VERIFY
/* Print helpful debugging info. */
static void
verify_print (const char *ts, struct tm *tm)
{
  char buf[64];

  strftime (buf, 64, "%d %b %Y %H:%M:%S", tm);
  log ("\"%s\" parsed as \"%s\"", ts, buf);
}
#endif /* VERIFY */
