/* TeamSCOPE - Software for Collaborative Project Environments.
   Copyright (C) 1999, 2000, 2001 Michigan State University.
   Written by Ben Pfaff <pfaffben@msu.edu>.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To contact the TeamSCOPE team, send email to scope@cscw.msu.edu, or
   write to Charles Steinfield, Room 436 Communication Arts Building,
   Michigan State University, Department of Telecommunication, East
   Lansing MI 48824-1212. */

#include <config.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include "alloc.h"
#include "db.h"
#include "error.h"
#include "identity.h"
#include "session.h"
#include "std.h"
#include "str.h"

/* Idea:

   Add to db_query an interface that allows the return types to
   be specified; i.e., "sdnb" for each row a string (column 0), a date
   (column 1), a number (column 2), and a boolean (column 3), along with
   the possibility to store these directly somehow.

   Can this be done? */

static PGconn *conn;
static int transaction_depth;

/* Public functions. */

static db_result *execute_query (const char *query);

/* Open a connection to database DBNAME on the database backend, or
   "scope" by default.  The effective uid should be appropriately set
   by the caller. */
void
db_open (const char *dbname)
{
  assert (conn == NULL);
  conn = PQsetdb (NULL, NULL, NULL, NULL, dbname != NULL ? dbname : "scope");
  assert (conn != NULL); /* libpq(3) says PQsetdb() never returns NULL. */

  if (PQstatus (conn) != CONNECTION_OK)
    fail ("couldn't connect to database engine: %s", PQerrorMessage (conn));

  /* Defaults on newer PostgreSQL aren't acceptable. */
  db_query ("SET DATESTYLE = 'US,postgres'", 0, NULL);

  /* Ensure that database connection gets closed on exit.
     Otherwise we tend to get ugly error messages on server console. */
  {
    static int registered = 0;
    if (!registered)
      {
	atexit (db_close);
	registered = 1;
      }
  }
}

/* Close the database connection, committing all writes. */
void
db_close (void)
{
  if (transaction_depth > 0)
    {
      log ("transaction left open!");
      transaction_depth = 1;
      db_end ();
    }

  if (conn != NULL)
    {
      PQfinish (conn);
      conn = NULL;
    }
}

/* Returns nonzero only if connected to the database. */
int
db_open_p (void)
{
  return conn != NULL;
}

/* Transactions. */

/* Begin a transaction.  Transactions may nest. */
void
db_begin (void)
{
  assert (conn != NULL);
  if (transaction_depth++ == 0)
    db_query ("BEGIN", 0, 0);
}

/* End a transaction. */
void
db_end (void)
{
  assert (conn != NULL);
  assert (transaction_depth > 0);
  
  if (--transaction_depth == 0)
    db_query ("END", 0, 0);
}

/* Submits QUERY to the SQL backend.  If the query is successful,
   returns a nonnegative value that, depending on the query, may
   represent the number of tuples affected.  If the query is
   unsuccessful, aborts the program, unless DB_FAIL_OK is given in
   FLAGS, in which case it returns -1 on failure.

   If DB_TUPLES is set in FLAGS, then RESULT must be nonnull and
   failure is any command status other than PGRES_TUPLES_OK.  If
   DB_TUPLES is not set, then RESULT must be a null pointer and
   failure is any command status other than PGRES_COMMAND_OK.

   A nonnull RESULT will be set to the result set returned by a
   successful query and to a null pointer for an unsuccessful query
   (assuming that DB_FAIL_OK is given).

   If DB_FORMAT is set in FLAGS, then QUERY will be formatted with
   vformat() before being passed to the database backend. */
int
db_query (const char *query, int flags, db_result **result, ...)
{
  db_result *r;
  
  assert (conn != NULL);
  assert (query != NULL);
  assert (((flags & DB_TUPLES) && result != NULL)
	  || (!(flags & DB_TUPLES) && result == NULL));

  if (flags & DB_FORMAT)
    {
      static char *buf;
      static size_t buf_size;

      va_list args;
      size_t needed;

      va_start (args, result);
      needed = vformat (buf, buf_size, query, args);
      va_end (args);
      
      if (needed >= buf_size)
	{
	  buf_size = needed;
	  buf = xrealloc (buf, buf_size);
	  
	  va_start (args, result);
	  vformat (buf, buf_size, query, args);
	  va_end (args);
	}
      
      query = buf;
    }

#if GLOBAL_DEBUGGING
  log ("%s", query);
#endif

  r = execute_query (query);

  /* PQexec() supposedly never returns a null pointer. */
  if (r == NULL)
    fail ("query returned null result: %s", query);

  /* Deal with query failure. */
  {
    const int rs = PQresultStatus (r);
    if (((flags & DB_TUPLES) && rs != PGRES_TUPLES_OK)
	|| (!(flags & DB_TUPLES) && rs != PGRES_COMMAND_OK))
      {
	if (!(flags & DB_FAIL_OK))
	  {
	    char *error = xstrdup (PQerrorMessage (conn));
	    size_t len = strlen (error);
	    if (len > 0 && error[len - 1] == '\n')
	      error[len - 1] = '\0';
	  
	    log ("query failed with message \"%s\" for query \"%s\"",
		 error, query);

	    fail ("failed database query");
	  }
	else
	  {
	    PQclear (r);
	    r = NULL;

	    if (result != NULL)
	      *result = NULL;

	    return -1;
	  }
	  
      }
  }

  /* Return to caller. */
  {
    int retval = r ? stoi (PQcmdTuples (r)) : 0;
      
    if (result)
      *result = r;
    else if (r != NULL)
      PQclear (r);

    return retval;
  }
}
      
/* Return the latest error message from the SQL backend. */
const char *
db_error (void)
{
  assert (conn != NULL);
  return PQerrorMessage (conn);
}

/* Result functions. */

/* Return the number of rows in query result RESULT returned by the
   backend. */
size_t
dbr_nrows (const db_result *result)
{
  int n_tuples;

  assert (result != NULL);

  n_tuples = PQntuples ((db_result *) result);
  assert (n_tuples >= 0);

  return (size_t) n_tuples;
}

/* Return the number of columns in query result RESULT returned by the
   backend. */
size_t
dbr_ncols (const db_result *result)
{
  int n_fields;

  assert (result != NULL);

  n_fields = PQnfields ((db_result *) result);
  assert (n_fields >= 0);

  return (size_t) n_fields;
}

/* Return the variate at row ROW and column COL in query result RESULT
   returned by the backend, as a null-terminated string.
   The string is owned by the table. */
char *
dbr_value (const db_result *result, int row, int col)
{
  char *value;

  assert (result && row >= 0 && col >= 0);

  value = PQgetvalue ((db_result *) result, row, col);
  assert (value != NULL);

  return value;
}

/* Return the length of the variate at row ROW and column COL in query
   result RESULT returned by the backend, not including the null
   terminator. */
size_t
dbr_value_length (const db_result *result, int row, int col)
{
  int length;

  assert (result && row >= 0 && col >= 0);

  length = PQgetlength ((db_result *) result, row, col);
  assert (length >= 0);

  return (size_t) length;
}

/* Returns a boolean indicating whether the variate at row ROW and
   column COLUMN in query result RESULT is a null value.  */
int
dbr_value_is_null (const db_result *result, int row, int col)
{
  assert (result && row >= 0 && col >= 0);

  return PQgetisnull ((db_result *) result, row, col);
}

/* Free the data represented by query result RESULT. */
void
dbr_destroy (db_result *result)
{
  assert (result != NULL);
  PQclear (result);
}

/* Commonly needed operations. */

/* Get and return a new sequence number. */
int
get_seq (void)
{
  db_result *r;
  int seq;

  assert (conn != NULL);
  db_query ("SELECT nextval ('seq')", DB_TUPLES, &r);
  seq = stoi (dbr_value (r, 0, 0));
  dbr_destroy (r);

  return seq;
}

/* Private functions. */

/* Execute QUERY and return the result.  Since libpq only uses an 8K
   buffer internally for queries (and tends to segfault if passed
   anything longer), limit the length to 7K, which should be (more
   than) sufficient and won't be a security hole. */
static db_result *
execute_query (const char *query)
{
  assert (conn != NULL);
  if (strlen (query) > 1024 * 7)
    fail ("query too long: %s", query);

  return PQexec (conn, query);
}

/* Formats FORMAT into BUFFER of size SIZE.  Returns the number of
   characters, including the null terminator, that would be written if
   there were room for all of them.  If the return value is greater than
   or equal to SIZE, and this function hasn't already been called with
   identical arguments to determine the buffer size necessary, then
   the contents of BUFFER on exit are undefined.

   Format specifiers:
   %% - literal percent sign
   %d - signed decimal integer
   %ss - standard string
   %sq - string with ' and \ quoted with \ prefix

   Other format specifiers may be added if needed in the future.
   Format specifiers' prefixes must match an existing printf() format
   specification so that gcc can check them at compile time. */
size_t
vformat (char *buffer, size_t size, const char *format, va_list args)
{
  ssize_t left = size;

  for (;;)
    switch (*format++)
      {
      case '\0':
	if (left-- > 0)
	  *buffer++ = '\0';
	return size - left;
	
      case '%':
	switch (*format++)
	  {
	  case '%':
	    if (left-- > 0)
	      *buffer++ = '%';
	    break;

	  case 'd':
            {
              /* Another example of fscking glibc silently fscking
                 changing return value behavior between two fscking
                 versions.  See str.c:st_printf() for more details. */
              char tmp[INT_DIGITS + 1];
              int count;

              count = snprintf (tmp, INT_DIGITS + 1, "%d", va_arg (args, int));
              left -= count;
              if (left > 0)
                {
                  strcpy (buffer, tmp);
                  buffer += count;
                }
            }
	    break;

	  case 's':
	    {
	      char *s = va_arg (args, char *);
	      assert (s != NULL);

	      if (*format == 's')
		{
		  size_t len = strlen (s);
		  
		  if (left >= (ssize_t) len)
		    {
		      memcpy (buffer, s, len);
		      buffer += len;
		    }

		  left -= len;
		}
	      else
		{
		  assert (*format == 'q');

		  for (;;)
		    switch (*s++)
		      {
		      case '\0':
			goto out;

		      case '\'':
			if (left-- > 0)
			  *buffer++ = '\\';
			if (left-- > 0)
			  *buffer++ = '\'';
			break;

		      case '\\':
			if (left-- > 0)
			  *buffer++ = '\\';
			if (left-- > 0)
			  *buffer++ = '\\';
			break;

		      default:
			if (left-- > 0)
			  *buffer++ = s[-1];
			break;
		      }
		out: ;
		}
	      format++;
	    }
	    break;

	  default:
	    assert (0);
	  }
	break;

      default:
	if (left-- > 0)
	  *buffer++ = format[-1];
	break;
      }
}

/* Retrieve and return the sequence number for the file named
   FILENAME. */
int
get_file_seq (struct context *cx, const char *filename)
{
  db_result *r;
  int seq;

  assert (cx != NULL);
  if (filename == NULL || filename[0] == '\0')
    return 0;

  db_begin ();
  db_query ("SELECT seq FROM strings WHERE string = '%sq'",
	    DB_FORMAT | DB_TUPLES, &r, filename);
  if (dbr_nrows (r) == 0)
    {
      seq = get_seq ();

      db_query ("INSERT INTO strings VALUES ('%sq', %d, '%sq')",
		DB_FORMAT, NULL, filename, seq,
		(has_prefix (filename, cx->shared) ? cx->grp : ""));
    }
  else
    seq = stoi (dbr_value (r, 0, 0));
  db_end ();
  dbr_destroy (r);

  return seq;
}
