/* TeamSCOPE - Software for Collaborative Project Environments.
   Copyright (C) 1999, 2000, 2001 Michigan State University.
   Written by Ben Pfaff <pfaffben@msu.edu>.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To contact the TeamSCOPE team, send email to scope@cscw.msu.edu, or
   write to Charles Steinfield, Room 436 Communication Arts Building,
   Michigan State University, Department of Telecommunication, East
   Lansing MI 48824-1212. */

#include <config.h>
#include <assert.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdio.h>
#include <pwd.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <time.h>
#include <unistd.h>
#include "activity.h"
#include "alloc.h"
#include "cgi.h"
#include "conf.h"
#include "datetime.h"
#include "db.h"
#include "error.h"
#include "files.h"
#include "html.h"
#include "identity.h"
#include "messages.h"
#include "priv.h"
#include "scope.cgi.h"
#include "session.h"
#include "str.h"
#include "std.h"

/* Public functions. */

/* Operation flags. */
enum
  {
    OPF_FILE = 001,	/* Operation applies to files. */
    OPF_DIR = 002,	/* Operation applies to directories. */
  };

typedef void op_func (struct request *, const char *filename,
		      const struct stat *);

/* An operation. */
struct operation
  {
    int flags;		/* One or more of OPF_*. */
    op_func *func;	/* Display handler function. */
    char *name;		/* Name. */
    char *label;	/* Label. */
    char *caption;	/* Explanation. */
  };

/* Operation indexes. */
enum
  {
    OP_COMMENT,
    OP_ACTIVITY,
    OP_VIEW,
    OP_DOWNLOAD,
    OP_UPLOAD,
    OP_REPLACE,
    OP_MOVE,
    OP_RENAME,
    OP_COPY,
    OP_DELETE,
    OP_MKDIR,
    OP_RMDIR,
    OP_CHMOD,

    OP_COUNT,		/* Number of operations. */
  };

/* Operation display types. */
enum
  {
    OPD_NORMAL = 'n',	/* Show normally. */
    OPD_DISABLE = 'd',	/* Show disabled. */
    OPD_SELECTED = 's',	/* Show as selected. */
    OPD_MOUSEOVER = 'm',/* Mouseover highlight. */
    OPD_OMIT = 0,	/* Don't display at all. */
  };

static op_func op_activity, op_comment, op_upload, op_download;
static op_func op_move, op_rename, op_mkdir, op_rmdir, op_copy;
static op_func op_delete, op_chmod;

static const struct operation op_table[OP_COUNT + 1] =
  {
    {OPF_FILE | OPF_DIR,
     op_comment,
     "comment",
     N_("comment"),
     N_("view comments about %s")},
    {OPF_FILE,
     op_activity,
     "activity",
     N_("activity"),
     N_("view activity on %s")},
    {OPF_FILE,
     op_download,
     "view",
     N_("view"),
     N_("view contents of %s")},
    {OPF_FILE | OPF_DIR,
     op_download,
     "download",
     N_("download"),
     N_("download contents of %s for save to disk")},
    {OPF_DIR,
     op_upload,
     "upload",
     N_("upload"),
     N_("upload a new file in %s")},
    {OPF_FILE,
     op_upload,
     "replace",
     N_("replace"),
     N_("upload a replacement copy of %s to this server")},
    {OPF_FILE | OPF_DIR,
     op_move,
     "move",
     N_("move"),
     N_("move %s to a different directory")},
    {OPF_FILE | OPF_DIR,
     op_rename,
     "rename",
     N_("rename"),
     N_("change the name of %s")},
    {OPF_FILE,
     op_copy,
     "copy",
     N_("copy"),
     N_("make another copy of %s under a different name")},
    {OPF_FILE,
     op_delete,
     "delete",
     N_("delete"),
     N_("delete %s")},
    {OPF_DIR,
     op_mkdir,
     "mkdir",
     N_("new<BR>folder"),
     N_("create a new folder under %s")},
    {OPF_DIR,
     op_rmdir,
     "rmdir",
     N_("remove<BR>folder"),
     N_("delete folder %s")},
    {OPF_FILE | OPF_DIR,
     op_chmod,
     "chmod",
     N_("access"),
     N_("change access privileges for %s")},
    {0,
     NULL,
     NULL,
     NULL,
     NULL},
  };

/* Types of sorting for directories. */
enum
  {
    FM_SORT_NAME,
    FM_SORT_SIZE,
    FM_SORT_DATE,

    FM_N_SORTS
  };

/* Under glibc, versionsort is the preferred sorting function for
   filenames, but alphasort is more common. */
#if __GLIBC__ > 1 && __GLIBC_MINOR__ > 0
#define sort_by_name versionsort
#else
#define sort_by_name alphasort
#endif

/* Sorting functions to corresponding to sorting types. */
static int sort_by_size (const void *, const void *);
static int sort_by_date (const void *, const void *);
static int (*sort_functions[FM_N_SORTS])(const void *, const void *) =
  {
    sort_by_name,
    sort_by_size,
    sort_by_date,
  };

/* Describes a file permission bit. */
struct permission
  {
    const char name[3];		/* Two-character name for the bit. */
    int value;			/* The bit's value. */
  };

/* All the typical permission bits. */
#define N_PERMS 12
static const struct permission perms[N_PERMS] = 
  {
    {"ur", S_IRUSR},
    {"uw", S_IWUSR},
    {"ux", S_IXUSR},
    {"gr", S_IRGRP},
    {"gw", S_IWGRP},
    {"gx", S_IXGRP},
    {"or", S_IROTH},
    {"ow", S_IWOTH},
    {"ox", S_IXOTH},
    {"su", S_ISUID},
    {"sg", S_ISGID},
    {"st", S_ISVTX},
  };

static void show_file (struct request *, const char *filename);

typedef void action_handler (struct request *);
static action_handler handle_move, handle_rename, handle_mkdir, handle_rmdir;
static action_handler handle_copy, handle_delete, handle_upload, handle_chmod;
static int handle_download (struct request *);

static void javascript_header (void);

static void do_show_reg (struct request *rq, const char *filename,
			 struct stat *st, op_func *);
static void do_show_dir (struct request *rq, const char *filename,
			 struct stat *st, op_func *);

static void show_name (struct request *, const char *filename, int is_dir);
static void show_perms (mode_t mode, mode_t r, mode_t w, mode_t x,
			const char *label);
static op_func *show_toolbar (struct request *,
			      const char *filename, const struct stat *st);

static void show_dir (struct request *rq,
		      const char *dir, const char *top_dir, int nest,
		      const struct std *, const char *start,
		      int (*sort_func) (const void *, const void *),
		      int *line);

static const char *summarize_symbolic_perms (struct request *,
					     mode_t, uid_t, gid_t);
static const char *pick_file_icon (const char *filename);
static void indent (int depth);

static int show_details;

/* File manager main function. */
void
scr_files (struct request *rq)
{
  const char *file_src;
  struct st file_dest;
  
  /* Downloads handle the HTTP and HTML headers on their own. */
  if (handle_download (rq))
    return;

  http_header ();
  html_banner (rq);

  javascript_header ();

  /* Perform all the possible file operations. */
  handle_move (rq);
  handle_rename (rq);
  handle_mkdir (rq);
  handle_rmdir (rq);
  handle_copy (rq);
  handle_delete (rq);
  handle_upload (rq);
  handle_chmod (rq);

  /* Figure out what file or directory we're displaying. */
  file_src = std_get_null (rq->cgi, "file");
  if (file_src == NULL)
    {
      file_src = rq->cx.shared;
      if (file_src == NULL)
	fail ("null shared directory");
      std_put (rq->cgi, "file", rq->cx.shared);
    }

  /* Copy file_src to file_dest.  Collapse all multiple slashes into
     single slashes, remove trailing slashes, and ensure that the
     filename begins with a slash.  We could also collapse .. and
     . entries. */
  st_init (&file_dest, 64);
  st_putchar (&file_dest, '/');
  if (*file_src == '/')
    file_src++;
  for (; *file_src; file_src++)
    {
      if (*file_src == '/'
	  && (file_src[1] == '/' || file_src[1] == '\0'))
	continue;
      st_putchar (&file_dest, *file_src);
    }

  /* Display the page. */
  show_file (rq, st_value (&file_dest));

  st_destroy (&file_dest);

  html_trailer (rq);
}

/* Private functions called from public functions. */

/* Display information about file FILENAME. */
static void
show_file (struct request *rq, const char *filename)
{
  /* Get file info. */
  struct stat st;

  /* Handler function. */
  op_func *func;

  if (-1 == stat (filename, &st))
    {
      html ("<H2 ALIGN=CENTER>%s</H2>\n"
	    "<P ALIGN=CENTER>\n"
	    "%s\n",
	    _("Error accessing file"),
	    _("The file you requested,"));
      html_escape (filename);
      html ("%s: %s",
	    _(", could not be accessed.  The error returned by the\n"
	      "the system was this"),
	    strerror (errno));
      return;
    }

  /* Handle sort order change. */
  if (std_contains (rq->cgi, "sort"))
    std_replace (rq->untrusted, "file-sort", std_get (rq->cgi, "sort"));

  func = show_toolbar (rq, filename, &st);

  /* Depending on file type, pass it on to an appropriate function.
     Note that more handlers could/should be added for symlinks,
     pipes, and so on. */
  if (S_ISDIR (st.st_mode))
    do_show_dir (rq, filename, &st, func);
  else 
    do_show_reg (rq, filename, &st, func);
}

/* Functions for showing different types of files. */

/* Display information about a regular file. */
static void
do_show_reg (struct request *rq, const char *filename, struct stat *st,
	     op_func *func)
{
  /* Show filename. */
  htmls ("<TABLE WIDTH=\"100%\" BGCOLOR=\"#CCCC99\">\n"
	 "  <TR VALIGN=TOP>\n"
	 "    <TD>\n\n");
  show_name (rq, filename, 0);

  html ("\n"
	"      <TABLE WIDTH=\"100%%\" BORDER=1 BGCOLOR=\"#FFFFFF\">\n"
	"        <TR><TH>%s</TH><TH>%s</TH>\n",
	_("Attribute"),
	_("Value"));

  /* Display full name of user who owns the file. */
  {
    struct passwd *pw;
    char name[64];

    pw = getpwuid (st->st_uid);
    if (pw)
      {
	char *p;
	strncpy (name, pw->pw_gecos, 64);
	name[63] = '\0';
	p = strchr (name, ',');
	if (p)
	  *p = '\0';
      }
    else
      sprintf (name, "%d", st->st_uid);

    html ("        <TR><TD>%s:</TD><TD>%s</TD>\n", _("User"), name);
  }

  /* Display file size in `123.45 kB' format. */
  html ("        <TR><TD>%s:</TD><TD>", _("Size"));
  if (st->st_size >= 1024 * 1024 * 1024)
    html ("%.2f %s", st->st_size / 1024. / 1024. / 1024., _("GB"));
  else if (st->st_size >= 1024 * 1024)
    html ("%.2f %s", st->st_size / 1024. / 1024., _("MB"));
  else if (st->st_size >= 1024)
    html ("%.2f %s", st->st_size / 1024., _("kB"));
  else
    html ("%d %s", (int) st->st_size, _("bytes"));
  htmls ("</TD></TR>\n");
  
  /* Show file times. */
  html (_("        <TR><TD>%s:</TD><TD>%.1f %s</TD>\n"
	  "        <TR><TD>%s:</TD><TD>%.1f %s</TD>\n"),
	_("Last access"),
	difftime (cur_time, st->st_atime) / 86400.,
	_("days ago"),
	_("Last change"),
	difftime (cur_time, st->st_mtime) / 86400.,
	_("days ago"));

  /* Show file permissions in `friendly' manner. */
  show_perms (st->st_mode, S_IRUSR, S_IWUSR, S_IXUSR, _("User privileges:"));
  show_perms (st->st_mode, S_IRGRP, S_IWGRP, S_IXGRP, _("Team privileges:"));
  show_perms (st->st_mode, S_IROTH, S_IWOTH, S_IXOTH,
	      _("Outsider privileges:"));

  htmls ("      </TABLE>\n"
	 "    </TD>\n"
	 "    <TD>\n");

  func (rq, filename, st);

  htmls ("    </TD>\n"
	 "  </TR>\n"
	 "</TABLE>\n");
}

/* Display information about a directory. */
static void 
do_show_dir (struct request *rq, const char *filename, struct stat *st,
	     op_func *func)
{
  int sort_order;
  
  {
    const char *sort_string = std_get (rq->untrusted, "file-sort");

    if (!strcmp (sort_string, "size"))
      sort_order = FM_SORT_SIZE;
    else if (!strcmp (sort_string, "date"))
      sort_order = FM_SORT_DATE;
    else
      sort_order = FM_SORT_NAME;
  }

  /* Display directory name. */
  htmls ("<TABLE BGCOLOR=\"#CCCC99\">\n"
	 "  <TR VALIGN=TOP>\n"
	 "    <TD>\n");
  show_name (rq, filename, 1);

  /* Handle details request. */
  show_details = stoi (std_get (rq->untrusted, "details"));

  {
    const char *details = std_get_null (rq->cgi, "details");

    if (details != NULL)
      {
	show_details = !strcmp (details, "yes");
	std_replace (rq->cgi, "details", show_details ? "1" : "0");
      }
  }

  /* Show/hide details link. */
  {
    struct url url;

    url_create (&url, "files");
    url_arg (&url, "file", filename);
    if (show_details)
      url_arg (&url, "details", "no");
    else
      url_arg (&url, "details", "yes");
    html ("      <DIV ALIGN=CENTER><A HREF=\"%s\">%s</A></DIV>\n\n",
	  url_string (&url),
	  show_details ? _("Fewer Details...") : _("More Details..."));
    url_destroy (&url);
  }

  html ("\n"
	"      <TABLE WIDTH=\"100%%\" BORDER=1>\n"
	"        <TR VALIGN=BOTTOM>\n");
  
  {
    static const char *columns[] =
      {N_("Name"), N_("Size"), N_("Date"), N_("User"),
       N_("Group"), N_("Permissions")};
    static const char *sort_name[FM_N_SORTS] =
      {"name", "size", "date"};

    const int n_columns = sizeof columns / sizeof *columns;
    int i;

    for (i = 0; i < n_columns; i++)
      {
	if (!show_details && i != 0 && i != 2)
	  continue;

	htmls ("          <TH>");
	if (i == 0)
	  html ("<IMG SRC=\"%s/f/f.gif\" ALT=\"&nbsp;\">", www_root);

	if (i < FM_N_SORTS && i != sort_order)
	  {
	    struct url sort_url;

	    url_create (&sort_url, "files");
	    url_copy_abbrev (&sort_url, rq);
	    url_arg (&sort_url, "file", filename);
	    url_arg (&sort_url, "sort", sort_name[i]);

	    html ("<A HREF=\"%s\">%s</A>",
		  url_string (&sort_url), gettext (columns[i]));

	    url_destroy (&sort_url);
	  }
	else
	  htmls (gettext (columns[i]));

	htmls ("</TH>\n");
      }
  }

  htmls ("        </TR>\n\n");
  
  /* Update the list of directories than should be shown rather than
     summarized, and display the directory's contents. */
  {
    struct std *expand_dirs;
    const char *start;
    int line;
    
    expand_dirs = std_get_subdict (rq->untrusted, "expand-dirs");

    start = std_get_null (rq->cgi, "expand");
    if (start != NULL)
      std_put (expand_dirs, start, NULL);
    else
      {
	start = std_get_null (rq->cgi, "collapse");
	if (start != NULL)
	  std_delete (expand_dirs, start);
      }

    if (std_modified (expand_dirs))
      std_put_subdict (rq->untrusted, "expand-dirs", expand_dirs);

    line = 0;
    show_dir (rq, filename, filename, 0, expand_dirs, start,
	      sort_functions[sort_order], &line);
  }

  htmls ("      </TABLE>\n\n"
	 "    </TD>\n"
	 "    <TD>\n");

  func (rq, filename, st);

  htmls ("    </TD>\n"
	 "  </TR>\n"
	 "</TABLE>\n\n");
}

/* Functions for showing files. */

static int
one (const struct dirent *foo unused)
{
  return 1;
}
     
/* Displays directory DIRNAME as a subdirectory of directory
   TOP_DIRNAME, under which it is nested NEST levels.  In turn,
   subdirectories of DIRNAME should be displayed in full if their
   names are present in EXPAND_DIRS. */
static void
show_dir (struct request *rq,
	  const char *dirname, const char *top_dirname, int nest,
	  const struct std *expand_dirs, const char *start,
	  int (*sort_func) (const void *, const void *),
	  int *line)
{
  /* Enough &nbsp; spaces to indent files in this directory. */
  struct st prefix;

  /* URL for files and directories under this one. */
  struct url file_url;

  /* URL for expanding and collapsing directory contents. */
  struct url dir_url;

  /* Used to store the full filename of a file in this directory. */
  struct st filename;

  /* Length of DIRNAME for truncation of `filename' above. */
  size_t dirname_len;

  /* Number of files actually displayed (doesn't include dotfiles). */
  int count_displayed_files;

  /* Directory entries for files in this directory. */
  struct dirent **files;
  int n_files;

  /* Instead of using the full filename including path to each file,
     which can be slow, chdir to the proper directory and use the
     short filenames. */
  if (-1 == chdir (dirname))
    {
      html ("<TR><TD COLSPAN=%d>%s ", 2 + 4 * show_details,
	    _("Can't chdir to"));
      html_escape (dirname);
      html (": %s</TD></TR>\n", strerror (errno));
      return;
    }

  /* Read files. */
  n_files = scandir (".", &files, one, sort_func);
  if (n_files == -1)
    {
      html ("<TR><TD COLSPAN=%d>Error reading directory ",
	     2 + 4 * show_details);
      html_escape (dirname);
      html (": %s</TD></TR>\n", strerror (errno));

      return;
    }

  /* Construct prefix. */
  {
    int i;
    
    st_init (&prefix, nest * 4 * 6);
    for (i = 0; i < nest * 4; i++)
      st_concat (&prefix, "&nbsp;");
  }

  /* Initialize variables. */
  url_create (&file_url, "files");
  url_copy_abbrev (&file_url, rq);
  url_mark (&file_url);

  st_create (&filename, dirname);
  st_putchar (&filename, '/');
  dirname_len = strlen (dirname) + 1;

  url_create (&dir_url, "files");
  url_copy_abbrev (&dir_url, rq);
  url_arg (&dir_url, "file", top_dirname);
  url_mark (&dir_url);

  count_displayed_files = 0;

  /* Display all the files. */
  {
    int i;
    
    for (i = 0; i < n_files; i++)
      {
	struct stat s;
	struct dirent *const entry = files[i];
	int expand_subdir;

	/* Skip dotfiles. */
	if (entry->d_name[0] == '.' && !show_details)
	  continue;

	count_displayed_files++;
	
	/* Alternate rows are in alternate shades of gray, in order to
           screw with your eyes. */
	{
	  const char *bg_color = *line ? "#F0F0F0" : "#C0C0C0";
	  *line ^= 2;
	  html ("        <TR BGCOLOR=\"%s\" VALIGN=BOTTOM>\n", bg_color);
	}
	
	if (-1 == lstat (entry->d_name, &s))
	  {
	    html ("<TR><TD COLSPAN=%d>%s ",
		  2 + 4 * show_details, _("Can't stat"));
	    html_escape (entry->d_name);
	    html (": %s</TD></TR>\n", strerror (errno));
	    continue;
	  }

	/* Append the name of this file to filename and use it to
           construct file_url. */
	st_truncate (&filename, dirname_len);
	st_concat (&filename, entry->d_name);
	
	url_release (&file_url);
	url_arg (&file_url, "file", st_value (&filename));

	/* Tell human readers of HTML what file this is. */
	htmls ("        <!-- ");
	html_escape (st_value (&filename));
	htmls (" -->\n");

	htmls ("          <TD NOWRAP>\n");
	if (st_length (&prefix))
	  html ("            %s\n", st_value (&prefix));

	/* Display file or directory name. */
	if (S_ISDIR (s.st_mode))
	  {
	    int gif;
	    const char *alt;

	    expand_subdir = std_contains (expand_dirs, st_value (&filename));
	    
	    url_release (&dir_url);
	    if (expand_subdir)
	      {
		url_arg (&dir_url, "collapse", st_value (&filename));
		gif = 'o';
		alt = "V";
	      }
	    else
	      {
		url_arg (&dir_url, "expand", st_value (&filename));
		gif = 'c';
		alt = "&gt;";
	      }
	    
	    html ("            <A HREF=\"%s#start\">\n"
		  "              <IMG BORDER=0 SRC=\"%s/f/%c.gif\" "
		  "ALT=\"%s\">\n"
		  "            </A>\n"
		  "            <IMG BORDER=0 SRC=\"/icons/folder.gif\" "
		  "ALT=\"\">\n",
		  url_string (&dir_url), www_root, gif, alt);
	  }
	else
	  {
	    expand_subdir = 0;
	    html ("            <IMG BORDER=0 SRC=\"%s/f/f.gif\" "
		  "ALT=\"&nbsp;\">\n"
		  "            <IMG BORDER=0 SRC=\"%s\" ALT=\"\">\n",
		  www_root, pick_file_icon (entry->d_name));
	  }
	
	html ("            <A HREF=\"%s\"%s>", url_string (&file_url),
	      (start != NULL && strcmp (st_value (&filename), start)
	       ? "" : " NAME=start"));
	html_escape (entry->d_name);
	htmls ("</A>");
	if (S_ISDIR (s.st_mode))
	  htmlc ('/');
	else if (S_ISLNK (s.st_mode))
	  {
	    char buf[129];
	    int count;

	    /* Display the name of the file pointed to by the link. */
	    html ("\n"
		  "            <IMG BORDER=0 SRC=\"%s/f/rarrow.gif\" "
		  "ALT=\"->\">\n",
		  www_root);
	    count = readlink (st_value (&filename), buf, 128);
	    if (count > 0 && count <= 128)
	      {
		buf[count] = '\0';
		
		html ("            <A HREF=\"%s\">", url_string (&file_url));
		html_escape (buf);
		htmls ("</A>\n");
	      }
	    else
	      html ("            (%s: %s)", _("couldn't read"),
		    strerror (errno));

	    /* Re-stat the file to get the real info, not the link
               info. */
	    if (-1 == stat (entry->d_name, &s))
	      {
		html ("</TD><TD COLSPAN=%d>(%s: %s)</TD></TR>\n",
		      1 + 4 * show_details,
		      _("error reading link"), strerror (errno));
		continue;
	      }
	  }

	htmls ("\n"
	       "          </TD>\n");

	/* Display size. */
	if (show_details)
	  {
	    if (S_ISDIR (s.st_mode))
	      htmls ("          <TD>&nbsp;</TD>\n");
	    else
	      html ("          <TD ALIGN=RIGHT>%.1f %s</TD>\n",
		    s.st_size / 1024., _("kB"));
	  }

	/* Display modification date. */
	{
	  struct tm *tm;

	  tm = localtime (&s.st_mtime);
	  if (tm->tm_year != cur_tm.tm_year)
	    html ("          <TD>%s</TD>\n",
		  datetime_format ("%b %d %H:%M %Y", tm));
	  else
	    html ("          <TD>%s</TD>\n",
		  datetime_format ("%b %d %H:%M", tm));
	}

	/* Display owner, group, permissions. */
	if (show_details)
	  {
	    html ("          <TD>%s</TD>\n"
		  "          <TD>%s</TD>\n"
		  "          <TD>%s</TD>\n",
		  uid_to_username (s.st_uid), gid_to_groupname (s.st_gid),
		  summarize_symbolic_perms (rq, s.st_mode, s.st_uid,
					    s.st_gid));
	  }
	
	htmls ("        </TR>\n\n");

	/* If it's a subdirectory that should be expanded, do so. */
	if (expand_subdir)
	  {
	    show_dir (rq, st_value (&filename), top_dirname,
		      nest + 1, expand_dirs, start, sort_func, line);
	    if (-1 == chdir (dirname))
	      {
		log ("can't chdir back to %s: %s", dirname, strerror (errno));
		break;
	      }
	  }
      }
  }

  if (count_displayed_files == 0)
    html ("          <TR><TD COLSPAN=%d>%s(%s)</TD></TR>\n",
	  2 + 4 * show_details,
	  st_value (&prefix), _("folder is empty"));

  free (files);
  
  url_destroy (&file_url);
  url_destroy (&dir_url);
  st_destroy (&prefix);
}

/* Break FILENAME into its components.  FILENAME must begin with a
   slash, it must not end with a slash, and it must not contain any
   doubled slashes.  

   Each call puts the next component of FILENAME into LABEL, and the
   entire accumulated filename so far into FULL.  Besides an actual
   filename component, LABEL may also be a user-friendly label such as
   "shared folder".

   On the first call, *FP should be NULL, and LABEL and FULL should
   not have already been initialized.  The value of *FP must be
   preserved between calls.

   Returns nonzero only as long as there are components left.  After
   zero is returned, a final call should be in order to discard
   storage. */
static int
decompose_filename (struct context *cx,
		    const char *filename, const char **fp,
		    struct st *label, struct st *full)
{
  if (!*fp)
    {
      *fp = filename;
      
      if (has_prefix (filename, cx->shared))
	{
	  *fp += strlen (cx->shared);
	  st_create (label, _("shared folder"));
	  st_create (full, cx->shared);
	}
      else if (has_prefix (filename, cx->home))
	{
	  *fp += strlen (cx->home);
	  st_create (label, _("personal folder"));
	  st_create (full, cx->home);
	}
      else
	{
	  st_create (label, _("root"));
	  st_init (full, 64);
	}
    }
  else
    {
      if (**fp == '\0')
	{
	  st_destroy (label);
	  st_destroy (full);
	  return 0;
	}

      /* Skip leading slash. */
      assert (**fp == '/');
      st_putchar (full, '/');
      (*fp)++;
      
      st_clear (label);
      for (; ; (*fp)++)
	{
	  if (**fp == '/' || **fp == '\0')
	    break;
	  
	  st_putchar (label, **fp);
	  st_putchar (full, **fp);
	}
    }

  return (**fp != '\0');
}

/* Display FILENAME, which is the name of a file if IS_DIR is zero or
   the name of a directory otherwise. */
static void
show_name (struct request *rq, const char *filename, int is_dir)
{
  /* URL to the accumulated filename. */
  struct url url;

  /* Label for this filename component. */
  struct st label;

  /* Full accumulated filename. */
  struct st full;

  /* Used by decompose_filename. */
  const char *fp;

  /* Number of components so far. */
  int depth;

  url_create (&url, "files");
  url_copy_abbrev (&url, rq);
  url_mark (&url);

  html ("      <FONT SIZE=\"+1\">%s:</FONT>\n",
	(is_dir
	 ? _("Viewing contents of folder")
	 : _("Viewing attributes of file")));

  /* We don't want a table for any a priori reason.  Instead, we just
     want to keep the lines from wrapping.  Oh well... */
  htmls ("      <TABLE WIDTH=\"100%\" CELLPADDING=0 "
	 "CELLSPACING=0>\n"
	 "        <TR><TD NOWRAP>\n");

  fp = NULL;
  for (depth = 0; ; depth++)
    {
      int more = decompose_filename (&rq->cx, filename, &fp, &label, &full);

      htmls ("          ");
      indent (depth);
      if (depth)
	html ("<IMG SRC=\"%s/f/drarrow.gif\" ALT=\"\">", www_root);
      if (more)
	{
	  url_release (&url);
	  url_arg (&url, "file", st_value (&full));
	  html ("          <A HREF=\"%s\">", url_string (&url));
	}
      htmls ("<STRONG>");
      html_escape (st_value (&label));
      htmls ("</STRONG>");
      if (more)
	htmls ("</A>");
      htmls ("<BR>\n");

      if (!more)
	{
	  decompose_filename (&rq->cx, filename, &fp, &label, &full);
	  break;
	}
    }
  htmls ("        </TD></TR>\n"
	 "      </TABLE>\n");

  url_destroy (&url);
}

/* Miscellaneous utilities. */

/* Display as text the permissions offered by MODE given read bit R,
   write bit W, and execute bit X.  Label the permissions with
   LABEL. */
static void 
show_perms (mode_t mode, mode_t r, mode_t w, mode_t x, const char *label)
{
  html ("        <TR><TD>%s</TD><TD>", label);
  if ((mode & (r | w)) == (r | w))
    htmls (_("read/write"));
  else if (mode & r)
    htmls (_("read-only"));
  else if (mode & w)
    htmls (_("write-only"));

  if (mode & x)
    {
      if (mode & (r | w))
	htmls (" + execute");
      else
	htmls ("execute-only");
    }

  if ((mode & (r | w | x)) == 0)
    htmls ("none");

  htmls ("</TD></TR>\n");
}

/* Given file permissions MODE, owner FILE_UID, and group FILE_GID,
   display the permissions we are allowed for our uid and primary and
   supplementary gids. */
static const char *
summarize_symbolic_perms (struct request *rq,
			  mode_t mode, uid_t file_uid, gid_t file_gid)
{
  static const char *mode_names[] = 
    {
      N_("none"), N_("execute"), N_("write-only"), N_("write + execute"),
      N_("read-only"), N_("read + execute"), N_("read/write"),
      N_("read/write + execute"),
      N_("none"), N_("none"), N_("write-only"), N_("write-only"),
      N_("none"), N_("read-only"), N_("write-only"), N_("read/write"),
    };
  
  mode_t r, w, x;
  int index;
  
  if (rq->cx.uid == file_uid)
    r = S_IRUSR, w = S_IWUSR, x = S_IXUSR;
  else
    {
      int g = 0;

      if (rq->cx.gid == file_gid)
	g = 1;
      else
	{
	  int n_gids;
	  const char **gids;

	  int i;

	  gids = std_get_many (rq->cgi, "supp-gids", &n_gids);
	  for (i = 0; i < n_gids; i++)
	    if (strtoul (gids[i], NULL, 10) == file_gid)
	      {
		g = 1;
		break;
	      }
	}
      
      if (g)
	r = S_IRGRP, w = S_IWGRP, x = S_IXGRP;
      else
	r = S_IROTH, w = S_IWOTH, x = S_IXOTH;
    }
  
  index = 0;
  if (S_ISDIR (mode))
    index += 010;
  if (mode & r)
    index += 004;
  if (mode & w)
    index += 002;
  if (mode & x)
    index += 001;

  return gettext (mode_names[index]);
}

/* Picks an icon for the file named FILENAME and returns its name. */
/* FIXME: save cache into rq->untrusted. */
static const char *
pick_file_icon (const char *filename)
{
  static struct std *file_icons;

  if (file_icons == NULL)
    {
      FILE *f;

      file_icons = std_create (0);

      f = fopen (ICON_FILE, "r");
      if (f != NULL)
	{
	  char *buf = NULL;
	  size_t buf_size = 0;

	  for (;;)
	    {
	      char *first;
	      char *icon;

	      if (-1 == getline (&buf, &buf_size, f))
		{
		  if (ferror (f))
		    log ("error reading " ICON_FILE ": %s", strerror (errno));
		  break;
		}

	      first = strtok (buf, " \t");
	      if (first == NULL || strcmp (first, "AddIcon"))
		continue;

	      icon = strtok (NULL, " \t");
	      if (icon == NULL)
		continue;

	      for (;;)
		{
		  char *ext = strtok (NULL, " \t");
		  if (ext == NULL)
		    break;

		  std_put (file_icons, ext, icon);
		}
	    }

	  free (buf);

	  if (-1 == fclose (f))
	    log ("error closing " ICON_FILE ": %s", strerror (errno));
	}
    }
  
  {
    const char *cp = strrchr (filename, '.');
    const char *icon = "/icons/unknown.gif";
    
    if (cp)
      {
	const char *s = std_get_null (file_icons, cp);
	if (s != NULL)
	  icon = s;
      }

    return icon;
  }
}

/* Displays a menu of operations for file FILENAME.  The operations
   are supplied as label/description string pairs, and terminated by a
   NULL character pointer. */
static op_func *
show_toolbar (struct request *rq, const char *filename, const struct stat *st)
{
  /* Display state for each operation, one of OPD_*. */
  char state[OP_COUNT];

  /* URL to the specified operation. */
  struct url file_url;

  /* FILENAME quoted for HTML consumption. */
  struct st file_quoted;

  /* URL to a download link for the file. */
  struct url download_url;

  /* User-selected operation. */
  int selected;

  /* Current operation. */
  int i;

  /* Display the toolbar. */
  html ("<TABLE ALIGN=CENTER>\n"
	"  <TR VALIGN=TOP>\n\n"
	"    <TD>\n"
	"      %s:<BR>\n"
	"      ",
	_("Folders"));
  
  if (strcmp (filename, rq->cx.shared))
    html ("<A HREF=\"%s\"><FONT SIZE=\"-1\">%s</FONT></A>",
	  url_this ("files", (char *) NULL),
	  _("shared"));
  else
    html ("<FONT SIZE=\"-1\">%s</FONT>", _("shared"));
  htmls ("<BR>\n"
	 "      ");
  if (strcmp (filename, rq->cx.home))
    html ("<A HREF=\"%s\"><FONT SIZE=\"-1\">%s</FONT></A>",
	  url_this ("files", "file", rq->cx.home, (char *) NULL),
	  _("personal"));
  else
    html ("<FONT SIZE=\"-1\">%s</FONT>", _("personal"));
  htmls ("<P>\n"
	 "    </TD>\n\n");

  url_create (&file_url, "files");
  url_copy_abbrev (&file_url, rq);
  url_arg (&file_url, "file", filename);
  url_mark (&file_url);

  {
    const char *filebasename = strrchr (filename, '/');
    if (filebasename)
      filebasename++;
    else
      filebasename = filename;
    url_create (&download_url, filebasename);
  }
  url_arg (&download_url, "screen", "files");
  url_arg (&download_url, "file", filename);
  url_mark (&download_url);

  st_init (&file_quoted, 64);
  st_concat_html_quoted (&file_quoted, filename);

  memset (state, OPD_NORMAL, sizeof state);

  if (S_ISDIR (st->st_mode))
    {
      state[OP_ACTIVITY] = OPD_OMIT;
      state[OP_VIEW] = OPD_OMIT;
      state[OP_REPLACE] = OPD_OMIT;
      state[OP_COPY] = OPD_OMIT;
      state[OP_DELETE] = OPD_OMIT;

      if (!strcmp (filename, rq->cx.shared) || !strcmp (filename, rq->cx.home))
	{
	  state[OP_MOVE] = OPD_DISABLE;
	  state[OP_RENAME] = OPD_DISABLE;
	  state[OP_RMDIR] = OPD_DISABLE;
	}
    }
  else
    {
      state[OP_UPLOAD] = OPD_OMIT;
      state[OP_MKDIR] = OPD_OMIT;
      state[OP_RMDIR] = OPD_OMIT;
    }

  selected = OP_COMMENT;
  for (i = 0; i < OP_COUNT; i++)
    if (!strcmp (op_table[i].name, rq->op))
      selected = i;

  for (i = 0; i < OP_COUNT; i++)
    {
      const char *name;
      const char *label;
      const char *caption;
      const char *url;
      int this_state;

      this_state = state[i];
      if (this_state == OPD_OMIT)
	continue;
      if (i == selected && this_state == OPD_NORMAL)
	this_state = OPD_SELECTED;

      name = op_table[i].name;
      label = op_table[i].label;
      caption = op_table[i].caption;

      if (!strcmp (name, "download"))
	{
	  url_release (&download_url);
	  url_arg (&download_url, "op", "download");
	  url = url_string (&download_url);
	}
      else if (!strcmp (name, "view"))
	{
	  url_release (&download_url);
	  url_arg (&download_url, "op", "view");
	  url = url_string (&download_url);
	}
      else
	{
	  url_release (&file_url);
	  url_arg (&file_url, "op",
		   strcmp (name, "replace") ? name : "upload");
	  url = url_string (&file_url);
	}
	      
      htmls ("    <TD WIDTH=48 ALIGN=CENTER>\n");

      if (this_state != OPD_DISABLE)
	html ("      <A HREF=\"%s\"\n"
	      "       ONMOUSEOVER=\"t(i_%s,'%s','%c');\"\n"
	      "       ONMOUSEOUT=\"t(i_%s,'%s','%c');\">\n",
	      url,
	      name, name, OPD_MOUSEOVER,
	      name, name, this_state);
      
      html ("        <IMG WIDTH=48 HEIGHT=34 BORDER=0 NAME=i_%s\n"
	    "         SRC=\"%s/f/%c/%s.gif\"\n"
	    "         ALT=\"[",
	    name,
	    www_root, this_state, name);
      html (gettext (caption), st_value (&file_quoted));
      html ("]\"><BR>\n"
	    "        <FONT SIZE=\"-1\">%s</FONT><BR>",
	    gettext (label));
	    
      if (this_state != OPD_DISABLE)
	htmls ("</A>");

      htmls ("</TD>\n\n");
    }

  htmls ("  </TR>\n"
	 "</TABLE>\n\n");

  url_destroy (&download_url);
  url_destroy (&file_url);
  st_destroy (&file_quoted);

  return op_table[selected].func;
}

/* Indent DEPTH levels. */
static void
indent (int depth)
{
  int i;
  
  for (i = 0; i < depth; i++)
    html ("<IMG SRC=\"%s/f/f.gif\" ALT=\"&nbsp;\">", www_root);
}

/* Allocate a new string equivalent to DIRNAME with the last filename
   component replaced by FILENAME.  DIRNAME must not end with a
   slash. */
static char *
replace_filename (const char *dirname, const char *filename)
{
  char *dst;
  
  const char *cp = strrchr (dirname, '/');
  if (cp == NULL)
    fail ("bad value for dirname: %s", dirname);

  dst = xmalloc (strlen (dirname) + strlen (filename) + 8);
  memcpy (dst, dirname, cp - dirname + 1);
  strcpy (&dst[cp - dirname + 1], filename);

  return dst;
}

/* Perform a file move operation if necessary.  Returns nonzero only
   if successful. */
/* FIXME: doesn't work for files not on the same disk volume. */
static int
do_move (const char *src, const char *dst)
{
  /* For an ordinary file, we want to make a link to the new name and
     unlink the old one.  This is superior to rename() since it won't
     overwrite an existing file.  Some systems won't link() a
     directory, so if EPERM, the appropriate error for that condition,
     is returned, then try a standard rename() as well. */
  if ((-1 == link (src, dst) || -1 == unlink (src))
      && (errno != EPERM || -1 == rename (src, dst)))
    {
      html ("<H2>%s</H2>\n"
	    "%s\n",
	    _("Error moving file"),
	    _("An error was encountered while attempting to move"));
      html_escape (src);
      htmls (_(" to "));
      html_escape (dst);
      html (": %s\n", strerror (errno));

      return 0;
    }
  else
    return 1;
}

/* Copy the contents of the file with handle S to the file with handle
   D.  Returns nonzero only if successful. */
static int
copy_fd (int s, int d)
{
  for (;;)
    {
      ssize_t nbytes;
      char buf[4096];

      nbytes = read (s, buf, 4096);
      if (nbytes == -1
	  || nbytes != write (d, buf, nbytes))
	return 0;

      if (nbytes != 4096)
	return 1;
    }
}

/* Copies the file named SRC to a new file named DST.  Returns nonzero
   only if successful. */
static int
do_copy (const char *src, const char *dst)
{
  int s, d;

  s = open (src, O_RDONLY);
  if (s == -1)
    goto lossage;
  
  {
    
    struct stat st;
    if (-1 == fstat (s, &st))
      goto lossage;
    
    {
      mode_t old_mask = umask (0);
      d = open (dst, O_WRONLY | O_CREAT | O_EXCL, st.st_mode);
      umask (old_mask);
    }
    
    if (s == -1 || d == -1)
      goto lossage;
  }
  
  if (!copy_fd (s, d) || -1 == close (s) || -1 == close (d))
    goto lossage;
  return 1;

 lossage:
  html ("<H2>%s</H2>\n"
	"%s\n",
	_("Error copying file"),
	_("An error was encountered while attempting to copy"));
  html_escape (src);
  htmls (_(" to "));
  html_escape (dst);
  html (": %s\n", strerror (errno));
  return 0;
}

/* Allocates and returns a new string containing all but the last
   component of FILENAME, i.e., the directory name.  FILENAME must
   contain at least one slash, and it must not be at the end of the
   filename.  It must not contain doubled slashes. */
char *
compute_dirname (const char *filename)
{
  const char *cp;
  char *dir;

  cp = strrchr (filename, '/');
  assert (cp != NULL);

  dir = xmalloc (cp - filename + 1);
  memcpy (dir, filename, cp - filename);
  dir[cp - filename] = '\0';

  return dir;
}

/* Calculate permissions to use in creating a file named FILENAME. */
int
calc_mode (struct context *cx, const char *filename)
{
  if (strstr (filename, "www")
      || has_prefix (filename, cx->web))
    /* 06775 */
    return S_ISUID | S_ISGID | S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH; 
  else
    /* 06770 */
    return S_ISUID | S_ISGID | S_IRWXU | S_IRWXG;
}

/* File operation handlers. */

/* Display comments on this file or directory. */
static void
op_comment (struct request *rq, const char *filename unused,
	    const struct stat *st unused)
{
  rq->abbrev = 0;
  scr_messages (rq);
}

/* Display activity for this file, passing off most of the work to
   activity.c. */
static void
op_activity (struct request *rq, const char *filename,
	     const struct stat *st unused)
{
  struct activity_query query;
  struct url url;

  activity_query_init (&query);
  query.search = 1;
  query.filename = xstrdup (filename);
  activity_query_run (&rq->cx, &query);

  url_create (&url, "files");
  url_copy_abbrev (&url, rq);
  url_arg (&url, "file", filename);
  url_arg (&url, "op", "activity");

  activity_query_present (&rq->cx, &query, &url, NULL);

  url_destroy (&url);
  activity_query_destroy (&query);
}

/* Display a form for uploading files.  The current file or directory
   is FILENAME and its stat results are in ST. */
static void
op_upload (struct request *rq, const char *filename,
	   const struct stat *st unused)
{
  int i;
  
  html_start_form ("files", HT_POST, HT_FORM_DATA);
  html_hidden_abbrev (rq);
  html ("<P><FONT SIZE=\"+1\"><B>%s</B></FONT></P>\n",
	_("Upload File"));

  if (S_ISDIR (st->st_mode))
    {
      html ("<P>%s",
	    _("Use this form to upload up to 8 files at a time into the\n"
	      "current folder.  Consider using an FTP program to upload\n"
	      "larger numbers of files or files in multiple folders: this\n"
	      "folder corresponds to FTP directory "));
      html_escape (filename);
      html ("%s</P>\n",
	    _("."));
      
      html ("<TABLE>\n"
	    "  <TR>\n"
	    "    <TD></TD>\n"
	    "    <TH>%s</TH>\n"
	    "    <TH>%s</TH>\n"
	    "  </TR>\n",
	    _("Local File"),
	    _("Remote Name (optional)"));

      for (i = 0; i < 8; i++)
	html ("  <TR>\n"
	      "    <TH>%d.</TH>\n"
	      "    <TD><INPUT NAME=upload%d TYPE=file></TD>\n"
	      "    <TD><INPUT NAME=upfile%d TYPE=text SIZE=20</TD>\n"
	      "  </TR>\n",
	      i + 1, i, i);

      htmls ("</TABLE>\n");
      html_hidden ("updir", filename);
    }
  else
    {
      html ("<P>%s:</P>\n",
	    _("Select <EM>local</EM> file to upload"));
      html_file_field ("upload0");
      html_hidden ("file", filename);
    }
  
  html ("<P>%s<BR>\n",
	_("Select desired treatment for existing file(s) with same names:"));
  html_radio ("exist", "versioning",
	      _("Keep old versions of uploaded file(s), if any, as backups.\n"
		"Rename old versions by appending a semicolon\n"
		"and a version number.\n"), 1);
  htmls ("<BR>\n");
  html_radio ("exist", "overwrite",
	      _("Delete and overwrite existing file(s) with identical names.\n"),
	      0);
  htmls ("<BR>\n");
  html_radio ("exist", "error",
	      _("Flag an error if file(s) with the same names exist.\n"), 0);
  htmls ("</P>\n");

  html_submit (_("Submit"));
  html_end_form ();
}

/* Display a form for downloading a directory as an archive.  The
   current directory is FILENAME. */
static void
op_download (struct request *rq, const char *filename,
	     const struct stat *st unused)
{
  html_start_form ("files", HT_POST, HT_FORM_DATA);
  html_hidden_abbrev (rq);
  html_hidden ("file", filename);
  html ("<P><FONT SIZE=\"+1\"><B>%s</B></FONT></P>\n"
	"<P>%s</P>\n"
	"<P>%s:\n"
	"<SELECT NAME=op>\n"
	"<OPTION VALUE=\"zip\">%s</INPUT>\n"
	"<OPTION VALUE=\"tgz\">%s</INPUT>\n"
	"</SELECT>\n"
	"</P>\n"
	"<P>",
	_("Download Folder as Archive"),
	_("Select archive format"),
	_("To download the contents of this folder,\n"
	  "including the contents of all subfolders, if any,\n"
	  "select an archive format below, then click on\n"
	  "<STRONG>Download</STRONG>"),
	_("compressed ZIP archive"),
	_("gzip'd tar archive"));
  html_submit (_("Download"));
  htmls ("</P>\n");
  html_end_form ();
}

/* scandir() select function that selects only files that are
   directories. */
static int 
select_dirs (const struct dirent *e)
{
  struct stat s;
  int include;

  include = (e->d_name[0] != '.'
	     && stat (e->d_name, &s) == 0
	     && S_ISDIR (s.st_mode));

  errno = 0;
  return include;
}

static void
display_move_line (int depth, int checked,
		   const char *label, const char *full)
{
  html ("<LABEL><INPUT TYPE=radio NAME=dir VALUE=\"");
  html_escape (full);
  html ("\"%s>", checked ? " CHECKED" : "");
  indent (depth);
  if (depth)
    html ("<IMG SRC=\"%s/f/drarrow.gif\" ALT=\"\">", www_root);
  html_escape (label);
  htmls ("</LABEL><BR>\n");
}

static void
display_folder_tree (struct request *rq, const char *dir)
{
  /* Number of directories so far. */
  int depth;
    
  /* Offer the personal folder or shared folder as appropriate. */
  if (!has_prefix (dir, rq->cx.shared))
    display_move_line (0, 0, _("shared folder"), rq->cx.shared);
  
  /* Display all the directories up to the one that contains
     FILENAME. */
  {
    /* Label for the current component of FILENAME. */
    struct st label;

    /* Full accumulated filename so far. */
    struct st full;

    /* Used by decompose_filename(). */
    const char *fp;

    fp = NULL;
    for (depth = 0; ; depth++)
      {
	int more = decompose_filename (&rq->cx, dir, &fp, &label, &full);

	display_move_line (depth, !more, st_value (&label), st_value (&full));

	if (!more)
	  {
	    decompose_filename (&rq->cx, dir, &fp, &label, &full);
	    break;
	  }
      }
  }
  
  /* Display the directories directly under the one that contains
     FILENAME. */
  {
    struct st full;
    struct dirent **file_list;
    size_t len;
    int n_files;
    int i;
    
    {
      int ok;

      ok = chdir (dir) == 0;
      if (ok)
	n_files = scandir (dir, &file_list, select_dirs, sort_by_name);

      if (!ok || n_files == -1)
	{
	  html ("%s: %s<BR>", _("error reading folder"), strerror (errno));
	  goto done;
	}
    }
    
    st_create (&full, dir);
    len = st_length (&full);
    
    depth++;
    for (i = 0; i < n_files; i++)
      {
	st_truncate (&full, len);
	st_printf (&full, "/%s", file_list[i]->d_name);
	
	display_move_line (depth, 0, file_list[i]->d_name, st_value (&full));
      }

  done:
    st_destroy (&full);

    free (file_list);
  }

  if (!has_prefix (dir, rq->cx.home))
    display_move_line (0, 0, _("personal folder"), rq->cx.home);
}

/* Display a form for moving files and directories.  The current file
   or directory is FILENAME and its stat results are in ST. */
static void
op_move (struct request *rq, const char *filename,
	 const struct stat *st)
{
  /* Directory that we've currently got selected for moving the
     file/folder to, and the basename of the file/folder. */
  char *dir, *basename;

  assert (rq != NULL && filename != NULL && st != NULL);

  dir = (char *) std_get_null (rq->cgi, "dir");
  basename = (char *) std_get_null (rq->cgi, "basename");
  if (dir == NULL || basename == NULL)
    {
      dir = xstrdup (filename);
      basename = strrchr (dir, '/');
      if (basename != NULL)
	{
	  *basename = '\0';
	  basename = xstrdup (basename + 1);
	}
      else
	{
	  /* Something's really screwed up here. */
	  basename = dir;
	  dir = xstrdup ("");
	}
    }
  else
    {
      dir = xstrdup (dir);
      basename = xstrdup (basename);
    }

  html ("<FONT SIZE=\"+1\"><B>%s</B></FONT>\n",
	S_ISDIR (st->st_mode) ? _("Move Folder") : _("Move File"));
  html ("<P>%s\n"
	"%s\n"
	"</P>\n"
	"<P>\n",
	(S_ISDIR (st->st_mode)
	 ? _("Click on a folder to move this folder into.")
	 : _("Click on a folder to move this file into.")),
	_("If the folder you want is not listed, then click on a folder\n"
	  "above it in the hierarchy and click on <STRONG>Update</STRONG>\n"
	  "to update the display.\n"
	  "When the folder you want is\n"
	  "selected, click on <STRONG>Move</STRONG> to perform the move."));

  html_start_form ("files", HT_GET, HT_URL_ENCODED);
  html_hidden_abbrev (rq);
  html_hidden ("op", "move");
  html_hidden ("file", filename);
  html_hidden ("basename", basename);

  display_folder_tree (rq, dir);

  html_submit (_("Update"));
  html_submit_alt ("domove", _("Move"));

  htmls ("</P>\n");

  html_end_form ();

  free (dir);
  free (basename);
}

/* Display a form for renaming a file or directory.  The current file
   or directory is FILENAME and its stat results are in ST. */
static void
op_rename (struct request *rq, const char *filename,
	   const struct stat *st)
{
  html ("<FONT SIZE=\"+1\"><B>%s</B></FONT>\n",
	S_ISDIR (st->st_mode) ? _("Rename Folder") : _("Rename File"));

  html_start_form ("files", HT_POST, HT_URL_ENCODED);
  html_hidden_abbrev (rq);
  htmls ("<P>\n");
  html_hidden ("rnfrom", filename);
  html_label ("rnto", _("Enter a new name for this file or folder:"));
  html_text_field ("rnto", "", 20);
  htmls ("<BR>\n");
  html_submit (_("Rename"));
  html_end_form ();
}

/* Display a form for creating a new directory.  The current directory
   is FILENAME. */
static void
op_mkdir (struct request *rq, const char *filename,
	  const struct stat *st unused)
{
  html ("<FONT SIZE=\"+1\"><B>%s</B></FONT>\n", _("Create Folder"));

  html_start_form ("files", HT_POST, HT_URL_ENCODED);
  html_hidden_abbrev (rq);
  htmls ("<P>\n");
  html_hidden ("mkdirin", filename);
  html_label ("mkdiras", _("Enter a name for the folder:"));
  html_text_field ("mkdiras", "", 20);
  htmls ("<BR>\n");
  html_submit (_("Create Folder"));
  html_end_form ();
}

/* Display a form for removing the current directory.  The current
   directory is FILENAME. */
static void
op_rmdir (struct request *rq, const char *filename,
	  const struct stat *st unused)
{
  html ("<FONT SIZE=\"+1\"><B>%s</B></FONT>\n", _("Remove Folder"));

  html ("<P>%s</P>\n",
	_("If this folder is empty and you wish to remove it,\n"
	  "click on the button below.  To remove a folder that\n"
	  "has contents, you must first delete its contents."));

  html_start_form ("files", HT_POST, HT_URL_ENCODED);
  html_hidden_abbrev (rq);
  htmls ("<P>\n");
  html_hidden ("rmdir", filename);

  {
    char *dir = compute_dirname (filename);
    html_hidden ("file", dir);
    free (dir);
  }

  html_submit (_("Remove Folder"));
  html_end_form ();
}

/* Display a form for copying the current file, which is FILENAME. */
static void
op_copy (struct request *rq, const char *filename,
	 const struct stat *st unused)
{
  char *dir, *basename;

  dir = (char *) std_get_null (rq->cgi, "dir");
  basename = (char *) std_get_null (rq->cgi, "basename");
  if (dir == NULL || basename == NULL)
    {
      struct st basename_string;
      
      dir = xstrdup (filename);
      basename = strrchr (dir, '/');
      if (basename != NULL)
	*basename = '\0';
      else
	{
	  /* Something is very wrong. */
	  basename = dir;
	  dir = xstrdup ("");
	}

      st_create (&basename_string, "Copy of ");
      st_concat (&basename_string, basename + 1);
      basename = st_value (&basename_string);
    }
  else
    {
      dir = xstrdup (dir);
      basename = xstrdup (basename);
    }

  html ("<FONT SIZE=\"+1\"><B>%s</B></FONT>\n", _("Copy File"));

  html_start_form ("files", HT_POST, HT_URL_ENCODED);
  html_hidden ("op", "copy");
  html_hidden ("file", filename);
  html_hidden ("basename", basename);
  html_hidden_abbrev (rq);

  html ("<P>\n"
	"<LABEL>%s:\n"
	"<INPUT TYPE=text NAME=name VALUE=\"",
	_("Enter a name for the new copy of this file"));
  html_escape (basename);
  html ("\" SIZE=%d></LABEL>\n"
	"</P>\n"
	"<P>%s<BR>",
	strlen (basename) > 20 ? strlen (basename) : 20,
	_("By default, the new copy will be in the same directory\n"
	  "as the original copy.  If this is what you want, click\n"
	  "on <STRONG>Copy</STRONG> below.\n"
	  "Otherwise, select a folder from the list below.\n"
	  "If the folder you want is not listed, select a folder\n"
	  "above it in the hierarchy, then click\n"
	  "<STRONG>Update</STRONG>.\n"));

  html_submit_alt ("docopy", _("Copy"));
  htmls ("<P>\n");

  display_folder_tree (rq, dir);

  html_submit (_("Update"));
  htmls ("</P>\n");

  html_end_form ();
}

/* Display a form for deleting the current file, which is FILENAME. */
static void
op_delete (struct request *rq, const char *filename,
	   const struct stat *st unused)
{
  html ("<FONT SIZE=\"+1\"><B>%s</B></FONT>\n", _("Delete File"));

  html_start_form ("files", HT_POST, HT_URL_ENCODED);
  html_hidden_abbrev (rq);
  htmls ("<P>\n");
  html_hidden ("rmfile", filename);
  html_hidden ("op", "delete");

  html ("<P>%s<BR>\n"
	"<LABEL><INPUT TYPE=radio NAME=rmconfirm VALUE=no CHECKED>%s"
	"</LABEL><BR>\n"
	"<LABEL><INPUT TYPE=radio NAME=rmconfirm VALUE=YES>%s</LABEL>\n"
	"</P>\n",
	_("To delete this file, confirm the operation by selecting\n"
	  "<STRONG>Yes</STRONG> below, then click on\n"
	  "<STRONG>Delete File</STRONG> to perform the operation.\n"),
	_("<STRONG>No</STRONG>, do not delete this file.\n"),
	_("<STRONG>Yes</STRONG>, delete this file.  I realize that\n"
	  "there is no way to recover a deleted file other than from\n"
	  "any personal backups I may have made."));

  {
    char *dir = compute_dirname (filename);
    html_hidden ("file", dir);
    free (dir);
  }

  htmls ("<BR>\n");
  html_submit (_("Delete File"));
  html_end_form ();
}

/* Display form for changing file or directory permissions. */
static void
op_chmod (struct request *rq, const char *filename unused,
	  const struct stat *st)
{
  html ("<FONT SIZE=\"+1\"><B>%s</B></FONT>\n", _("Change file permissions"));

  html ("<P>%s</P>\n",
	_("Use the check boxes below to select permission bits for this\n"
	  "file or directory, then click on <STRONG>Change\n"
	  "Permissions</STRONG> to make your changes take effect.\n"
	  "Recommended for advanced users only."));
  
  html_start_form ("files", HT_POST, HT_URL_ENCODED);
  html_hidden_abbrev (rq);
  html ("<TABLE ALIGN=CENTER BORDER=1>\n"
	"  <TR>\n"
	"    <TD></TD>\n"
	"    <TH>%s</TH>\n"
	"    <TH>%s</TH>\n"
	"    <TH>%s</TH>\n"
	"  </TR>\n",
	S_ISDIR (st->st_mode) ? _("List Files") : _("Read"),
	_("Write"),
	S_ISDIR (st->st_mode) ? _("Access") : _("Execute"));

  {
    const struct permission *p = perms;
    int row;
    
    for (row = 0; row < 3; row++)
      {
	static const char *row_heads[3]
	  = {N_("Owner"), N_("Team"), N_("Outsiders")};
	int col;
      
	html ("  <TR>\n"
	      "    <TD>%s</TH>\n",
	      row_heads[row]);

	for (col = 0; col < 3; col++, p++)
	  html ("    <TD><INPUT TYPE=checkbox NAME=%s VALUE=on%s></TD>",
		p->name, st->st_mode & p->value ? " CHECKED" : "");

	htmls ("  </TR>\n");
      }

    html ("</TABLE>\n"
	  "<P>\n"
	  "  <STRONG>%s:</STRONG><BR>\n"
	  "  <LABEL><INPUT TYPE=checkbox NAME=%s VALUE=on%s>%s</LABEL><BR>\n"
	  "  <LABEL><INPUT TYPE=checkbox NAME=%s VALUE=on%s>%s</LABEL><BR>\n"
	  "  <LABEL><INPUT TYPE=checkbox NAME=%s VALUE=on%s>%s</LABEL><BR>\n"
	  "</P>\n",
	  _("Special permissions"),
	  p[0].name, st->st_mode & p[0].value ? " CHECKED" : "",
	  (S_ISDIR (st->st_mode)
	   ? _("Default file owner is folder owner")
	   : _("Set-user-id execution")),
	  p[1].name, st->st_mode & p[1].value ? " CHECKED" : "",
	  (S_ISDIR (st->st_mode)
	   ? _("Default file group is folder group")
	   : _("Set-group-id execution")),
	  p[2].name, st->st_mode & p[2].value ? " CHECKED" : "",
	  (S_ISDIR (st->st_mode)
	   ? _("Sticky file permissions")
	   : _("Save text")));
  }

  html_hidden ("file", filename);
  html_hidden ("op", "chmod");
  html_hidden ("change", "yes");

  html_submit (_("Change Permissions"));
  html_end_form ();
}

/* Actually do operations. */

/* Perform a move operation. */
static void
handle_move (struct request *rq)
{
  const char *src = std_get_null (rq->cgi, "file");
  const char *dstdir = std_get_null (rq->cgi, "dir");
  const char *basename = std_get_null (rq->cgi, "basename");
  struct st dst;

  if (src == NULL || dstdir == NULL || basename == NULL
      || !std_contains (rq->cgi, "domove"))
    return;

  st_init (&dst, 128);
  st_printf (&dst, "%s/%s", dstdir, basename);

  if (do_move (src, st_value (&dst)))
    {
      log_activity_by_name (&rq->cx, st_value (&dst), src,
			    AC_CAT_FILE, AC_FILE_MOVE, 0);
      std_replace (rq->cgi, "file", st_value (&dst));
      rq->op = "comment";
    }

  st_destroy (&dst);
}

/* Perform a rename operation. */
static void
handle_rename (struct request *rq)
{
  const char *src = std_get_null (rq->cgi, "rnfrom");
  const char *dst_suffix = std_get_null (rq->cgi, "rnto");
  char *dst;

  if (src == NULL || dst_suffix == NULL)
    return;
  
  dst = replace_filename (src, dst_suffix);

  if (do_move (src, dst))
    {
      log_activity_by_name (&rq->cx, dst, src, AC_CAT_FILE, AC_FILE_RENAME, 0);
      std_replace (rq->cgi, "file", dst);
    }
  
  free (dst);
}

/* Perform a mkdir operation. */
static void
handle_mkdir (struct request *rq)
{
  const char *in = std_get_null (rq->cgi, "mkdirin");
  const char *as = std_get_null (rq->cgi, "mkdiras");
  struct st new_dir;
  
  if (in == NULL || as == NULL)
    return;

  st_create (&new_dir, in);
  st_printf (&new_dir, "/%s", as);

  if (-1 == mkdir (st_value (&new_dir),
		   calc_mode (&rq->cx, st_value (&new_dir))))
    {
      html ("<H2>%s</H2>\n"
	    "<P>%s\n",
	    _("Error creating folder"),
	    _("An error was encountered while creating folder"));
      html_escape (st_value (&new_dir));
      html (": %s</P>\n", strerror (errno));
    }
  else 
    {
      log_activity_by_name (&rq->cx, st_value (&new_dir), NULL,
			    AC_CAT_FILE, AC_FILE_MKDIR, 0);
      std_replace (rq->cgi, "file", st_value (&new_dir));
    }

  st_destroy (&new_dir);
}

/* Perform a rmdir operation. */
static void
handle_rmdir (struct request *rq)
{
  const char *rm = std_get_null (rq->cgi, "rmdir");
  if (rm == NULL)
    return;

  if (-1 == rmdir (rm))
    {
      html ("<H2>%s</H2>\n"
	    "<P>%s\n",
	    _("Error removing folder"),
	    _("An error was encountered while removing folder"));
      html_escape (rm);
      html (": %s</P>\n", strerror (errno));
    }
  else
    log_activity_by_name (&rq->cx, rm, NULL, AC_CAT_FILE, AC_FILE_RMDIR, 0);
}

/* Perform a copy operation. */
static void
handle_copy (struct request *rq)
{
  const char *src = std_get_null (rq->cgi, "file");
  const char *dstdir = std_get_null (rq->cgi, "dir");
  const char *basename = std_get_null (rq->cgi, "basename");
  struct st dst;

  if (src == NULL || dstdir == NULL || basename == NULL
      || !std_contains (rq->cgi, "docopy"))
    return;

  st_init (&dst, 128);
  st_printf (&dst, "%s/%s", dstdir, basename);

  if (do_copy (src, st_value (&dst)))
    {
      log_activity_by_name (&rq->cx, st_value (&dst), src,
			    AC_CAT_FILE, AC_FILE_COPY, 0);
      std_replace (rq->cgi, "file", st_value (&dst));
      rq->op = "comment";
    }

  st_destroy (&dst);
}

/* Perform a delete operation. */
static void
handle_delete (struct request *rq)
{
  const char *rmfile = std_get_null (rq->cgi, "rmfile");
  const char *rmconfirm = std_get_null (rq->cgi, "rmconfirm");

  if (rmfile == NULL || rmconfirm == NULL)
    return;

  /* Make sure the user is as paying attention. */
  if (strcmp (rmconfirm, "YES"))
    {
      html ("<H2>%s</H2>\n"
	    "<P>%s</P>\n",
	    _("Error Deleting File"),
	    _("You failed to provide proper confirmation for file\n"
	      "deletion.\n"
	      "You must select the radio button labeled\n"
	      "<STRONG>YES</STRONG> for your file to be deleted.\n"));
      std_replace (rq->cgi, "file", rmfile);
      return;
    }

  if (-1 == unlink (rmfile))
    {
      html ("<H2>%s</H2>\n"
	    "<P>%s\n",
	    _("Error deleting file"),
	    _("An error was encountered while deleting file"));
      html_escape (rmfile);
      html (": %s</P>\n", strerror (errno));
    }
  else
    {
      log_activity_by_name (&rq->cx, rmfile, NULL, AC_CAT_FILE,
			    AC_FILE_DELETE, 0);
      rq->op = "comment";
    }
}
  
/* Handle chmod. */
static void
handle_chmod (struct request *rq)
{
  mode_t mode;
  const struct permission *p;
  const char *filename;

  filename = std_get_null (rq->cgi, "file");
  if (strcmp (rq->op, "chmod")
      || !std_contains (rq->cgi, "change")
      || filename == NULL)
    return;

  mode = 0;
  for (p = perms; p < perms + N_PERMS; p++)
    if (std_contains (rq->cgi, p->name))
      mode |= p->value;

  if (chmod (filename, mode) == -1)
    html ("<H1>%s</H1>\n"
	  "<P>%s: %s</P>\n",
	  _("Error changing file permissions"),
	  _("The system returned error changing file permissions"),
	  strerror (errno));
}

/* Guesses the MIME type of the file FN by reading the mime.types
   file, then prints an appropriate Content-Type: header. */
static void
guess_media_type (const char *fn)
{
  size_t fn_len;
  char *mime_type;
  FILE *f;

  char *line;
  size_t line_size;

  fn_len = strlen (fn);
  mime_type = NULL;

  f = fopen (MIME_TYPES_FILE, "r");
  if (f == NULL)
    {
      log ("opening %s: %s", MIME_TYPES_FILE, strerror (errno));
      goto exit;
    }

  line = NULL;
  line_size = 0;
  
  for (;;)
    {
      char *type;

      /* Read a line. */
      if (-1 == getline (&line, &line_size, f))
	{
	  if (ferror (f))
	    log ("reading %s: %s", MIME_TYPES_FILE, strerror (errno));

	  break;
	}

      /* Remove comments. */
      {
	char *hash = strchr (line, '#');
	if (hash)
	  *hash = '\0';
      }

      /* Determine MIME type for this line. */
      type = strtok (line, " \t\r\n");
      if (type == NULL)
	continue;

      /* Compare extension. */
      for (;;)
	{
	  char *ext;
	  size_t ext_len;

	  ext = strtok (NULL, " \t\r\n");
	  if (ext == NULL)
	    break;

	  ext_len = strlen (ext);
	  if (fn_len > ext_len + 1
	      && fn[fn_len - ext_len - 1] == '.'
	      && !strcmp (fn + fn_len - ext_len, ext))
	    {
	      free (mime_type);
	      mime_type = xstrdup (type);
	      break;
	    }
	}
    }

  if (fclose (f) == EOF)
    log ("closing %s: %s", MIME_TYPES_FILE, strerror (errno));

 exit:
  if (mime_type == NULL)
    htmls ("Content-Type: application/octet-stream\n");
  else
    {
      html ("Content-Type: %s\n", mime_type);
      free (mime_type);
    }
}

/* Handle a download operation, which includes zip and tar formats.
   Prints appropriate HTTP headers as well. */
static int
handle_download (struct request *rq)
{
  const char *filename;
  ssize_t size;
  FILE *f;
  int type;

  filename = std_get_null (rq->cgi, "file");
  if (rq->op == NULL || filename == NULL)
    return 0;
  if (!strcmp (rq->op, "view"))
    type = 0;
  else if (!strcmp (rq->op, "download"))
    type = 1;
  else if (!strcmp (rq->op, "tgz"))
    type = 2;
  else if (!strcmp (rq->op, "zip"))
    type = 3;
  else
    return 0;

  switch (type)
    {
    case 0:
    case 1:
      {
	struct stat st;
	if (0 == stat (filename, &st))
	  {
	    size = st.st_size;

	    /* You're not allowed to just download a directory.
	       Instead, you'll be shown the download screen and
	       allowed to select an archive format. */
	    if (S_ISDIR (st.st_mode))
	      return 0;

	    f = fopen (filename, "rb");
	  }
      }
      break;

    case 2:
    case 3:
      {
	struct st cmd;
	st_create (&cmd, "cd ");
	st_concat_quoted (&cmd, filename, '"');
	st_concat (&cmd, type == 2 ? "; tar cfz - ." : "; zip -r - .");
	f = popen (st_value (&cmd), "r");
	st_destroy (&cmd);

	size = -1;
      }
      break;

    default:
      assert (0);
    }

  if (f == NULL)
    {
      http_header ();
      html_banner (rq);

      html ("<H2>%s</H2>\n"
	    "<P>%s: %s.</P>\n",
	    _("Error in download"),
	    _("A system error occurred during the download operation"),
	    strerror (errno));

      html_trailer (rq);
      return 1;
    }

  switch (type)
    {
    case 0:
      guess_media_type (filename);
      break;
    case 1:
      htmls ("Content-Type: application/octet-stream\n");
      break;
    case 2:
      htmls ("Content-Type: application/x-gtar\n");
      break;
    case 3:
      htmls ("Content-Type: application/zip\n");
      break;
    default:
      assert (0);
    }

  if (size != -1)
    html ("Content-Length: %lu\n", (unsigned long) size);
  htmlc ('\n');

  for (;;)
    {
      char buf[4096];
      size_t nbytes;

      nbytes = fread (buf, 1, 4096, f);
      if (nbytes != fwrite (buf, 1, nbytes, stdout)
	  || nbytes != 4096)
	break;
    }
  
  log_activity_by_name (&rq->cx, filename, NULL, AC_CAT_FILE,
			AC_FILE_DOWNLOAD, 0);

  return 1;
}

static void
upload_error_header (struct context *cx, int *okay, const char *filename)
{
  if (*okay != 0)
    {
      *okay = 0;

      html ("<H2>%s</H2>\n"
	    "<P>%s</P><UL>\n",
	    _("Upload Error"),
	    _("One or more error(s) occurred during the upload process:"));
    }
  else
    htmls ("</LI>");

  htmls ("<LI>");
  html_a_file (cx, filename, NULL);
  htmls (":");
}

static void
do_upload (struct request *rq,
	   const char *dst_fn, FILE *src, 
	   int *okay)
{
  /* Possible behaviors when a file with this name already exists. */
  enum 
    {
      EXIST_VERSIONING,		/* Make numeric backups. */
      EXIST_OVERWRITE,		/* Overwrite. */
      EXIST_ERROR		/* Flag an error. */
    }
  exist;

  int dst;	/* Destination file handle. */
  
  {
    const char *s_exist = std_get (rq->cgi, "exist");
    
    if (!strcmp (s_exist, "versioning"))
      exist = EXIST_VERSIONING;
    else if (!strcmp (s_exist, "overwrite"))
      exist = EXIST_OVERWRITE;
    else
      exist = EXIST_ERROR;
  }
  
  for (;;)
    {
      if (exist == EXIST_VERSIONING)
	{
	  struct st v;
	  size_t len;
	  int i;

	  st_create (&v, dst_fn);
	  len = st_length (&v);
	  for (i = 1; ; i++)
	    {
	      st_truncate (&v, len);
	      st_printf (&v, ";%d", i);
	      if (0 == link (dst_fn, st_value (&v)))
		{
		  unlink (dst_fn);
		  break;
		}
	      else if (errno != EEXIST)
		break;
	    }
	}

      dst = open (dst_fn,
		  O_WRONLY | O_CREAT | O_TRUNC
		  | (exist != EXIST_OVERWRITE ? O_EXCL : 0),
		  calc_mode (&rq->cx, dst_fn));

      if (dst == -1 && errno == EEXIST)
	{
	  if (exist == EXIST_VERSIONING)
	    continue;
	  
	  upload_error_header (&rq->cx, okay, dst_fn);
	  htmls (_("A file by this name already exists.\n"));
	  return;
	}

      break;
    }
  
  if (dst == -1 || !copy_fd (fileno (src), dst) || -1 == close (dst))
    {
      upload_error_header (&rq->cx, okay, dst_fn);
      html ("%s: %s\n",
	    _("A system error occurred during upload"),
	    strerror (errno));
    }
  else
    log_activity_by_name (&rq->cx, dst_fn, NULL, AC_CAT_FILE,
			  AC_FILE_UPLOAD, 0);
}

/* Handle an upload operation. */
static void
handle_upload (struct request *rq)
{
  const char *updir = std_get_nonempty (rq->cgi, "updir");
  struct st filename;
  int okay = 1;

  int i;

  st_init (&filename, 64);
  for (i = 0; i < 8; i++)
    {
      const char *upload, *upfile;
      FILE *stream;
      
      {
	char tmp[16];
	
	sprintf (tmp, "upload%d", i);
	stream = cgi_file (tmp, &upload);
	if (stream == NULL)
	  continue;
	
	sprintf (tmp, "upfile%d", i);
	upfile = std_get_nonempty (rq->cgi, tmp);
      }

      /* Figure out destination filename. */
      if (std_contains (rq->cgi, "file") && i == 0)
	st_concat (&filename, std_get (rq->cgi, "file"));
      else
	{
	  /* Construct filename from partial filenames. */
	  if (updir == NULL || (upfile == NULL && upload == NULL))
	    continue;

	  /* If a filename was provided on the upload form, append it
	     to the directory name; otherwise, use the one provided by
	     the web browser itself. */
	  st_clear (&filename);
	  if (upfile != NULL)
	    st_printf (&filename, "%s/%s", updir, upfile);
	  else
	    {
	      const char *cp;
	  
	      /* Strip all but the last component. */
	      cp = strrchr (upload, '/');
	      if (cp == NULL)
		cp = strrchr (upload, '\\');
	      if (cp == NULL)
		cp = upload;
	      else
		cp++;
	      if (*cp == '\0')
		cp = upload;

	      st_printf (&filename, "%s/%s", updir, cp);
	    }
	}
      
      do_upload (rq, st_value (&filename), stream, &okay);
      rq->op = "upload";
    }

  if (!okay)
    html ("</LI></UL>\n");
  
  st_destroy (&filename);

  if (updir != NULL)
    std_put (rq->cgi, "file", updir);
}

/* Sort functions. */

static int 
sort_by_size (const void *va, const void *vb)
{
  struct dirent *a = *(struct dirent **) va;
  struct dirent *b = *(struct dirent **) vb;
  struct stat sa, sb;

  if (-1 == stat (a->d_name, &sa))
    sa.st_size = 0;
  if (-1 == stat (b->d_name, &sb))
    sb.st_size = 0;
  
  if (sa.st_size < sb.st_size)
    return -1;
  else if (sa.st_size > sb.st_size)
    return 1;
  else
    return 0;
}
  
static int 
sort_by_date (const void *va, const void *vb)
{
  struct dirent *a = *(struct dirent **) va;
  struct dirent *b = *(struct dirent **) vb;
  struct stat sa, sb;

  if (-1 == stat (a->d_name, &sa))
    sa.st_mtime = 0;
  if (-1 == stat (b->d_name, &sb))
    sb.st_mtime = 0;
  
  if (sa.st_mtime < sb.st_mtime)
    return 1;
  else if (sa.st_mtime > sb.st_mtime)
    return -1;
  else
    return 0;
}

/* Yes, we are using JavaScript to do mouseover highlighting.
   Shoot me now... */
static void
javascript_header (void)
{
  int i;
  
  htmls
    ("<SCRIPT LANGUAGE=\"JavaScript\">\n"
     "<!--\n"
     "function t(img,name,state) {\n"
     "    if (document.images) {\n"
     "        img.src = '/scope/f/' + state + '/' + name + '.gif'\n"
     "    }\n"
     "}\n\n"

     "if (document.images) {\n"
     "    var ops = new Object ()\n");

  for (i = 0; i < OP_COUNT; i++)
    html ("    ops['%s'] = 1\n", op_table[i].name);

  htmls
    ("\n"
     "    var images = new Object ()\n"
     "    for (op in ops) {\n"
     "        images[op] = new Image (48, 34)\n"
     "        images[op].src = '/scope/f/m/' + op + '.gif'\n"
     "    }\n"
     "}\n"
     "// -->\n"
     "</SCRIPT>\n");
}
