/* TeamSCOPE - Software for Collaborative Project Environments.
   Copyright (C) 1999, 2000, 2001 Michigan State University.
   Written by Ben Pfaff <pfaffben@msu.edu>.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To contact the TeamSCOPE team, send email to scope@cscw.msu.edu, or
   write to Charles Steinfield, Room 436 Communication Arts Building,
   Michigan State University, Department of Telecommunication, East
   Lansing MI 48824-1212. */

#include <config.h>
#include <assert.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include "alloc.h"
#include "db.h"
#include "html.h"
#include "identity.h"
#include "options.h"
#include "scope.cgi.h"
#include "session.h"
#include "std.h"

/* Output our standard HTTP headers for the benefit of the web server
   that we're talking to.  Note that the session manager might have
   already issued a Set-Cookie: header (see
   session.c:new_session()). */
void
http_header (void)
{
  htmls ("Content-Type: text/html\n\n");
}

static void user_activity_meter (struct request *);

/* Output our standard HTML banner. */
void
html_banner (struct request *rq)
{
  if (rq->abbrev == 1)
    return;

  html_header ();

  if (rq->abbrev == 2)
    return;

  if (server_name)
    html ("<FORM ACTION=\"%s\" METHOD=GET>\n", script_name);

  if (rq->cx.grp[0] != '\0'
      && rq->cx.options != NULL
      && stol (std_get (rq->cx.options, "chat")) == 0
      && !std_contains (rq->untrusted, "sent-chat"))
    {
      std_put (rq->untrusted, "sent-chat", "yes");

      html ("  <SCRIPT LANGUAGE=\"JavaScript\">\n"
	    "  <!--\n"
	    "  scopeChatWindow = open (\"%s\", \"ScopeChat\",\n"
	    "        \"toolbar=no,location=no,directories=no,status=no,"
	    "menubar=no,scrollbars=no,resizable=no,width=40,height=28\");\n"
	    "  // -->\n"
	    "  </SCRIPT>\n",
	    url_this ("chat", (char *) NULL));
    }
  
  html
    ("<TABLE CELLPADDING=0 CELLSPACING=0 WIDTH=\"100%%\">\n"
     "  <TR VALIGN=BOTTOM>\n"
     "    <TD ALIGN=LEFT>\n"
     "      <IMG ALT=\"%s\" WIDTH=227 HEIGHT=31\n"
     "       SRC=\"%s/t/scope.gif\"><BR>\n"
     "    </TD>\n"
     "    <TD ALIGN=CENTER>\n",
     _("Team Software for Collaborative Project Environments"),
     www_root);

  if (server_name)
    {
      static const struct screen
	{
	  char *name;
	  char *title;
	}
      screens[] = 
	{
	  {"", N_("Main Menu...")},
          {"welcome", N_("Welcome")},
	  {"activity", N_("Activity")},
	  {"files", N_("Files")},
	  {"messages", N_("Messages")},
	  {"calendar", N_("Calendar")},
	  {"options", N_("Options")},
	  {"admin", N_("Administration")},
	  {"search", N_("Search")},
	  {"info", N_("Info")},
	  {"teamsite", N_("Team Site")},
	  {"help", N_("Documentation")},
	  {"logout", N_("Log Out")},
	};

#define N_SCREENS ((int) (sizeof screens / sizeof *screens))

      int i;
      int selected = 2;

      if (rq->screen != NULL)
	{
	  for (i = 1; i < N_SCREENS; i++)
	    if (!strcmp (rq->screen, screens[i].name)
		|| (!strncmp (rq->screen, screens[i].name,
			      strlen (screens[i].name))
		    && rq->screen[strlen (screens[i].name)] == '-'))
	      selected = i;
	}

      html ("      <SELECT NAME=screen"
	    " ONCHANGE=\"if (selectedIndex && selectedIndex != %d) {"
	    " var i = selectedIndex; selectedIndex = %d; "
	    "window.location.href = '%s/' + options[i].value; }\">\n",
	    N_SCREENS, selected, our_abs_url);
      
      for (i = 0; i < N_SCREENS; i++)
	html ("        <OPTION VALUE=\"%s\"%s>%s</OPTION>\n",
	      screens[i].name, i == selected ? " SELECTED" : "",
	      gettext (screens[i].title));

      /* List of other groups. */
      if (rq->cx.grp[0])
	{
	  int n;
	  const char **groups;

	  groups = std_get_many (rq->trusted, "groups", &n);
	  if (n > 1)
	    {
	      int i;
	    
	      html ("        <OPTION DISABLED>--%s--</OPTION>\n",
		    _("Change Workgroup:"));
	      for (i = 0; i < n; i++)
		{
		  char *s = local_strdup (groups[i]);
		  char *grp;

		  grp = strtok (s, ";");
		  if (grp != NULL && strcmp (grp, rq->cx.grp))
		    {
		      char *name = strtok (NULL, "");

		      html ("        <OPTION VALUE=\"grp-%s-%s\">%s"
			    "</OPTION>\n",
			    grp, rq->screen, name != NULL ? name : grp);
		    }

		  local_free (s);
		}
	    }
	}
	  
      htmls ("      </SELECT>\n"
	     "      <INPUT TYPE=submit VALUE=\"Go!\">\n");

      if (rq->screen != NULL)
	{
	  /* Link to context-sensitive help. */
	  struct st s;
	  int in_help = has_prefix (rq->screen, "help");

	  st_create (&s, "help");
	  if (!in_help)
	    st_printf (&s, "-%s",
		       rq->screen[0] == 0 ? "activity" : rq->screen);
    
	  html ("<A HREF=\"%s\"%s>%s</A>\n",
		url_this (st_value (&s), (char *) NULL),
		in_help ? "" : " TARGET=\"_blank\"",
		_("Help"));

	  st_destroy (&s);
	}
    }

  if (rq->cx.options != NULL && stol (std_get (rq->cx.options, "chat")) == 1)
    html_chat_applet (rq);
  
  html
    ("    </TD>\n"
     "    <TD ALIGN=RIGHT>\n");

  if (rq->cx.grp[0])
    html ("      <FONT SIZE=\"+2\">%s Team</FONT><BR>\n", rq->cx.fullgrp);

  htmls ("    </TD>\n"
	 "  </TR>\n");

  if (rq->cx.usr[0] && rq->cx.grp[0])
    user_activity_meter (rq);

  htmls ("</TABLE>\n");
  if (server_name)
    htmls ("</FORM>\n");
  htmls ("<HR NOSHADE>\n\n");
}

/* Send the user some colorful advertising materials for other free
   software. */
void
html_powered_by (struct request *rq)
{
  if (rq->abbrev)
    return;

  if (std_contains (conf, "login_banner"))
    {
      const char *filename = std_get (conf, "login_banner");
      FILE *f = fopen (filename, "r");

      if (f != NULL)
	{
	  for (;;)
	    {
	      char buffer[4096];

	      if (fwrite (buffer, 1, fread (buffer, 1, 4096, f), stdout) != 4096)
		{
		  fclose (f);
		  return;
		}
	    }
	}
    }
  else
    {
      static const char *pgm[]
	= {"apache", "debian", "gnu", "linux", "gimp", "proftpd"};
      static const char *title[sizeof pgm / sizeof *pgm]
	= {"Apache", "Debian", "GNU", "Linux", "GIMP", "ProFTPd"};
      size_t i;

      htmls ("<HR NOSHADE>\n"	 
	     "<P ALIGN=CENTER>\n");

      for (i = 0; i < sizeof pgm / sizeof *pgm; i++)
	html ("<A HREF=\"http://www.%s.org\">"
	      "<IMG BORDER=0 SRC=\"%s/pb/pb_%s.gif\" "
	      "ALT=\"[Powered by %s]\"></A>\n",
	      pgm[i], www_root, pgm[i], title[i]);
      
      htmls ("</P>\n");
    }
}

#if GLOBAL_DEBUGGING
/* Subtract the `struct timeval' values X and Y,
   storing the result in RESULT.
   Return 1 if the difference is negative, otherwise 0.

   From libc.info.  */
static int
timeval_subtract (struct timeval *result,
		  struct timeval *x, struct timeval *y)
{
  /* Perform the carry for the later subtraction by updating Y. */
  if (x->tv_usec < y->tv_usec)
    {
      int nsec = (y->tv_usec - x->tv_usec) / 1000000 + 1;
      y->tv_usec -= 1000000 * nsec;
      y->tv_sec += nsec;
    }
  
  if (x->tv_usec - y->tv_usec > 1000000)
    {
      int nsec = (x->tv_usec - y->tv_usec) / 1000000;
      y->tv_usec += 1000000 * nsec;
      y->tv_sec -= nsec;
    }
     
  /* Compute the time remaining to wait.
     `tv_usec' is certainly positive. */
  result->tv_sec = x->tv_sec - y->tv_sec;
  result->tv_usec = x->tv_usec - y->tv_usec;
     
  /* Return 1 if result is negative. */
  return x->tv_sec < y->tv_sec;
}
#endif /* GLOBAL_DEBUGGING */

/* Send the user the HTML trailer.  Not that his web-user-agent cares,
   but it's good form to do so, and if we're lucky our output will
   even be validating HTML 4.0. */
void
html_trailer (struct request *rq)
{
  if (rq->abbrev)
    return;

  /* Link to feedback form. */
  {
    struct url url;

    url_create (&url, "feedback");
    url_arg (&url, "fscreen", rq->screen);
    url_arg (&url, "fquery", getenv ("QUERY_STRING"));
    
    html ("\n"
	  "<HR NOSHADE>\n"
	  "<P ALIGN=CENTER>\n"
	  "<A HREF=\"%s\">Feedback</A>\n"
	  "</P>\n",
	  url_string (&url));

    url_destroy (&url);
  }

#if GLOBAL_DEBUGGING
  {
    struct timeval cur;
    struct timeval diff;

    gettimeofday (&cur, NULL);
    timeval_subtract (&diff, &start_timeval, &cur);
    
    html ("<P>TeamSCOPE compiled " __DATE__ " " __TIME__ 
#if __CHECKER__
	  " including Checker memory debugging"
#endif
	  ".\n<BR>Time elapsed: CPU %.02f, clock: %.02f.</P>\n",
	  (double) (clock () - start_clock) / CLOCKS_PER_SEC,
	  cur.tv_sec - start_timeval.tv_sec
	  + (cur.tv_usec - start_timeval.tv_usec) / 1000000.);
  }
#endif

  htmls ("</BODY>\n"
	 "</HTML>\n");
}

/* Output an HTML hidden text field that reflects the current
   abbreviation level. */
void
html_hidden_abbrev (struct request *request)
{
  if (request->abbrev == 1)
    html_hidden ("abbrev", "yes");
  else if (request->abbrev == 2)
    html_hidden ("abbrev", "partial");
}

/* Output HTML code for the TeamSCOPE Java-based chat applet. */
void
html_chat_applet (struct request *request)
{
  html ("<APPLET CODE=\"ScopeChat.class\" CODEBASE=\"/scope\"\n"
	"        ARCHIVE=\"ScopeChat.zip\" WIDTH=40 HEIGHT=28>\n"
	"  <PARAM NAME=session VALUE=\"%s\">\n"
	"</APPLET>\n",
	request->session_id);
}

/* Copy the current screen's abbrev value into the specified URL. */
void
url_copy_abbrev (struct url *url, struct request *rq)
{
  if (rq->abbrev == 1)
    url_arg (url, "abbrev", "yes");
  else if (rq->abbrev == 2)
    url_arg (url, "abbrev", "partial");
}

/* Private functions. */

/* Tell the user what other users are logged in. */
static void
user_activity_meter (struct request *rq)
{
  db_result *r;
  int count;
  int i;

  assert (rq->cx.usr[0] && rq->cx.grp[0]);

  db_query ("SELECT DISTINCT users.name FROM users, user_group "
	    "WHERE users.usr IN "
	    "(SELECT DISTINCT usr FROM user_group WHERE grp IN "
	    "(SELECT grp FROM user_group WHERE usr = '%sq')) "
	    "AND last_usage > datetime (timenow () - reltime '5 minutes') "
	    "AND users.usr <> '%sq' "
	    "ORDER by users.name",
	    DB_TUPLES | DB_FORMAT, &r, rq->cx.usr, rq->cx.usr);

  count = dbr_nrows (r);

  if (count)
    html ("      %s: ", _("Also logged in"));

  for (i = 0; i < count; i++)
    {
      if (i)
	htmlc (',');
	  
      html ("\n%s", dbr_value (r, i, 0));
    }

  htmls ("<BR>\n"
	 "    </TD>\n"
	 "  </TR>\n");

  dbr_destroy (r);
}
