/* TeamSCOPE - Software for Collaborative Project Environments.
   Copyright (C) 1999, 2000, 2001 Michigan State University.
   Written by Ben Pfaff <pfaffben@msu.edu>.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To contact the TeamSCOPE team, send email to scope@cscw.msu.edu, or
   write to Charles Steinfield, Room 436 Communication Arts Building,
   Michigan State University, Department of Telecommunication, East
   Lansing MI 48824-1212. */

#include <config.h>
#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <grp.h>
#include <pwd.h>
#include <stdlib.h>
#include <sys/stat.h>
#include "alloc.h"
#include "conf.h"
#include "db.h"
#include "error.h"
#include "id-list.h"
#include "identity.h"
#include "session.h"
#include "str.h"
#include "std.h"

/* Returns the name of resource having sequence number SEQ. */
const char *
seq_no_to_resource_name (int seq)
{
  db_result *r;
  char *s;

  db_query ("SELECT name FROM resources WHERE seq = %d",
	    DB_FORMAT | DB_TUPLES, &r, seq);

  s = xstrdup (dbr_value (r, 0, 0));
  dbr_destroy (r);

  return s;
}

/* Number of items to cache at once. */
#define CACHE_SIZE 8

/* A cache for gids or uids. */
struct id_cache
  {
    int n;			/* Number of items in cache. */
    uid_t id[CACHE_SIZE];	/* gids or uids in the cache. */
    char *name[CACHE_SIZE];	/* Names associated with gids/uids. */
  };

/* For cache CACHE, swap items indexed A and B. */
static void
swap (struct id_cache *cache, int a, int b)
{
  if (a != b)
    {
      uid_t t_id = cache->id[a];
      char *t_name = cache->name[a];

      cache->id[a] = cache->id[b];
      cache->id[b] = t_id;

      cache->name[a] = cache->name[b];
      cache->name[b] = t_name;
    }
}

/* Look in CACHE for item with uid or gid ID, as indicated by TYPE,
   where a zero TYPE means uid, nonzero TYPE means gid.  If not found,
   loads it from the appropriate file, stores it into the cache.  In
   either case the name associated with the id is returned. */
static const char *
get_from_id_cache (struct id_cache *cache, uid_t id, int type)
{
  /* Look for id in cache. */
  {
    int i;
  
    for (i = 0; i < cache->n; i++)
      if (cache->id[i] == id)
	{
	  /* Swap item with most-recently-used item in order to get
	     the list into order of mru...lru. */
	  swap (cache, 0, i);
	  return cache->name[0];
	}
  }

  /* If cache is full, delete the least-recently-used item. */
  if (cache->n == CACHE_SIZE)
    {
      free (cache->name[CACHE_SIZE - 1]);
      cache->n--;
    }

  /* Retrieve uncached from appropriate file and make it the mru. */
  {
    const char *name = "unknown";

    if (type == 0)
      {
	struct passwd *pw = getpwuid (id);
	if (pw)
	  name = pw->pw_name;
      }
    else
      {
	struct group *gr = getgrgid (id);
	if (gr)
	  name = gr->gr_name;
      }
      
    cache->id[cache->n] = id;
    cache->name[cache->n] = xstrdup (name);
    swap (cache, 0, cache->n);
    cache->n++;
  }

  return cache->name[0];
}

/* Returns the username associated with uid UID. */
const char *
uid_to_username (uid_t uid)
{
  static struct id_cache cache;

  return get_from_id_cache (&cache, uid, 0);
}

/* Returns the name of the group associated with gid GID. */
const char *
gid_to_groupname (gid_t gid)
{
  static struct id_cache cache;

  assert (sizeof (gid_t) == sizeof (uid_t));
  return get_from_id_cache (&cache, gid, 1);
}

/* Returns the full name of the user in the current user's workgroup
   with username USERNAME. */
const char *
username_to_fullname (struct workgroup *wg, const char *username)
{
  const char *fullname;
  assert (wg != NULL && wg->users != NULL && username != NULL);
  fullname = std_get_null (wg->users, username);
  return fullname != NULL ? fullname : username;
}

/* Read the list of users from the database into `users', `n_users'. */
void
read_user_list (struct workgroup **wg, const char *grp)
{
  db_result *result;
  
  if (*wg != NULL)
    return;

  *wg = xmalloc (sizeof **wg);
  (*wg)->users = std_create (0);

  db_query ("SELECT users.usr, name, last_word (name) as last_name "
	    "FROM users, user_group "
	    "WHERE grp = '%sq' and users.usr = user_group.usr "
	    "ORDER BY last_name, name",
	    DB_FORMAT | DB_TUPLES, &result, grp);
  
  {
    int i;

    (*wg)->n = dbr_nrows (result);
    (*wg)->name = xmalloc (sizeof *(*wg)->name * (*wg)->n);
    (*wg)->fullname = xmalloc (sizeof *(*wg)->fullname * (*wg)->n);
    for (i = 0; i < (*wg)->n; i++)
      {
	std_put ((*wg)->users, dbr_value (result, i, 0),
		 dbr_value (result, i, 1));
	(*wg)->name[i] = xstrdup (dbr_value (result, i, 0));
	(*wg)->fullname[i] = xstrdup (dbr_value (result, i, 1));
      }
    dbr_destroy (result);
  }
}

char *
relevant_file_crit (struct context *cx)
{
  return blp_asprintf ("strings.grp = '%s'", cx->grp);
}

void
init_context (struct context *cx)
{
  cx->usr[0] = '\0';
  cx->fullusr = NULL;
  cx->uid = -1;
  cx->home = NULL;
  cx->options = NULL;

  cx->grp[0] = '\0';
  cx->fullgrp = NULL;
  cx->gid = -1;
  cx->shared = NULL;
  cx->web = NULL;
  cx->wg = NULL;
}

void
destroy_context (struct context *cx)
{
  free (cx->fullusr);
  free (cx->home);
  if (cx->options != NULL)
    std_destroy (cx->options);

  free (cx->fullgrp);
  free (cx->shared);
  free (cx->web);
  if (cx->wg != NULL)
    {
      int i;

      std_destroy (cx->wg->users);
      for (i = 0; i < cx->wg->n; i++)
	{
	  free (cx->wg->name[i]);
	  free (cx->wg->fullname[i]);
	}
      free (cx->wg->name);
      free (cx->wg->fullname);
      free (cx->wg);
    }
}

/* Verifies that file F, named FILENAME, is owned by the user with uid
   UID. */
void
verify_file_owner (const char *filename, FILE *f, uid_t uid)
{
  struct stat s;

  if (fstat (fileno (f), &s) == -1)
    fail ("%s: fstat: %s", filename, strerror (errno));
  if (s.st_uid != uid)
    fail ("%s: owner is not uid %d", filename, uid);
}

/* Returns the row of R that contains USR in the first column,
   or -1 if USR is not in R. */
static int
find_user (db_result *r, const char *usr)
{
  int l = 0;
  int h = dbr_nrows (r) - 1;

  while (l <= h)
    {
      int m = (l + h) / 2;
      int cmp = strcmp (usr, dbr_value (r, m, 0));

      if (cmp > 0)
	l = m + 1;
      else if (cmp < 0)
	h = m - 1;
      else
	return m;
    }

  return -1;
}

/* Reads the TeamSCOPE password file to determine which users'
   accounts are locked.
   Sets the appropriate `locked' fields in INFO, corresponding to
   IDS. */
static void
read_locks (const struct id_list *ids, struct user_info *info)
{
  char *line = NULL;
  size_t size = 0;

  FILE *f = fopen (PASSWD_FILE, "r");
  if (f == NULL)
    fail ("%s: open: %s", PASSWD_FILE, strerror (errno));
  verify_file_owner (PASSWD_FILE, f, 0);

  for (;;)
    {
      char usr[9];
      char *sp, *dp;
      
      long len = getline (&line, &size, f);
      if (len == -1)
	break;

      dp = usr;
      for (sp = line; *sp != '\0' && *sp != ':' && sp < line + 8; sp++)
	*dp++ = *sp;
      *dp = '\0';

      if (sp[0] == ':' && sp[1] == '!')
	{
	  int index = id_list_find (ids, usr);
	  if (index != -1)
	    info[index].locked = 1;
	}
    }
  free (line);

  fclose (f);
}

/* Fills in INFO corresponding to IDS.
   At least the requested FIELDS are filled, but others may be as
   well.  Pointer fields not filled will be set to NULL. */
void 
get_user_info (const struct id_list *ids, struct user_info *info,
	       enum ui_fields fields)
{
  db_result *r;
  struct st query;
  size_t i;

  st_create (&query, "SELECT usr, name, recordable, last_usage, email "
	     "FROM users ");
  if (ids->n > 0 && ids->n <= 50)
    {
      st_concat (&query, "WHERE usr IN (");
      for (i = 0; i < ids->n; i++)
	{
	  if (i != 0)
	    st_concat (&query, ", ");
	  st_concat_quoted (&query, ids->s[i], '\'');
	}
      st_concat (&query, ") ");
    }
  st_concat (&query, "ORDER BY usr");

  db_query (st_value (&query), DB_TUPLES, &r);

  for (i = 0; i < ids->n; i++)
    {
      int row;
      info[i].usr = xstrdup (ids->s[i]);
      info[i].locked = 0;

      row = find_user (r, ids->s[i]);
      if (row != -1)
	{
	  info[i].fullname = xstrdup (dbr_value (r, row, 1));
	  info[i].recordable = dbr_value (r, row, 2)[0] == 't';
	  info[i].last_usage = xstrdup (dbr_value (r, row, 3));
	  info[i].email = xstrdup (dbr_value (r, row, 4));
	}
      else
	{
	  info[i].fullname = xstrdup ("");
	  info[i].recordable = 0;
	  info[i].last_usage = xstrdup ("");
	  info[i].email = xstrdup ("");
	}
    }

  if (fields & UI_LOCKED)
    read_locks (ids, info);

  dbr_destroy (r);
}

/* Free memory associated with INFO, which contains COUNT elements. */
void
free_user_info (struct user_info *info, size_t count)
{
  size_t i;

  for (i = 0; i < count; i++)
    {
      free (info[i].usr);
      free (info[i].fullname);
      free (info[i].last_usage);
      free (info[i].email);
    }
}

