/* TeamSCOPE - Software for Collaborative Project Environments.
   Copyright (C) 1999, 2000, 2001 Michigan State University.
   Written by Ben Pfaff <pfaffben@msu.edu>.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To contact the TeamSCOPE team, send email to scope@cscw.msu.edu, or
   write to Charles Steinfield, Room 436 Communication Arts Building,
   Michigan State University, Department of Telecommunication, East
   Lansing MI 48824-1212. */

#include <config.h>
#include <assert.h>
#include <ctype.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <pwd.h>
#include "activity.h"
#include "activityP.h"
#include "alloc.h"
#include "calendar.h"
#include "datetime.h"
#include "db.h"
#include "error.h"
#include "html.h"
#include "identity.h"
#include "mail.h"
#include "messages.h"
#include "scope-db.h"
#include "std.h"
#include "str.h"

/* Results of a query for calendar events. */
struct event_query
  {
    db_result *result;		/* Result table. */
    size_t count;		/* Number of results. */
  };

static void event_query_run (struct event_query *, struct context *, int len);
static void event_query_present_text (struct event_query *);
static void event_query_present_html (struct event_query *);
static void event_query_destroy (struct event_query *);

/* Information needed to process received mail, a mishmash of
   information that mostly goes into the `notes' table.  These are in
   the order that they are figured out. */
struct message
  {
    const char *title;	/* Subject:. */
    char *usr;		/* From: or Reply-To:. */
    int followup;	/* X-Followup: or In-Reply-To: or References:. */

    char *grp;		/* Obtained from message being followed up. */
    int seq;		/* get_seq(). */
    int file;		/* Obtained from message being followed up. */
    char *thread;	/* Derived from followed up message, seq. */

    char *content;	/* From parsing body of message. */
  };

/* A MIME part, including its content type and contents. */
struct part
  {
    /* A MIME type. */
    struct mime_type
      {
	const char *name;	/* Content type; i.e., "text/plain". */
	int boundary;		/* Boundary, if multipart. */
      }
    type;

    /* Content. */
    char *content;
  };

static void send_update_mail (struct context *cx, int last_read_mail,
			      struct std *filter,
			      int format, int body, int freq);
static void compose_text_update (struct context *, struct st *,
				 struct activity_query *,
				 struct event_query *);
static void compose_html_update (struct context *cx, struct st *s,
				 struct activity_query *,
				 struct event_query *);
static void compose_bodies (struct context *cx, struct mime_type *overall_type,
			    struct st *bodies, db_result *result,
			    int format, int body);
static int is_post (db_result *result, int row);
static void text_present_activities (struct context *cx,
				     struct activity_query *query);

static int compose_rfc822_message (struct context *cx, struct st *mail,
				   int seq, int format);

static struct std *parse_headers (struct st *hdr_txt);
static char *read_content (struct std *headers, const char *hdr_txt);
static char *read_until (const char *boundary);

static void text_fill (struct st *dst, const char *src);
static void text_hr (struct st *s);
static void text_center_line (struct st *s, const char *text);
static void text_escape (const char *text);
static void text_file (struct context *cx, const char *filename);

static void make_boundary (struct context *cx, char boundary[128], int n);
static void print_mime_type (struct context *cx,
			     struct st *s, const struct mime_type *type);
static void make_mime_mail (struct context *cx, struct st *mail,
			    struct mime_type *overall_type,
			    const struct mime_type *multipart_type,
			    const struct part part[], int n_parts);
static void send_mail (const char *mail);

static int get_user_context (struct context *cx,
			     const char *usr, const char *grp);

static char *parse_boundary (const char *content_type);
static int content_type_eq (const char *a, const char *b);
static int extract_email (const char *beg, const char **p, struct st *email);
static int st_casecmp (const char *a, const char *b);
static int determine_usr_email (struct message *msg, const char *email);
static int determine_usr (struct message *msg,
			  struct std *headers, const char *hdr_txt);
static int extract_message_id (const char *msg_id);
static int determine_followup (struct message *msg,
			       struct std *headers, const char *hdr_txt);
static void send_reply (struct std *headers, const char *hdr_txt,
			const char *subject, const char *content, ...);

/* scope-db hourly-mail. */

/* Users can request to have summary emails sent to them as often as
   hourly.  `scope-db hourly-mail', implemented by this function,
   should be called once an hour. */
void
cmd_hourly_mail (int argc unused, char *argv[] unused)
{
  /* Current level.  An email is sent if the user's frequency is set
     to this value or higher. */
  int level;

  /* Sequence number. */
  int seq;

  db_result *r;
  size_t i;

  /* Figure out the level. */
  {
    time_t t;
    struct tm *tm;

    time (&t);
    tm = localtime (&t);
    if (tm->tm_hour == 0 && tm->tm_wday == 1)
      level = 1;
    else if (tm->tm_hour == 0 && (tm->tm_wday == 3 || tm->tm_wday == 5))
      level = 2;
    else if (tm->tm_hour == 0)
      level = 3;
    else if (tm->tm_hour == 12)
      level = 4;
    else
      level = 5;
  }

  /* Initialize database. */
  connect_db ();
  seq = get_seq ();
  db_begin ();

  /* Initialize HTML module because quite a bit of code for displaying
     activity is in activity.c, shared between scope-db and scope.cgi,
     and the display functions use html(), etc., which we redirect to
     string variables. */
  html_initialize ();

  /* Iterate across all the users interested in being sent mail. */
  db_query ("SELECT usr, groups.grp, last_read_mail "
	    "FROM user_group, groups "
	    "WHERE user_group.grp = groups.grp",
	    DB_TUPLES, &r);
  for (i = 0; i < dbr_nrows (r); i++)
    {
      struct context cx;
      int last_read_mail;
      int freq;

      /* Make sure that the user should be sent mail at this time. */
      last_read_mail = stoi (dbr_value (r, i, 2));

      if (!get_user_context (&cx, dbr_value (r, i, 0), dbr_value (r, i, 1)))
	goto next;

      freq = stoi (std_get (cx.options, "emailfreq"));
      if (freq < level)
	goto next;
      
      datetime_init_timezone (&cx);

      /* Send email to the user. */
      {
	struct std *filter = std_get_subdict (cx.options, "emailfilter");
	send_update_mail (&cx, last_read_mail, filter,
			  stoi (std_get (cx.options, "emailfmt")),
			  stoi (std_get (cx.options, "emailbody")),
			  freq);
	std_destroy (filter);
      }

      /* Update last_read_mail (and last_read_web if appropriate). */
      db_query ("UPDATE user_group SET last_read_mail = %d "
		"WHERE usr = '%sq' AND grp = '%sq' "
		"AND last_read_mail < %d",
		DB_FORMAT, NULL, seq, cx.usr, cx.grp, seq);

      if (!strcmp (std_get (cx.options, "mailresetsweb"), "yes"))
	db_query ("UPDATE user_group SET last_read_web = %d "
		  "WHERE usr = '%sq' AND grp = '%sq' "
		  "AND last_read_web < %d",
		  DB_FORMAT, NULL, seq, cx.usr, cx.grp, seq);

    next:
      destroy_context (&cx);

      db_end ();
      db_begin ();
    }
  dbr_destroy (r);

  db_end ();
}

/* Sends a activity summary email to the user represented by CX, who
   last saw the activity with sequence number LAST_READ_MAIL.  The
   user is interested in activities described by FILTER.  The email
   should be in plain text for FORMAT == 0, HTML for FORMAT == 1, or
   both for FORMAT == 2.  For messages included in these activities,
   their bodies should be included if BODY == 1, otherwise just the
   subjects.  FREQ is the user's selected email frequency. */
static void
send_update_mail (struct context *cx,
		  int last_read_mail, struct std *filter,
		  int format, int body, int freq)
{
  struct activity_query activities;
  struct event_query events;
  struct mime_type summary_type;
  struct st mail;
  struct st summary;
  struct st bodies;
  struct part part[2];
  int n_parts;
  int i;

  /* New activities. */
  activity_query_init (&activities);
  activity_query_from_dict (&activities, filter, cx->usr);
  activities.last_read_seq = last_read_mail;
  activities.search = 1;
  activity_query_run (cx, &activities);

  /* Upcoming calendar events. */
  event_query_run (&events, cx, freq);

  if (activities.count == 0 && events.count == 0)
    goto done;

  st_init (&mail, 4096);
  st_printf (&mail,
  	     "From: TeamSCOPE <%s>\n"
	     "Subject: TeamSCOPE activity update for workgroup %s\n"
#if 0
	     "CC: root\n"
#endif
	     "To: %s@%s\n",
	     std_get (conf, "admin_email"),
	     cx->grp,
	     cx->usr, std_get (conf, "mail_domain"));

  /* Construct summary in `summary'. */
  n_parts = 0;

  if (format != 1)
    {
      struct st text;

      st_init (&text, 1024);
      compose_text_update (cx, &text, &activities, &events);
      part[n_parts].type.name = "text/plain";
      part[n_parts].type.boundary = -1;
      part[n_parts].content = st_value (&text);
      n_parts++;
    }
  
  if (format != 0)
    {
      struct st html;

      st_init (&html, 1024);
      compose_html_update (cx, &html, &activities, &events);
      part[n_parts].type.name = "text/html";
      part[n_parts].type.boundary = -1;
      part[n_parts].content = st_value (&html);
      n_parts++;
    }
  st_init (&summary, 4096);

  {
    static const struct mime_type type = {"multipart/alternative", 1};
    make_mime_mail (cx, &summary, &summary_type, &type, part, n_parts);
  }
  
  for (i = 0; i < n_parts; i++)
    free (part[i].content);

  n_parts = 0;
  
  part[0].type = summary_type;
  part[0].content = st_value (&summary);
  n_parts++;

  st_init (&bodies, 1024);
  compose_bodies (cx, &part[1].type, &bodies, activities.result, format, body);
  if (st_length (&bodies) > 0)
    {
      part[1].content = st_value (&bodies);
      n_parts++;
    }
  
  {
    static const struct mime_type type = {"multipart/mixed", 0};
    make_mime_mail (cx, &mail, NULL, &type, part, n_parts);
  }

  send_mail (st_value (&mail));
  st_destroy (&mail);
  
 done:
  activity_query_destroy (&activities);
  event_query_destroy (&events);
}

/* Composes a plain text email activity summary for user content CX within
   string S given the results from activity query ACTIVITIES. */
static void
compose_text_update (struct context *cx,
		     struct st *s, struct activity_query *activities,
		     struct event_query *events)
{
  struct st p;

  st_init (&p, 256);

  if (events->count > 0)
    {
      st_printf (&p, "There %s %d upcoming scheduled calendar %s, "
		 "described below.  ",
		 events->count > 1 ? "are" : "is",
		 events->count,
		 events->count > 1 ? "events" : "event");
    }
  
  if (activities->count > 0)
    {
      st_printf (&p, "There %s %s%d new %s since the last time the email "
		 "activity log was reset, described below.  ",
		 activities->count > 1 ? "are" : "is",
		 activities->count >= activities->limit ? "at least " : "",
		 activities->count,
		 activities->count > 1 ? "activities" : "activity");

      if (activities->count >= activities->limit)
	st_printf (&p, "There may be additional new activities, but only the "
		   "first %d are displayed, as requested on your TeamSCOPE "
		   "Options screen.  ", activities->limit);
    }

  st_printf (&p, "\n\n"
	     "This message has been sent because of an explicit user "
	     "request.  If you do not know why you have received it, "
	     "or for more information on how to use TeamSCOPE, visit "
	     "the TeamSCOPE webpage at\n\t%s://%s%s\nor "
	     "send email to the local TeamSCOPE administrator "
	     "at\n\t%s\n",
	     server_protocol,
	     std_get (conf, "server_name"), std_get (conf, "www_root"),
	     std_get (conf, "admin_email"));
  text_fill (s, st_value (&p));
  st_destroy (&p);

  html_push_buffer (s);

  if (events->count > 0)
    event_query_present_text (events);

  if (events->count > 0 && activities->count > 0)
    htmls ("\n\n");

  if (activities->count > 0)
    text_present_activities (cx, activities);

  {
    char *s;
    s = blp_asprintf ("\n(All dates and time above are expressed in "
		      "your currently selected time zone, %s.)\n",
		      getenv ("TZ") ? getenv ("TZ") : tzname[0]);
    text_fill (html_buffer, s);
    free (s);
  }

  html_pop_buffer ();
}

/* Composes an HTML email activity summary for user content CX within
   string S given the results from activity query ACTIVITIES. */
static void
compose_html_update (struct context *cx,
		     struct st *s,
		     struct activity_query *activities,
		     struct event_query *events)
{
  html_push_buffer (s);
  html_header ();

  htmls ("<P>\n");
  if (events->count > 0)
    {
      html ("There %s %d upcoming scheduled calendar %s,\n"
	    "described below.\n",
	    events->count > 1 ? "are" : "is",
	    events->count,
	    events->count > 1 ? "events" : "event");
    }
  
  if (activities->count > 0)
    {
      html ("There %s %s%d new %s since the last time the email\n"
	    "activity log was reset, described below.\n",
	    activities->count > 1 ? "are" : "is",
	    activities->count >= activities->limit ? "at least " : "",
	    activities->count,
	    activities->count > 1 ? "activities" : "activity");

      if (activities->count >= activities->limit)
	{
	  struct url options_url;
	  
	  url_create (&options_url, "emailfilter");
	  html ("There may be additional new activities, but only the\n"
		"first %d are displayed, as requested on the\n"
		"<A HREF=\"%s\">TeamSCOPE Email Filter screen</A>.\n",
		activities->limit, url_string (&options_url));
	  url_destroy (&options_url);
	}
    }
  html ("</P>\n\n"
	"<P>This message has been sent because of an explicit user\n"
	"request.  If you do not know why you have received it,\n"
	"or for more information on how to use TeamSCOPE, visit\n"
	"the <A HREF=\"%s://%s%s\">TeamSCOPE webpage</A> or\n"
	"send email to the local TeamSCOPE administrator at\n"
	"<A HREF=\"mailto:%s\">%s</A>.</P>\n\n",
	server_protocol,
	std_get (conf, "server_name"), std_get (conf, "www_root"),
	std_get (conf, "admin_email"), std_get (conf, "admin_email"));

  if (events->count > 0)
    {
      htmls ("<HR>\n\n"
	     "<H2>Upcoming Calendar Events</H2>\n");
      event_query_present_html (events);
    }

  if (activities->count > 0)
    {
      struct url base_url;
      struct std *extra;

      url_create (&base_url, "activity");
      url_arg (&base_url, "needgrp", cx->grp);
      
      extra = std_create (0);
      std_put (extra, "needgrp", cx->grp);

      htmls ("<HR>\n\n"
	     "<H2>New Activities</H2>\n");
      activity_query_present (cx, activities, &base_url, extra);
      
      url_destroy (&base_url);
      std_destroy (extra);
    }

  htmls ("</BODY>\n"
	 "</HTML>\n");

  html_pop_buffer ();
}

/* Composes the bodies of all the emails given within RESULT, which
   should be a database query result produced by activity_query_run(),
   into BODIES.  The corresponding user is represented by CX.  The
   MIME type of the result is stored into OVERALL_TYPE, probably
   "multipart/digest", but if there is only one email then it could
   also be "multipart/alternative", "text/plain", or "text/html".
   FORMAT and BODY are as passed to send_update_mail(). */
static void
compose_bodies (struct context *cx, struct mime_type *overall_type,
		struct st *bodies, db_result *result, int format, int body)
{
  struct part *part;
  int n_parts;
  size_t count = dbr_nrows (result);
  
  if (body != 1)
    return;

  {
    size_t i;

    count = dbr_nrows (result);
    n_parts = 0;
    for (i = 0; i < count; i++)
      if (is_post (result, i))
	n_parts++;
  }

  if (n_parts == 0)
    return;
  
  part = xmalloc (sizeof *part * n_parts);
  n_parts = 0;
  
  {
    size_t i;
    struct st s;

    st_init (&s, 1024);
    for (i = 0; i < count; i++)
      {
	int ref_seq = stoi (dbr_value (result, i, 6));

	if (is_post (result, i)
	    && compose_rfc822_message (cx, &s, ref_seq, format))
	  {
	    part[n_parts].type.name = "message/rfc822";
	    part[n_parts].type.boundary = -1;
	    part[n_parts].content = st_value (&s);
	    n_parts++;
	    
	    st_init (&s, 1024);
	  }
	else
	  st_clear (&s);
      }
    st_destroy (&s);
  }

  if (n_parts > 0)
    {
      static const struct mime_type digest_type = {"multipart/digest", 2};
      make_mime_mail (cx, bodies, overall_type, &digest_type, part, n_parts);
    }
}

/* Returns nonzero iff the ROW'th row of RESULT, as produced by
   activity_query_run(), represents a message board post or a file
   comment post. */
static int
is_post (db_result *result, int row)
{
  int cat = stoi (dbr_value (result, row, 3));
  int type = stoi (dbr_value (result, row, 4));
	
  return ((cat == AC_CAT_FILE && type == AC_FILE_POST)
	  || (cat == AC_CAT_MSG_BOARD && type == AC_MSG_POST));
}

/* Equivalent to activity_query_present() but outputs results in plain
   text instead of HTML. */
static void
text_present_activities (struct context *cx, struct activity_query *query)
{
  struct activity a, b;
  
  int i;

  text_hr (html_buffer);
  text_center_line (html_buffer, "New Activities");
  text_hr (html_buffer);
  
  read_user_list (&cx->wg, cx->grp);
  for (i = 0; i < query->count; i++)
    {
      /* Keep track of the current and the previous query results by
	 alternating between using `a' and `b' for storage and
	 alternating the pointers to them as well. */
      struct activity *cur, *prev;
      prev = i % 2 ? &a : &b;
      cur = i % 2 ? &b : &a;

      interpret_query (query, cur, i);

      if ((query->sort == AC_SORT_FILENAME
	   || query->sort == AC_SORT_ACTIVITY)
	  && (i == 0 || cur->cat != prev->cat))
	text_center_line (html_buffer, activity_category_name (cur->cat));

      html ("%s %s\n\t",
	    username_to_fullname (cx->wg, cur->user),
	    activity_type_name (cur->cat, cur->type));
      switch (cur->cat)
	{
	case AC_CAT_FILE:
	  text_file (cx, cur->s1);
	  if (cur->type == AC_FILE_COPY
	      || cur->type == AC_FILE_RENAME
	      || cur->type == AC_FILE_MOVE)
	    {
	      htmls ("\n\tfrom ");
	      text_file (cx, cur->s2);
	    }
	  break;

	case AC_CAT_MSG_BOARD:
	  text_escape (cur->s2);
	  break;

	case AC_CAT_EVENT:
	  text_escape (cur->s1);
	  break;

	default:
	  assert (0);
	}
	
      {
	struct tm date;

	if (datetime_parse (cur->date, 0, 0, &date))
	  {
	    const char *format_string;

	    if (date.tm_year == cur_tm.tm_year)
	      format_string = "%b %d %H:%M";
	    else
	      format_string = "%b %d %H:%M %Y";
	  
	    html ("\n\tat %s\n",
		  datetime_format (format_string, &date));
	  }
	else
	  html ("\n\tat %s\n", cur->date);
      }

      text_hr (html_buffer);
    }
}

/* scope-db sync-mail. */

/* Users can request that new messages in message boards be emailed to
   them.  `scope-db sync-mail', which this function implements, should
   be executed often (every 1-10 minutes) in order to do this. */
void
cmd_sync_mail (int argc unused, char *argv[] unused)
{
  db_result *r;
  size_t i, count;
  struct context cx;
  struct st mail;
  int seq;

  connect_db ();

  /* Get a list of the messages to send. */
  db_query ("SELECT notes.seq, users.usr, notes.grp "
	    "FROM notes, users, sync_mail "
	    "WHERE notes.seq > sync_mail.seq "
	    "AND emailbody = 2"
	    "AND notes.grp IN "
	    "(SELECT grp FROM user_group WHERE user_group.usr = users.usr) "
	    "ORDER by users.usr",
	    DB_TUPLES, &r);
  seq = get_seq ();

  /* Iterate across the messages to send. */
  st_init (&mail, 1024);
  count = dbr_nrows (r);
  for (i = 0; i < count; i++)
    {
      int mail_seq = stoi (dbr_value (r, i, 0));
      char *usr = dbr_value (r, i, 1);
      char *grp = dbr_value (r, i, 2);

      /* We grab a user context for the first message a user is sent
	 in a particular run, and retain that user context while the
	 same user is being sent messages.  Unfortunately I doubt
	 whether this is a useful optimization in practice: since this
	 function is run so often, typically few messages will pass
	 through in any one run. */
      int new_usr = i == 0 || strcmp (usr, dbr_value (r, i - 1, 1));

      if (new_usr)
	{
	  if (i != 0)
	    destroy_context (&cx);
	  if (!get_user_context (&cx, usr, grp))
	    {
	      destroy_context (&cx);
	      while (i < count - 1 && !strcmp (usr, dbr_value (r, i + 1, 1)))
		i++;
	      continue;
	    }
	}

      st_clear (&mail);
      if (compose_rfc822_message (&cx, &mail, mail_seq,
				  stoi (std_get (cx.options, "emailfmt"))))
	send_mail (st_value (&mail));

      if (new_usr)
	db_query ("UPDATE user_group SET last_read_mail = %d "
		  "WHERE usr = '%sq'", DB_FORMAT, NULL, seq, usr);
    }
  if (count != 0)
    destroy_context (&cx);
  st_destroy (&mail);
  
  db_query ("UPDATE sync_mail SET seq = %d", DB_FORMAT, NULL, seq);
}

/* Composes an RFC822-conformant message for user CX into MAIL from
   the message in the `notes' table having sequence number SEQ.
   FORMAT is as passed to send_update_mail(). */
static int
compose_rfc822_message (struct context *cx,
			struct st *mail, int seq, int format)
{
  struct part part[2];
  int n_parts;
  
  const char *file, *usr, *content, *title, *date, *thread;
  const char *mail_domain;
  
  db_result *r;

  db_query ("SELECT T1.string, usr, content, T2.string, date, thread, file, title "
	    "FROM notes, strings T1, strings T2 "
	    "WHERE notes.seq = %d AND T1.seq = notes.file "
	    "AND T2.seq = notes.title AND notes.grp = '%sq'",
	    DB_FORMAT | DB_TUPLES, &r, seq, cx->grp);
  if (dbr_nrows (r) < 1)
    {
      dbr_destroy (r);
      return 0;
    }

  file = dbr_value (r, 0, 0);
  usr = dbr_value (r, 0, 1);
  content = dbr_value (r, 0, 2);
  title = dbr_value (r, 0, 3);
  date = dbr_value (r, 0, 4);
  thread = dbr_value (r, 0, 5);
  
  log_activity (cx, stoi (dbr_value (r, 0, 6)), stoi (dbr_value (r, 0, 7)),
		strncmp (file, "//", 2) ? AC_CAT_FILE : AC_CAT_MSG_BOARD,
		strncmp (file, "//", 2) ? AC_FILE_POST_MAILED : AC_MSG_MAILED,
		seq);

  mail_domain = std_get (conf, "mail_domain");

  html_push_buffer (mail);
  st_printf (mail,
	     "From: %s <scope-gw@%s>\n"
	     "To: %s <%s@%s>\n"
	     "Subject: %s\n"
	     "Message-ID: <%d-scope@%s>\n"
	     "Reply-To: TeamSCOPE Gateway <scope-gw@%s>\n"
	     "Organization: %s",
	     username_to_fullname (cx->wg, usr), mail_domain,
	     username_to_fullname (cx->wg, cx->usr), cx->usr, mail_domain,
	     title,
	     seq, mail_domain,
	     mail_domain,
	     cx->grp);
  if (strncmp (file, "//", 2))
    {
      st_concat (mail, " (");
      text_file (cx, file);
      st_putchar (mail, ')');
    }
  st_putchar (mail, '\n');
  html_pop_buffer ();

  /* In-Reply-To: and References: headers. */
  {
    const char *cp;
    int x = -1;
    
    for (cp = thread; *cp; )
      {
	if (!isdigit ((unsigned char) *cp))
	  {
	    cp++;
	    continue;
	  }

	if (x == -1)
	  st_concat (mail, "References:");
	x = strtol (cp, (char **) &cp, 10);
	st_printf (mail, " <%d-scope@%s>", x, mail_domain);
	if (x == seq)
	  break;
      }

    if (x != -1)
      st_printf (mail, "\nIn-Reply-To: <%d-scope@%s>\n", x, mail_domain);
  }

  n_parts = 0;

  if (format != 1)
    {
      struct st text;

      st_init (&text, 1024);
      html_push_buffer (&text);
      st_printf (&text,
		 "-*- This is a TeamSCOPE message posting from %s\n"
		 "-*- within team %s",
		 std_get (conf, "server_name"), cx->grp);
      if (strncmp (file, "//", 2))
	{
	  st_concat (&text, "\n-*- for file ");
	  text_file (cx, file);
	}
      st_concat (&text, ".\n");
      html_pop_buffer ();

      text_fill (&text, content);
      st_putchar (&text, '\n');

      part[n_parts].type.name = "text/plain";
      part[n_parts].type.boundary = -1;
      part[n_parts].content = st_value (&text);
      n_parts++;
    }

  if (format != 0)
    {
      struct st h;

      st_init (&h, 1024);
      html_push_buffer (&h);
      st_concat (&h, "<HTML><HEAD>\n<TITLE>");
      html_pretty_print (title);
      st_printf (&h, "</TITLE>\n<BASE HREF=\"%s\">\n</HEAD><BODY>\n",
		 our_abs_url);

      html ("<BLOCKQUOTE>\n"
	    "This is a TeamSCOPE message posting from <STRONG>%s</STRONG>\n"
	    "within team <STRONG>%s</STRONG>",
	    std_get (conf, "server_name"), cx->grp);
      if (strncmp (file, "//", 2))
	{
	  struct std *extra = std_create (0);
	  std_put (extra, "needgrp", cx->grp);

	  htmls ("\nfor file ");
	  html_a_file (cx, file, extra);

	  std_destroy (extra);
	}
      htmls (".\n"
	     "</BLOCKQUOTE>\n\n");

      html_pretty_print (content);
      st_concat (&h, "\n</BODY></HTML>\n");
      html_pop_buffer ();
      
      part[n_parts].type.name = "text/html";
      part[n_parts].type.boundary = -1;
      part[n_parts].content = st_value (&h);
      n_parts++;
    }
  
  {
    static const struct mime_type type = {"multipart/alternative", 3};
    make_mime_mail (cx, mail, NULL, &type, part, n_parts);
  }
  
  {
    int i;

    for (i = 0; i < n_parts; i++)
      free (part[i].content);
  }
  dbr_destroy (r);

  return 1;
}

/* scope-db received-mail. */

/* Users can follow up to messages mailed to them by sync-mail or
   hourly-mail.  If they do, then `scope-db received-mail' is called,
   running as user `scope', with the message waiting on stdin. */
void
cmd_received_mail (int argc unused, char *argv[] unused)
{
  struct std *headers;
  struct message msg;
  struct context *cx;
  struct st hdr_txt;

  msg.title = NULL;
  msg.usr = NULL;
  msg.followup = 0;
  msg.grp = NULL;
  msg.seq = 0;
  msg.file = 0;
  msg.thread = NULL;
  msg.content = NULL;
  cx = NULL;
  st_init (&hdr_txt, 512);

  if (!db_open_p ())
    db_open (NULL);

#if 0
  open_log_file ("/tmp/foo");
#endif
  headers = parse_headers (&hdr_txt);
  msg.content = read_content (headers, st_value (&hdr_txt));
  if (msg.content == NULL)
    goto done;

  /* Determine title, usr, followup. */
  msg.title = std_get (headers, "subject");
  if (!determine_usr (&msg, headers, st_value (&hdr_txt))
      || !determine_followup (&msg, headers, st_value (&hdr_txt)))
    goto done;

  /* Determine grp, seq, file, thread. */
  {
    db_result *r;
    
    db_query ("SELECT grp, file, thread FROM notes WHERE notes.seq = %d",
	      DB_FORMAT | DB_TUPLES, &r, msg.followup);
    if (dbr_nrows (r) == 0)
      {
	send_reply (headers, st_value (&hdr_txt), NULL,
		    "The message being followed up, article number %d, "
		    "is not present in the TeamSCOPE database.  This "
		    "means that the message never existed or has been "
		    "deleted by its author.",
		    msg.followup);
	dbr_destroy (r);
	goto done;
      }
    msg.grp = xstrdup (dbr_value (r, 0, 0));
    msg.file = stoi (dbr_value (r, 0, 1));
    msg.seq = get_seq ();
    
    {
      const char *p = dbr_value (r, 0, 2);
      assert (*p == '{' && p[strlen (p) - 1] == '}');
	
      msg.thread = xmalloc (strlen (p) + INT_DIGITS + 8);
      strcpy (msg.thread, p);
      sprintf (&msg.thread[strlen (p) - 1], ",%d}", msg.seq);
    }
    
    dbr_destroy (r);
  }

  /* Verify that usr belongs to grp. */
  {
    db_result *r;
    int nrows;
    db_query ("SELECT usr, grp FROM user_group "
	      "WHERE usr = '%sq' AND grp = '%sq'",
	      DB_FORMAT | DB_TUPLES, &r, msg.usr, msg.grp);
    nrows = dbr_nrows (r);
    dbr_destroy (r);
    if (!nrows)
      {
	send_reply (headers, st_value (&hdr_txt), NULL,
		    "Your email headers identify you as user `%s'.  However, "
		    "the message being followed up is owned by group `%s', "
		    "of which you are not a member.  You are not allowed to "
		    "post to groups to which you do not belong.",
		    msg.usr, msg.grp);
	goto done;
      }
  }

  /* Grab a user context. */
  cx = xmalloc (sizeof *cx);
  if (!get_user_context (cx, msg.usr, msg.grp))
    {
      send_reply (headers, st_value (&hdr_txt), NULL,
		  "An internal error occurred while processing your email.  "
		  "If the error persists, please contact the TeamSCOPE "
		  "administrator by replying to this message, enclosing "
		  "this message's complete body and headers.  Sorry for "
		  "the inconvenience.");
      goto done;
    }

  /* Post the message. */
  {
    int title_seq = get_file_seq (cx, msg.title);
    
    db_query ("INSERT INTO notes VALUES "
	      "(%d, '%sq', '%sq', '%sq', %d, datetime (timenow ()), '%sq', %d)",
	      DB_FORMAT, NULL, msg.file, msg.grp, msg.usr, msg.content,
	      title_seq, msg.thread, msg.seq);
  }

  /* Send verification reply. */
  {
    char *subject;
    subject = blp_asprintf ("Message posted in group %s (was: %s)",
			    msg.grp, std_get (headers, "subject"));
    send_reply (headers, st_value (&hdr_txt), subject,
		"Your message titled \"%s\" has been posted in group %s "
		"as user %s, with the following content:\n\n"
		"-------------------------------------------------------\n"
		"%s"
		"-------------------------------------------------------\n",
		msg.title, msg.grp, msg.usr, msg.content);
    free (subject);
  }

  /* Clean up. */
 done:
  if (cx != NULL)
    {
      destroy_context (cx);
      free (cx);
    }
  free (msg.usr);
  free (msg.grp);
  free (msg.thread);
  free (msg.content);
  st_destroy (&hdr_txt);
  std_destroy (headers);
  db_close ();
  return;
}

/* Parse RFC822 message headers from stdin.  Headers are stored into a
   dictionary based on field name, which is returned.  Multiple
   headers of the same type are concatenated separated by commas; this
   appears to actually be RFC822-compliant, believe it or not.
   Headers are also stored in their original form into TEXT. */
static struct std *
parse_headers (struct st *text)
{
  struct std *headers;
  struct st key, value;

  headers = std_create (0);
  st_init (&key, 32);
  st_init (&value, 128);
  for (;;)
    {
      st_clear (&key);
      st_clear (&value);

      /* Check for end of headers. */
      {
	int c = getchar ();
	if (c == '\n')
	  break;
	ungetc (c, stdin);
      }

      /* Parse field name. */
      for (;;)
	{
	  int c = getchar ();
	  st_putchar (text, c);
	  switch (c)
	    {
	    case EOF:
	    case '\n':
	      goto next_field;

	    case ':':
	      goto end_of_field_name;

	    default:
	      st_putchar (&key, tolower ((unsigned char) c));
	      break;
	    }
	}
    end_of_field_name:;

      /* Parse field content. */
      for (;;)
	{
	  int c = getchar ();
	  st_putchar (text, c);
	  switch (c)
	    {
	    case EOF:
	      goto end_of_field;

	    case '\n':
	      c = getchar ();
	      if (c == ' ' || c == '\t')
		{
		  st_putchar (text, c);
		  break;
		}
	      ungetc (c, stdin);
	      goto end_of_field;

	    case ' ':
	    case '\t':
	    case '\r':
	      /* Skip leading spaces. */
	      if (st_length (&value) == 0)
		break;
	      /* Fall through. */

	    default:
	      st_putchar (&value, c);
	      break;
	    }
	}
    end_of_field:;

      /* Insert parsed fields into headers dictionary. */
      if (!std_contains (headers, st_value (&key)))
	std_put (headers, st_value (&key), st_value (&value));
      else
	{
	  st_printf (&value, ", %s", std_get (headers, st_value (&key)));
	  std_replace (headers, st_value (&key), st_value (&value));
	}

    next_field:;
    }

  st_destroy (&value);
  st_destroy (&key);

  return headers;
}

/* Reads the body of an RFC822 message from stdin and attempts to
   extract the interesting parts.  HEADERS are the headers from the
   message as read by parse_headers().

   The basic strategy is to read text/plain parts, skipping other
   parts.  In fact the default Content-Type: for multipart/alternative
   *is* text/plain, but I'm too lazy to fix this until someone
   actually complains that it rejects their mail wrongly <g,d&r>. */
static char *
read_content (struct std *headers, const char *hdr_txt)
{
  const char *ct = std_get (headers, "content-type");
  char *boundary;
  size_t boundary_len;

  if (content_type_eq (ct, "") || content_type_eq (ct, "text/plain"))
    return read_until (NULL);

  if (!content_type_eq (ct, "multipart/alternative"))
    {
      const char *end = strchr (ct, ';');
      if (end == NULL)
	end = strchr (ct, '\0');
      
      send_reply (headers, hdr_txt, NULL,
		  "Email contents are in an unknown format \"%.*s\".  "
		  "TeamSCOPE messages must be in plain text format.  "
		  "If your email client is configured to send email in "
		  "HTML format, please reconfigure it to also send "
		  "plain text within the same email.", end - ct, ct);
      return NULL;
    }

  boundary = parse_boundary (ct);
  if (boundary == NULL)
    {
      send_reply (headers, hdr_txt, NULL,
		  "The MIME RFC 2045 Content-Type: line supplied by "
		  "your mail client does not contain a valid "
		  "boundary string.  Please reconfigure your mail "
		  "client to send emails as plain text.");
      return NULL;
    }
  boundary_len = strlen (boundary);
  
  {
    char *line = NULL;
    size_t size = 0;
    int state = 0;
    
    for (;;)
      {
	long len = getline (&line, &size, stdin);
	if (len == -1)
	  break;

	state &= 0xf;
	if (!strncmp (line, boundary, boundary_len))
	  {
	    if (boundary[boundary_len] == '-'
		&& boundary[boundary_len + 1] == '-')
	      break;

	    state |= 0x10;
	  }

	switch (state)
	  {
	  case 0x00:
	    /* We are in the preamble or some part that we're not
               interested in. */
	    break;

	  case 0x10:
	    /* We are in the preamble or some part that we're not
               interested in, but we hit a boundary. */
	    state = 1;
	    break;
	    
	  case 0x01:
	    /* We are in a header of one of the parts.
	       Check for an interesting header or end-of-headers. */
	    if (!strncasecmp (line, "Content-Type:", strlen ("Content-Type:"))
		&& content_type_eq (line + strlen ("Content-Type:"),
				    "text/plain"))
	      state = 2;
	    else if (empty (line))
	      state = 1;
	    break;

	  case 0x11:
	    /* We hit a boundary in a header of one of the parts.
	       Weird.  Nothing to do, anyhow. */
	    break;

	  case 0x02:
	    /* We're in the headers of a text/plain part.
	       We're not interested in any other headers. */
	    if (empty (line))
	      {
		char *content = read_until (boundary);
		free (line);
		free (boundary);
		return content;
	      }
	    break;

	  case 0x12:
	    /* We're hit a boundary in a text/plain header.
	       Weird.  Start over on parsing headers. */
	    state = 1;
	    break;

	  default:
	    abort ();
	  }
      }

    free (line);
  }

  free (boundary);

  send_reply (headers, hdr_txt, NULL,
	      "Your mail contained multiple parts, but none of these "
	      "parts was in plain text format.  TeamSCOPE can only "
	      "accept plain text for posting.  Please reconfigure "
	      "your mail client to send a plain text version of "
	      "outgoing emails.");
  return NULL;	    
}

/* Reads and accumulates text from stdin until it encounters: (1) end
   of file, (2) a signature separator "--" or "-- ", or (3) boundary
   line specified by BOUNDARY if BOUNDARY is non-null. */
static char *
read_until (const char *boundary)
{
  char *line;
  size_t size;

  size_t boundary_len;

  struct st s;

  int blank;

  line = NULL;
  size = 0;
  if (boundary != NULL)
    boundary_len = strlen (boundary);
  st_init (&s, 512);
  blank = 0;

  for (;;)
    {
      long len = getline (&line, &size, stdin);
      if (len == -1)
	break;

      blp_strip_trailing_spaces (line);
      if (!strcmp (line, "--")
	  || (boundary && !strncmp (boundary, line, boundary_len)))
	break;

      if (strstr (line, "-*-") || empty (line))
	{
	  blank = 1;
	  continue;
	}

      if (blank)
	{
	  if (st_length (&s) > 0)
	    st_putchar (&s, '\n');
	  blank = 0;
	}

      st_concat (&s, line);
      st_putchar (&s, '\n');
    }
  free (line);

  return st_value (&s);
}

/* Text handling. */

/* Max width of a text line. */
#define TEXT_LINE_WIDTH 65	/* As specified by RFC 1855. */

/* Fills paragraphs to width TEXT_LINE_WIDTH from SRC into DST. */
static void
text_fill (struct st *dst, const char *src)
{
  while (*src != '\0')
    {
      size_t ll;
      
      /* Indented lines are not filled, unless the first
	 non-whitespace character is punctuation. */
      if (*src == ' ' || *src == '\t' || *src == '\n' || *src == '\r')
	{
	  while (*src == ' ' || *src == '\t')
	    st_putchar (dst, *src++);

	  if (isalnum ((unsigned char) *src))
	    {
	      while (*src && *src != '\n' && *src != '\r')
		st_putchar (dst, *src++);
	      while (*src == '\r')
		src++;
	      if (*src == '\n')
		st_putchar (dst, *src++);
	      continue;
	    }
	}
      
      ll = TEXT_LINE_WIDTH;
      for (;;)
	{
	  size_t wl = strcspn (src, " \t\r\n");
	  if (wl > TEXT_LINE_WIDTH)
	    {
	      st_concat_buffer (dst, src, ll);
	      src += ll;
	      break;
	    }
	  else if (wl == 0 || wl > ll)
	    break;

	  st_concat_buffer (dst, src, wl);
	  src += wl;
	  ll -= wl;

	  for (; *src == ' ' || *src == '\t' || *src == '\r'; src++)
	    if (ll > 0)
	      {
		st_putchar (dst, *src);
		ll--;
	      }

	  if (*src == '\n' || ll == 0)
	    break;
	}

      while (*src == '\r')
	src++;
      if (*src == '\n')
	src++;
      st_putchar (dst, '\n');
    }
}

/* Draws a horzizontal rule of width TEXT_LINE_WIDTH in S. */
static void
text_hr (struct st *s)
{
  st_concat_dup (s, '_', TEXT_LINE_WIDTH);
  st_concat (s, "\n\n");
}

/* Centers TEXT within a line of width TEXT_LINE_WIDTH in S. */
static void
text_center_line (struct st *s, const char *text)
{
  st_concat_dup (s, ' ', (TEXT_LINE_WIDTH - strlen (text)) / 2);
  st_concat (s, text);
  st_putchar (s, '\n');
}

/* Text equivalent of html_escape(); i.e., it attempts to write the
   contents of TEXT without breaking any plain text conventions, by
   escaping high-bit characters and so on. */
static void
text_escape (const char *text)
{
  const char *cp;
  
  for (cp = text; ; cp++)
    if (*cp == '\0')
      {
	htmls (text);
	return;
      }
    else if (!isgraph ((unsigned char) *cp))
      break;

  htmlc ('"');
  for (cp = text; *cp; cp++)
    if (isprint ((unsigned char) *cp))
      htmlc (*cp);
    else
      html ("\\%03o", (unsigned char) *cp);
  htmlc ('"');
}

/* Text equivalent of html_file(). */
static void
text_file (struct context *cx, const char *filename)
{
  if (has_prefix (filename, cx->shared))
    {
      filename += strlen (cx->shared);
      if (*filename == '/')
	filename++;
      if (*filename == '\0')
	html ("(%s)", _("shared folder"));
    }
  else if (has_prefix (filename, cx->home))
    {
      filename += strlen (cx->home);
      if (*filename == '/')
	filename++;
      html ("(%s) ", _("personal folder"));
    }
  text_escape (filename);
}

/* Mail and MIME composition. */

/* Makes a unique MIME subpart boundary within BOUNDARY for user CX
   and unique integer N. */
static void
make_boundary (struct context *cx, char boundary[128], int n)
{
  sprintf (boundary, ":%s:%s:%d:--", cx->usr, cx->grp, n);
}

/* Prints MIME Content-Type: header for TYPE into S for user CX. */
static void
print_mime_type (struct context *cx,
		 struct st *s, const struct mime_type *type)
{
  if (type->boundary != -1)
    {
      char boundary[128];

      make_boundary (cx, boundary, type->boundary);
      st_printf (s, "Content-Type: %s; boundary=\"%s\"\n",
		 type->name, boundary);
    }
  else
    st_printf (s, "Content-Type: %s\n", type->name);
}

/* Constructs a MIME RFC2045 email for user CX in MAIL.  The email has
   N_PARTS parts in PART[].  If N_PARTS > 1 then MULTIPART_TYPE is the
   MIME type that should encapsulate PARTS[].  OVERALL_TYPE receives
   the type of the constructed message, either MULTIPART_TYPE or
   PART[0]. */
static void
make_mime_mail (struct context *cx, struct st *mail,
		struct mime_type *overall_type,
		const struct mime_type *multipart_type,
		const struct part part[], int n_parts)
{
  char boundary[128];

  int i;

  assert (n_parts > 0);
  
  make_boundary (cx, boundary, multipart_type->boundary);
  if (overall_type != NULL)
    {
      if (n_parts != 1)
	*overall_type = *multipart_type;
      else
	{
	  overall_type->name = part[0].type.name;
	  overall_type->boundary = -1;
	}
    }
  else
    {
      st_concat (mail,
		 "MIME-Version: 1.0\n"
		 "Content-Transfer-Encoding: 7bit\n");

      if (n_parts > 1)
	print_mime_type (cx, mail, multipart_type);
      else
	print_mime_type (cx, mail, &part[0].type);

      st_concat (mail, "\nThis is a multipart message in MIME format.\n");
    }

  for (i = 0; i < n_parts; i++)
    {
      if (n_parts > 1)
	{
	  st_printf (mail, "\n--%s\n", boundary);
	  print_mime_type (cx, mail, &part[i].type);
	  st_putchar (mail, '\n');
	}
	  
      st_concat (mail, part[i].content);
    }

  if (n_parts > 1)
    st_printf (mail, "\n--%s--\n", boundary);
}

/* Sends out MAIL, which must be an RFC822-compliant email message. */
static void
send_mail (const char *mail)
{
  static const char command[] = "sendmail -bm -t";
  FILE *metasend = open_pipe (command, "w");
  fputs (mail, metasend);
  close_pipe (metasend, command);
}

/* Constructs user context CX for user USR within team GRP. */
static int 
get_user_context (struct context *cx, const char *usr, const char *grp)
{
  db_result *r;
  
  init_context (cx);
  
  strncpy (cx->usr, usr, 8);
  cx->usr[8] = '\0';

  strncpy (cx->grp, grp, 8);
  cx->grp[8] = '\0';

  /* Get group's shared directory, web directory. */
  db_query ("SELECT shareddir, webdir FROM groups WHERE grp = '%sq'",
	    DB_FORMAT | DB_TUPLES, &r, grp);
  if (dbr_nrows (r) != 1)
    {
      dbr_destroy (r);
      return 0;
    }
  cx->shared = xstrdup (dbr_value (r, 0, 0));
  cx->web = xstrdup (dbr_value (r, 0, 1));
  dbr_destroy (r);

  /* Get user's home directory, UID. */
  {
    struct passwd *pwd = getpwnam (usr);
    if (pwd == NULL)
      {
	log ("can't find UNIX user %s", usr);
	return 0;
      }
    cx->home = xstrdup (pwd->pw_dir);
    cx->uid = pwd->pw_uid;
  }
  
  /* Read user's .scope/options file. */
  {
    char *options_file = blp_asprintf ("%s/.scope/options", cx->home);
    cx->options = std_create (1);
    if (file_exists (options_file))
      std_read (cx->options, options_file, 0, cx->uid);
    free (options_file);
  }

  read_user_list (&cx->wg, cx->grp);

  return 1;
}

/* Attempt to parse the Content-Type: header as supplied in
   CONTENT_TYPE to determine the MIME multipart boundary.  If
   successful, returns the boundary as a malloc()'d string; otherwise,
   returns a null pointer.  */
static char *
parse_boundary (const char *content_type)
{
  /* FIXME: This is *not* fully RFC 2045 compliant.  However, I
     suspect that no one else is, either, so it's not a big deal. */

  const char *boundary_beg;
  char *boundary;

  boundary_beg = blp_strcasestr (content_type, "boundary=");
  if (boundary_beg)
    {
      size_t boundary_len;
      
      boundary_beg += strlen ("boundary=");
      if (*boundary_beg == '"')
	boundary_beg++;
      
      boundary_len = strcspn (boundary_beg, " \t\r\n\"");
      boundary = xmalloc (2 + boundary_len + 1);
      boundary[0] = boundary[1] = '-';
      memcpy (&boundary[2], boundary_beg, boundary_len);
      boundary[boundary_len + 2] = '\0';
    }
  else
    boundary = NULL;
  
  return boundary;
}

/* Compares two Content-Type: strings A and B, possibly prefaced by
   spaces, for equality.  Returns nonzero iff A and B are equal as
   content types. */
static int
content_type_eq (const char *a, const char *b)
{
  while (isspace ((unsigned char) *a))
    a++;

  while (isspace ((unsigned char) *b))
    b++;
  
  for (;;)
    {
      int ae = isspace ((unsigned char) *a) || *a == ';' || *a == '\0';
      int be = isspace ((unsigned char) *b) || *b == ';' || *b == '\0';
      if (ae ^ be)
	return 0;
      else if (ae && be)
	return 1;
      else if (tolower ((unsigned char) *a) != tolower ((unsigned char) *b))
	return 0;

      a++;
      b++;
    }
}

/* Attempts to extract the first email address into EMAIL from a
   string, beginning at *P.  Stores the next start location into *P.
   The string starts at BEG.  Returns nonzero only if an email address
   was successfully extracted.  Downcases the domain-part of the email
   address. */
static int 
extract_email (const char *beg, const char **p, struct st *email)
{
  const char *b, *e;
  
  /* FIXME: Ohmigod this is *so* RFC822 nonconformant. */
  *p = strchr (*p, '@');
  if (*p == NULL)
    return 0;

  for (b = *p; b > beg; b--)
    if (strchr (" ,'\"<:", b[-1]))
      break;
  for (e = *p; *e; e++)
    if (strchr (" ,'\">;", *e))
      break;

  st_clear (email);
  st_concat_buffer (email, b, *p - b);
  for (b = *p; b < e; b++)
    st_putchar (email, tolower ((unsigned char) *b));

  *p = e;
  return 1;
}

/* Case-insensitively compares strings A and B.  Returns zero if A and
   B are equal, nonzero otherwise. */
static int
st_casecmp (const char *a, const char *b)
{
  for (;;)
    {
      int ca = tolower ((unsigned char) *a++);
      int cb = tolower ((unsigned char) *b++);
      if (ca != cb)
	return ca - cb;
      if (ca == '\0')
	return 0;
    }
}

/* Attempts to match email address EMAIL with a username.  If
   successful fills in MSG->usr and returns nonzero. */
static int
determine_usr_email (struct message *msg, const char *email)
{
  /* An email address of foo@our.mail.domain is user `foo'. */
  {
    const char *domain = strchr (email, '@');
    assert (domain != NULL);
    domain++;

    if (!st_casecmp (domain, std_get (conf, "mail_domain")))
      {
	msg->usr = malloc (domain - email);
	memcpy (msg->usr, email, domain - email - 1);
	msg->usr[domain - email - 1] = '\0';
	return 1;
      }
  }

  /* Otherwise attempt to match it against the database. */
  {
    db_result *r;
    
    db_query ("SELECT * FROM users WHERE email = '%sq'",
	      DB_FORMAT | DB_TUPLES, &r, email);
    if (dbr_nrows (r))
      msg->usr = xstrdup (dbr_value (r, 0, 0));
    dbr_destroy (r);
  }

  return msg->usr != NULL;
}
	
/* Attempts to determine the user who sent the message we're
   processing, by running extract_email() and determine_usr_mail() on
   the From: and Reply-To: headers.  Returns nonzero only if
   successful. */
static int
determine_usr (struct message *msg, struct std *headers, const char *hdr_txt)
{
  static const char *header_name[] = {"from", "reply-to"};
  static const char *proper_name[] = {"From", "Reply-To"};
  struct st email;
  size_t i;

  st_init (&email, 32);
  for (i = 0; i < sizeof header_name / sizeof *header_name; i++)
    {
      const char *beg = std_get (headers, header_name[i]);
      const char *p = beg;
      while (extract_email (beg, &p, &email))
	{
	  if (determine_usr_email (msg, st_value (&email)))
	    {
	      st_destroy (&email);
	      return 1;
	    }
	}
    }
  st_destroy (&email);
  
  {
    struct st s;

    st_init (&s, 256);
    st_printf (&s, "None of the email addresses in the headers were "
	       "recognized as valid TeamSCOPE users.  To correct this "
	       "problem, set your Reply-To: and/or From: headers to "
	       "an email address known by TeamSCOPE.  This can be "
	       "either your TeamSCOPE email address, of the form "
	       "<username>@%s, or your upstream email address as "
	       "registered with your TeamSCOPE administrator.\n\n",
	       std_get (conf, "mail_domain"));

    for (i = 0; i < sizeof header_name / sizeof *header_name; i++)
      {
	const char *name = header_name[i];
	const char *proper = proper_name[i];
	
	if (std_contains (headers, name))
	  st_printf (&s, "Your %s: header was set to:\n\t%s\n",
		     proper, std_get (headers, name));
	else
	  st_printf (&s, "Your %s: header was unset.\n", proper);
      }
    
    send_reply (headers, hdr_txt, NULL, st_value (&s));

    st_destroy (&s);
  }

  return 0;
}

/* Attempts to extract a TeamSCOPE message sequence number from a
   string that putatively contains a message ID. */
static int
extract_message_id (const char *msg_id)
{
  const char *p = strstr (msg_id, "-scope@");
  if (p == NULL)
    return 0;
  
  while (p > msg_id && isdigit ((unsigned char) p[-1]))
    p--;
  return stoi (p);
}

/* Attempts to determine which message it is that the message we're
   currently processing follows up, first by trying for an X-Followup:
   header, then by looking at In-Reply-To: and References:.  Returns
   nonzero only if successful.  If successful sets MSG->followup.
   HEADERS contains the headers of the message being processed. */
static int
determine_followup (struct message *msg,
		    struct std *headers, const char *hdr_txt)
{
  if (std_contains (headers, "x-followup"))
    {
      const char *fu = std_get (headers, "x-followup"); 
      msg->followup = stoi (fu);
      if (msg->followup == 0)
	msg->followup = extract_message_id (fu);
      if (msg->followup == 0)
	{
	  send_reply (headers, hdr_txt, NULL,
		      "The X-Followup: header supplied could not "
		      "be parsed.  If supplied, X-Followup: must "
		      "consist either of a bare article number "
		      "or the Message-ID: from the article being "
		      "replied to, of the form <#-scope@%s>, where "
		      "# is an article number.\n\n"
		      "The X-Followup: that you supplied was \"%s\"\n",
		      std_get (conf, "mail_domain"),
		      std_get (headers, "x-followup"));
	  return 0;
	}
    }
  
  {
    static const char *header_name[] = {"in-reply-to", "references"};
    struct st email;
    size_t i;

    st_init (&email, 32);
    for (i = 0; i < sizeof header_name / sizeof *header_name; i++)
      {
	const char *beg = std_get (headers, header_name[i]);
	const char *p = beg;
	
	while (extract_email (beg, &p, &email))
	  {
	    int x = extract_message_id (st_value (&email));
	    if (x > msg->followup)
	      msg->followup = x;
	  }
      }
    st_destroy (&email);
  }

  if (msg->followup)
    return 1;

  send_reply (headers, hdr_txt, NULL,
	      "TeamSCOPE was unable to determine from your email "
	      "which message you were attempting to follow up.  "
	      "(Good email clients normally supply this "
	      "information through the In-Reply-To: header, set to "
	      "\"%s\" in your email.)  To inform TeamSCOPE of the "
	      "message being followed up to, add a header, X-Followup:, "
	      "to your email.  The contents of this header should be "
	      "the Message-ID: from the email corresponding to the "
	      "message to which you are replying.  Overall, it should look "
	      "like this:\n"
	      "\tX-Followup: <#-scope@%s>\n"
	      "where # is an article number.\n",
	      std_get (headers, "in-reply-to"),
	      std_get (conf, "mail_domain"));

  return 0;
}

/* Sends an email reply to the originator of the message being
   processed, which has headers given by HEADERS.  The subject of the
   reply is SUBJECT and its contents are CONTENT, which latter is
   formatted as with printf().  If SUBJECT is a null pointer then the
   subject is made an error message and error boilerplate is added to
   the email content.  In any case some generic boilerplate is added
   too. */
static void
send_reply (struct std *headers, const char *hdr_txt,
	    const char *subject, const char *content, ...)
{
  struct st reply;
  int error;

  error = subject == NULL;
  if (error)
    subject = blp_asprintf ("TeamSCOPE mail processing error (was: %s)",
			    std_get (headers, "subject"));

  st_init (&reply, 1024);
  st_printf (&reply,
	     "To: %s\n"
	     "From: TeamSCOPE Administrator <scope@%s>\n"
	     "Sender: TeamSCOPE Email Gateway Daemon <scope-gw@%s>\n",
	     (std_contains (headers, "reply-to")
	      ? std_get (headers, "reply-to")
	      : std_get (headers, "from")),
	     std_get (conf, "mail_domain"),
	     std_get (conf, "mail_domain"));
  if (std_contains (headers, "message-id"))
    st_printf (&reply, "In-Reply-To: %s\n", std_get (headers, "message-id"));
  st_printf (&reply, "Subject: %s\n\n", subject);

  if (error)
    {
      char *s;
      s = blp_asprintf ("An error occurred while processing your mail dated "
			"\"%s\" with subject \"%s\".  Your message was not "
			"posted.  To post it, you must resubmit it after "
			"correcting the problem described below.\n\n",
			std_get (headers, "date"),
			std_get (headers, "subject"));
      text_fill (&reply, s);
      free (s);
    }
  
  {
    int size = 64;
  
    for (;;)
      {
	va_list args;
	char *buf;
	int n;
      
	buf = xmalloc (size);

	va_start (args, content);
	n = vsnprintf (buf, size, content, args);
	va_end (args);
      
	if (n < 0)
	  size *= 2;
	else if (n < size)
	  {
	    text_fill (&reply, buf);
	    free (buf);
	    break;
	  }
	else
	  size = n + 1;
	free (buf);
      }
  }
  
  text_fill (&reply,
	     "\nThis email was sent by the TeamSCOPE email "
	     "gateway daemon.  Replies to this message will go "
	     "to a human, the TeamSCOPE administrator at your site.  "
	     "Consequently, please do not reply to this message "
	     "unless you wish to report a bug or misconfiguration "
	     "in TeamSCOPE.\n");

  if (error)
    {
      text_fill (&reply,
		 "\n"
		 "For your reference, here are the complete headers "
		 "from your email.  Please quote this entire message, "
		 "including all of these headers, in any "
		 "correspondence regarding this error reply.\n\n");

      st_concat (&reply, hdr_txt);
    }

  send_mail (st_value (&reply));
  st_destroy (&reply);

  if (error)
    free ((char *) subject);
}

/* Event queries. */

/* Retrieve all the events between BEGIN and END and
   return the database result. */
static db_result *
calendar_query (struct context *cx,
		const struct tm *begin, const struct tm *end)
{
  char b[32], e[32];
  db_result *r;

  if (!strftime (b, 32, "%b %d %Y %H:%M:%S", begin)
      || !strftime (e, 32, "%b %d %Y %H:%M:%S", end))
    fail ("date too long");

  db_query ("SELECT start, finish, string AS title, events.seq, owner "
	    "FROM events, strings "
	    "WHERE events.grp = '%sq' AND events.title = strings.seq "
	    "AND (abstime (start) <#> abstime (start)) && ('%sq' <#> '%sq') "
	    "ORDER BY start",
	    DB_FORMAT | DB_TUPLES, &r, cx->grp, b, e);

  return r;
}

static void 
event_query_run (struct event_query *qy, struct context *cx, int len)
{
  struct tm start, finish;

  start = cur_tm;
  start.tm_min = 0;
  start.tm_sec = 0;
  mktime (&start);
  
  finish = start;
  switch (len)
    {
    case 1:
      finish.tm_hour += 24 * 7;
      break;

    case 2:
      if (start.tm_wday == 6)
	finish.tm_hour += 24 * 3;
      else
	finish.tm_hour += 24 * 2;
      break;

    case 3:
      finish.tm_hour += 24;
      break;

    case 4:
      finish.tm_hour += 12;
      break;

    case 5:
    default:
      finish.tm_hour++;
      break;
    }
  mktime (&finish);
  
  qy->result = calendar_query (cx, &start, &finish);
  qy->count = dbr_nrows (qy->result);
}

static void
event_query_destroy (struct event_query *qy)
{
  dbr_destroy (qy->result);
}

static void 
event_query_present_text (struct event_query *qy)
{
  size_t i;

  text_hr (html_buffer);
  text_center_line (html_buffer, "Upcoming Calendar Events");
  text_hr (html_buffer);
  
  for (i = 0; i < qy->count; i++)
    {
      html ("%s\n"
	    " begins: %s\n"
	    " ends:   %s\n",
	    dbr_value (qy->result, i, 2),
	    dbr_value (qy->result, i, 0),
	    dbr_value (qy->result, i, 1));

      text_hr (html_buffer);
    }
}

static void
event_query_present_html (struct event_query *qy)
{
  size_t i;
  struct url url;
  
  htmls ("<TABLE BGCOLOR=\"#FFFFFF\" WIDTH=\"100%\" BORDER=1>\n\n"
	 "<TR>\n"
	 "  <TH>Title</TH>\n"
	 "  <TH>Begins</TH>\n"
	 "  <TH>Ends</TH>\n"
	 "</TR>\n\n");

  url_create (&url, "calendar");
  url_mark (&url);
  for (i = 0; i < qy->count; i++)
    {
      url_release (&url);
      url_arg (&url, "seq", dbr_value (qy->result, i, 3));
      
      html ("<TR>\n"
	    "  <TD><A HREF=\"%s\">%s</A></TD>\n"
	    "  <TD>%s</TD>\n"
	    "  <TD>%s</TD>\n"
	    "</TR>\n\n",
	    url_string (&url), dbr_value (qy->result, i, 2),
	    dbr_value (qy->result, i, 0),
	    dbr_value (qy->result, i, 1));
    }
  url_destroy (&url);

  htmls ("</TABLE>\n");
}
