/* TeamSCOPE - Software for Collaborative Project Environments.
   Copyright (C) 1999, 2000, 2001 Michigan State University.
   Written by Ben Pfaff <pfaffben@msu.edu>.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To contact the TeamSCOPE team, send email to scope@cscw.msu.edu, or
   write to Charles Steinfield, Room 436 Communication Arts Building,
   Michigan State University, Department of Telecommunication, East
   Lansing MI 48824-1212. */

#include <config.h>
#include <assert.h>
#include <stdlib.h>
#include <time.h>
#include "alloc.h"
#include "activity.h"
#include "auth.h"
#include "cgi.h"
#include "datetime.h"
#include "db.h"
#include "error.h"
#include "html.h"
#include "identity.h"
#include "messages.h"
#include "scope.cgi.h"
#include "str.h"
#include "std.h"

/* A message. */
struct message
  {
    char *user;
    char *content;
    char *title;
    char *date;
    int nest;
    int seq;
    int thread1;
  };

db_result *retrieve_message (struct request *, int seq, struct message *msg);

static void show_messages_summary (struct request *);
static void show_message_board (struct request *, const char *filename);
static void show_post (struct request *);
static void show_single_message (struct request *, int seq,
				 const char *filename);

static void handle_postrm (struct request *);
static void handle_reset (struct request *, const char *filename);
static void submit (struct request *);

static db_result *get_message (struct context *, int seq);
static void interpret_message_result (const db_result *, int row,
				      struct message *);

static void show_message_index (struct request *,
				const db_result *, const char *filename,
				int seq_exclude, int full);
static void show_message (struct context *, const struct message *);

static int virtual_file_seq (struct context *, const char *filename);

/* Public functions. */

/* Handles the message board screen and message board subscreens under
   the file manager. */
void
scr_messages (struct request *rq)
{
  const char *filename = std_get_null (rq->cgi, "file");
  read_user_list (&rq->cx.wg, rq->cx.grp);

  if (rq->abbrev)
    {
      show_messages_summary (rq);
      return;
    }

  handle_postrm (rq);
  handle_reset (rq, filename);
  if (std_contains (rq->cgi, "seq"))
    show_single_message (rq, stoi (std_get (rq->cgi, "seq")), filename);
  else if (!strcmp (rq->op, "post"))
    show_post (rq);
  else
    show_message_board (rq, filename);
}

/* Private functions. */

/* Show a summary of the messages posted over the last week,
   for the abbreviated message board used on the team site. */
static void
show_messages_summary (struct request *rq)
{
  db_result *r;
  
  {
    int file_seq;

    {
      const size_t size = 2 + strlen (rq->cx.grp) + 1;
      char *file = local_alloc (size);
      snprintf (file, size, "//%s", rq->cx.grp);
      file_seq = get_file_seq (&rq->cx, file);
      local_free (file);
    }

    db_query ("SELECT usr, string AS title, date, notes.seq "
	      "FROM notes, strings "
	      "WHERE file = %d "
	      "AND date >= 'now'::datetime - '1 week'::timespan "
	      "AND notes.title = strings.seq "
	      "ORDER BY date DESC",
	      DB_FORMAT | DB_TUPLES, &r, file_seq);
  }
  
  {
    struct url msg_url;
    size_t n_rows;
    size_t i;
    
    url_create (&msg_url, "messages");
    url_mark (&msg_url);

    n_rows = dbr_nrows (r);
    if (n_rows > 15)
      n_rows = 15;
    
    for (i = 0; i < n_rows; i++)
      {
	struct tm when;
	
	url_release (&msg_url);
	url_arg (&msg_url, "seq", dbr_value (r, i, 3));

	datetime_parse (dbr_value (r, i, 2), 0, 0, &when);
	
	html ("<LI><A HREF=\"%s\">", url_string (&msg_url));
	html_escape (dbr_value (r, i, 1));
	html ("</A> [%s %s %s %s %s ",
	      _("posted by"),
	      username_to_fullname (rq->cx.wg, dbr_value (r, i, 0)),
	      _("on"),
	      datetime_format ("%d %b %Y", &when),
	      _("at"));
	html ("%s]\n", datetime_format ("%H:%M", &when));
      }

    if (n_rows == 0)
      html ("<LI>%s\n", _("No messages in previous week."));
    else if (dbr_nrows (r) > n_rows)
      html ("<LI>%s\n", _("(additional messages omitted for brevity)"));
  }
}
  
/* Displays the message board. */
static void
show_message_board (struct request *rq, const char *filename)
{
  struct url url;
  struct st banner;
  db_result *r;
  int last_read;
  int file_seq;

  db_begin ();

  file_seq = virtual_file_seq (&rq->cx, filename);

  last_read = 0;
  if (!std_contains (rq->cgi, "full"))
    {
      db_result *catchup_r;
    
      db_query ("SELECT last_read FROM catchup "
		"WHERE file = %d AND usr = '%sq' AND grp = '%sq'",
		DB_FORMAT | DB_TUPLES, &catchup_r, file_seq,
		rq->cx.usr, rq->cx.grp);
      if (dbr_nrows (catchup_r))
	last_read = stoi (dbr_value (catchup_r, 0, 0));
      dbr_destroy (catchup_r);
    }

  db_query ("SELECT usr, content, string, date, "
	    "_int4_depth(thread), notes.seq, notes.thread[1] AS thread1, "
	    "thread " /* We don't want thread but older Postgresqls need it
			 since we're ordering by it. */
	    "FROM notes, strings "
	    "WHERE notes.seq > %d AND file = %d "
	    "AND notes.title = strings.seq "
	    "ORDER BY thread",
	    DB_FORMAT | DB_TUPLES, &r, last_read, file_seq);

  html ("<H2>%s</H2>\n"
	"<P>\n",
	filename != NULL ? _("Comments") : _("Messages"));
  html (_("To avoid seeing the %s below again, click on\n"
	  "<STRONG>Reset</STRONG> below.  You can always see\n"
	  "a complete index by clicking on <STRONG>Full\n"
	  "Contents</STRONG>."),
	filename != NULL ? _("comments") : _("messages"));

  url_create (&url, rq->screen);
  if (filename != NULL)
    url_arg (&url, "file", filename);
  url_mark (&url);

  /* Build banner as something like this:
     [ Reset | Full Contents | Post Comment ]
     with links as appropriate. */

  /* Make reset link by computing maximum sequence number. */
  {
    int n_rows;
    int reset_seq;
    int i;

    reset_seq = 0;
    n_rows = dbr_nrows (r);
    for (i = 0; i < n_rows; i++)
      {
	int seq = stoi (dbr_value (r, i, 5));
	if (seq > reset_seq)
	  reset_seq = seq;
      }
    
    st_init (&banner, 256);
    url_arg_int (&url, "reset", reset_seq);
    st_printf (&banner,
	       "\n"
	       "<P ALIGN=CENTER>\n"
	       "[ <A HREF=\"%s\">%s</A>\n",
	       url_string (&url), _("Reset"));
  }
  
  url_release (&url);
  url_arg (&url, "full", "yes");
  st_printf (&banner, "| <A HREF=\"%s\">%s</A>\n",
	     url_string (&url),
	     _("Full Contents"));

  url_release (&url);
  url_arg (&url, "op", "post");
  st_printf (&banner,
	     "| <A HREF=\"%s\">%s</A> ]\n"
	     "</P>\n"
	     "\n",
	     url_string (&url), _("Post"));

  /* Show the message index with the banner above and below it. */
  htmls (st_value (&banner));
  show_message_index (rq, r, filename, 0, std_contains (rq->cgi, "full"));
  htmls (st_value (&banner));

  dbr_destroy (r);

  if (filename == NULL)
    {
      char *crit;
      int n_rows;

      crit = relevant_file_crit (&rq->cx);
      db_query ("SELECT DISTINCT string "
		"FROM notes, strings "
		"WHERE notes.file = strings.seq AND (%ss)",
		DB_FORMAT | DB_TUPLES, &r, crit);
      free (crit);
      
      n_rows = dbr_nrows (r);

      if (n_rows)
	{
	  int i;
	  
	  html ("<HR>\n"
		"<H2 ALIGN=CENTER>%s</H2>\n",
		_("Files with Comments"));

	  for (i = 0; i < n_rows; i++)
	    {
	      html_list (1);
	      html_a_file (&rq->cx, dbr_value (r, i, 0), NULL);
	      htmlc ('\n');
	    }
	  html_list (0);
	}
    }

  url_destroy (&url);
  st_destroy (&banner);

  db_end ();
}

/* Displays the message board posting screen. */
static void
show_post (struct request *rq)
{
  /* What the user wants to do. */
  enum
    {
      POST_ENTRY,		/* Enter a message for posting. */
      POST_PREVIEW,		/* Show a preview of a message for posting. */
      POST_SUBMIT		/* Actually post the message. */
    }
  what;
  
  int followup;
  db_result *r;
  struct message msg;

  if (std_contains (rq->cgi, "post"))
    what = POST_SUBMIT;
  else if (std_contains (rq->cgi, "preview"))
    what = POST_PREVIEW;
  else
    what = POST_ENTRY;

  html ("<H2>%s</H2>",
	std_contains (rq->cgi, "file") ? _("Post Comment") : _("Post Message"));

  if (what != POST_ENTRY && empty (std_get (rq->cgi, "title")))
    {
      html ("<P>\n"
	    "%s\n"
	    "</P>\n",
	    _("Sorry, you may not specify an empty title.\n"
	      "Please fill in the title field and re-submit.\n"));

      what = POST_ENTRY;
    }
  
  followup = stoi (std_get (rq->cgi, "followup"));
  if (what == POST_ENTRY && followup)
    {
      r = get_message (&rq->cx, followup);
      interpret_message_result (r, 0, &msg);

      html ("<P>%s:</P>\n", _("In reply to"));
      html_list (1);
      show_message (&rq->cx, &msg);
      html_list (0);
    }
  else
    {
      r = NULL;
      msg.title = (char *) std_get (rq->cgi, "title");
    }

  if (what == POST_PREVIEW && empty (std_get (rq->cgi, "content")))
    html ("<P>%s</P>\n",
	  _("<STRONG>Warning:</STRONG>\n"
	    "You specified no content."));

  if (what == POST_PREVIEW)
    {
      struct message msg;

      msg.user = (char *) rq->cx.usr;
      msg.content = (char *) std_get (rq->cgi, "content");
      msg.title = (char *) std_get (rq->cgi, "title");
      msg.date = asctime (&cur_tm);
      
      html_list (1);
      show_message (&rq->cx, &msg);
      html_list (0);

      html (_("To post this %s as-is,\n"
	      "click on the <STRONG>Post</STRONG> button.\n"
	      "Otherwise, revise the %s's title and/or contents using the\n"
	      "form below, then click on <STRONG>Preview</STRONG>\n"
	      "or <STRONG>Post</STRONG> to continue.\n"),
	    std_contains (rq->cgi, "file") ? _("comment") : _("message"),
	    std_contains (rq->cgi, "file") ? _("comment") : _("message"));
    }

  if (what == POST_SUBMIT)
    submit (rq);

  if (what != POST_SUBMIT)
    {
      html_start_form (rq->screen, HT_POST, HT_URL_ENCODED);

      html_hidden ("op", "post");
      if (std_contains (rq->cgi, "file"))
	html_hidden ("file", std_get (rq->cgi, "file"));
      if (followup)
	html_hidden ("followup", std_get (rq->cgi, "followup"));

      html ("<TABLE>\n"
	    "<TR>\n"
	    "<TD ALIGN=RIGHT><LABEL FOR=title>%s</LABEL></TD>\n"
	    "<TD>",
	    _("Title"));
      html_text_field ("title", msg.title, 40);
      html ("</TD></TR>\n"
	    "<TR><TD ALIGN=RIGHT><LABEL FOR=content>%s</LABEL></TD>\n"
	    "<TD>",
	    _("Content"));
      html_text_area ("content", std_get (rq->cgi, "content"), 10, 60);
      htmls ("</TD></TR>\n"
	     "</TABLE>\n");
    }
  
  if (what == POST_ENTRY)
    html ("<P>%s</P>\n",
	  _("Fill in the form above.\n"
	    "When you're done, click on <STRONG>Preview</STRONG> to\n"
	    "see a preview of the message's presentation, or on\n"
	    "<STRONG>Post</STRONG> to post it immediately.\n"));

  if (what != POST_SUBMIT)
    {
      html ("<P>%s</P>\n",
	    _("You may include URLs in title and content.\n"
	      "Use the syntax <TT>&lt;URL:<i>url</i>&gt;</TT>.\n"
	      "For instance, for a link to\n"
	      "<STRONG>http://www.example.com</STRONG>,\n"
	      "write <TT>&lt;URL:http://www.example.com&gt;</TT>.\n"
	      "If <TT>http://</TT> is omitted, it will be supplied\n"
	      "automatically.  You may also include email addresses with the\n"
	      "same syntax."));

      html_submit_alt ("preview", _("Preview"));
      html_submit_alt ("post", _("Post"));
    }
  else
    show_message_board (rq, std_get_null (rq->cgi, "file"));
  
  if (r != NULL)
    dbr_destroy (r);
}

/* Posts to the message board based on the CGI parameters. */
static void
submit (struct request *rq)
{
  int seq;
  char *thread;
  int for_file;
  int file_seq;
  int title_seq;

  db_begin ();

  /* Get a sequence number for the post. */
  seq = get_seq ();
  
  /* Determine the thread value for the post, based on the thread
     value of the post that it follows up, if any, along with the
     sequence number of the post itself. */
  {
    int followup = stoi (std_get (rq->cgi, "followup"));
    if (followup)
      {
	db_result *r;
	const char *p;

	db_query ("SELECT thread FROM notes WHERE grp = '%sq' AND seq = %d",
		  DB_FORMAT | DB_TUPLES, &r, rq->cx.grp, followup);
	
	p = dbr_value (r, 0, 0);
	assert (*p == '{' && p[strlen (p) - 1] == '}');
	
	thread = xmalloc (strlen (p) + INT_DIGITS + 8);
	strcpy (thread, p);
	sprintf (&thread[strlen (p) - 1], ",%d}", seq);
	
	dbr_destroy (r);
      }
    else
      {
	thread = xmalloc (INT_DIGITS + 8);
	sprintf (thread, "{%d}", seq);
      }
  }
  
  {
    const char *filename = std_get_null (rq->cgi, "file");

    if (filename == NULL)
      {
	char tmp[16];
	
	for_file = 0;

	sprintf (tmp, "//%s", rq->cx.grp);
	file_seq = get_file_seq (&rq->cx, tmp);
      }
    else
      {
	for_file = 1;
	file_seq = get_file_seq (&rq->cx, filename);
      }
  }

  title_seq = get_file_seq (&rq->cx, std_get (rq->cgi, "title"));

  db_query ("INSERT INTO notes VALUES (%d, '%sq', '%sq', '%sq', %d, "
	    "datetime (timenow ()), '%sq', %d)",
	    DB_FORMAT, NULL,
	    file_seq, rq->cx.grp, rq->cx.usr, std_get (rq->cgi, "content"),
	    title_seq, thread, seq);

  log_activity (&rq->cx, file_seq, title_seq, 
		for_file ? AC_CAT_FILE : AC_CAT_MSG_BOARD,
		for_file ? AC_FILE_POST : AC_MSG_POST,
		seq);

  free (thread);

  db_end ();

  html ("<P>%s</P>\n",
	_("Your message has been posted."));
}

/* Show the message with sequence number SEQ.  FILENAME should be set
   to the file that this message applies to, or to a null pointer if
   none. */
static void
show_single_message (struct request *rq, int seq, const char *filename)
{
  struct message msg;
  db_result *r_msg;
  
  {
    struct url url;

    url_create (&url, rq->screen);
    if (filename != NULL)
      {
	url_arg (&url, "file", filename);
	url_arg (&url, "op", "comment");
      }

    html ("<P ALIGN=CENTER>[ <A HREF=\"%s\">%s</A> ]</P>\n\n",
	  url_string (&url), _("Index"));
  }

  db_begin ();
  r_msg = get_message (&rq->cx, seq);
  if (dbr_nrows (r_msg))
    {
      interpret_message_result (r_msg, 0, &msg);
      show_message (&rq->cx, &msg);
    }
  else
    {
      htmls ("<P>\n");
      html (_("%s not found.\n"
	      "This usually indicates that the %s\n"
	      "was deleted by its original poster.\n"),
	    filename != NULL ? _("Comment") : _("Message"),
	    filename != NULL ? _("comment") : _("message"));
      db_end ();
      return;
    }
  
  {
    char *f = (char *) filename;

    if (filename == NULL)
      f = blp_asprintf ("//%s", rq->cx.grp);
      
    show_viewed_by (&rq->cx, seq, f, msg.title,
		    filename != NULL ? AC_CAT_FILE : AC_CAT_MSG_BOARD,
		    filename != NULL ? AC_FILE_POST_READ : AC_MSG_READ);

    if (filename == NULL)
      free (f);
  }
  
  {
    struct url url;

    url_create (&url, rq->screen);
    if (filename != NULL)
      url_arg (&url, "file", filename);
    url_mark (&url);
    
    url_arg (&url, "op", "post");
    url_arg_int (&url, "followup", msg.seq);
    html ("<BR>[ <A HREF=\"%s\">Reply</A>\n", url_string (&url));

    if (!strcmp (msg.user, rq->cx.usr))
      {
	url_release (&url);
	url_arg_int (&url, "postrm", msg.seq);
	html ("| <A HREF=\"%s\">Delete</A> ]\n", url_string (&url));
      }
    else
      htmls ("]\n");

    url_destroy (&url);
  }

  htmls ("<HR NOSHADE>\n");

  {
    db_result *r;

    db_query ("SELECT usr, content, string, date, "
	      "_int4_depth(thread), notes.seq, notes.thread[1] AS thread1, "
	      "thread " /* See comment above about old Postgresql. */
	      "FROM notes, strings "
	      "WHERE notes.thread[1] = %d AND notes.title = strings.seq "
	      "ORDER BY thread",
	      DB_FORMAT | DB_TUPLES, &r, msg.thread1);

    
    html ("<P>%s:</P>", _("Thread Index"));
    show_message_index (rq, r, filename, seq, 0);

    dbr_destroy (r);
  }

  dbr_destroy (r_msg);

  db_end ();
}

void
show_viewed_by (struct context *cx, int seq, const char *s1, const char *s2,
		int cat, int type)
{
  int user_present;

  db_begin ();
  
  {
    db_result *r;
    int n_rows;

    db_query ("SELECT DISTINCT users.usr, name, last_word (name) AS last_name "
	      "FROM activities, users "
	      "WHERE ref_seq = %d AND users.usr = activities.usr "
	      "ORDER BY last_name, name",
	      DB_FORMAT | DB_TUPLES, &r, seq);
    n_rows = dbr_nrows (r);
    user_present = 0;
    if (n_rows)
      {
	int i;
	
	html ("<P>%s: ", _("Viewed by"));
	for (i = 0; i < n_rows; i++)
	  {
	    if (i != 0)
	      htmls (", ");
	    htmls (dbr_value (r, i, 1));
	    if (!strcmp (dbr_value (r, i, 0), cx->usr))
	      user_present = 1;
	  }
	html ("</P>\n");
      }
  }
  
  if (!user_present)
    log_activity_by_name (cx, s1, s2, cat, type, seq);

  db_end ();
}

/* Display message MSG, with links to an index if INDEXED is
   nonzero. */
static void
show_message (struct context *cx, const struct message *msg)
{
  htmls ("<B>");
  html_pretty_print (msg->title);
  html ("</B><BR>\n"
	"%s %s "
	"&lt;<A HREF=\"mailto:%s@%s\">%s@%s</A>&gt; "
	"%s %s<BR>\n",
	_("by"), username_to_fullname (cx->wg, msg->user),
	msg->user, mail_domain,
	msg->user, mail_domain,
	_("at"), msg->date);
  html_pretty_print (msg->content);
}

/* Show a message index for the messages in database result R. */
static void
show_message_index (struct request *rq,
		    const db_result *r, const char *filename, int seq_exclude,
		    int full)
{
  struct url url;
  
  int count;
  int i;

  url_create (&url, rq->screen);
  if (filename != NULL)
    url_arg (&url, "file", filename);
  url_mark (&url);

  count = dbr_nrows (r);
  for (i = 0; i < count; i++)
    {
      struct message msg;

      interpret_message_result (r, i, &msg);
      html_list (msg.nest);

      if (seq_exclude != msg.seq)
	{
	  url_release (&url);
	  url_arg_int (&url, "seq", msg.seq);
	  html ("<A HREF=\"%s\">", url_string (&url));
	}
      html_escape (msg.title);
      if (seq_exclude != msg.seq)
	htmls ("</A>");
      
      html ("\n%s %s %s %s\n",
	    _("by"), username_to_fullname (rq->cx.wg, msg.user),
	    _("at"), msg.date);
    }

  if (count == 0)
    html ("<P>%s</P>\n", full ? _("No messages.") : _("No new messages."));

  html_list (0);

  url_destroy (&url);
}

/* Copy the message in row I of database results R into MSG. */
void 
interpret_message_result (const db_result *r, int i, struct message *msg)
{
  msg->user = dbr_value (r, i, 0);
  msg->content = dbr_value (r, i, 1);
  msg->title = dbr_value (r, i, 2);
  msg->date = dbr_value (r, i, 3);
  msg->nest = stoi (dbr_value (r, i, 4));
  msg->seq = stoi (dbr_value (r, i, 5));
  msg->thread1 = stoi (dbr_value (r, i, 6));
}

/* Retrieve the message with sequence number SEQ and return the
   database result. */
db_result *
get_message (struct context *cx, int seq)
{
  db_result *r;
  db_query ("SELECT usr, content, string, date, "
	    "_int4_depth(thread), notes.seq, notes.thread[1] AS thread1 "
	    "FROM notes, strings "
	    "WHERE notes.seq = %d AND strings.seq = notes.title "
	    "AND notes.grp = '%sq'",
	    DB_FORMAT | DB_TUPLES, &r, seq, cx->grp);

  return r;
}

/* Retrieves message with sequence number SEQ into MSG.  Returns
   nonnull if successful. */
db_result *
retrieve_message (struct request *rq, int seq, struct message *msg)
{
  db_result *r = get_message (&rq->cx, seq);
  if (dbr_nrows (r) < 1)
    {
      dbr_destroy (r);
      return NULL;
    }
  interpret_message_result (r, 0, msg);
  return r;
}

/* Handle message deletion. */
static void
handle_postrm (struct request *rq)
{
  int seq;

  seq = stoi (std_get (rq->cgi, "postrm"));
  if (seq == 0)
    return;

  db_begin ();
  
  {
    db_result *r;
    
    db_query ("SELECT file, title, seq "
	      "FROM notes "
	      "WHERE seq = %d AND grp = '%sq' AND usr = '%sq'",
	      DB_FORMAT | DB_TUPLES, &r, seq, rq->cx.grp, rq->cx.usr);

    if (dbr_nrows (r) != 0)
      {
	int for_file = !strncmp (dbr_value (r, 0, 2), "//", 2);
	
	log_activity (&rq->cx, stoi (dbr_value (r, 0, 0)),
		      stoi (dbr_value (r, 0, 1)),
		      for_file ? AC_CAT_FILE : AC_CAT_MSG_BOARD,
		      for_file ? AC_FILE_POST_DELETE : AC_MSG_DELETE,
		      seq);
      }
    dbr_destroy (r);
  }

  db_query ("DELETE FROM notes "
	    "WHERE grp = '%sq' AND usr = '%sq' AND seq = %d",
	    DB_FORMAT, NULL, rq->cx.grp, rq->cx.usr, seq);

  db_end ();
}

/* Handle reset. */
static void
handle_reset (struct request *rq, const char *filename)
{
  int n_affected;
  int file_seq;
  int reset;

  reset = stoi (std_get (rq->cgi, "reset"));
  if (reset == 0)
    return;

  file_seq = virtual_file_seq (&rq->cx, filename);

  db_begin ();
  n_affected = db_query ("UPDATE catchup "
			 "SET last_read = int4larger (%d, last_read), "
			 "seq = nextval ('seq') "
			 "WHERE usr = '%sq' AND grp = '%sq' AND file = %d",
			 DB_FORMAT, NULL,
			 reset, rq->cx.usr, rq->cx.grp, file_seq);

  if (n_affected == 0)
    db_query ("INSERT INTO catchup "
	      "VALUES ('%sq', '%sq', %d, %d, nextval ('seq'))",
	      DB_FORMAT, NULL, rq->cx.usr, rq->cx.grp, file_seq, reset);
  db_end ();
}

/* Returns the value that should be used to identity file FILENAME, or
   to identity the current workgroup if FILENAME is null. */
static int
virtual_file_seq (struct context *cx, const char *filename)
{
  int seq;

  if (filename != NULL)
    seq = get_file_seq (cx, filename);
  else
    {
      char *f = local_alloc (2 + strlen (cx->grp) + 1);
      sprintf (f, "//%s", cx->grp);
      seq = get_file_seq (cx, f);
      local_free (f);
    }

  return seq;
}
