/* TeamSCOPE - Software for Collaborative Project Environments.
   Copyright (C) 1999, 2000, 2001 Michigan State University.
   Written by Ben Pfaff <pfaffben@msu.edu>.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To contact the TeamSCOPE team, send email to scope@cscw.msu.edu, or
   write to Charles Steinfield, Room 436 Communication Arts Building,
   Michigan State University, Department of Telecommunication, East
   Lansing MI 48824-1212. */

#include <config.h>
#include <ctype.h>
#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include "activity.h"
#include "cgi.h"
#include "conf.h"
#include "db.h"
#include "error.h"
#include "html.h"
#include "identity.h"
#include "options.h"
#include "scope.cgi.h"
#include "session.h"
#include "std.h"
#include "str.h"

static void handle_changes (struct request *);
static void show_options (struct request *);

static void show_timezone_names (const char *timezone);

static int inited;

void
options_initialize (struct request *rq)
{
  struct st f;

  assert (inited++ == 0);
  
  st_init (&f, 64);
  st_printf (&f, "%s/.scope/options", rq->cx.home);

  rq->cx.options = std_create (1);
  std_read (rq->cx.options, st_value (&f), 0, rq->cx.uid);

  st_destroy (&f);
}

void
options_close (struct request *rq)
{
  assert (--inited == 0);

  if (std_modified (rq->cx.options))
    {
      struct st f;

      st_init (&f, 64);
      st_printf (&f, "%s/.scope/options", rq->cx.home);

      if (std_write (rq->cx.options, st_value (&f), S_IRWXU | S_IRGRP) == 0)
	{
	  st_clear (&f);
	  st_printf (&f, "%s/.scope", rq->cx.home);
	  if (mkdir (st_value (&f), S_IRWXU) == 0)
	    {
	      st_concat (&f, "/options");
	      std_write (rq->cx.options, st_value (&f), S_IRWXU | S_IRGRP);
	    }
	}

      st_destroy (&f);
    }
}
      
void
scr_options (struct request *rq)
{
  handle_changes (rq);
  show_options (rq);
}

void
filter_screen (struct request *rq, const char *name, const char *title)
{
  struct activity_query query;
  
  activity_query_init (&query);
  activity_query_from_dict (&query, std_get_subdict (rq->cx.options, name),
			    rq->cx.usr);

  if (std_contains (rq->cgi, "update"))
    {
      struct std *d;

      activity_query_modify_from_cgi (&query, rq->cgi, rq->cx.usr);
      d = activity_query_to_dict (&query);
      std_put_subdict (rq->cx.options, name, d);
      std_destroy (d);
    }

  html ("<H2 ALIGN=CENTER>%s</H2>\n", title);
  html_start_form (name, HT_POST, HT_URL_ENCODED);
  html_hidden ("update", "yes");

  activity_show_filter (rq, &query);
  activity_query_destroy (&query);
}

void 
scr_webfilter (struct request *rq)
{
  filter_screen (rq, "webfilter", _("Activity Filter"));
}

void 
scr_emailfilter (struct request *rq)
{
  filter_screen (rq, "emailfilter", _("Email Filter"));
}

static void
handle_changes (struct request *rq)
{
  if (!std_contains (rq->cgi, "change"))
    return;
  
  if (std_contains (rq->cgi, "timezone"))
    std_replace (rq->cx.options, "timezone", std_get (rq->cgi, "timezone"));
  
  std_replace (rq->cx.options, "email",
	       std_contains (rq->cgi, "email") ? "yes" : "no");
  std_replace (rq->cx.options, "emailfmt", std_get (rq->cgi, "emailfmt"));
  std_replace (rq->cx.options, "emailfreq", std_get (rq->cgi, "emailfreq"));
  std_replace (rq->cx.options, "emailbody", std_get (rq->cgi, "emailbody"));
  std_replace (rq->cx.options, "webresetsmail",
	       std_contains (rq->cgi, "webresetsmail") ? "yes" : "no");
  std_replace (rq->cx.options, "mailresetsweb",
	       std_contains (rq->cgi, "mailresetsweb") ? "yes" : "no");

  db_query ("UPDATE users SET emailbody = %d, primarygrp = '%sq' "
	    "WHERE usr = '%sq'",
	    DB_FORMAT, NULL,
	    stoi (std_get (rq->cgi, "emailbody")),
	    std_get (rq->cgi, "primarygrp"),
	    rq->cx.usr);
  
  {
    int idle = stoi (std_get (rq->cgi, "idle"));

    if (idle != 0)
      {
	char buf[INT_DIGITS + 1];
	
	db_query ("UPDATE users SET idle = %d WHERE usr = '%sq'",
		  DB_FORMAT, NULL, idle, rq->cx.usr);

	sprintf (buf, "%d", idle);
	std_replace (rq->trusted, "idle", buf);
      }
  }

  std_replace (rq->cx.options, "chat", std_get (rq->cgi, "chat"));
}

static void
show_options (struct request *rq)
{
  html
    ("<H2 ALIGN=CENTER>%s</H2>\n"
     "\n"
     "<P>%s</P>\n"
     "\n",
     _("Account Options"),
     _("You can use the options below to customize how your\n"
       "TeamSCOPE account works.\n"));

  html_start_form ("options", HT_POST, HT_URL_ENCODED);
  html_hidden ("change", "yes");

  html 
    ("<TABLE CELLPADDING=0>\n"
     "\n"
     "<TR VALIGN=BOTTOM>\n"
     "<TD WIDTH=\"45%%\">\n"
     "<FONT SIZE=\"+1\">%s</FONT><BR>\n"
     "<FONT SIZE=\"-1\">%s</FONT>\n"
     "<HR NOSHADE>\n"
     "\n"
     "<TD WIDTH=\"10%%\">&nbsp;\n"
     "<TD WIDTH=\"45%%\">\n"
     "<FONT SIZE=\"+1\">%s</FONT><BR>\n"
     "<FONT SIZE=\"-1\">%s</FONT>\n"
     "<HR NOSHADE>\n"
     "\n"
     "<TR VALIGN=TOP>\n"
     "<TD>%s\n"
     "<BR><FONT SIZE=\"-1\">%s<BR>\n",
     _("Settings"),
     _("After changing these options to your desired\n"
       "settings, click on <STRONG>Set Options</STRONG> below to save\n"
       "your changes."),
     _("Additional Options"),
     _("Each of these links will bring up a new\n"
       "screenful of options.  Be sure to save any changes in the\n"
       "lefthand column before proceeding."),
     _("Time Zone"),
     _("Select your preferred time zone for input\n"
       "and display of dates and times.\n"));

  if (std_get_nonempty (rq->cx.options, "timezone") != NULL)
    html ("%s <STRONG>%s</STRONG>.",
	  _("Your current time zone is"), std_get (rq->cx.options, "timezone"));
  else
    htmls (_("Your time zone is currently unconfigured.\n"));

  htmls ("</FONT><BR>\n"
	 "<SELECT NAME=timezone SIZE=8>\n");
  show_timezone_names (std_get (rq->cx.options, "timezone"));
  htmls ("</SELECT>\n"
	 "<HR NOSHADE>\n"
	 "\n");

  html
    ("%s<BR>\n"
     "<FONT SIZE=\"-1\">%s</FONT>\n",
     _("Email Options"),
     _("Optionally, you can have summaries of activities sent\n"
       "to your email account from time to time.\n"
       "To enable this feature, select from the list below one\n"
       "of the options for summary email frequency.\n"));

  {
    int emailfreq = stoi (std_get (rq->cx.options, "emailfreq"));
    
    htmls ("<BR>");
    html_radio ("emailfreq", "0", _("Do not send periodic emails."),
		emailfreq == 0);
    htmls ("<BR>");
    html_radio ("emailfreq", "1", _("Monday only."),
		emailfreq == 1);
    htmls ("<BR>");
    html_radio ("emailfreq", "2", _("Monday, Wednesday, Friday."),
		emailfreq == 2);
    htmls ("<BR>");
    html_radio ("emailfreq", "3", _("Daily."),
		emailfreq == 3);
    htmls ("<BR>");
    html_radio ("emailfreq", "4", _("Twice a day."),
		emailfreq == 4);
    htmls ("<BR>");
    html_radio ("emailfreq", "5", _("Hourly."),
		emailfreq == 5);
  }

  {
    static const char *formats[3] = 
      {
	N_("text"), N_("HTML"), N_("both"),
      };
    
    int format = stoi (std_get (rq->cx.options, "emailfmt"));
    int i;

    html ("\n<P>%s<BR>\n", _("Send emails as"));
    for (i = 0; i < 3; i++)
      html ("<INPUT TYPE=radio NAME=emailfmt VALUE=%d%s>%s</INPUT>\n",
	    i, format == i ? " CHECKED" : "", gettext (formats[i]));
  }
    
  html ("\n<P><FONT SIZE=\"-1\">%s</FONT>\n",
	_("By default, only the subject line of messages and comments\n"
	  "are included in summary emails.  Messages' bodies can be\n"
	  "included as well.  Alternatively, messages can be mailed\n"
	  "individually.  The latter option takes effect even if\n"
	  "summary emails are not enabled.\n"));
  
  {
    int emailbody = stoi (std_get (rq->cx.options, "emailbody"));
    
    htmls ("<BR>");
    html_radio ("emailbody", "0", _("Do not include message bodies."),
		emailbody == 0);
    htmls ("<BR>");
    html_radio ("emailbody", "1", _("Include message bodies."),
		emailbody == 1);
    htmls ("<BR>");
    html_radio ("emailbody", "2", _("Mail new messages individually."),
		emailbody == 2);
  }

  html
    ("<P>%s<BR>\n"
     "<LABEL><INPUT NAME=webresetsmail TYPE=checkbox VALUE=\"yes\"%s>\n"
     "%s</LABEL>\n"
     "<BR><LABEL><INPUT NAME=mailresetsweb TYPE=checkbox VALUE=\"yes\"%s>\n"
     "%s</LABEL>\n"
     "<HR NOSHADE>\n"
     "\n",
     _("Web/Email Interaction"),
     (strcmp (std_get (rq->cx.options, "webresetsmail"), "no")
      ? " CHECKED" : ""),
     _("Emails reset the <STRONG>Activity</STRONG> list "
       "automatically\n"),
     (!strcmp (std_get (rq->cx.options, "mailresetsweb"), "yes")
      ? " CHECKED" : ""),
     _("Resetting the <STRONG>Activity</STRONG> list prevents "
       "them from being sent in email as well"));

  {
    static const char *names[9] = 
      {
	N_("5 minutes"),
	N_("10 minutes"),
	N_("20 minutes"),
	N_("30 minutes"),
	N_("1 hour"),
	N_("2 hours"),
	N_("4 hours"),
	N_("8 hours"),
	N_("1 day"),
      };

    static int values[9] = 
      {
	5 * 60,
	10 * 60,
	20 * 60,
	30 * 60,
	1 * 60 * 60,
	2 * 60 * 60,
	4 * 60 * 60,
	8 * 60 * 60,
	24 * 60 * 60,
      };

    int idle;

    int i;

    html
      ("%s<BR>\n"
       "<FONT SIZE=\"-1\">%s</FONT><BR>\n"
       "<SELECT NAME=\"idle\">",
       _("Session Idle Time"),
       _("Select the length of time after which an idle session\n"
	 "should be logged out automatically.\n"
	 "Closing your web browser will also log out your session.\n"));

    idle = stoi (std_get (rq->trusted, "idle"));
    for (i = 0; i < 9; i++)
      html ("<OPTION VALUE=%d%s>%s</OPTION>\n",
	    values[i],
	    idle == values[i] ? " SELECTED" : "",
	    names[i]);
    htmls ("</SELECT>\n"
	   "<HR NOSHADE>\n"
	   "\n");
  }
  
  {
    int selection = stoi (std_get (rq->cx.options, "chat"));

    html
      ("%s<BR>\n"
       "<FONT SIZE=\"-1\">%s</FONT><BR>\n",
       _("Chat Window"),
       _("TeamSCOPE incorporates a simple Java-based chat client that\n"
	 "can allow you to easily talk to other members of your team\n"
	 "when multiple team members are logged in at once.\n"
	 "Select one of the configuration possibilities for this feature\n"
	 "from the list below."));

    html_radio ("chat", "0",
		_("Display a separate window with the chat client\n"
		  "activation button.\n"
		  "Requires Java and JavaScript support in your browser.\n"),
		selection == 0);
    htmls ("<BR>\n");
    html_radio ("chat", "1",
		_("Put the chat client activation button in the TeamSCOPE\n"
		  "header that appears on each page.\n"
		  "Requires Java support in your browser.\n"
		  "This can be bandwidth-intensive, so it is recommended\n"
		  "only for users with a fast Internet connection.\n"
		  "<STRONG>Warning:</STRONG> This option is known to\n"
		  "eventually crash many versions of Netscape Navigator.\n"),
		selection == 1);
    htmls ("<BR>\n");
    html_radio ("chat", "2",
		_("Disable TeamSCOPE chat functions."),
		selection == 2);
    htmls ("<HR NOSHADE>\n"
	   "\n");
  }
  
  {
    const char **groups;
    int n;

    groups = std_get_many (rq->trusted, "groups", &n);
    if (n > 1)
      {
	int i;
	const char *primarygrp;
	db_result *r;

	db_query ("SELECT primarygrp FROM users WHERE usr = '%sq'",
		  DB_FORMAT | DB_TUPLES, &r, rq->cx.usr);
	primarygrp = dbr_value (r, 0, 0);

	html ("%s<BR>\n"
	      "<FONT SIZE=\"-1\">%s</FONT><BR>\n"
	      "<SELECT NAME=primarygrp SIZE=1>\n",
	      _("Primary Group"),
	      _("Select one of the groups from the list "
		"below.  This group will be your primary group, the group "
		"entered by default when you log in."));
	for (i = 0; i < n; i++)
	  {
	    char *p = strchr (groups[i], ';');
	    size_t len;
	    if (p == NULL)
	      continue;

	    len = p - groups[i];
	    html ("  <OPTION VALUE=\"%.*s\"%s>%s</OPTION>\n",
		  (int) len, groups[i],
		  (len == strlen (primarygrp)
		   && !memcmp (primarygrp, groups[i], len) ? "SELECTED " : ""),
		  p + 1);
	  }
	htmls ("</SELECT>\n"
	       "<HR NOSHADE>\n\n");
      }
  }
  
  html 
    ("<TD>&nbsp;\n"
     "\n"
     "<TD>\n"
     "<P>\n"
     "<A HREF=\"%s\">%s</A><BR>\n"
     "<FONT SIZE=\"-1\">%s\n"
     "</FONT>\n"
     "<HR NOSHADE>\n"
     "\n",
     url_this ("emailfilter", (char *) NULL),
     _("Email Filter"),
     _("By default, if you have requested recent activity to be\n"
       "emailed to you, all activities will be emailed.\n"
       "You can optionally select only certain activities for inclusion.\n"
       "To do so, click the link above."));
  
  html
    ("<A HREF=\"%s\">%s</A><BR>\n"
     "<FONT SIZE=\"-1\">%s\n"
     "</FONT>\n"
     "</P>\n"
     "<HR NOSHADE>\n"
     "\n",
     url_this ("webfilter", (char *) NULL),
     _("<STRONG>Activity</STRONG> Filter"),
     _("By default, the <STRONG>Activity</STRONG> page\n"
       "will show all new activity.\n"
       "You can optionally select only certain activities for inclusion.\n"
       "To do so, click the link above."));

  html
    ("<A HREF=\"%s\">%s</A><BR>\n"
     "<FONT SIZE=\"-1\">%s\n"
     "</FONT>\n"
     "</P>\n"
     "<HR NOSHADE>\n"
     "\n",
     url_this ("password", (char *) NULL),
     _("Change Password"),
     _("You can change your account password from this web interface.\n"
       "To do so, click the link above."));

  htmls ("</TABLE>");

  html_submit (_("Set Options"));

  html_end_form ();
}

static void
show_timezone_names (const char *timezone)
{
  size_t line_size;
  char *line;
  FILE *f;
  
  f = fopen (TIMEZONES_FILE, "r");
  if (f == NULL)
    {
      log ("error opening " TIMEZONES_FILE ": %s", strerror (errno));
      return;
    }

  line = NULL;
  line_size = 0;
  for (;;)
    {
      ssize_t len = getline (&line, &line_size, f);
      if (len == -1)
	{
	  if (ferror (f))
	    log ("error reading " TIMEZONES_FILE ": %s", strerror (errno));

	  fclose (f);
	  free (line);

	  return;
	}

      while (len > 0 && isspace ((unsigned char) line[len - 1]))
	len--;
      line[len] = '\0';

      if (len > 0)
	html ("<OPTION%s>%s</OPTION>\n",
	      strcmp (timezone, line) ? "" : " SELECTED",
	      line);
    }
}
