/* TeamSCOPE - Software for Collaborative Project Environments.
   Copyright (C) 1999, 2000, 2001 Michigan State University.
   Written by Ben Pfaff <pfaffben@msu.edu>.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To contact the TeamSCOPE team, send email to scope@cscw.msu.edu, or
   write to Charles Steinfield, Room 436 Communication Arts Building,
   Michigan State University, Department of Telecommunication, East
   Lansing MI 48824-1212. */

#include <config.h>
#include <assert.h>
#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <termios.h>
#include <unistd.h>
#include <time.h>
#include <grp.h>
#include <pwd.h>
#include "alloc.h"
#include "conf.h"
#include "db.h"
#include "error.h"
#include "mail.h"
#include "scope-db.h"
#include "std.h"
#include "str.h"

/* Latest database version. */
#define DB_VERSION_MAJOR 12
#define DB_VERSION_MINOR 0

/* Name program invoked as. */
char *pgmname;

/* Time at beginning of execution. */
time_t cur_time;
struct tm cur_tm;

/* Dictionary of configuration variables. */
struct std *conf;

/* Common functions. */

#define su_postgres() ((void) su (postgres_uid (), "postgres"))
#define su_scope() ((void) su (scope_uid (), "scope"))
#define su_nobody() ((void) su (nobody_uid (), "nobody"))
#define su_root() ((void) su (0, "root"))

static void su (int uid, const char *username);
static int scope_uid (void);
static int postgres_uid (void);
static int nobody_uid (void);
static uid_t cache_uid (const char *username, uid_t *uid);

static int user_is_member (const char *usr, const char *grp);
static void get_users_groups (const char *usr, struct st *s,
			      const char *add, const char *del);
static uid_t usr_to_uid (const char *usr);
static gid_t grp_to_gid (const char *grp);
static void first_avail_id (unsigned long min, unsigned long max,
			    uid_t *uid, gid_t *gid);

static void create_db (void);
static void init_db (void);
static void dump_database (const char *dst_fn);

static void copy (FILE *dst, FILE *src);
static void make_dir_tree (const char *tree, mode_t mode,
			   uid_t uid, gid_t gid);
static void replace_file (const char *base, const char *with);
static int check_inited (const char *fn);

typedef void foreign_config_callback (FILE *, char *, void *);
static void update_foreign_config (const char *src_fn, const char *mod,
				   foreign_config_callback *, void *);

static int lock_file (const char *fn, int block);
static void unlock_file (int handle, const char *fn);
static void scope_lock (void);
static void mail_lock (void);

static void validate_name (const char *name, const char *type);
static void add_user_to_unix_group (const char *usr, const char *grp);
static void add_user_to_teamscope_group (const char *usr, const char *grp);
static void add_group (const char *grp, const char *fullname);
static void add_user (const char *grp, const char *usr,
		      const char *fullname, const char *email);
static void del_user_from_teamscope_group (const char *usr, const char *grp);
static void del_user_from_unix_group (const char *usr, const char *grp);
static void del_group (const char *grp);
static void de_luser (const char *usr);
static void send_user_mail (const char *usr, int reset);
static void update_team_site (const char *grp);

static void randpass (char buf[9]);
static void set_password (const char *usr, const char *pwd, int lock_it);

static void shell (int expand, const char *format, ...);
static void check_system_value (int result, const char *s);
static char *prompt_input (int *pargc, char ***pargv, const char *prompt);

/* Main program and command dispatcher. */

/* Dispatch function type. */
typedef void command_func (int argc, char *argv[]);

/* A command that might be given on the command line. */
struct command
  {
    const char *name;		/* Name. */
    int n_args;			/* Number of arguments, or -1 for "any". */
    int args_optional;		/* For n_args > 0, are args optional? */ 
    command_func *func;		/* Dispatch function. */
  };

/* Prototypes for dispatch functions. */
static command_func cmd_create, cmd_remove, cmd_dump, cmd_restore, cmd_upgrade;
static command_func cmd_init, cmd_uninit, cmd_update;
static command_func cmd_addgroup, cmd_delgroup, cmd_adduser, cmd_deluser;
static command_func cmd_addusertogroup, cmd_deluserfromgroup, cmd_batch;
static command_func cmd_passwd, cmd_mail, cmd_lock, cmd_unlock;
static command_func cmd_email, cmd_fullname, cmd_group_admin;
static command_func cmd_daily_maint, cmd_daily_backup, cmd_daily_cleanup;
static command_func cmd_help, cmd_version;

/* Command dispatch table. */
static const struct command commands[] =
  {
    {"create", 0, 0, cmd_create},
    {"remove", 0, 0, cmd_remove},
    {"dump", 1, 0, cmd_dump},
    {"restore", 1, 0, cmd_restore},
    {"upgrade", 0, 0, cmd_upgrade},

    {"init", 1, 0, cmd_init},
    {"uninit", 1, 0, cmd_uninit},
    {"update", 0, 0, cmd_update},

    {"addgroup", 2, 1, cmd_addgroup},
    {"delgroup", 1, 0, cmd_delgroup},
    {"adduser", 4, 1, cmd_adduser},
    {"deluser", -1, 0, cmd_deluser},

    {"addusertogroup", 2, 1, cmd_addusertogroup},
    {"deluserfromgroup", 2, 1, cmd_deluserfromgroup},
    {"batch", 0, 0, cmd_batch},

    {"passwd", 1, 1, cmd_passwd},
    {"mail", -1, 0, cmd_mail},
    {"lock", -1, 0, cmd_lock},
    {"unlock", -1, 0, cmd_unlock},
    {"info", -1, 0, NULL},
    {"email", 2, 1, cmd_email},
    {"fullname", 2, 1, cmd_fullname},
    {"group-admin", 3, 1, cmd_group_admin},

    {"daily-maint", 0, 0, cmd_daily_maint},
    {"daily-backup", 0, 0, cmd_daily_backup},
    {"daily-cleanup", 0, 0, cmd_daily_cleanup},

    {"hourly-mail", 0, 0, cmd_hourly_mail},
    {"sync-mail", 0, 0, cmd_sync_mail},
    {"received-mail", 0, 0, cmd_received_mail},

    {"help", 0, 0, cmd_help},
    {"version", 0, 0, cmd_version},

    {NULL, 0, 0, NULL},
  };

static void usage (int exit_code) attribute ((noreturn));
static const struct command *parse_command_line (int *pargc, char ***pargv);
static char *host_fqdn (void);
static void get_configuration (void);
static void check_options (void);

int
main (int argc, char *argv[])
{
  const struct command *command;
  
  /* Determine program name. */
  {
    char *p = strrchr (argv[0], '/');
    pgmname = p ? &p[1] : argv[0];
  }
  
  /* Read clock. */
  time (&cur_time);
  cur_tm = *localtime (&cur_time);

  /* Set a secure path for execution of subprograms. */
  if (putenv ("PATH=/bin:/sbin:/usr/bin:/usr/sbin") == -1)
    fail ("can't set PATH environment variable");

  /* Initialize. */
  command = parse_command_line (&argc, &argv);
  get_configuration ();

  if (command->func != cmd_help && command->func != cmd_version
      && command->func != cmd_received_mail)
    {
      setuid (0);
      setgid (0);
      setgroups (0, NULL);
      umask (0022);
      if (command->func == cmd_sync_mail)
	mail_lock ();
      else	
	scope_lock ();
    }

  /* Run command. */
  if (command->func == NULL)
    fail ("%s: not yet implemented", command->name);
  command->func (argc, argv);

  /* Finish. */
  if (db_open_p ())
    db_close ();
  return EXIT_SUCCESS;
}

/* Sets up `conf' configuration dictionary. */
static void
get_configuration (void)
{
  /* Create conf dictionary. */
  conf = std_create (0);

  /* Insert FQDN of localhost into conf dictionary. */
  {
    char *fqdn = host_fqdn ();
    std_put (conf, "server_name", fqdn);
    free (fqdn);
  }

  /* Read configuration file. */
  if (!std_read (conf, CONF_FILE, STDF_SUBST | STDF_TRIM_SPACES, 0))
    fail ("error reading " CONF_FILE ", aborting");

  /* Verify we have all the settings we need. */
  check_options ();
}

/* Checks that all the expected options are present in conf
   dictionary. */
static void
check_options (void)
{
  static const char *vars[] =
    {
      "server_name",

      "conf_dir",
      "share_data_dir",
      "arch_data_dir",
      "state_dir",
      "log_dir",

      "conf_file",
      "passwd_file",
      "reload_script",

      "www_root",
      "script_name",

      "chat_log_file",

      "trusted_state_dir",
      "untrusted_state_dir",

      "httpd_conf_file",
      "exim_conf_file",
      "icon_file",
      "aliases_file",

      "timezones_file",
      "zoneinfo_dir",

      "home_style",
      "home_base",

      "first_uid",
      "last_uid",

      "backup_dir",
      "backup_count",

      NULL
    };

  const char **var;

  for (var = vars; *var; var++)
    if (std_get_null (conf, *var) == NULL)
      fail ("configuration file doesn't define required variable %s", *var);

  if (strncmp (std_get (conf, "home_style"), "hier", 4)
      && strncmp (std_get (conf, "home_style"), "flat", 4))
    fail ("bad configuration: home_style must be `hierarchical' or `flat'");
}

/* Parses the main command name on the command line and returns the
   corresponding command structure.  Modifies *PARGC and *PARGV to
   exclude the program name and command name. */
static const struct command *
parse_command_line (int *pargc, char ***pargv)
{
  int argc = *pargc;
  char **argv = *pargv;

  const struct command *command;
  
  if (argc < 2)
    fail ("missing action argument; use `%s help' for usage", pgmname);

  if (!strncmp (argv[1], "--", 2))
    argv[1] += 2;
  for (command = commands; ; command++)
    if (command->name == NULL)
      usage (EXIT_FAILURE);
    else if (!strcmp (command->name, argv[1]))
      break;

  argv += 2;
  argc -= 2;

  if (command->n_args == -1)
    /* Okay. */;
  else if (command->args_optional && argc > command->n_args)
    fail ("too many arguments for %s command", command->name);
  else if (!command->args_optional && argc != command->n_args)
    fail ("%s command takes %d arguments, but given %d",
	  command->name, command->n_args, argc);

  *pargc = argc;
  *pargv = argv;

  return command;
}

static const char usage_message[] =
"scope-db, a program for TeamSCOPE configuration and database maintenance.\n"
"\nUsage: %s COMMAND [ARGS...]\n"
"\nIn the command descriptions below, CAPITALS denote arguments, square\n"
"brackets [] enclose optional arguments, and ellipses... specify that\n"
"arguments may be repeated as many times as desired.  See the TeamSCOPE\n"
"Installation Manual for complete details.\n" 
"\nDatabase maintenance:\n"
"  create            Create the TeamSCOPE database.\n"
"  remove            Remove the TeamSCOPE database.  All data will be lost!\n"
"  dump DUMP         Dump a snapshot of the TeamSCOPE database to DUMP.\n"
"  restore DUMP      Restore from previous database snapshot DUMP.\n"
"  upgrade           Upgrade the SCOPE database.  Backup first with `dump'!\n"
"\nConfiguration maintenance:\n"
"  init FUNCTION     Initialize or enable specified FUNCTION.\n"
"  uninit FUNCTION   Uninitialize or disable specified FUNCTION.\n"
"  update            Put configuration files in synch with database.\n"
"\n  The currently supported values for FUNCTION are:\n"
"    apache          Apache setup.\n"
"    timezones       Timezones setup.\n"
"    exim            Exim setup.\n"
"    aliases         Aliases setup.\n"
"\nUser maintenance:\n"
"  addgroup [GROUP [FULLNAME]]\n"
"                    Create new group named GROUP with full name FULLNAME.\n"
"  delgroup GROUP... Delete GROUP.\n"
"  adduser [GROUP [USER [FULLNAME [EMAIL]]]]\n"
"                    Create USER with full name FULLNAME in GROUP with email\n"
"                    address EMAIL.  If USER is null then it is generated.\n"
"  deluser USER...   Delete USER.\n"
"  addusertogroup [USER [GROUP]]\n"
"                    Add USER to GROUP.\n"
"  deluserfromgroup [USER [GROUP]]\n"
"                    Remove USER from GROUP.\n"
"  batch < FILE      Add/remove users/groups as described in FILE.\n"
"  passwd USER       Set password for USER.\n"
"  mail [--no-reset] [USER | --pending]\n"
"                    Send TeamSCOPE introductory email to USER.\n"
"                    Use `--pending' for USER to send to all new users.\n"
"                    If `--pending' or `--no-reset' specified, users with\n"
"                    existing passwords will not be assigned new ones.\n"
"  lock USER...      Lock USER's TeamSCOPE account.\n"
"  unlock USER...    Unlock USER's TeamSCOPE account.\n"
"  info USER...      Show information on USER if given, all users otherwise.\n"
"  email [USER [EMAIL]]  Set upstream email address for USER to EMAIL.\n"
"  fullname [USER [NAME]]  Set full name for USER to NAME.\n"
"  group-admin [USER [GROUP [SUBCOMMAND]]]\n"
"                    Gets or sets admin privileges for USER within GROUP.\n"
"                    SUBCOMMAND must be `yes', `no', or `query'.\n"
"\nScheduled maintenance (normally run from cron):\n"
"  daily-backup      Dump database backup and rotate old backups.\n"
"  daily-cleanup     Remove old session data files.\n"
"  daily-maint       Perform each of the tasks above.\n"
"  hourly-mail       Send notification emails to users who requested them.\n"
"  sync-mail         Send out sychronous email.\n"
"\nGeneral:\n"
"  help              Display this help screen.\n"
"  version           Report version and copyright information.\n"
"\nReport bugs to scope@cscw.msu.edu.\n";

/* Prints a usage message and exits to the OS with code EXIT_CODE. */
static void
usage (int exit_code)
{
  printf (usage_message, pgmname);
  exit (exit_code);
}

/* Determines the local fully-qualified-domain-name of this machine
   and returns a copy of it in a malloc()'d buffer. */
static char *
host_fqdn (void)
{
  struct hostent *hep;
  char hostname[1024];

  /* If the hostname doesn't fit in 1024 bytes, tough. */
  gethostname (hostname, 1024);
  hostname[1023] = '\0';

  /* Retrieve DNS record for this hostname. */
  if (getuid () == 0)
    su_nobody ();
  hep = gethostbyname (hostname);
  if (getuid () == 0)
    su_root ();
  return xstrdup (hep != NULL ? hep->h_name : hostname);
}

static void
cmd_help (int argc unused, char *argv[] unused)
{
  usage (EXIT_SUCCESS);
}

static void
cmd_version (int argc unused, char *argv[] unused)
{
  fputs
    ("scope-db (TeamSCOPE) 0.9\n"
     "Copyright (C) 1999, 2000, 2001 Michigan State University.\n"
     "TeamSCOPE comes with NO WARRANTY, not even for MERCHANTABILITY\n"
     "or FITNESS FOR A PARTICULAR PURPOSE.  You may redistribute\n"
     "copies of TeamSCOPE under the terms of the GNU General Public\n"
     "License.  For more information, refer to file COPYING.\n",
     stdout);
}

/* Database functionality. */

static void create_db (void);
static void init_db (void);

static void
cmd_create (int argc unused, char *argv[] unused)
{
  init_db ();
}

static void
cmd_remove (int argc unused, char *argv[] unused)
{
  su_postgres ();
  db_open ("template1");
  db_query ("DROP DATABASE scope", 0, NULL);
  db_query ("DROP USER scope", 0, NULL);
  db_close ();
}

static void
cmd_dump (int argc unused, char *argv[])
{
  dump_database (argv[0]);
}

static void
cmd_restore (int argc unused, char *argv[])
{
  /* The `postgres' user probably won't have permissions to read from
     the file that we want to restore from, so we can't simply run
     `psql -e scope < file'.  This alternate method should work
     better.  */
  
  FILE *in, *out;

  in = open_file (argv[0], "rb");
  create_db ();
  su_scope ();
  out = open_pipe ("psql scope", "w");
  
  copy (out, in);

  close_pipe (out, "psql scope");
  close_file (in, argv[0]);  
}

static void
cmd_upgrade (int argc unused, char *argv[] unused)
{
  db_result *r;
  int major, minor;

  connect_db ();
  db_begin ();
  
  db_query ("SELECT major, minor FROM version", DB_TUPLES, &r);
  if (dbr_nrows (r) < 1)
    fail ("database has empty version table");
  major = stoi (dbr_value (r, 0, 0));
  minor = stoi (dbr_value (r, 0, 1));
  dbr_destroy (r);

  if (major < 1 || major > DB_VERSION_MAJOR || minor != 0)
    fail ("unknown database version %d.%d", major, minor);

  if (major < 2)
    {
      db_query ("ALTER TABLE users ADD COLUMN last_usage datetime", 0, NULL);
      db_query ("UPDATE users SET last_usage = 'Jan 1 1970'", 0, NULL);
    }
  
  if (major < 3)
    {
      db_query
	("CREATE TABLE uselog (\n"
	 "usr		varchar(8)	not null,\n"
	 "when		datetime	not null,\n"
	 "loc		text);\n", 0, NULL);
    }

  if (major < 4)
    db_query ("ALTER TABLE groups ADD COLUMN md5 text[]", 0, NULL);

  if (major < 5)
    db_query ("ALTER TABLE uselog RENAME when TO t", 0, NULL);

  if (major < 6)
    {
      db_query ("ALTER TABLE users ADD COLUMN emailbody int4", 0, NULL);
      db_query ("UPDATE users SET emailbody = 0", 0, NULL);
      db_query ("CREATE TABLE sync_mail (seq int4)", 0, NULL);
      db_query ("INSERT INTO sync_mail VALUES (nextval ('seq'))", 0, NULL);
    }

  if (major < 7)
    db_query ("ALTER TABLE users ADD COLUMN primarygrp varchar(8)", 0, NULL);

  if (major < 8)
    {
      db_query ("ALTER TABLE uselog ADD COLUMN grp varchar(8)", 0, NULL);
      db_query ("ALTER TABLE uselog ADD COLUMN query text", 0, NULL);
    }

  if (major < 9)
    {
      db_query ("ALTER TABLE user_group ADD COLUMN may_admin bool", 0, NULL);
      db_query ("UPDATE user_group SET may_admin = 'f'", 0, NULL);
    }

  if (major < 10)
    {
      db_end ();
      db_close ();

      su_postgres ();
      db_open (NULL);
      su_root ();

      db_query ("CREATE FUNCTION last_word(text) RETURNS text "
		"AS '%sq/text.o' LANGUAGE 'c';",
		DB_FORMAT, NULL,
		std_get (conf, "arch_data_dir"));
      db_close ();

      connect_db ();
      db_begin ();
    }

  if (major < 11)
    {
      db_query ("ALTER TABLE users ADD COLUMN welcome bool", 0, NULL);
      db_query ("UPDATE users SET welcome = 't'", 0, NULL);
    }

  if (major < 12)
    {
      db_query ("ALTER TABLE events ADD COLUMN source int4", 0, NULL);
      db_query ("UPDATE events SET source = 0", 0, NULL);
    }

  if (major != DB_VERSION_MAJOR || minor != DB_VERSION_MINOR)
    db_query ("UPDATE version SET major = %d, minor = %d",
	      DB_FORMAT, NULL, DB_VERSION_MAJOR, DB_VERSION_MINOR);

  db_end ();
  db_close ();
}

/* Foreign configuration file handling. */ 

static void init_apache (void);
static void init_timezones (void);
static void init_update_file (const char *fn);
static void update_apache_conf (void);
static void update_exim_conf (void);
static void update_aliases (void);
static void restart_stuff (void);

static void
cmd_init (int argc unused, char *argv[])
{
  if (!strcmp (argv[0], "apache"))
    init_apache ();
  else if (!strcmp (argv[0], "timezones"))
    init_timezones ();
  else if (!strcmp (argv[0], "exim"))
    init_update_file (std_get (conf, "exim_conf_file"));
  else if (!strcmp (argv[0], "aliases"))
    init_update_file (std_get (conf, "aliases_file"));
  else
    fail ("don't know how to initialize %s", argv[0]);
}

static void
cmd_uninit (int argc unused, char *argv[])
{
  if (!strcmp (argv[0], "apache"))
    update_foreign_config (std_get (conf, "httpd_conf_file"),
			   NULL, NULL, NULL);
  else if (!strcmp (argv[0], "timezones"))
    {
      const char *tz_file = std_get (conf, "timezones_file");
      if (unlink (tz_file) == -1 && errno != ENOENT)
	fail ("%s: unlink: %s", tz_file, strerror (errno));
    }
  else if (!strcmp (argv[0], "exim"))
    update_foreign_config (std_get (conf, "exim_conf_file"), NULL, NULL, NULL);
  else if (!strcmp (argv[0], "aliases"))
    update_foreign_config (std_get (conf, "aliases_file"), NULL, NULL, NULL);
  else
    fail ("don't know how to initialize %s", argv[0]);
}

static void
cmd_update (int argc unused, char *argv[] unused)
{
  update_apache_conf ();
  update_exim_conf ();
  update_aliases ();
  restart_stuff ();
}

/* Initializes apache.conf by inserting a command that causes the
   TeamSCOPE-specific Apache configuration file to be included. */
static void
init_apache (void)
{
  struct st s;

  st_init (&s, 256);
  st_printf (&s, "Include %s/apache.conf\n", std_get (conf, "conf_dir"));
  update_foreign_config (std_get (conf, "httpd_conf_file"), st_value (&s),
			 NULL, NULL);
  st_destroy (&s);
}

/* Initializes the TeamSCOPE timezones file. */
static void
init_timezones (void)
{
  struct st s;
  char *dst_fn;
  FILE *src, *dst;

  /* FIXME: shouldn't depend on external programs for this. */
  st_init (&s, 128);
  st_printf (&s,
	     "find '%s' -type f -print | sed -e 's#%s/\\?##;' "
	     "| grep ^[A-Z] | sort",
	     std_get (conf, "zoneinfo_dir"), std_get (conf, "zoneinfo_dir"));

  src = open_pipe (st_value (&s), "r");
  st_clear (&s);

  dst_fn = blp_asprintf ("%s.new.%ld",
			 std_get (conf, "timezones_file"), (long) getpid ());
  dst = open_file (dst_fn, "w");

  copy (dst, src);
  close_file (dst, dst_fn);
  close_pipe (src, "subprocess");
  replace_file (std_get (conf, "timezones_file"), NULL);

  free (dst_fn);
  st_destroy (&s);
}

/* Initializes file FN to have TeamSCOPE-specific content. */
static void
init_update_file (const char *fn)
{
  if (check_inited (fn))
    log ("%s: already set up - use `scope-db update' to update contents", fn);
  else
    update_foreign_config (fn, "", NULL, NULL);

  log ("** don't forget to run `scope-db update'");
}

/* Updates the TeamSCOPE-specific Apache configuration file. */
static void
update_apache_conf (void)
{
  const char *www_root;
  struct st s;

  www_root = std_get (conf, "www_root");

  st_init (&s, 1024);
  st_printf
    (&s,
     "# TeamSCOPE Apache authentication module.\n"
     "LoadModule auth_TeamSCOPE_module %s/mod_auth_scope.so\n\n"
     "# TeamSCOPE directories.\n"
     "Alias %s/doc %s\n"
     "Alias %s/f/ %s/www/f/\n"
     "Alias %s/pb/ %s/www/pb/\n"
     "Alias %s/t/ %s/www/t/\n"
     "Alias %s %s/www\n\n"
     "<Directory \"%s/www\">\n"
     "Options +ExecCGI\n"
     "DirectoryIndex scope.cgi\n"
     "</Directory>\n\n"
     "<Directory \"%s\">\n"
     "Options Indexes\n"
     "AllowOverride None\n"
     "allow from all\n"
     "</Directory>\n\n"
     "<Directory \"%s/www\">\n"
     "Options Indexes\n"
     "AllowOverride None\n"
     "allow from all\n"
     "</Directory>\n\n"
     "<Directory \"%s/*/shared/www\">\n"
     "Options +Includes\n"
     "AllowOverride AuthConfig Indexes\n"
     "</Directory>\n\n"
     "# Team directories.\n\n",
     std_get (conf, "arch_data_dir"),
     www_root, std_get (conf, "doc_dir"),
     www_root, std_get (conf, "share_data_dir"),
     www_root, std_get (conf, "share_data_dir"),
     www_root, std_get (conf, "share_data_dir"),
     www_root, std_get (conf, "arch_data_dir"),
     std_get (conf, "arch_data_dir"),
     std_get (conf, "doc_dir"),
     std_get (conf, "share_data_dir"),
     std_get (conf, "home_base"));

  connect_db ();
  
  {
    db_result *r;
    size_t i;
    
    db_query ("SELECT grp, webdir FROM groups", DB_TUPLES, &r);

    for (i = 0; i < dbr_nrows (r); i++)
      {
	update_team_site (dbr_value (r, i, 0));
	st_printf (&s, "Alias /teams/%s %s\n",
		   dbr_value (r, i, 0), dbr_value (r, i, 1));
      }
    dbr_destroy (r);
  }
  
  {
    char *fn = blp_asprintf ("%s/apache.conf", std_get (conf, "conf_dir"));
    log ("updating %s...", fn);
    replace_file (fn, st_value (&s));
    free (fn);
  }

  st_destroy (&s);
}

/* Status of exim configuration file parsing. */
struct exim_state 
  {
    int count;		/* Number of section ends encountered. */
    char *aliases;	/* Aliases to print into rewrite section. */
  };

static void
exim_config_callback (FILE *dst, char *line, void *param)
{
  const char *p = line + strspn (line, " \t\r\n");
  if (!strncmp (p, "end", 3) && p[strspn (p + 3, " \t\r\n") + 3] == '\0')
    {
      struct exim_state *state = param;
      state->count++;
      if (state->count == 2)
	{
	  fprintf (dst,
		   "#SCOPE-BEG#\n"
		   "teamscope:\n"
		   "  driver = pipe\n"
		   "  path = /bin:/sbin:/usr/bin:/usr/sbin\n"
		   "  command = %s/scope-db --received-mail\n"
		   "  user = scope\n"
		   "  group = scope\n"
		   "#SCOPE-END#\n\n"
		   "end\n\n"
		   "#SCOPE-BEG#\n"
		   "teamscope:\n"
		   "  local_parts = scope-gw\n"
		   "  driver = smartuser\n"
		   "  transport = teamscope\n"
		   "#SCOPE-END#\n",
		   std_get (conf, "sbin_dir"));
	}
      else if (state->count == 5)
	{
	  fputs ("end\n\n"
		 "#SCOPE-BEG#\n", dst);
	  fputs (state->aliases, dst);
	  fputs ("#SCOPE-END#\n", dst);
	}
      else
	fputs (line, dst);
    }
  else
    fputs (line, dst);
}

/* Only certain characters are allowed in names output to mailer
   configuration files, because we do not want the user fscking up
   these files and causing all sorts of security problems.  This
   function returns nonzero only if NAME is okay. */
int
exim_name_okay (const char *name)
{
  int okay = (strspn (name,
		      "abcdefghijklmnopqrstuvwxyz"
		      "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
		      "0123456789%()[].-_") == strlen (name));
  if (!okay)
    log ("name `%s' not okay for mailer configuration", name);

  return okay;
}

/* Some characters, such as space, in the email address will screw up
   exim, which does not have any allowances for escapes there.
   These characters should never appear in an email address anyway,
   but who knows what the k1ddies will try.
   Returns nonzero only if EMAIL is okay. */
int
exim_email_okay (const char *email)
{
  int okay = (strspn (email,
		      "abcdefghijklmnopqrstuvwxyz"
		      "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
		      "0123456789@!$%&()+-./=?[]_{}~") == strlen (email));

  if (!okay)
    log ("email `%s' not okay for mailer configuration", email);

  return okay;
}

/* Updates /etc/exim.conf. */
static void
update_exim_conf (void)
{
  struct st s;
  const char *mail_domain;

  db_result *r;
  size_t i;

  if (!check_inited (std_get (conf, "exim_conf_file")))
    return;
  log ("updating %s...", std_get (conf, "exim_conf_file"));

  st_init (&s, 1024);
  mail_domain = std_get (conf, "mail_domain");

  db_query ("SELECT usr, email FROM users WHERE recordable ORDER by usr",
	    DB_TUPLES, &r);
  for (i = 0; i < dbr_nrows (r); i++)
    {
      const char *usr = dbr_value (r, i, 0);
      const char *email = dbr_value (r, i, 1);
      
      if (email == NULL || email[0] == '\0'
	  || strchr (email, '@') == NULL)
	continue;

      /* Only certain characters allowed in usernames, for the same
	 reason.  Actually we could just escape $ and \ with \ here,
	 but this is easier and more paranoid at the same time. */
      if (!exim_email_okay (email) || !exim_name_okay (usr))
	continue;		  

      /* Kluge! */
      if (strstr (email, "@msu.edu")
	  || strstr (email, "@pilot.msu.edu")
	  || strstr (email, "@egr.msu.edu"))
	{
	  int len = strchr (email, '@') - email;
	  st_printf (&s,
		     "%.*s@msu.edu\t%s@%s\tfr\n"
		     "%.*s@pilot.msu.edu\t%s@%s\tfr\n"
		     "%.*s@egr.msu.edu\t%s@%s\tfr\n",
		     len, email, usr, mail_domain,
		     len, email, usr, mail_domain,
		     len, email, usr, mail_domain);
	}
      else
	st_printf (&s, "%s\t%s@%s\tfr\n", email, usr, mail_domain);
    }
  dbr_destroy (r);

  {
    struct exim_state state;

    state.count = 0;
    state.aliases = st_value (&s);
    
    update_foreign_config (std_get (conf, "exim_conf_file"), NULL,
			   exim_config_callback, &state);
  }
  
  st_destroy (&s);
}

/* Updates /etc/aliases. */
static void
update_aliases (void)
{
  struct st s;

  const char *CC;
  db_result *r;
  size_t i;
  size_t last;

  if (!check_inited (std_get (conf, "aliases_file")))
    return;
  log ("updating %s...", std_get (conf, "aliases_file"));

  CC = std_get_nonempty (conf, "CC");

  st_init (&s, 1024);

  db_query ("SELECT grp, usr FROM user_group ORDER BY grp, usr",
	    DB_TUPLES, &r);
  last = (size_t) -1;
  for (i = 0; i < dbr_nrows (r); i++)
    {
      const char *grp = dbr_value (r, i, 0);
      const char *usr = dbr_value (r, i, 1);

      if (!exim_name_okay (grp) || !exim_name_okay (usr))
	continue;
      
      if (last == (size_t) -1 || strcmp (grp, dbr_value (r, last, 0)))
	{
	  if (last != (size_t) -1)
	    st_putchar (&s, '\n');
	  st_printf (&s, "%s: %s", grp, usr);
	}
      else
	st_printf (&s, ", %s", usr);
      last = i;
    }
  st_concat (&s, "\n\n");
  dbr_destroy (r);

  db_query ("SELECT usr, email, recordable FROM users ORDER BY usr",
	    DB_TUPLES, &r);
  for (i = 0; i < dbr_nrows (r); i++)
    {
      const char *usr = dbr_value (r, i, 0);
      const char *email = dbr_value (r, i, 1);
      const char *recordable = dbr_value (r, i, 2);

      if (!exim_name_okay (usr) || !exim_email_okay (email))
	continue;
      
      st_printf (&s, "%s: %s",
		 usr, 
		 (email != NULL && email[0] != '\0') ? email : usr);
      if (recordable[0] == 't' && CC != NULL)
	st_printf (&s, ", %s", CC);
      st_putchar (&s, '\n');
    }
  dbr_destroy (r);

  update_foreign_config (std_get (conf, "aliases_file"), st_value (&s),
			 NULL, NULL);
  st_destroy (&s);
}

static void
restart_stuff (void)
{
  char *restart = blp_asprintf ("%s/reload", CONF_DIR);

  if (file_exists (restart))
    shell (0, restart);
  else
    {
      log ("** don't forget to restart Apache and exim (if you use it)");
      log ("** `/etc/init.d/apache reload' and `/etc/init.d/exim reload'");
      log ("** might work, depending on your system setup.");
    }
}

/* FIXME: CONTINUE REVIEW HERE!!! */

/* User and group management front-end. */

static void
cmd_addgroup (int argc, char *argv[])
{
  char *grp = prompt_input (&argc, &argv, "groupname");
  char *fullname = prompt_input (&argc, &argv, "group full name");

  add_group (grp, fullname);
  log ("** don't forget to run `scope-db update'");

  free (fullname);
  free (grp);
}

static void
cmd_delgroup (int argc, char *argv[])
{
  for (; argc--; argv++)
    del_group (argv[0]);
  log ("** don't forget to run `scope-db update'");
}

/* Tests whether NAME is available as a username.  Returns nonzero
   only if it is available. */
static int
name_available (const char *name)
{
  return getpwnam (name) == NULL && getgrnam (name) == NULL;
}

/* Generate and return (in a static buffer) a username for a user with
   real name FULLNAME. */
static const char *
generate_username (const char *fullname)
{
  static struct st *usr;
  const char *last_name;

  if (usr == NULL)
    {
      usr = xmalloc (sizeof *usr);
      st_init (usr, 32);
    }

  /* Swallow leading spaces in full name. */
  while (isspace ((unsigned char) *fullname))
    fullname++;
  
  /* Find last name. */
  {
    const char *p = strchr (fullname, '\0');
    while (p > fullname && isspace ((unsigned char) p[-1]))
      p--;
    while (p > fullname && !isspace ((unsigned char) p[-1]))
      p--;
    last_name = p;
  }

  /* Try last name by itself first. */
  {
    const char *p;
    
    for (p = last_name; *p; p++)
      if (isalpha ((unsigned char) *p))
	st_putchar (usr, tolower ((unsigned char) *p));
    st_truncate (usr, 8);
    if (st_length (usr) > 0 && name_available (st_value (usr)))
      return st_value (usr);
  }

  /* Append first name. */
  if (st_length (usr) < 8 && fullname != last_name)
    {
      const char *p;
      for (p = fullname; *p != '\0' && !isspace ((unsigned char) *p); p++)
	if (isalpha ((unsigned char) *p))
	  st_putchar (usr, tolower ((unsigned char) *p));
      st_truncate (usr, 8);
      if (st_length (usr) > 0 && name_available (st_value (usr)))
	return st_value (usr);
    }

  /* Try numeric suffixes. */
  {
    int i;

    for (i = 2; i < 1000; i++)
      {
	/* `1's are too easily confused with `l's. */
	if (i % 10 == 1 || i / 10 % 10 == 1 || i / 100 == 1)
	  continue;
	st_truncate (usr, i < 10 ? 7 : i < 100 ? 6 : 5);
	st_printf (usr, "%d", i);
	if (name_available (st_value (usr)))
	  return st_value (usr);
      }
  }

  /* Give up. */
  fail ("can't find available username for `%s'", fullname);
}

static void
cmd_adduser (int argc, char *argv[])
{
  const char *grp = prompt_input (&argc, &argv, "groupname");
  const char *usr = prompt_input (&argc, &argv, "username, if preference");
  const char *fullname = prompt_input (&argc, &argv, "user full name");
  const char *email
    = prompt_input (&argc, &argv, "upstream email address, if any");

  if (!strcmp (usr, ""))
    {
      usr = generate_username (fullname);
      printf ("## Username: %s\n", usr);
    }
  
  add_user (grp, usr, fullname, email);
  log ("** don't forget to run `scope-db update'");
  log ("** use `scope-db mail --pending' to notify the user by email");
  log ("** or `scope-db passwd %s' to set a password by hand", usr);
}

static void
cmd_deluser (int argc, char *argv[])
{
  for (; argc--; argv++)
    de_luser (argv[0]);
  
  log ("** don't forget to run `scope-db update'");
}

static void
cmd_addusertogroup (int argc, char *argv[])
{
  char *usr = prompt_input (&argc, &argv, "username");
  char *grp = prompt_input (&argc, &argv, "groupname");

  connect_db ();
  add_user_to_teamscope_group (usr, grp);
  add_user_to_unix_group (usr, grp);
  log ("** don't forget to run `scope-db update'");
  log ("** use `scope-db mail --pending' to notify the user by email");

  free (grp);
  free (usr);
}

static void
cmd_deluserfromgroup (int argc, char *argv[])
{
  char *usr = prompt_input (&argc, &argv, "username");
  char *grp = prompt_input (&argc, &argv, "groupname");

  connect_db ();
  del_user_from_teamscope_group (usr, grp);
  del_user_from_unix_group (usr, grp);
  log ("** don't forget to run `scope-db update'");

  free (grp);
  free (usr);
}

static void
set_trim (const char *s, const char *e, struct st *dst)
{
  while (s < e && isspace ((unsigned char) *s))
    s++;
  while (s < e && isspace ((unsigned char) e[-1]))
    e--;

  st_clear (dst);
  st_concat_buffer (dst, s, e - s);
}

static int
parse_batch_line (const char *line, struct st *a, struct st *b, int *type)
{
  const char *s, *e;

  printf ("received \"%s\"\n", line);
  s = line + strspn (line, " \t\r\n");
  e = s + strcspn (s, ":,");
  printf ("first field: \"%.*s\"\n", e - s, s);
  if (*e != ':' && *e != ',')
    return 0;
  *type = *e;

  set_trim (s, e, a);

  s = e + 1;
  e = strchr (s, '\0');
  set_trim (s, e, b);

  return st_length (a) > 0 && st_length (b) > 0;
}

static void
cmd_batch (int argc unused, char *argv[] unused)
{
  char *line = NULL;
  size_t line_size = 0;
  int ln = 0;

  struct st a, b;
  int type;
  
  char grp[9] = "";

  st_init (&a, 32);
  st_init (&b, 32);

  connect_db ();
  
  for (;;)
    {
      ssize_t line_len = getline (&line, &line_size, stdin);
      if (line_len == -1)
	break;
      ln++;
      
      {
	char *p = strchr (line, '#');
	if (p != NULL)
	  *p = '\0';
      }

      if (empty (line))
	continue;

      if (!parse_batch_line (line, &a, &b, &type))
	fail ("line %d: malformed input", ln);

      if (type == ':')
	{
	  validate_name (st_value (&a), "group name");
	  add_group (st_value (&a), st_value (&b));
	  strcpy (grp, st_value (&a));
	}
      else
	{
	  db_result *r;
	  
	  if (grp[0] == '\0')
	    fail ("line %d: group not yet selected", ln);

	  db_query ("SELECT * FROM users WHERE "
		    "name = '%sq' AND email = '%sq'",
		    DB_FORMAT | DB_TUPLES, &r,
		    st_value (&a), st_value (&b));
	  if (dbr_nrows (r) > 0)
	    log ("line %d: user named `%s' with email `%s' already exists, "
		 "not re-creating", ln, st_value (&a), st_value (&b));
	  else
	    add_user (grp, generate_username (st_value (&a)), st_value (&a), st_value (&b));
	  dbr_destroy (r);	  
	}
    }

  free (line);
  st_destroy (&a);
  st_destroy (&b);
}

static void
cmd_mail (int argc, char *argv[])
{
  int reset = 1;
  int i;
  
  if (argc == 0)
    fail ("`mail' command requires arguments");

  connect_db ();
  db_begin ();

  for (i = 0; i < argc; i++)
    {
      if (!strcmp (argv[i], "--no-reset"))
	reset = 0;
      else if (!strcmp (argv[i], "--pending"))
	{
	  db_result *r;
	  size_t j;

	  db_query ("SELECT DISTINCT usr FROM user_group "
		    "WHERE sent_email = 'f'",
		    DB_TUPLES, &r);
	  for (j = 0; j < dbr_nrows (r); j++)
	    send_user_mail (dbr_value (r, j, 0), 0);
	  dbr_destroy (r);
	}
      else
	send_user_mail (argv[i], reset);
    }

  db_end ();
}

/* Modified from GNU libc manual example. */
char *
my_getpass (char *prompt)
{
  char *line = NULL;
  size_t n = 0;
  int len;

  struct termios old;
  int got_termios;

  fputs (prompt, stdout);
  fflush (stdout);
  
  /* Turn echoing off, but don't worry if we can't; we might be talking to a pipe. */
  if (tcgetattr (fileno (stdin), &old) == 0)
    {
      struct termios new = old;
      new.c_lflag &= ~ECHO;
      tcsetattr (fileno (stdin), TCSAFLUSH, &new);

      got_termios = 1;
    }
  else
    got_termios = 0;
     
  /* Read the password.  Remove trailing newline. */
  len = getline (&line, &n, stdin);
  if (len == -1)
    return xstrdup ("");
  if (len > 0 && line[len - 1] == '\n')
    line[--len] = '\0';
  if (len > 0 && line[len - 1] == '\r')
    line[--len] = '\0';
     
  /* Restore terminal.  Print newline. */
  if (got_termios)
    tcsetattr (fileno (stdin), TCSAFLUSH, &old);
  putchar ('\n');
     
  return line;
}

static void
cmd_passwd (int argc, char *argv[])
{
  char *usr = prompt_input (&argc, &argv, "user");
  char *passwd = my_getpass ("new password: ");
  char *passwd2 = my_getpass ("again: ");

  if (strcmp (passwd, passwd2))
    fail ("no match");
  if (!strcmp (passwd, ""))
    log ("warning: empty password");

  set_password (usr, passwd, 0);

  free (usr);
  free (passwd);
  free (passwd2);
}

static void
cmd_lock (int argc, char *argv[])
{
  for (; argc--; argv++)
    {
      set_password (argv[0], NULL, 1);
      shell (1, "passwd -l -- '%s'", argv[0]);
    }
}

static void
cmd_unlock (int argc, char *argv[])
{
  for (; argc--; argv++)
    {
      set_password (argv[0], NULL, 0);
      shell (1, "passwd -u -- '%s'", argv[0]);
    }
}

static void
cmd_email (int argc, char *argv[])
{
  char *usr = prompt_input (&argc, &argv, "user");
  char *email = prompt_input (&argc, &argv, "upstream email address, if any");

  connect_db ();
  db_query ("UPDATE users SET email = '%sq' WHERE usr = '%sq'",
	    DB_FORMAT, NULL, email, usr);
}

static void
cmd_fullname (int argc, char *argv[])
{
  char *usr = prompt_input (&argc, &argv, "user");
  char *fullname = prompt_input (&argc, &argv, "user full name");

  connect_db ();
  db_query ("UPDATE users SET name = '%sq' WHERE usr = '%sq'",
	    DB_FORMAT, NULL, fullname, usr);
}

static void
cmd_group_admin (int argc, char *argv[])
{
  char *usr = prompt_input (&argc, &argv, "user");
  char *grp = prompt_input (&argc, &argv, "groupname");
  char *cmd = prompt_input (&argc, &argv, "command (yes/no/query)");

  usr_to_uid (usr);
  grp_to_gid (grp);
  connect_db ();
  if (!strcmp (cmd, "yes"))
    {
      db_query ("UPDATE user_group SET may_admin = 't' "
		"WHERE usr = '%sq' AND grp = '%sq'",
		DB_FORMAT, NULL, usr, grp);
      log ("added group admin privileges for %s in %s", usr, grp);
    }
  else if (!strcmp (cmd, "no"))
    {
      db_query ("UPDATE user_group SET may_admin = 'f' "
		"WHERE usr = '%sq' AND grp = '%sq'",
		DB_FORMAT, NULL, usr, grp);
      log ("revoked group admin privileges for %s in %s", usr, grp);
    }
  else if (!strcmp (cmd, "query"))
    {
      db_result *r;

      db_query ("SELECT may_admin FROM user_group "
		"WHERE usr = '%sq' AND grp = '%sq'",
		DB_FORMAT | DB_TUPLES, &r, usr, grp);
      log ("%s %s admin privileges in %s",
	   usr, dbr_value (r, 0, 0)[0] == 't' ? "has" : "lacks", grp);
      dbr_destroy (r);
    }
  else
    fail ("unrecognized subcommand `%s'", cmd);
}

/* Daily maintenance front-end. */

static void backup_database (void);
static void cleanup_dir (const char *dir);
static void cleanup_sessions (void);

static void
cmd_daily_backup (int argc unused, char *argv[] unused)
{
  backup_database ();
}

static void
cmd_daily_cleanup (int argc unused, char *argv[] unused)
{
  cleanup_sessions ();
}

static void
cmd_daily_maint (int argc unused, char *argv[] unused)
{
  backup_database ();
  cleanup_sessions ();
}

/* Backs up the TeamSCOPE database into /var/backups, rotating out old
   backups. */
static void
backup_database (void)
{
  struct st a, b;
  size_t len;
  int count;
  int i;

  count = stoi (std_get (conf, "backup_count"));
  st_init (&a, 64);
  st_printf (&a, "%s/scope", std_get (conf, "backup_dir"));
  len = st_length (&a);
  st_create (&b, st_value (&a));

  st_printf (&a, ".%d.gz", count);
  unlink (st_value (&a));

  for (i = count; i >= 0; i--)
    {
      st_truncate (&a, len);
      st_truncate (&b, len);
      if (i != 0)
	{
	  st_printf (&a, ".%d.gz", i);
	  st_printf (&b, ".%d.gz", i + 1);
	}
      else
	st_printf (&b, ".1");

      if (rename (st_value (&a), st_value (&b)) == -1 && errno != ENOENT)
	fail ("renaming %s to %s: %s",
	      st_value (&a), st_value (&b), strerror (errno));
    }

  /* At this point we have ".../scope" in A and ".../scope.1" in B.
     This turns out to be convenient. */
  if (file_exists (st_value (&b)))
    shell (1, "gzip -9 '%s'", st_value (&b));
  dump_database (st_value (&a));
}

/* Cleans up the contents of DIR, removing files over a day old. */
static void
cleanup_dir (const char *dir)
{
  DIR *d;
  time_t t;
  
  if (chdir (dir) == -1)
    fail ("%s: chdir: %s", dir, strerror (errno));

  d = opendir (".");
  if (d == NULL)
    fail ("%s: opendir: %s", dir, strerror (errno));

  time (&t);
  for (;;)
    {
      struct stat s;
      
      struct dirent *ent = readdir (d);
      if (ent == NULL)
	break;
      if (ent->d_name[0] == '.'
	  && (ent->d_name[1] == '\0'
	      || (ent->d_name[1] == '.' && ent->d_name[2] == '\0')))
	continue;

      if (stat (ent->d_name, &s) == -1)
	{
	  log ("%s/%s: stat: %s", dir, ent->d_name, strerror (errno));
	  continue;
	}

      if (difftime (t, s.st_atime) > 60.0 * 60 * 24)
	if (unlink (ent->d_name) == -1)
	  log ("%s/%s: unlink: %s", dir, ent->d_name, strerror (errno));
    }
  
  closedir (d);
}

/* Cleans up the TeamSCOPE state directories, removing old files. */
static void
cleanup_sessions (void)
{
  char *s;

  s = blp_asprintf ("%s/trusted", std_get (conf, "state_dir"));
  cleanup_dir (s);
  free (s);
  
  s = blp_asprintf ("%s/untrusted", std_get (conf, "state_dir"));
  cleanup_dir (s);
  free (s);
}

/* Foreign module helper functions. */

/* Returns the uid for the `scope' user, the one who is allowed to
   access the database.  This is a helper function for db.c. */
int
get_scope_uid (void)
{
  return scope_uid ();
}

/* Print a helpful log message prefix to stderr of the form `May 13
   19:31:30 1999 scope[31230]'.  Helper function for error.c. */
void
msg_prefix (FILE *file, const char *function_name unused)
{
  fprintf (file, "%s: ", pgmname);
}

/* Setting our uid. */

/* Sets our effective uid to UID, corrresponding to user USERNAME. */
static void
su (int uid, const char *username)
{
  if (seteuid (uid) != 0)
    fail ("can't set euid to `%s'", username);

  /* putenv() retains a reference to its string argument, so we must
     preserve it.  We can't use a struct st because that the
     string's location might change, leading to undefined behavior
     when putenv() looks through the environment. */
  {
    static char *s;

    if (s == NULL)
      {
	s = xmalloc (64);
	sprintf (s, "USER=%.32s", username);
	putenv (s);
      }
    else
      sprintf (s, "USER=%.32s", username);
  }
}

/* Returns the uid for user `scope'. */
static int
scope_uid (void)
{
  static uid_t uid = 0;
  return cache_uid ("scope", &uid);
}

/* Returns the uid for user `postgres'. */
static int
postgres_uid (void)
{
  static uid_t uid = 0;
  return cache_uid ("postgres", &uid);
}

/* Returns the uid for user `nobody'. */
static int
nobody_uid (void)
{
  static uid_t uid = 0;
  return cache_uid ("nobody", &uid);
}

/* Sets *UID to the uid for USERNAME and returns the same uid.  If
   *UID is nonzero, assumes that is the correct uid for that user. */
static uid_t
cache_uid (const char *username, uid_t *uid)
{
  if (*uid == 0)
    {
      *uid = usr_to_uid (username);
      if (*uid == 0)
	fail ("user %s has uid 0", username);
    }
  
  return *uid;
}

/* User and group queries. */

static int list_contains_string (char **list, const char *search);
static int compare_id (const void *pa, const void *pb);
static void add_to_list (unsigned long x, unsigned long **list,
			 int *n, int *m);

/* Returns nonzero only if USR is in GRP. */
static int
user_is_member (const char *usr, const char *grp)
{
  struct group *group = getgrnam (grp);
  
  if (group == NULL)
    fail ("group %s doesn't really exist", grp);

  return list_contains_string (group->gr_mem, usr);
}

/* Puts a comma-separated list of groups that USR belongs to into S.
   If ADD is non-null then it is appended to the list.  If DEL is
   non-null then it is omitted from the list. */
static void
get_users_groups (const char *usr, struct st *s,
		  const char *add, const char *del)
{
  int n = 0;
  
  setgrent ();
  for (;;)
    {
      struct group *group = getgrent ();
      if (group == NULL)
	break;

      if (list_contains_string (group->gr_mem, usr)
	  && (del == NULL || strcmp (del, group->gr_name)))
	{
	  if (n++ > 0)
	    st_putchar (s, ',');
	  st_concat (s, group->gr_name);
	}
    }

  if (add != NULL)
    {
      if (n++ > 0)
	st_putchar (s, ',');
      st_concat (s, add);
    }

  endgrent ();
}

/* Returns the uid for USR. */
static uid_t
usr_to_uid (const char *usr)
{
  struct passwd *passwd = getpwnam (usr);
  if (passwd == NULL)
    fail ("no such user `%s'", usr);
  return passwd->pw_uid;
}

/* Returns the gid for GRP. */
static gid_t
grp_to_gid (const char *grp)
{
  struct group *group = getgrnam (grp);
  if (group == NULL)
    fail ("no such group `%s'", grp);
  return group->gr_gid;
}

/* Finds the first unused uid/gid pair between MIN and MAX.  Results
   are put into *UID and *GID. */
static void
first_avail_id (unsigned long min, unsigned long max, uid_t *uid, gid_t *gid)
{
  unsigned long *ids, *p;
  size_t n_ids, m_ids;

  n_ids = m_ids = 0;
  ids = NULL;
  setpwent ();
  for (;;)
    {
      struct passwd *passwd = getpwent ();
      if (passwd == NULL)
	break;
      add_to_list (passwd->pw_uid, &ids, &n_ids, &m_ids);
    }
  endpwent ();
  setgrent ();
  for (;;)
    {
      struct group *group = getgrent ();
      if (group == NULL)
	break;

      add_to_list (group->gr_gid, &ids, &n_ids, &m_ids);
    }
  endgrent ();
  qsort (ids, n_ids, sizeof *ids, compare_id);
  
  p = ids;
  while (min <= max)
    {
      if ((size_t) (p - ids) >= n_ids || min < *p)
	break;
      if (min > *p)
	{
	  p++;
	  continue;
	}
      if (min == *p)
	min++;
    }
  free (ids);

  if (min > max)
    fail ("out of uid/gid pairs");

  *uid = min;
  *gid = min;
}

/* Returns nonzero if SEARCH is in LIST, which must be terminated by a
   null pointer sentinel. */
static int
list_contains_string (char **list, const char *search)
{
  for (; *list; list++)
    if (!strcmp (*list, search))
      return 1;
  return 0;
}

/* Compares unsigned longs pointed to by PA and PB and
   returns a strcmp()-type result. */
static int
compare_id (const void *pa, const void *pb)
{
  const unsigned long *a = pa;
  const unsigned long *b = pb;

  if (*a < *b)
    return -1;
  else if (*a > *b)
    return 1;
  else
    return 0;
}

/* Adds X to the list of unsigned longs in *LIST, reallocating memory
   as necessary.  *LIST is length *N, with room for *M elements. */
static void
add_to_list (unsigned long x, unsigned long **list, int *n, int *m)
{
  if (*n >= *m)
    {
      *m = *m * 2 + 16;
      *list = xrealloc (*list, sizeof **list * *m);
    }

  (*list)[(*n)++] = x;
}

/* Database functions. */

/* Connects to the database as user `scope', if not already
   connected. */
void
connect_db (void)
{
  if (db_open_p ())
    return;

  su_scope ();
  db_open (NULL);
  su_root ();
}

/* Creates the TeamSCOPE database. */
static void
create_db (void)
{
  /* Must be done as the PostgreSQL superuser. */
  su_postgres ();
  db_open ("template1");
  db_query ("CREATE USER scope", 0, NULL);
  db_query ("CREATE DATABASE scope", 0, NULL);
  db_close ();

  /* Open TeamSCOPE database and add our special functions and
     operators, which must also be done as the PostgreSQL
     superuser. */
  db_open ("scope");
  db_query
    ("DROP OPERATOR <(int4[],int4[]);\n\n"
     "DROP FUNCTION _int4_lt(_int4,_int4);\n\n"
     "DROP FUNCTION _int4_depth(_int4);\n\n",
     DB_FAIL_OK, NULL);
  db_query 
    ("CREATE FUNCTION _int4_lt(_int4,_int4) RETURNS bool "
     "AS '%sq/_int4.o' LANGUAGE 'c';\n\n"
     "CREATE OPERATOR <(leftarg=_int4,rightarg=_int4,procedure=_int4_lt);\n\n"
     "CREATE FUNCTION _int4_depth(_int4) RETURNS int "
     "AS '%sq/_int4.o' LANGUAGE 'c';\n\n"
     "CREATE FUNCTION last_word(text) RETURNS text "
     "AS '%sq/text.o' LANGUAGE 'c';\n\n",
     DB_FORMAT, NULL,
     std_get (conf, "arch_data_dir"),
     std_get (conf, "arch_data_dir"),
     std_get (conf, "arch_data_dir"));
  db_close ();

  su_root ();
}

/* Creates and initializes the TeamSCOPE database. */
static void
init_db (void)
{
  create_db ();
  
  connect_db ();
  db_query
    ("CREATE TABLE activities\n"
     "(\n"
     "	date		datetime	not null,\n"
     "	s1		int4		not null,\n"
     "	usr		varchar(8)	not null,\n"
     "	grp		varchar(8)	not null,\n"
     "	cat		int2		not null,\n"
     "	type		int2		not null,\n"
     "	s2		int4		not null,\n"
     "	ref_seq		int4,\n"
     "	seq		int4		primary key\n"
     ");\n\n"
     "CREATE TABLE catchup \n"
     "(\n"
     "	usr		varchar(8)	not null, \n"
     "	grp		varchar(8)	not null,\n"
     "	file		int4		not null, \n"
     "	last_read	int4		not null,\n"
     "	seq		int4		primary key,\n"
     "\n"
     "	unique (usr, grp, file)\n"
     ");\n\n"
     "CREATE TABLE conflicts \n"
     "(\n"
     "	seq1		int4		not null, \n"
     "	seq2		int4		not null, \n"
     "	reported	bool		not null,\n"
     "	seq		int4		primary key,\n\n"
     "	unique (seq1, seq2),\n"
     "	check (seq1 < seq2)\n"
     ");\n\n"
     "CREATE TABLE event_resource\n"
     "(\n"
     "	event		int4		not null,\n"
     "	resource	int4		not null,\n"
     "	seq		int4		primary key,\n"
     "\n"
     "	unique (event, resource)\n"
     ");\n\n"
     "CREATE TABLE events\n"
     "(\n"
     "	start		datetime	not null,\n"
     "	finish		datetime	not null,\n"
     "	title		int4		not null,\n"
     "	note		text		not null,\n"
     "	owner		varchar(8)	not null,\n"
     "	grp		varchar(8)	not null,\n"
     "	seq		int4		primary key,\n"
     "  source		int4		not null\n"
     ");\n\n"
     "CREATE TABLE groups\n"
     "(\n"
     "	grp		varchar(8)	primary key,\n"
     "	name		text		not null,\n"
     "	shareddir	text		not null,\n"
     "	webdir		text		not null,\n"
     "	md5		text[]\n"
     ");\n\n"
     "CREATE TABLE notes\n"
     "(\n"
     "	file		int4		not null,\n"
     "	grp		varchar(8)	not null,\n"
     "	usr		varchar(8)	not null,\n"
     "	content		text		not null,\n"
     "	title		int4		not null,\n"
     "	date		datetime	not null,\n"
     "	thread		int4[]		not null,\n"
     "	seq		int4		primary key\n"
     ");\n\n"
     "CREATE TABLE resources\n"
     "(\n"
     "	name		text		not null,\n"
     "	seq		int4		primary key\n"
     ");\n\n"
     "CREATE TABLE strings\n"
     "(\n"
     "	string		text		not null unique,\n"
     "	seq		int4		primary key,\n"
     "	grp		varchar(8)\n"
     ");\n\n"
     "CREATE TABLE sync_mail\n"
     "(\n"
     "seq int4\n"
     ");\n\n"
     "CREATE TABLE uselog\n"
     "(\n"
     "	usr		varchar(8)	not null,\n"
     "	t		datetime	not null,\n"
     "	loc		text,\n"
     "  grp             varchar(8),\n"
     "  query           text\n"
     ");\n\n"
     "CREATE TABLE user_group\n"
     "(\n"
     "	grp		varchar(8)	not null,\n"
     "	usr		varchar(8)	not null,\n"
     "	last_read_web	int4,\n"
     "	last_read_mail	int4,\n"
     "	seq		int4		primary key,\n"
     "	sent_email	bool		not null,\n"
     "  may_admin	bool		not null,\n"
     "	unique (grp, usr)\n"
     ");\n\n"
     "CREATE TABLE users\n"
     "(\n"
     "	usr		varchar(8)	primary key,\n"
     "	name		text		not null,\n"
     "	email		text,\n"
     "	recordable	bool		not null,\n"
     "	idle		int4,\n"
     "	last_usage	datetime,\n"
     "  emailbody	int4,\n"
     "  primarygrp	varchar(8),\n"
     "  welcome		bool\n"
     ");\n\n"
     "CREATE TABLE version\n"
     "(\n"
     "	major		int2,\n"
     "	minor		int2\n"
     ");\n\n",
     0, NULL);

  db_query ("INSERT INTO version VALUES (%d, %d)",
	    DB_FORMAT, NULL, DB_VERSION_MAJOR, DB_VERSION_MINOR);
  db_query ("CREATE SEQUENCE seq", 0, NULL);
  db_query ("INSERT INTO sync_mail VALUES (nextval ('seq'))", 0, NULL);

  /* These indexes speed up the ``Viewed By'' query in
     messages.c:show_single_message() incredibly, from ~2 seconds to
     negligible. */
  db_query ("CREATE INDEX activities_usr ON activities (usr);\n\n"
	    "CREATE UNIQUE INDEX users_user ON users (usr);\n",
	    0, NULL);

  /* Often we want files that belong to a particular group. */
  db_query ("CREATE INDEX strings_grp ON strings (grp)", 0, NULL);

  db_close ();
}

/* Dumps a backup of the TeamSCOPE database into DST_FN. */
static void
dump_database (const char *dst_fn)
{
  /* The `postgres' user probably doesn't have permissions to write to
     the file that we want to dump to, so we can't simply run `pg_dump
     > file'.  This alternate method should work better.  */
  
  FILE *in, *out;

  out = open_file (dst_fn, "wb");
  if (fchmod (fileno (out), S_IRUSR | S_IWUSR) == -1)
    fail ("%s: fchmod: %s", dst_fn, strerror (errno));

  su_postgres ();
  in = open_pipe ("pg_dump scope", "r");
  su_root ();
  
  copy (out, in);

  close_file (out, dst_fn);
  close_pipe (in, "pg_dump scope");
}

/* Files and directories. */

/* Opens file FN with mode MODE and returns a pointer to a FILE.
   Handles errors internally. */
FILE *
open_file (const char *fn, const char *mode)
{
  FILE *f = fopen (fn, mode);
  if (f != NULL)
    return f;

  if (mode[0] == 'r')
    fail ("%s: open for reading: %s", fn, strerror (errno));
  else
    fail ("%s: open for writing: %s", fn, strerror (errno));
}

/* Opens a pipe to command FN with mode MODE and returns a pointer to
   a FILE.  Handles errors internally. */
FILE *
open_pipe (const char *cmd, const char *mode)
{
  FILE *f = popen (cmd, mode);
  if (f != NULL)
    return f;

  if (mode[0] == 'r')
    fail ("%s: open subprocess for reading: %s", cmd, strerror (errno));
  else
    fail ("%s: open subprocess for writing: %s", cmd, strerror (errno));
}

/* Closes file F opened on file FN, handling errors internally. */
void
close_file (FILE *f, const char *fn)
{
  if (ferror (f) || fclose (f))
    fail ("%s: close: %s", fn, strerror (errno));
}

/* Closes pipe file F opened for command CMD with popen() or
   open_pipe(), handling errors internally. */
void
close_pipe (FILE *f, const char *cmd)
{
  if (ferror (f))
    fail ("%s: %s", cmd, strerror (errno));
  check_system_value (pclose (f), cmd);
}

/* Copies the contents of SRC into DST. */
static void
copy (FILE *dst, FILE *src)
{
  char buf[BUFSIZ];
  while (fwrite (buf, 1, fread (buf, 1, BUFSIZ, src), dst) == BUFSIZ)
    ;
}

/* Creates all the directories from the root up to TREE, if they don't
   already exist.  Directories are created with permissions MODE and
   are owned by user UID and group GID.  TREE should be an absolute
   path, beginning with `/'. */
static void
make_dir_tree (const char *tree, mode_t mode, uid_t uid, gid_t gid)
{
  struct st dir;

  st_init (&dir, strlen (tree) + 1);
  for (;;)
    {
      st_putchar (&dir, '/');
      while (*tree == '/')
	tree++;
      if (*tree == '\0')
	break;

      while (*tree != '\0' && *tree != '/')
	st_putchar (&dir, *tree++);

      if (-1 == mkdir (st_value (&dir), mode))
	{
	  if (errno == EEXIST)
	    continue;
	  else
	    fail ("%s: mkdir: %s", st_value (&dir), strerror (errno));
	}

      if (chmod (st_value (&dir), mode) == -1)
	fail ("%s: chmod: %s", st_value (&dir), strerror (errno));
      if (chown (st_value (&dir), uid, gid) == -1)
	fail ("%s: chown: %s", st_value (&dir), strerror (errno));
    }

  st_destroy (&dir);
}

/* Atomically replaces file `$base' by `$base.new.$$', where $base is
   BASE and $$ is this process's PID.  Backs up the old content of
   `$base' to `$base.old'.  If WITH is non-null, then `$base.new.$$'
   is first created with the provided content. */
static void
replace_file (const char *base, const char *with)
{
  char *new = blp_asprintf ("%s.new.%ld", base, (long) getpid ());
  char *old = blp_asprintf ("%s.old", base);

  if (with != NULL)
    {
      FILE *f = open_file (new, "w");
      if (fputs (with, f) < 0)
	fail ("%s: write: %s", new, strerror (errno));
      close_file (f, new);
    }

  /* Make a backup of the old file. */
  unlink (old);		/* Ignore error. */
  if (link (base, old) == -1 && errno != ENOENT)
    fail ("linking %s to %s: %s\n", base, old, strerror (errno));

  /* Move the new file into place. */
  if (rename (new, base) != 0)
    fail ("renaming %s to %s: %s\n", new, base, strerror (errno));

  free (old);
  free (new);
}

/* Checks whether foreign configuration file FN has content owned by
   TeamSCOPE; i.e., whether `scope-db --init' has been run for the
   file. */
static int
check_inited (const char *fn)
{
  size_t size = 0;
  char *line = NULL;

  int result = 0;

  FILE *f = fopen (fn, "r");
  if (f == NULL)
    {
      log ("%s: open: %s", fn, strerror (errno));
      return 0;
    }

  for (;;)
    {
      long len = getline (&line, &size, f);
      if (len == -1)
	break;
      
      if (strstr (line, "#SCOPE-BEG#"))
	{
	  result = 1;
	  break;
	}
    }

  free (line);
  close_file (f, fn);

  return result;
}

/* Updates foreign configuration file SRC_FN, replacing any
   TeamSCOPE-owned content by MOD.  If MOD is a null pointer, removes
   any TeamSCOPE-owned content.  If CALLBACK is nonnull then ordinary
   lines in the file are passed through that function with parameter
   PARAM instead of being written literally.
   Squeezes multiple blank lines to single blank lines. */
static void 
update_foreign_config (const char *src_fn, const char *mod,
		       foreign_config_callback *callback, void *param)
{
  size_t size = 0;
  char *line = NULL;
  
  FILE *src, *dst;
  char *dst_fn;

  int print;
  int blank;

  src = open_file (src_fn, "r");
  
  dst_fn = blp_asprintf ("%s.new.%ld", src_fn, (long) getpid ());
  dst = open_file (dst_fn, "w");

  print = 1;
  blank = 0;
  for (;;)
    {
      long len = getline (&line, &size, src);
      if (len == -1)
	{
	  if (ferror (src))
	    fail ("%s: read: %s", src_fn, strerror (errno));

	  free (line);
	  break;
	}

      if (strstr (line, "#SCOPE-BEG#"))
	print = 0;
      if (print)
	{
	  if (empty (line))
	    blank = 1;
	  else
	    {
	      if (blank)
		{
		  putc ('\n', dst);
		  blank = 0;
		}
	      
	      if (callback != NULL)
		callback (dst, line, param);
	      else
		fputs (line, dst);
	    }
	}
      if (strstr (line, "#SCOPE-END#"))
	print = 1;
    }

  if (mod != NULL)
    {
      fprintf (dst,
	       "\n#SCOPE-BEG#\n"
	       "# These lines added by scope-db at user request.\n"
	       "%s"
	       "#SCOPE-END#\n",
	       mod);
    }

  close_file (dst, dst_fn);
  close_file (src, src_fn);
  free (dst_fn);

  replace_file (src_fn, NULL);
}

/* Creates and locks a lockfile for FN, and returns a handle that can
   be passed to unlock_file() to unlock the file.  If BLOCK is
   nonzero, success is guaranteed because the function will wait until
   the file is available.  Otherwise, if another process has the lock
   or for some other reason the file cannot be locked, returns -1. */
static int
lock_file (const char *fn, int block)
{
  char *lock_fn = blp_asprintf ("%s.lock", fn);
  int handle = open (lock_fn, O_WRONLY | O_CREAT, 0666);
  if (handle == -1)
    fail ("%s: create: %s", lock_fn, strerror (errno));

  {
    struct flock lock;

    lock.l_type = F_WRLCK;
    lock.l_whence = SEEK_SET;
    lock.l_start = 0;
    lock.l_len = 0;

    if (fcntl (handle, block ? F_SETLKW : F_SETLK, &lock) == -1)
      {
	if (!block && (errno == EACCES || errno == EAGAIN))
	  {
	    close (handle);
	    handle = -1;
	  }
	else
	  fail ("%s: lock: %s", lock_fn, strerror (errno));
      }
  }
  
  free (lock_fn);
  return handle;
}

/* Unlocks lock for file FN.  HANDLE is the handle returned by
   lock_file(). */
static void
unlock_file (int handle, const char *fn)
{
  char *lock_fn = blp_asprintf ("%s.lock", fn);
  unlink (lock_fn);
  free (lock_fn);
  
  close (handle);
}

/* Obtains the scope-db global lock. */
static void
scope_lock (void)
{
  char *fn = blp_asprintf ("%s/%s", std_get (conf, "conf_dir"), "scope-db");
  lock_file (fn, 1);
  free (fn);
}

/* Obtains the scope-db mail lock.
   This lock is used to make sure only one --hourly-mail or
   --sync-mail process runs at a time, not because they need mutual
   exclusion but because otherwise if something blocks mail going out
   then they'll pile up and fill up the process table by running from
   cron often. */
static void
mail_lock (void)
{
  char *fn = blp_asprintf ("%s/%s", std_get (conf, "conf_dir"), "mail");
  if (lock_file (fn, 0) == -1)
    exit (EXIT_FAILURE);
  free (fn);
}

/* Returns nonzero only if a file named FN exists. */
int
file_exists (const char *fn)
{
  struct stat s;

  return stat (fn, &s) == 0;
}

/* Copies from file SRC_FN to DST.  Substitutions from keys to values
   in SUBST are made for each copied line.  Substitution takes the
   form $(KEY).  Lines that contain a substitution key for which the
   corresponding value is a null pointer are removed.  SUBST must be
   terminated by a sentinel formed of null pointers. */
static void
copy_file_subst (const char *src_fn, FILE *dst, struct std *subst)
{
  size_t size = 0;
  char *line = NULL;
  struct st translated;
  int ln = 0;

  FILE *src = open_file (src_fn, "r");

  st_init (&translated, 128);
  for (;;)
    {
      const char *p;
      
      if (-1 == getline (&line, &size, src))
	break;
      ln++;

      p = line;
      for (;;) 
	switch (*p)
	  {
	  case '$':
	    if (p[1] == '(')
	      {
		char *var;
		const char *value;
		const char *end = strchr (p, ')');
		if (end == NULL)
		  fail ("%s:%d: missing `)' in $(var) substitution",
			src_fn, ln);

		var = local_alloc (end - p - 1);
		memcpy (var, p + 2, end - p - 2);
		var[end - p - 2] = '\0';

		value = std_get_null (subst, var);
		if (value != NULL)
		  st_concat (&translated, value);
		else if (!std_contains (subst, var))
		  fail ("%s:%d: substitution for unknown variable %s",
			src_fn, ln, var);
		else
		  {
		    local_free (var);
		    goto skip_line;
		  }
		p = end + 1;
		
		local_free (var);
	      }
	    else if (p[1] == '$')
	      {
		st_putchar (&translated, '$');
		p += 2;
	      }
	    else
	      fail ("%s:%d: syntax error following `$'", src_fn, ln);
	    break;
	    
	  case '\0':
	    goto done;
	    
	  default:
	    st_putchar (&translated, *p++);
	    break;
	  }

    done:
      fputs (st_value (&translated), dst);

    skip_line:
      st_clear (&translated);
    }

  free (line);
  st_destroy (&translated);

  close_file (src, src_fn);
}

/* If file DST_FN doesn't already exist, copies file SRC_FN to DST_FN,
   setting DST_FN's permissions to MODE, its uid to 0, its gid to GID,
   and making substitutions in SUBST. */
static void
copy_file_conditionally (mode_t mode, gid_t gid,
			 const char *src_fn, const char *dst_fn,
			 struct std *subst, int index unused)
{
  FILE *dst;

  /* If the file already exists then don't replace it.
     FIXME: Use md5 like Perl version. */
  if (file_exists (dst_fn))
    return;

  log ("replacing %s...", dst_fn);

  dst = fopen (dst_fn, "wb");
  if (dst == NULL)
    {
      log ("can't open %s: %s", dst_fn, strerror (errno));
      return;
    }

  if (-1 == chown (dst_fn, 0, gid))
    fail ("%s: chown: %s", dst_fn, strerror (errno));
  if (-1 == chmod (dst_fn, mode))
    fail ("%s: chmod: %s", dst_fn, strerror (errno));

  copy_file_subst (src_fn, dst, subst);
  close_file (dst, dst_fn);
}

/* User and group management. */

/* Adds USR to the UNIX system group GRP. */
static void
add_user_to_unix_group (const char *usr, const char *grp)
{
  if (!user_is_member (usr, grp))
    {
      struct st command;

      log ("adding %s to UNIX group %s...", usr, grp);
      
      st_create (&command, "usermod -G ");
      get_users_groups (usr, &command, grp, NULL);
      st_printf (&command, " -- '%s'", usr);

      shell (0, st_value (&command));

      st_destroy (&command);
    }
}

/* Adds USR to TeamSCOPE group GRP. */
static void
add_user_to_teamscope_group (const char *usr, const char *grp)
{
  db_result *r;

  db_query ("SELECT count(*) FROM user_group "
	    "WHERE usr = '%sq' AND grp = '%sq'",
	    DB_FORMAT | DB_TUPLES, &r, usr, grp);
  if (stol (dbr_value (r, 0, 0)) == 0)
    {
      log ("inserting %s/%s into database USER_GROUP table...", usr, grp);
      db_query ("INSERT INTO user_group "
		"VALUES ('%sq', '%sq', 0, 0, nextval ('seq'), 'f', 'f')",
		DB_FORMAT, NULL, grp, usr);
    }
  dbr_destroy (r);
}

/* Verify that NAME is an acceptable user or group name.  TYPE
   contains "username" or "group name". */
static void
validate_name (const char *name, const char *type)
{
  if (strlen (name) > 8)
    fail ("%s `%s' is longer than 8 characters", type, name);
  
  {
    int capital = 0;
    const char *cp;
    for (cp = name; *cp; cp++)
      {
	if (!isalnum ((unsigned char) *cp) && *cp != '-' && *cp != '_')
	  fail ("invalid character `%c' in %s `%s'", *cp, type, name);
	if (isupper ((unsigned char) *cp) && !capital)
	  {
	    log ("warning: %s `%s' contains uppercase", type, name);
	    capital = 1;
	  }
      }
  }
}

/* Creates a new TeamSCOPE group and associated UNIX system group GRP
   with full name FULLNAME. */
static void
add_group (const char *grp, const char *fullname)
{
  gid_t gid;
  struct st shared_dir, web_dir;

  validate_name (grp, "group name");
  log ("creating group %s with full name \"%s\"...", grp, fullname);

  if (getgrnam (grp) == NULL)
    {
      log ("creating UNIX group %s...", grp);
      shell (1, "groupadd -- '%s'", grp);
    }

  gid = grp_to_gid (grp);

  /* Shared directory. */
  st_init (&shared_dir, 64);
  st_printf (&shared_dir, "%s/%s", std_get (conf, "home_base"), grp);
  if (!strncmp (std_get (conf, "home_style"), "hier", 4)) 
    {
      make_dir_tree (st_value (&shared_dir), 0755, 0, 0);
      st_concat (&shared_dir, "/shared");
    }
  make_dir_tree (st_value (&shared_dir), 06775, 0, gid);

  /* Web directory. */
  st_create (&web_dir, st_value (&shared_dir));
  st_concat (&web_dir, "/www");
  make_dir_tree (st_value (&web_dir), 06775, 0, grp_to_gid (grp));
  if (chown (st_value (&web_dir), 0, gid) == -1)
    fail ("%s: chown: %s", st_value (&web_dir), strerror (errno));

  /* Public web directory. */
  {
    size_t len = st_length (&web_dir);
    
    st_concat (&web_dir, "/public");
    make_dir_tree (st_value (&web_dir), 06775, 0, grp_to_gid (grp));
    if (chown (st_value (&web_dir), 0, gid) == -1)
      fail ("%s: chown: %s", st_value (&web_dir), strerror (errno));

    st_truncate (&web_dir, len);
  }
  
  connect_db ();
  
  {
    db_result *r;

    db_begin ();
    db_query ("SELECT count(*) FROM groups WHERE grp = '%sq'",
	      DB_FORMAT | DB_TUPLES, &r, grp);
    if (stol (dbr_value (r, 0, 0)) == 0)
      {
	log ("inserting %s into database GROUPS table...", grp);
	db_query ("INSERT INTO groups VALUES ('%sq', '%sq', '%sq', '%sq')",
		  DB_FORMAT, NULL,
		  grp, fullname, st_value (&shared_dir), st_value (&web_dir));
      }
    dbr_destroy (r);
    db_end ();
  }

  st_destroy (&shared_dir);
  st_destroy (&web_dir);

  update_team_site (grp);
}

/* Creates a new TeamSCOPE and UNIX system user USR in group GRP with
   full name FULLNAME and email address EMAIL. */
static void
add_user (const char *grp, const char *usr,
	  const char *fullname, const char *email)
{
  /* Fail out if GRP doesn't exist. */
  grp_to_gid (grp);

  validate_name (usr, "username");
  log ("adding %s (%s) to %s with email %s...", usr, fullname, grp, email);
  if (getpwnam (usr) != NULL)
    log ("system user named %s already exists", usr);
  else
    {
      struct st homedir;

      uid_t uid;
      gid_t gid;
	
      log ("creating system user %s...", usr);

      st_init (&homedir, 64);
      if (!strncmp (std_get (conf, "home_style"), "hier", 4))
	st_printf (&homedir, "%s/%s/%s", std_get (conf, "home_base"), grp, usr);
      else
	st_printf (&homedir, "%s/%s", std_get (conf, "home_base"), usr);

      first_avail_id (strtoul (std_get (conf, "first_uid"), NULL, 10),
		      strtoul (std_get (conf, "last_uid"), NULL, 10),
		      &uid, &gid);

      shell (1, "groupadd -g %lu -- '%s'", (unsigned long) uid, usr);
      shell (1, "useradd -d '%s' -g '%s' -G '%s','%s' -u %lu -m "
	     "-s /bin/false -c '%s',,,, -- '%s'",
	     st_value (&homedir), usr, usr, grp, (unsigned long) uid,
	     fullname, usr);
    }

  connect_db ();
  
  {
    db_result *r;

    db_query ("SELECT count(*) FROM users WHERE usr = '%sq'",
	      DB_FORMAT | DB_TUPLES, &r, usr);
    if (stol (dbr_value (r, 0, 0)) == 0)
      {
	log ("inserting %s into database USERS table...", usr);
	db_query ("INSERT INTO users "
		  "VALUES ('%sq', '%sq', '%sq', 't', "
		  "60 * 20, 'Jan 1 1970', 0, '', 't')",
		  DB_FORMAT, NULL, usr, fullname, email);
      }
    dbr_destroy (r);
  }

  add_user_to_teamscope_group (usr, grp);
}

/* Removes USR from TeamSCOPE group GRP. */
static void
del_user_from_teamscope_group (const char *usr, const char *grp)
{
  db_query ("DELETE FROM user_group WHERE usr = '%sq' AND grp = '%sq'",
	    DB_FORMAT, NULL, usr, grp);
}

/* Removes USR from UNIX system group GRP. */
static void
del_user_from_unix_group (const char *usr, const char *grp)
{
  if (user_is_member (usr, grp))
    {
      struct st command;

      st_create (&command, "usermod -G ");
      get_users_groups (usr, &command, NULL, grp);
      st_printf (&command, " -- '%s'", usr);
      shell (0, st_value (&command));
      st_destroy (&command);
    }
}

/* Removes TeamSCOPE and UNIX system group GRP. */
static void
del_group (const char *grp)
{
  struct group *group;
  
  log ("deleting group %s", grp);

  group = getgrnam (grp);
  if (group == NULL)
    {
      log ("no such group %s", grp);
      return;
    }

  if (*group->gr_mem && strcmp (*group->gr_mem, "scope"))
    fail ("can't delete group %s: still contains member %s",
	  grp, *group->gr_mem);

  connect_db ();
  
  db_query ("DELETE FROM groups WHERE grp = '%sq'", DB_FORMAT, 0, grp);

  log ("deleting UNIX group...");
  shell (1, "groupdel '%s'", grp);

  log ("leaving %s shared directory for manual removal...", grp);
}
  
/* Deletes TeamSCOPE and UNIX system user USR. */
static void
de_luser (const char *usr)
{
  log ("removing user %s...", usr);

  connect_db ();

  db_begin ();
  db_query ("DELETE FROM user_group WHERE usr = '%sq'", DB_FORMAT, NULL, usr);
  db_query ("DELETE FROM users WHERE usr = '%sq'", DB_FORMAT, NULL, usr);
  db_end ();

  if (getpwnam (usr) != NULL)
    {
      log ("deleting UNIX user %s...", usr);
      shell (1, "userdel -- '%s'", usr);
    }

  if (getgrnam (usr) != NULL)
    {
      log ("deleting UNIX group %s...", usr);
      shell (1, "groupdel '%s'", usr);
    }

  set_password (usr, "", 0);

  log ("leaving %s home directory for manual removal...", usr);
}

/* Sends a new-user mail to USR.  If RESET is nonzero, or if the user
   doesn't yet have a password, then the user is randomly assigned a
   new password. */
static void
send_user_mail (const char *usr, int reset)
{
  char password[9];
  struct std *subst;
  
  usr_to_uid (usr);
  if (!reset)
    {
      db_result *r;

      db_query ("SELECT count(*) FROM user_group "
		"WHERE usr = '%sq' AND sent_email = 't'",
		DB_FORMAT | DB_TUPLES, &r, usr);
      if (stol (dbr_value (r, 0, 0)) == 0)
	reset = 1;
      dbr_destroy (r);
    }

  subst = std_create (0);
  std_put (subst, "username", usr);
  std_put (subst, "servername", std_get (conf, "server_name"));

  if (reset)
    {
      randpass (password);
      std_put (subst, "password", password);
      set_password (usr, password, 0);
    }
  else
    std_put (subst, "password", NULL);

  
  {
    db_result *r;

    db_query ("SELECT email FROM users WHERE usr = '%sq'",
	      DB_FORMAT | DB_TUPLES, &r, usr);
    if (dbr_nrows (r) > 0)
      std_put (subst, "emailaddress", dbr_value (r, 0, 0));
    dbr_destroy (r);
  }

  {
    struct st s;
    db_result *r;
    size_t i;

    db_query ("SELECT grp FROM user_group WHERE usr = '%sq' ORDER BY grp",
	      DB_FORMAT | DB_TUPLES, &r, usr);
    /* FIXME: should check that db_nrows > 0. */

    std_put (subst, "teamname", dbr_value (r, 0, 0));

    st_init (&s, 32);
    for (i = 0; i < dbr_nrows (r); i++)
      {
	if (i != 0)
	  st_concat (&s, ", ");
	st_concat (&s, dbr_value (r, i, 0));
      }
    std_put (subst, "teams", st_value (&s));
    st_destroy (&s);
  }
  
  {
    db_result *r;
    struct st s;
    const char *lastgrp = "";
    size_t i;

    db_query ("SELECT grp, users.usr, name, last_word (name) AS last_name "
	      "FROM user_group, users "
	      "WHERE users.usr = user_group.usr "
	      "AND grp IN (SELECT grp FROM user_group "
	      "WHERE usr = '%sq') "
	      "ORDER BY grp, last_name, name",
	      DB_FORMAT | DB_TUPLES, &r, usr);
    st_init (&s, 128);
    for (i = 0; i < dbr_nrows (r); i++)
      {
	const char *grp = dbr_value (r, i, 0);
	const char *usr = dbr_value (r, i, 1);
	const char *name = dbr_value (r, i, 2);

	if (strcmp (grp, lastgrp))
	  {
	    st_printf (&s, "  Team `%s':\n", grp);
	    lastgrp = grp;
	  }
	if (strcmp (name, usr))
	  st_printf (&s, "    %s `%s'\n", name, usr);
	else
	  st_printf (&s, "    %s\n", usr);
      }
    st_truncate (&s, st_length (&s) - 1);

    std_put (subst, "others", st_value (&s));
    st_destroy (&s);
  }

  {
    FILE *dst;
    char *cmd, *src_fn;

    cmd = blp_asprintf ("mail -s '%s account information' -c scope %s",
			std_get (conf, "server_name"), usr);
    dst = open_pipe (cmd, "w");
    src_fn = blp_asprintf ("%s/new-user-mail", std_get (conf, "conf_dir"));

    copy_file_subst (src_fn, dst, subst);
    close_pipe (dst, cmd);
    free (src_fn);
    free (cmd);
  }
  
  db_query ("UPDATE user_group SET sent_email = 't' WHERE usr = '%sq'",
	    DB_FORMAT, NULL, usr);

  std_destroy (subst);
}

/* Updates the ``team site'' for GRP. */
static void
update_team_site (const char *grp)
{
  gid_t gid;
  const char *webdir, *fullname;
  struct st src, dst;
  db_result *r;
  struct std *subst;
  
  {
    struct group *group = getgrnam (grp);
    if (group == NULL)
      {
	log ("can't update team site for %s: no such UNIX group %s", grp, grp);
	return;
      }
    gid = group->gr_gid;
  }

  db_query ("SELECT webdir, name FROM groups WHERE grp = '%sq'",
	    DB_FORMAT | DB_TUPLES, &r, grp);
  webdir = dbr_value (r, 0, 0);
  fullname = dbr_value (r, 0, 1);

  subst = std_create (0);
  std_put (subst, "group_name", grp);
  std_put (subst, "group_full_name", fullname);
  {
    struct std_iterator iter;
    const char *key;
    char *value;
    
    std_init_iterator (conf, &iter);
    while (std_iterate (&iter, &key, &value))
      std_put (subst, key, value);
  }

  st_init (&src, 64);
  st_init (&dst, 64);
  
  st_printf (&src, "%s/index.shtml.default", std_get (conf, "conf_dir"));
  st_printf (&dst, "%s/index.shtml", webdir);
  copy_file_conditionally (0664, gid, st_value (&src), st_value (&dst),
			   subst, 1);

  st_clear (&src);
  st_clear (&dst);

  st_printf (&src, "%s/htaccess.default", std_get (conf, "conf_dir"));
  st_printf (&dst, "%s/.htaccess", webdir);
  copy_file_conditionally (0664, gid, st_value (&src), st_value (&dst),
			   subst, 2);

  st_clear (&src);
  st_clear (&dst);

  st_printf (&src, "%s/htaccess.public.default", std_get (conf, "conf_dir"));
  st_printf (&dst, "%s/public/.htaccess", webdir);
  copy_file_conditionally (0664, gid, st_value (&src), st_value (&dst),
			   subst, 2);

  st_destroy (&src);
  st_destroy (&dst);
  dbr_destroy (r);
  std_destroy (subst);
}

/* Password management. */

static void entropy_fill (char buf[], size_t count);
static void randsalt (char buf[3]);

/* Generates a random 8-character null-terminated password into
   BUF. */
static void
randpass (char buf[9])
{
  static const char alphabet[32]
    = "ABCDEFGHJKLMNPQRSTWXY3456789(*%)";

  int i;

  entropy_fill (buf, 8);

  for (i = 0; i < 8; i++)
    buf[i] = alphabet[((unsigned char) buf[i]) % 32];
  buf[8] = '\0';
}

/* For USR, sets the password to PWD if PWD is non-null, without
   affecting the account's locked/unlocked status.  If PWD is null,
   locks or unlocks the account, depending on LOCK_IT, without
   changing the password.  If PWD is the empty string (not a null
   pointer), then the line in the password file for USR is completely
   deleted.  Don't try to both lock/unlock an account and set its
   password at once. */
static void
set_password (const char *usr, const char *pwd, int lock_it)
{
  char *line = NULL;
  size_t size = 0;
  
  char salt[13];
  char *crypted;

  int lock;
  FILE *src, *dst;
  char *src_fn, *dst_fn;

  int found = 0;

  if (pwd != NULL)
    {
      randsalt (salt);
      crypted = crypt (pwd, salt);
    }

  src_fn = blp_asprintf ("%s/passwd", std_get (conf, "conf_dir"));
  dst_fn = blp_asprintf ("%s.new.%ld", src_fn, (long) getpid ());
  lock = lock_file (src_fn, 1);
  src = open_file (src_fn, "r");
  dst = open_file (dst_fn, "w");

  if (pwd != NULL && pwd[0] == '\0')
    found = 1;
  for (;;)
    {
      if (-1 == getline (&line, &size, src))
	break;

      if (!strncmp (line, usr, strlen (usr)) && line[strlen (usr)] == ':')
	{
	  found = 1;
	  if (pwd == NULL || pwd[0] != '\0')
	    {
	      fprintf (dst, "%s:", usr);
	      if (pwd != NULL)
		{
		  if (line[strlen (usr) + 1] == '!')
		    putc ('!', dst);
		  fprintf (dst, "%s\n", crypted);
		}
	      else
		{
		  char *p;

		  if (lock_it)
		    putc ('!', dst);

		  p = strchr (line, ':') + 1;
		  if (*p == '!')
		    p++;
		  fputs (p, dst);
		}
	    }
	}
      else
	fputs (line, dst);
    }
  if (!found)
    {
      if (pwd != NULL)
	fprintf (dst, "%s:%s\n", usr, crypted);
      else
	fail ("%s: user not found in password file", usr);
    }
  close_file (dst, dst_fn);
  close_file (src, src_fn);

  replace_file (src_fn, NULL);
  unlock_file (lock, src_fn);

  free (src_fn);
  free (dst_fn);
  free (line);
}

/* Fills COUNT bytes of BUF with cryptographically secure random
   values. */
static void
entropy_fill (char buf[], size_t count)
{
  FILE *f = open_file ("/dev/urandom", "rb");
  if (fread (buf, 1, count, f) != count)
    fail ("/dev/urandom: read: %s", strerror (errno));
  close_file (f, "/dev/urandom");
}

/* Generate a random MD5 salt into SALT. */
static void
randsalt (char salt[13])
{
  unsigned char bits[8];

  /* Get random bits. */
  entropy_fill (bits, sizeof bits);

  /* Convert bits to desired format. */
  {
    static const char salts[64]
      = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789./";
    int i;

    strcpy (salt, "$1$........$");
    for (i = 0; i < 8; i++)
      salt[i + 3] = salts[bits[i] & 63];
  }
}

/* Miscellaneous. */

/* Executes specified shell command FORMAT and checks its return
   value.  If EXPAND is nonzero then FORMAT is formatted with
   printf(). */
static void
shell (int expand, const char *format, ...)
{
  const char *s;
  char buf[4096];
  
  if (expand)
    {
      va_list args;
  
      int n;

      va_start (args, format);
      n = vsnprintf (buf, 4096, format, args);
      va_end (args);
  
      if (n < 0 || n >= 4095)
	fail ("shell command too long: %s", format);
      s = buf;
    }
  else
    s = format;

  check_system_value (system (s), s);
}

/* Given that S is a shell command and RESULT is the return value from
   running it, checks that everything went okay. */
static void
check_system_value (int result, const char *s)
{
  if (result == -1)
    fail ("couldn't create shell process: %s", s);
  else if (WEXITSTATUS (result) != 0)
    fail ("shell process returned error result %d: %s",
	  WEXITSTATUS (result), s);
}

/* Gets an input value from the command line in *PARGC and *PARGV, or,
   if command line arguments are exhausted, prompts the user with
   PROMPT and gets an answer from stdin.  Set PARGC and PARGV to null
   pointers to force input from stdin. */
static char *
prompt_input (int *pargc, char ***pargv, const char *prompt)
{
  char *result;

  if (pargc && *pargc > 0)
    {
      result = xstrdup ((*pargv)[0]);
      (*pargc)--;
      (*pargv)++;
    }
  else
    {
      printf ("Enter %s: ", prompt);
      fflush (stdout);

      {
	size_t size = 0;
	result = NULL;
	if (-1 == getline (&result, &size, stdin))
	  fail ("stdin: read: %s",
		ferror (stdin) ? strerror (errno) : "end of file");
      }
      
      {
	char *cp = strchr (result, '\n');
	if (cp != NULL)
	  *cp = '\0';
      }
    }

  return result;
}
