/* TeamSCOPE - Software for Collaborative Project Environments.
   Copyright (C) 1999, 2000, 2001 Michigan State University.
   Written by Ben Pfaff <pfaffben@msu.edu>.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To contact the TeamSCOPE team, send email to scope@cscw.msu.edu, or
   write to Charles Steinfield, Room 436 Communication Arts Building,
   Michigan State University, Department of Telecommunication, East
   Lansing MI 48824-1212. */

#include <config.h>
#include <assert.h>
#include <dirent.h>
#include <errno.h>
#include <fnmatch.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <time.h>
#include <unistd.h>
#include "activity.h"
#include "alloc.h"
#include "cgi.h"
#include "datetime.h"
#include "error.h"
#include "html.h"
#include "identity.h"
#include "scope.cgi.h"
#include "search.h"
#include "std.h"
#include "str.h"

static void file_search_form (void);
static void do_file_search (struct request *rq);

/* Handle the `search' screen, which just presents a menu. */
void
scr_search (struct request *rq unused)
{
  html ("<H2 ALIGN=CENTER>%s</H2>\n"
	"<P ALIGN=CENTER>%s</P>\n"
	"<DIV ALIGN=CENTER>\n",
	_("Search"),
	_("Select a search type from the options below."));

  html_button ("activitysearch", _("Activity Search"), (char *) NULL);
  html_button ("filesearch", _("Document Search"), (char *) NULL);

  htmls ("</DIV>\n");
}

/* Handle the `Activity Search' screen. */
void
scr_activitysearch (struct request *rq)
{
  struct activity_query query;
  
  html ("<H2 ALIGN=CENTER>%s</H2>\n",
	_("Activity Search"));
  html_start_form ("activity", HT_POST, HT_URL_ENCODED);

  activity_query_init (&query);
  query.search = 1;
  activity_show_filter (rq, &query);
  activity_query_destroy (&query);
}

/* Handle the `Document Search' screen. */
void
scr_filesearch (struct request *rq)
{
  if (std_contains (rq->cgi, "search"))
    do_file_search (rq);
  else
    file_search_form ();
}

/* Show a form offering various file search options. */
static void
file_search_form (void)
{
  html ("<H2 ALIGN=CENTER>%s</H2>\n",
	_("Document Search"));

  html_start_form ("filesearch", HT_POST, HT_URL_ENCODED);
  html_hidden ("search", "yes");

  html ("<P>%s:\n",
	_("Search for "));
  html_text_field ("string", "", 20);
  html_submit (_("Search"));

  html ("</P>\n\n"
	
	"<P>%s:<BR>\n"
	"<LABEL><INPUT TYPE=radio NAME=folder VALUE=shared CHECKED>%s"
	"</LABEL><BR>\n"
	"<LABEL><INPUT TYPE=radio NAME=folder VALUE=personal>%s</LABEL>\n"
	"</P>\n\n"

	"<P>%s:<BR>\n"
	"<LABEL><INPUT TYPE=radio NAME=match VALUE=substring CHECKED>%s"
	"</LABEL><BR>\n"
	"<LABEL><INPUT TYPE=radio NAME=match VALUE=wildcard>%s</LABEL><BR>"
	"<LABEL><INPUT TYPE=radio NAME=match VALUE=exact>%s</LABEL>\n\n"

	"<P><INPUT TYPE=checkbox NAME=caseinsensitive VALUE=yes CHECKED>%s\n"
	"</INPUT></P>\n\n"

	"<P>%s:<BR>\n"
	"<LABEL>%s <INPUT TYPE=text NAME=after></LABEL>\n"
	"<LABEL>%s <INPUT TYPE=text NAME=before></LABEL>\n"
	"</P>\n"

	"<P>%s:<BR>\n"
	"<LABEL><INPUT TYPE=radio NAME=sort VALUE=none CHECKED>%s"
	"</LABEL><BR>\n"
	"<LABEL><INPUT TYPE=radio NAME=sort VALUE=name>%s</LABEL><BR>\n"
	"<LABEL><INPUT TYPE=radio NAME=sort VALUE=date>%s</LABEL><BR>\n"
	"<LABEL><INPUT TYPE=radio NAME=sort VALUE=size>%s</LABEL><BR>\n"
	"</P>\n\n"

	"<P>%s:<BR>\n"
	"<LABEL><INPUT TYPE=radio NAME=limit VALUE=100 CHECKED>100"
	"</LABEL><BR>\n"
	"<LABEL><INPUT TYPE=radio NAME=limit VALUE=250>250</LABEL><BR>\n"
	"<LABEL><INPUT TYPE=radio NAME=limit VALUE=500>500</LABEL><BR>\n"
	"<LABEL><INPUT TYPE=radio NAME=limit VALUE=1000>1000</LABEL><BR>\n"
	"</P>\n\n",

	_("Search within"),
	_("Shared folder"),
	_("Personal folder"),

	_("Match type"),
	_("Substring match"),
	_("Wildcard match (? and *)"),
	_("Exact filename"),

	_("Match case-insensitively"),

	_("Modification date"),
	_("After"),
	_("but before"),

	_("Sort results by"),

	_("No sorting (fastest)"),
	_("Name"),
	_("Date"),
	_("Size"),

	_("Maximum number of results"));

  html_submit (_("Search"));

  html_end_form ();
}

/* Match types. */
enum
  {
    FS_MATCH_SUBSTRING,			/* Substring match. */
    FS_MATCH_WILDCARD,			/* Wildcard match. */
    FS_MATCH_EXACT,			/* Exact name match. */

    FS_MATCH_CASE_INSENSITIVE = 4	/* Applies to any above with |. */
  };

/* Sort types. */
enum
  {
    FS_SORT_NAME,			/* Sort by name. */
    FS_SORT_DATE,			/* Sort by date. */
    FS_SORT_SIZE,			/* Sort by size. */
    FS_SORT_NONE			/* No sorting. */
  };

/* Sort functions array indexed by FS_*. */
static int sort_by_name (const void *, const void *);
static int sort_by_date (const void *, const void *);
static int sort_by_size (const void *, const void *);

int (*sort_functions[FS_SORT_NONE]) (const void *, const void *) =
  {
    sort_by_name,
    sort_by_date,
    sort_by_size,
  };

/* Everything associated with a particular file search. */
struct file_search
  {
    const char *folder;		/* Name of folder to search, no '/' at end. */

    const char *string;		/* String to search for. */
    int match_type;		/* FS_MATCH_*. */
    
    int sort;			/* One of FS_SORT_*. */

    time_t after;		/* Start date/time. */
    time_t before;		/* End date/time. */

    struct file *results;	/* List of results. */
    int n_results;		/* Number of results so far. */
    int m_results;		/* Capicity of results array. */

    struct dir *dirlist;	/* List of directories seen so far. */
  };

/* A directory.  Used only so that we can keep track of directories
   separately from filenames in order to use less memory. */
struct dir
  {
    char *name;		/* Directory name. */
    struct dir *next;	/* Next directory in chain. */
  };

/* A file. */
struct file
  {
    char *dir;		/* Directory name, does not end in '/'. */
    char *name;		/* Filename within directory. */
    struct stat st;	/* File statistics. */
  };

static void file_search_from_cgi (struct request *, struct file_search *);
static int execute_file_search (struct request *, struct file_search *,
				const char *folder);
static void display_file (struct context *, struct file_search *,
			  const char *name, const char *dir, struct stat *);
static void display_search_results (struct request *, struct file_search *);

/* Perform a search for files. */
static void
do_file_search (struct request *rq)
{
  struct file_search search;
  
  /* Nice us down to the lowest level.  If we're lucky this will
     actually make a difference, but it probably won't since we're i/o
     heavy, not CPU heavy. */
  if (setpriority (PRIO_PROCESS, 0, PRIO_MAX) == -1)
    log ("setpriority failed: %s", strerror (errno));

  /* Obtain CGI form settings. */
  file_search_from_cgi (rq, &search);

  /* Warning messages. */
  html ("<P>%s</P>\n",
	_("Please wait.  This operation will take a while, especially\n"
	  "if there are many files and directories in the folder\n"
	  "to be searched.\n"));

  if (search.sort != FS_SORT_NONE)
    html ("<P>%s</P>\n",
	  _("<STRONG>Note:</STRONG> You have requested sorting of results.\n"
	    "Results may appear more slowly since all files and directories\n"
	    "must be read before any may be displayed.\n"));

  if (search.sort == FS_SORT_NONE)
    htmls ("<UL>\n");

  html_flush ();

  /* Turn off the soft CPU limit because this might take a while. */
  signal (SIGXCPU, SIG_IGN);
  
  /* Do the file search.
     If unsorted, also displays results. */
  if (chdir (search.folder) == -1)
    fail ("can't chdir to %s", search.folder);

  execute_file_search (rq, &search, search.folder);

  /* Display results if necessary. */
  if (search.sort == FS_SORT_NONE)
    htmls ("</UL>\n");
  else
    display_search_results (rq, &search);

  /* Finished! */
  html ("<P ALIGN=CENTER>%s</P>\n",
	search.n_results == 0
	? _("Search complete.  No matches found.")
	: _("Search complete."));
}

/* Extract form contents from CGI into SEARCH. */
static void
file_search_from_cgi (struct request *rq, struct file_search *search)
{
  search->string = std_get (rq->cgi, "string");
  
  if (!strcmp (std_get (rq->cgi, "folder"), "personal"))
    search->folder = rq->cx.home;
  else
    search->folder = rq->cx.shared;
  
  {
    const char *match = std_get (rq->cgi, "match");
    
    if (!strcmp (match, "substring"))
      search->match_type = FS_MATCH_SUBSTRING;
    else if (!strcmp (match, "wildcard"))
      search->match_type = FS_MATCH_WILDCARD;
    else
      search->match_type = FS_MATCH_EXACT;
  }

  if (std_contains (rq->cgi, "caseinsensitive"))
    search->match_type |= FS_MATCH_CASE_INSENSITIVE;
    
  {
    const char *sort = std_get (rq->cgi, "sort");

    if (!strcmp (sort, "none"))
      search->sort = FS_SORT_NONE;
    else if (!strcmp (sort, "name"))
      search->sort = FS_SORT_NAME;
    else if (!strcmp (sort, "date"))
      search->sort = FS_SORT_DATE;
    else if (!strcmp (sort, "size"))
      search->sort = FS_SORT_SIZE;
  }

  search->after = datetime_parse (std_get (rq->cgi, "after"), 0, 0, 0);
  search->before = datetime_parse (std_get (rq->cgi, "before"), 0, 0, 0);

  search->m_results = stoi (std_get (rq->cgi, "limit"));
  if (search->m_results < 100)
    search->m_results = 100;
  else if (search->m_results > 1000)
    search->m_results = 1000;

  search->results = xmalloc (sizeof *search->results * search->m_results);
  search->n_results = 0;

  search->dirlist = NULL;
}

static int select_file (struct file_search *, const char *name, struct stat *);
static int submit_file (struct context *, struct file_search *,
			const char *name, const char *dirname,
			struct stat *);
static void add_dir (struct file_search *search, char *dirname);

/* Search directory DIRNAME for files according to SEARCH.
   DIRNAME is the current directory.
   Returns nonzero if the caller may continue search,
   or zero if it should terminate the search.
   If SEARCH is sorted, then files found are added to the file list.
   Otherwise, files found are displayed.
   Either way, n_results is incremented appropriately. */
static int
execute_file_search (struct request *rq, struct file_search *search,
		     const char *dirname)
{
  DIR *dir;
  char *dircopy;

  dir = opendir (".");
  if (dir == NULL)
    {
      log ("error opening directory %s: %s", dirname, strerror (errno));
      return 1;
    }
  
  dircopy = NULL;
  for (;;)
    {
      struct stat st;
      struct dirent *de;

      de = readdir (dir);
      if (de == NULL)
	break;

      if (de->d_name[0] == '.'
	  && (de->d_name[1] == '\0'
	      || (de->d_name[1] == '.' && de->d_name[2] == '\0')))
	continue;

      if (lstat (de->d_name, &st) == -1)
	continue;
      
      if (select_file (search, de->d_name, &st))
	{
	  if (dircopy == NULL)
	    {
	      dircopy = xstrdup (dirname);
	      add_dir (search, dircopy);
	    }

	  if (submit_file (&rq->cx, search, de->d_name, dircopy, &st) == 0)
	    {
	      closedir (dir);
	      return 0;
	    }
	}

      if (S_ISDIR (st.st_mode) && chdir (de->d_name) == 0)
	{
	  char *s = blp_asprintf ("%s/%s", dirname, de->d_name);

	  if (execute_file_search (rq, search, s) == 0)
	    {
	      free (s);
	      closedir (dir);
	      return 0;
	    }
	  
	  if (chdir ("..") == -1)
	    fail ("can't back out of directory %s", s);
	  free (s);
	}

    }

  closedir (dir);

  return 1;
}

/* Returns nonzero only if a file named NAME with statistics ST should
   be selected for file search SEARCH. */
static int 
select_file (struct file_search *search, const char *name, struct stat *st)
{
  if (search->string[0] == 0)
    return 1;
  
  if ((search->after && st->st_mtime < search->after)
      || (search->before && st->st_mtime > search->before))
    return 0;

  switch (search->match_type)
    {
    case FS_MATCH_SUBSTRING:
      return strstr (name, search->string) != NULL;
    case FS_MATCH_SUBSTRING | FS_MATCH_CASE_INSENSITIVE:
      return blp_strcasestr (name, search->string) != NULL;

    case FS_MATCH_WILDCARD:
      return fnmatch (search->string, name, FNM_PERIOD);
    case FS_MATCH_WILDCARD | FS_MATCH_CASE_INSENSITIVE:
      return fnmatch (search->string, name, FNM_PERIOD | FNM_CASEFOLD);

    case FS_MATCH_EXACT:
      return !strcmp (search->string, name);
    case FS_MATCH_EXACT | FS_MATCH_CASE_INSENSITIVE:
      return !strcasecmp (search->string, name);

    default:
      abort ();
    }
}

/* File NAME in directory DIRNAME with statistics ST has been selected
   for file search S.  Deal with it by adding it to the list or
   displaying it, as appropriate for the sort type. */
static int 
submit_file (struct context *cx, struct file_search *s,
	     const char *name, const char *dirname,
	     struct stat *st)
{
  if (s->sort == FS_SORT_NONE)
    {
      display_file (cx, s, name, dirname, st);
      html_flush ();
    }
  else
    {
      struct file *file;

      if (s->n_results >= s->m_results)
	{
	  if (s->m_results == 0)
	    s->m_results = 16;
	  else
	    s->m_results *= 2;

	  s->results = xrealloc (s->results,
				 s->m_results * sizeof *s->results);
	}
      
      file = &s->results[s->n_results];
      file->name = xstrdup (name);
      file->dir = (char *) dirname;
      file->st = *st;
    }

  s->n_results++;
  return s->n_results < s->m_results;
}

/* Add directory DIRNAME to the list in file search S. */
static void
add_dir (struct file_search *s, char *dirname)
{
  struct dir *dir = xmalloc (sizeof *dir);
  dir->name = dirname;
  dir->next = s->dirlist;
  s->dirlist = dir;
}

/* Display the file with name NAME in directory DIR and statistics ST
   as appropriate for file search SEARCH. */
static void 
display_file (struct context *cx, struct file_search *search,
	      const char *name, const char *dir, struct stat *st)
{
  static struct st *s;
  struct tm *tm;
  const char *date;

  if (s == NULL)
    {
      s = xmalloc (sizeof *s);
      st_init (s, 128);
    }
  st_clear (s);
  st_printf (s, "%s/%s", dir, name);

  tm = localtime (&st->st_mtime);
  date = datetime_format ("%b %d %Y %H:%M:%S", tm);

  if (search->sort == FS_SORT_NONE)
    {
      htmls ("<LI>");
      html_a_file (cx, st_value (s), NULL);
      html (", %s %.1f %s, %s %s\n",
	    _("size"), st->st_size / 1024., _("kB"),
	    _("last modified"), date);
    }
  else
    {
      htmls ("  <TR>\n"
	     "    <TD>");
      html_a_file (cx, st_value (s), NULL);
      html ("</TD>\n"
	    "    <TD>%.1f kB</TD>\n"
	    "    <TD>%s</TD>\n"
	    "  </TR>\n",
	    st->st_size / 1024., date);
    }
}

/* Display the results of file search SEARCH. */
static void
display_search_results (struct request *rq, struct file_search *search)
{
  int i;
  
  html ("<TABLE BORDER=1>\n"
	"  <TR>\n"
	"    <TD>%s</TD>\n"
	"    <TD>%s</TD>\n"
	"    <TD>%s</TD>\n"
	"  </TR>\n",
	_("Filename"),
	_("Size"),
	_("Date"));

  qsort (search->results, search->n_results, sizeof *search->results,
	 sort_functions[search->sort]);

  for (i = 0; i < search->n_results; i++)
    {
      struct file *file = &search->results[i];
      
      display_file (&rq->cx, search, file->name, file->dir, &file->st);
    }
      
  htmls ("</TABLE>\n");
}

/* Returns a strcmp()-type value for the names of files represented by
   VA and VB. */
static int 
sort_by_name (const void *va, const void *vb)
{
  const struct file *a = va;
  const struct file *b = vb;
  int r;

  r = strcmp (a->dir, b->dir);
  if (r)
    return r;
  return strcmp (a->name, b->name);
}

/* Returns a strcmp()-type value for the dates of files represented by
   VA and VB. */
static int 
sort_by_date (const void *va, const void *vb)
{
  const struct file *a = va;
  const struct file *b = vb;

  if (a->st.st_mtime > b->st.st_mtime)
    return 1;
  else if (a->st.st_mtime < b->st.st_mtime)
    return -1;
  else
    return 0;
}

/* Returns a strcmp()-type value for the sizes of files represented by
   VA and VB. */
static int 
sort_by_size (const void *va, const void *vb)
{
  const struct file *a = va;
  const struct file *b = vb;

  if (a->st.st_size > b->st.st_size)
    return 1;
  else if (a->st.st_size < b->st.st_size)
    return -1;
  else
    return 0;
}

