/* TeamSCOPE - Software for Collaborative Project Environments.
   Copyright (C) 1999, 2000, 2001 Michigan State University.
   Written by Ben Pfaff <pfaffben@msu.edu>.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To contact the TeamSCOPE team, send email to scope@cscw.msu.edu, or
   write to Charles Steinfield, Room 436 Communication Arts Building,
   Michigan State University, Department of Telecommunication, East
   Lansing MI 48824-1212. */

#include <config.h>
#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <time.h>
#include <unistd.h>
#include "alloc.h"
#include "auth.h"
#include "cgi.h"
#include "conf.h"
#include "db.h"
#include "error.h"
#include "html.h"
#include "identity.h"
#include "scope.cgi.h"
#include "session.h"
#include "str.h"
#include "std.h"

/*#undef DEBUGGING*/

struct std *conf;

/* Public functions. */

static int resume_session (struct request *);
static void new_session (struct request *);
static void clear_data (struct request *);
static int file_exists (const char *filename, time_t *expiry);
static char *trusted_filename (const char *session_id);
static char *untrusted_filename (const char *session_id);

/* Initialize the session manager and load trusted data. */
void
session_initialize (struct request *rq)
{
  if (!resume_session (rq))
    new_session (rq);
  else
    log ("Continuing session %.20s-<censored>", rq->session_id);
}

/* Save the current session's trusted data. */
void
session_save_trusted (struct request *rq)
{
  if (std_modified (rq->trusted))
    {
      char *fn = trusted_filename (rq->session_id);
      if (!std_write (rq->trusted, fn, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH))
	log ("couldn't save session trusted data %.20s-<censored>",
	     rq->session_id);
      free (fn);
    }

#if DEBUGGING
  fprintf (stderr, "Trusted data:\n");
  std_print (rq->trusted);
#endif
}

/* Load untrusted data. */
void
session_read_untrusted (struct request *rq)
{
  char *fn = untrusted_filename (rq->session_id);
  rq->untrusted = std_create (1);
  if (file_exists (fn, NULL)
      && !std_read (rq->untrusted, fn, 0, rq->cx.uid))
    fail ("error reading untrusted data for session %.20s-<censored>",
	  rq->session_id);
  free (fn);
}

/* Save the current session's untrusted data. */
void
session_save_untrusted (struct request *rq)
{
  if (std_modified (rq->untrusted))
    {
      char *fn = untrusted_filename (rq->session_id);
      if (!std_write (rq->untrusted, fn, S_IRUSR | S_IWUSR | S_IROTH))
	log ("couldn't save untrusted data for session %.20s-<censored>",
	     rq->session_id);
      free (fn);
    }

#if DEBUGGING
  fprintf (stderr, "Untrusted data:\n");
  std_print (rq->untrusted);
#endif
}

/* Send a login screen to the client. */
void
scr_auth (struct request *rq)
{
  const char *groups;
  
  html ("<H2 ALIGN=CENTER>%s</H2>\n",
	std_contains (rq->cgi, "need")
	? _("Authorization Required") : _("TeamSCOPE Login"));
  if (!strcmp (rq->screen, "authfailed"))
    {
      html ("<P><B>%s</B></P>\n"
	    "<UL>\n"
	    "  <LI><B>%s</B>\n"
	    "      %s\n"
	    "  <LI><B>%s</B>\n"
	    "      %s\n"
	    "  <LI><B>%s</B>\n"
	    "      %s\n"
	    "  <LI><B>%s</B>\n"
	    "      %s\n"
	    "  <LI><B>%s</B>\n"
	    "      %s\n"
	    "</UL>\n",
	    _("Authorization failed.\n"
	      "You entered an unrecognized username or incorrect password.\n"
	      "Possible causes include:"),
	    _("You mistyped your username or password."),
	    _("Try again."),
	    _("Your browser does not support cookies or they are disabled."),
	    _("Enable cookies or switch to a browser that supports them."),
	    _("You have forgotten your username or password."),
	    _("Ask the TeamSCOPE administrator for assistance."),
	    _("Your account has been deleted or disabled."),
	    _("Ask the TeamSCOPE administrator for assistance."),
	    _("You have discovered a bug in TeamSCOPE."),
	    _("Ask the TeamSCOPE administrator for assistance."));
    }
  else
    {
      htmls ("<P ALIGN=CENTER>\n");
      htmls (_("Welcome to TeamSCOPE.\n"));
      if (rq->cx.usr[0] == '\0')
	htmls (_("You are not logged in, or your login information has expired.\n"
		 "To log in, enter your login name and password below,\n"
		 "then click on the <STRONG>Log In</STRONG> button to log in.\n"));
      else
	htmls (_("You have attempted to access a page for which you do not\n"
		 "have access credentials.\n"
		 "You may supply the necessary credentials below by\n"
		 "entering an alternate login name and password,\n"
		 "then clicking on\n"
		 "the <STRONG>Log In</STRONG> button.\n"
		 "After doing so, you will be redirected to the page\n"
		 "you were attempting to access.\n"));

      htmls (_("If you've already done that, then you may need to turn on\n"
	       "cookies in your browser.\n"));
      htmls ("</P>\n\n");
    }

  if (rq->cx.usr[0] != '\0')
    html ("<H2 ALIGN=CENTER>%s</H2>\n"
	  "<TABLE ALIGN=CENTER>\n"
	  "<TR><TD ALIGN=RIGHT>%s:<TD>%s\n"
	  "<TR><TD ALIGN=RIGHT>%s:<TD>%s\n"
	  "</TABLE>\n\n",
	  _("Current Credentials"),
	  _("Username"), rq->cx.usr,
	  _("Workgroup"), rq->cx.grp);

  groups = NULL;
  if (std_contains (rq->cgi, "need")
      || std_contains (rq->cgi, "needgrp"))
    {
      const char **need;
      
      html ("<H2 ALIGN=CENTER>%s</H2>\n"
	    "<P ALIGN=CENTER>%s</P>\n"
	    "<P ALIGN=CENTER>\n",
	    _("Required Credentials"),
	    _("The page you are attempting to access requires the following\n"
	      "credentials:\n"));

      for (need = std_get_many (rq->cgi, "need", 0);
	   need != NULL && *need != NULL; need++)
	{
	  char *s = local_strdup (*need);
	  char *tok = strtok (s, " \t\r\n\v");

	  if (tok != NULL)
	    {
	      if (!strcasecmp (tok, "valid-user"))
		html ("%s<BR>\n", _("Must be a validated TeamSCOPE user."));
	      else if (!strcasecmp (tok, "user"))
		{
		  int count;
	      
		  html ("%s:<BR>\n",
			_("Must be one of the following TeamSCOPE users"));

		  for (count = 0; ; count++)
		    {
		      tok = strtok (NULL, " \t\r\n\v");
		      if (tok == NULL)
			break;

		      if (count != 0)
			htmls (",\n");
		      html ("<STRONG>%s</STRONG>", tok);
		    }
		  html ("<BR>\n");
		}
	      else if (!strcasecmp (tok, "group"))
		{
		  int count;

		  html ("%s:<BR>\n",
			_("Must belong to one of the following "
			  "TeamSCOPE groups"));

		  for (count = 0; ; count++)
		    {
		      tok = strtok (NULL, " \t\r\n\v");
		      if (tok == NULL)
			break;

		      if (count != 0)
			htmls (",\n");
		      html ("<STRONG>%s</STRONG>", tok);
		    }
		  htmls ("<BR>\n");

		  groups = *need;
		}
	    }
	  
	  local_free (s);
	}

      if (std_contains (rq->cgi, "needgrp"))
	html ("%s %s %s.\n",
	      _("Must belong to"), std_get (rq->cgi, "needgrp"),
	      _("TeamSCOPE group"));
      
    }

  html_start_form ("activity", HT_POST, HT_URL_ENCODED);
  if (std_contains (rq->cgi, "redirect"))
    html_hidden ("redirect", std_get (rq->cgi, "redirect"));
  else
    html_hidden ("redirect", getenv ("REQUEST_URI"));
  if (std_contains (rq->cgi, "need"))
    html_hidden ("need", std_get (rq->cgi, "need"));
  if (std_contains (rq->cgi, "needgrp"))
    html_hidden ("needgrp", std_get (rq->cgi, "needgrp"));
  if (groups)
    html_hidden ("groups", groups);

  htmls ("<CENTER>\n");
  html_label ("username", "Username");
  html_text_field ("username", "", 8);
  htmls ("<BR>\n");
  html_label ("password", "Password");
  html_password_field ("password", 8);
  htmls ("<BR>\n");
  html_submit (_("Log In"));
  htmls ("</CENTER>\n");

  html_end_form ();

  html_powered_by (rq);
}

/* Send an information screen to the client. */
void
session_info (struct request *rq)
{
  html ("<H2 ALIGN=CENTER>%s</H2>\n", _("Session Info"));

  html ("<P ALIGN=CENTER>%s</P>\n"
	"<TABLE ALIGN=CENTER>\n"
	"<TR><TD ALIGN=RIGHT>%s:<TD>%s\n"
	"<TR><TD ALIGN=RIGHT>%s:<TD>%s\n"
	"</TABLE>\n",
	_("Welcome to TeamSCOPE.\n"
	  "You are currently logged in as follows:\n"),
	_("Username"), rq->cx.usr,
	_("Workgroup"),	rq->cx.grp);

  htmls ("<DIV ALIGN=CENTER>\n");
  html_button ("logout", _("Log Out"), (char *) NULL);
  htmls ("</DIV>\n");
}

/* Log out the client. */
void
session_logout (struct request *rq)
{
  char *fn;
  
  printf ("Set-Cookie: scope-session=\"\";Version=1;Path=/;Max-Age=0\n");

  fn = trusted_filename (rq->session_id);
  if (-1 == remove (fn) && errno != ENOENT)
    fail ("error removing trusted data for session %.20s-<censored>: %s",
	  fn, strerror (errno));
  free (fn);

  fn = untrusted_filename (rq->session_id);
  if (-1 == remove (fn) && errno != ENOENT)
    fail ("error removing untrusted data for session %.20s-<censored>: %s",
	  fn, strerror (errno));
  free (fn);
  
  http_header ();
  html_banner (rq);

  html ("<H2 ALIGN=CENTER>%s</H2>\n", _("Log Out"));

  html ("<P ALIGN=CENTER>\n"
	"%s\n"
	"</P>\n",
	_("Thanks for using TeamSCOPE.\n"
	  "You are now logged out."));

  htmls ("<DIV ALIGN=CENTER>\n");
  html_button ("", _("Log In"), (char *) NULL);
  htmls ("</DIV>\n");

  html_trailer (rq);

  exit (EXIT_SUCCESS);
}

/* Private functions. */

static void generate_session_id (char[42]);

/* Attempt to resume the session indicated by `session_id' by
   verifying that it exists, hasn't expired, and can be read.  Returns
   nonzero on success, zero for failure.
   If successful, RQ->trusted and RQ->session_id are initialized, as
   well as the global variable `conf'. */
static int
resume_session (struct request *rq)
{
  int fd;	/* File descriptor. */

  /* Find out what file descriptor will be used to send us the
     session-id. */
  {
    const char *string = getenv ("SCOPE_SESSION");
    if (string == NULL)
      string = getenv ("REDIRECT_SCOPE_SESSION");
    if (string == NULL)
      {
#if DEBUGGING
	log ("can't resume session: no SCOPE_SESSION or "
	     "REDIRECT_SCOPE_SESSION environment var");
#endif
	return 0;
      }

    fd = stoi (string);
  }

  /* Read the session-id. */
  for (;;)
    {
      ssize_t amt = read (fd, rq->session_id, 41);
      if (amt == 41)
	break;
      if (amt == -1 && errno == EINTR)
	continue;

      log ("can't resume session: error reading from SCOPE_SESSION file: %s",
	   strerror (errno));
      return 0;
    }

  rq->session_id[41] = 0;
  close (fd);

  /* Verify that the session-id is in the correct form. */
  {
    const char *cp;
    
    for (cp = rq->session_id; *cp; cp++)
      if (strchr ("0123456789ABCDEF-", *cp) == NULL)
	return 0;

    if (cp - rq->session_id != 41)
      {
#if DEBUGGING
	log ("can't resume session: scope-session in wrong form");
#endif
	return 0;
      }
  }

  /* Attempt to find and read the trusted session data. */
  {
    char *fn = trusted_filename (rq->session_id);
    time_t age;
    int idle;
    
    if (!file_exists (fn, &age))
      {
	clear_data (rq);
#if DEBUGGING
	log ("can't resume session: session data missing or expired");
#endif
	free (fn);
	return 0;
      }

    rq->trusted = std_create (1);

    if (!std_read (rq->trusted, fn, 0, 0))
      {
	log ("can't resume session: error reading session data");
	clear_data (rq);
	free (fn);
	return 0;
      }

    /* Determine maximum idle time. */
    idle = stoi (std_get (rq->trusted, "idle"));

    if (age > idle)
      {
#if DEBUGGING
	log ("can't resume session: too much idle time (%d > %d seconds)",
	     (int) age, idle);
#endif
	clear_data (rq);
	free (fn);
	return 0;
      }

    free (fn);
  }

  conf = std_get_subdict (rq->trusted, "conf");

  return 1;
}

/* Start a new session by generating a random session ID and sending
   out a cookie for it. */
static void
new_session (struct request *rq)
{
  generate_session_id (rq->session_id);

  /* For some reason Nutscrape doesn't accept cookies that have spaces
     between components.  Apparently jwz can't read RFCs. */
  printf ("Set-Cookie: scope-session=\"%s\";Version=1;Path=/\n",
	  rq->session_id);

  rq->trusted = std_create (1);
  rq->untrusted = std_create (1);

  /* Set default idle time of 20 minutes. */
  std_put (rq->trusted, "idle", "1200");

  session_read_conf ();
  std_put_subdict (rq->trusted, "conf", conf);
}

/* Reads the `conf' dictionary from disk. */
void
session_read_conf (void)
{
  conf = std_create (0);
  if (getenv ("SERVER_NAME"))
    std_put (conf, "server_name", getenv ("SERVER_NAME"));
  if (!std_read (conf, CONF_FILE, STDF_SUBST | STDF_TRIM_SPACES, 0))
    fail ("error reading " CONF_FILE ", can't continue");
}

/* Generate a cryptographically secure random session ID. */
static void
generate_session_id (char *id)
{
  unsigned char bits[20];

  /* Read 20 bytes of random noise into `bits'. */
  {
    int fd = open ("/dev/urandom", O_RDONLY);
    if (fd == -1)
      fail ("opening /dev/urandom: %s", strerror (errno));

    if (read (fd, bits, 20) != 20)
      fail ("reading /dev/urandom: %s", strerror (errno));

    if (close (fd) == -1)
      log ("closing /dev/urandom: %s", strerror (errno));
  }

  /* Format it as a pair of 20-digit hyphen-separated hexadecimal
     numbers. */
  {
    int i;

    for (i = 0; i < 20; i++)
      {
	static const char hex[16] = "0123456789ABCDEF";

	if (i == 10)
	  *id++ = '-';
	*id++ = hex[bits[i] / 16];
	*id++ = hex[bits[i] % 16];
      }
    *id++ = '\0';
  }

  log ("Starting new session %.20s-<censored>", id - 42);
}

/* Returns the filename used for trusted state of session named
   SESSION_ID. */
static char *
trusted_filename (const char *session_id)
{
  return blp_asprintf ("%s/%s", TRUSTED_STATE_DIR, session_id);
}

/* Returns the filename used for untrusted state of session named
   SESSION_ID. */
static char *
untrusted_filename (const char *session_id)
{
  return blp_asprintf ("%s/%s", UNTRUSTED_STATE_DIR, session_id);
}

/* Free trusted_data and untrusted_data and their filenames as
   well. */
static void
clear_data (struct request *rq)
{
  if (rq->trusted != NULL)
    {
      std_destroy (rq->trusted);
      rq->trusted = NULL;
    }

  if (rq->untrusted != NULL)
    {
      std_destroy (rq->untrusted);
      rq->untrusted = NULL;
    }
}

/* Returns nonzero if FILENAME exists.  If AGE is non-null then puts
   the number of seconds since the file was last modified into it.  */
static int
file_exists (const char *filename, time_t *age)
{
  struct stat s;
    
  if (-1 == stat (filename, &s))
    return 0;

  if (age != NULL)
    *age = cur_time - s.st_atime;

  return 1;
}
