/* TeamSCOPE - Software for Collaborative Project Environments.
   Copyright (C) 1998-9, 2000 Free Software Foundation, Inc.
   Copyright (C) 1999, 2000, 2001, 2001 Michigan State University.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To contact the TeamSCOPE team, send email to scope@cscw.msu.edu, or
   write to Charles Steinfield, Room 436 Communication Arts Building,
   Michigan State University, Department of Telecommunication, East
   Lansing MI 48824-1212. */

#include <config.h>
#include <ctype.h>
#include <assert.h>
#include <limits.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include "alloc.h"
#include "db.h"
#include "str.h"

/* Normal null-terminated string functions. */

/* Search for NEEDLE in HAYSTACK case-insensitively.  Returns a
   pointer to the leftmost instance of NEEDLE within HAYSTACK if
   successful; otherwise returns a null pointer. */
char *
blp_strcasestr (const char *haystack, const char *needle)
{
  size_t haystack_len = strlen (haystack);
  size_t needle_len = strlen (needle);
  size_t i;

  if (needle_len <= haystack_len)
    {
      for (i = 0; i <= haystack_len - needle_len; i++)
	if (!strncasecmp (haystack + i, needle, needle_len))
	  return (char *) haystack + i;
    }
  
  return NULL;
}

/* Formats FMT into a dynamically allocated string as with printf().
   Returns the string. */
char *
blp_asprintf (const char *fmt, ...)
{
  int size = 64;
  
  for (;;)
    {
      va_list args;
      char *buf;
      int n;
      
      buf = xmalloc (size);

      va_start (args, fmt);
      n = vsnprintf (buf, size, fmt, args);
      va_end (args);
      
      if (n < 0)
        size *= 2;
      else if (n < size)
        return buf;
      else
        size = n + 1;

      free (buf);
    }
}

/* Returns nonzero if STRING has PREFIX as a prefix. */
int
has_prefix (const char *string, const char *prefix)
{
  while (*prefix != '\0')
    if (*string++ != *prefix++)
      return 0;
  return 1;
}

/* Strips trailing spaces from STRING. */
void 
blp_strip_trailing_spaces (char *s)
{
  char *p = strchr (s, '\0');
  while (p > s && isspace ((unsigned char) p[-1]))
    p--;
  *p = '\0';
}

/* Returns nonzero only if S contains only whitespace characters. */
int
empty (const char *s)
{
  for (; *s != '\0'; s++)
    if (!isspace ((unsigned char) *s))
      return 0;
  return 1;
}

/* Returns the value of decimal integer STRING as an int. */
int
stoi (const char *string)
{
  long value = strtol (string, NULL, 10);
  return value > INT_MAX ? INT_MAX : value < INT_MIN ? INT_MIN : value;
}

/* st_* functions. */

/* Initializes ST with initial contents S. */
void
st_create (struct st *st, const char *s)
{
  st->len = strlen (s);
  st->size = 8 + st->len * 2;
  st->s = xmalloc (st->size + 1);
  strcpy (st->s, s);
}

/* Initializes ST, making room for at least SIZE characters. */
void
st_init (struct st *st, size_t size)
{
  st->len = 0;
  if (size > 8)
    st->size = size;
  else
    st->size = 8;
  st->s = xmalloc (st->size + 1);
}

/* Frees ST. */
void
st_destroy (struct st *st)
{
  free (st->s);
}

/* Truncates ST to zero length. */
void
st_clear (struct st *st)
{
  st->len = 0;
}

/* Trims off the last AMT characters of ST. */
void
st_trim (struct st *st, size_t amt)
{
  if (st->len >= amt)
    st->len -= amt;
  else
    st->len = 0;
}

/* Ensures that ST can hold at least MIN_SIZE characters plus a null
   terminator. */
void
st_extend (struct st *st, size_t min_size)
{
  assert (st->size >= st->len);
  if (min_size > st->size)
    {
      st->size *= 2;
      if (st->size < min_size)
	st->size = min_size * 2;
      
      st->s = xrealloc (st->s, st->size + 1);
      assert (st->size >= st->len);
    }
  assert (st->size >= min_size);
}

/* Truncates ST to at most LENGTH characters long. */
void
st_truncate (struct st *st, size_t length)
{
  if (length < st->len)
    st->len = length;
}

/* Returns the length of ST. */
size_t
st_length (const struct st *st)
{
  return st->len;
}

/* Returns the value of ST as a null-terminated string. */
char *
st_value (const struct st *st)
{
  ((char *) st->s)[st->len] = '\0';
  return st->s;
}

/* Concatenates S onto ST. */
void
st_concat (struct st *st, const char *s)
{
  size_t s_len = strlen (s);
  st_extend (st, st->len + s_len);
  memcpy (st->s + st->len, s, s_len);
  st->len += s_len;
}

/* Concatenates S onto ST, between quotes of type QUOTE ('
   or "), with instances of QUOTE, as well as \, in S quoted with a
   prefixed \. */
void
st_concat_quoted (struct st *st, const char *s, int quote)
{
  st_putchar (st, quote);
  for (; *s; s++)
    {
      if (*s == quote || *s == '\\')
	st_putchar (st, '\\');
      st_putchar (st, *s);
    }
  st_putchar (st, quote);
}

/* Appends S to ST, quoting characters which are reserved in URL query
   strings with %HH syntax. */
void
st_concat_url_query_quoted (struct st *st, const char *s)
{
  while (*s != '\0')
    {
      unsigned char ch = *s++;
      int quote = 0;
      
      switch (ch)
	{
	case ' ':
	  st_putchar (st, '+');
	  continue;
	  
	case '%': case '<': case '>': case '#': case '"':
	case '{': case '}': case '|': case '[': case ']':
	case '^': case '`': case ';': case '/': case '?':
	case ':': case '=': case ',': case '$': case '@':
	case '\\': case '&': case '+':
	  quote = 1;
	  break;

	default:
	  quote = !isprint (ch);
	  break;
	}

      if (!quote)
	st_putchar (st, ch);
      else
	st_printf (st, "%%%02X", ch);
    }
}

/* Appends S to ST, escaping HTML-reserved characters in HTML
   fashion. */
void
st_concat_html_quoted (struct st *st, const char *s)
{
  for (; ; s++)
    switch (*s)
      {
      case '\0':
	return;

      case '<':
	st_concat (st, "&lt;");
	break;
	
      case '>':
	st_concat (st, "&gt;");
	break;

      case '&':
	st_concat (st, "&amp;");
	break;

      case '"':
	st_concat (st, "&quot;");
	break;

      default:
	st_putchar (st, *s);
	break;
      }
}

/* Concatenates LEN characters from BUF onto ST. */
void
st_concat_buffer (struct st *st, const char *buf, size_t len)
{
  st_extend (st, st->len + len);
  memcpy (st->s + st->len, buf, len);
  st->len += len;
}

/* Formats FORMAT as a printf string and appends the result to ST. */
void
st_printf (struct st *st, const char *format, ...)
{
  /* Fscking glibc silently changed behavior between 2.0 and 2.1.
     Fsck fsck fsck.  Before, it returned -1 on buffer overflow.  Now,
     it returns the number of characters (not bytes) that would have
     been written. */
  va_list args;

  int avail, needed;

  assert (st != NULL && format != NULL && st->size >= st->len);

  va_start (args, format);
  avail = st->size - st->len + 1;
  needed = vsnprintf (st->s + st->len, avail, format, args);
  va_end (args);

  if (needed >= avail)
    {
      st_extend (st, st->len + needed);
      
      va_start (args, format);
      vsprintf (st->s + st->len, format, args);
      va_end (args);
    }
  else
    while (needed == -1)
      {
	st_extend (st, (st->size + 1) * 2);
	avail = st->size - st->len + 1;
	
	va_start (args, format);
	needed = vsnprintf (st->s + st->len, avail, format, args);
	va_end (args);
      }

  st->len += needed;
}

/* Formats FORMAT as a vformat string and appends the result to ST. */
void
st_format (struct st *st, const char *format, ...)
{
  va_list args;
  int avail, needed;

  va_start (args, format);
  avail = st->size - st->len + 1;
  needed = vformat (st->s + st->len, avail, format, args);
  va_end (args);

  if (needed > avail)
    {
      st_extend (st, st->len + needed);
      
      va_start (args, format);
      vformat (st->s + st->len, needed, format, args);
      va_end (args);
    }

  st->len += needed - 1;
}

/* Formats FORMAT as a strftime string for date-time DATE and appends
   the result to ST. */
void
st_strftime (struct st *st, const char *format, const struct tm *date)
{
  size_t needed;

  st_extend (st, st->len + 32);
  for (;;)
    {
      /* This assumes that you won't ever use a FORMAT string that
	 could result in a valid zero-length output (e.g., "%p" in a
	 24-hour time zone).  If that ever happens, you'll have to
	 take another approach. */
      size_t avail = st->size - st->len + 1;
      needed = strftime (st->s + st->len, avail, format, date);
      if (needed)
	break;

      st_extend (st, st->size * 2);
    }

  st->len += needed;
}

/* Appends character CH to ST. */
void
st_putchar (struct st *st, int ch)
{
  if (st->len >= st->size)
    st_extend (st, st->len + 1);

  /* Notice that we don't append a null byte.
     This is the responsibility of st_value(). */
  st->s[st->len++] = ch;
}

/* Appends to ST character CH repeated COUNT times. */
void
st_concat_dup (struct st *st, int ch, size_t count)
{
  st_extend (st, st->len + count);
  memset (st->s + st->len, ch, count);
  st->len += count;
}

/* Replaces the contents of ST by a line read from STREAM.
   Returns nonzero if successful. */
int
st_getline (struct st *st, FILE *stream)
{
  size_t n = st->size + 1;
  ssize_t len = getline (&st->s, &n, stream);
  st->size = n - 1;
  if (len != -1)
    {
      st->len = len;
      return 1;
    }
  else
    {
      st->len = 0;
      return 0;
    }
}

/* Returns the last character in ST,
   or EOF if it's empty. */
int
st_last_char (const struct st *st)
{
  return st->len > 0 ? st->s[st->len - 1] : EOF;
}
